#include <vcl_sstream.h>
#include <vcl_iostream.h>
#include <vcl_fstream.h>

#include <cstdlib>

#include <vil1/vil1_image.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_load.h>
#include <vil1/vil1_save.h>
#include <misc_utility/defines.h>                // def of color_buffer_t
#include <misc_utility/loaders.h>                // save_ints
#include <region_utility/Vitto_region_file.h>

#include <main_processing/mosaicpropagator.h>


void process_arguements(int argc, char** argv, bool& blade, bool& small, float& radius, char*& biasfile){
  for(int a = 8; a < argc; a++) {
    if(strcmp(argv[a],"-blade") == 0) {
      vcl_cout << "Blade propagation mode activated." << vcl_endl;
      blade = true;
    }

    if(strcmp(argv[a],"-small") == 0) {
      vcl_cout << "Small search space." << vcl_endl;
      small = true;
    }

    if(strcmp(argv[a],"-radius") == 0) {
      radius = atof(argv[a+1]);
    }

    if(strcmp(argv[a],"-bias") == 0) {
      vcl_cout << "Bias file " << argv[a+1] << vcl_endl;
      biasfile = argv[a+1];
    }
  }
}



int main(int argc, char** argv)
{
  if (argc < 8) {
    vcl_cout << "Usage: " << argv[0] 
	     << " image1.tif ref_regs1.regmV image2.tif ref_regs2.regmV cand_regs1.regmV cand_regs2.regmV ref_regs.dat ";
    vcl_cout << " [-blade] [-small] [-radius r] [-bias bias.dat]" << vcl_endl;
    vcl_cout << "Default mode = integrated propagation, not small refinement bounds, radius = 120, unbiased" << vcl_endl;
    return 1;
  }

  // Load reference regions
  vcl_cout << "Loading reference regions in Vitto format..." << vcl_endl;
  Vitto_region_file vrf;
  regions_list_t ref_rs1, ref_rs2;
  vrf.load_regions(ref_rs1, argv[2], (vil1_image*)NULL);  // last param just needed by signature
  vrf.load_regions(ref_rs2, argv[4], (vil1_image*)NULL);
  if(ref_rs1.size() != ref_rs2.size()) {
    vcl_cout << "Not the same number of reference regions in the two images. Improper data format." << vcl_endl;
    return -1;
  }
  vcl_cout << ref_rs1.size() << " reference regions loaded" << vcl_endl;

  // Load candidate regions
  vcl_cout << "Loading candidate regions in Vitto format..." << vcl_endl;
  regions_list_t cand_rs1;
  vrf.load_regions(cand_rs1, argv[5], (vil1_image*)NULL);
  vcl_cout << cand_rs1.size() << " candidate regions loaded" << vcl_endl;

  // Load images
  vcl_cout << "Loading images..." << vcl_endl;
  vil1_image image1 = vil1_load(argv[1]);
  vil1_image image2 = vil1_load(argv[3]);

  // Processing arguements
  bool  blade = false;                                    // default mode
  bool  small = false;
  float radius = 120.0;                                   // default int-prop radius
  char* biasfile = NULL;                                  // bias for int-prop supports
  process_arguements(argc, argv, blade, small, radius, biasfile);
  
	vcl_vector<float>* bias = NULL;
	if(biasfile) {
		bias = load_floats(biasfile);
		vcl_cout << "Load biasfile" << vcl_endl;
	}

	MosaicPropagator prop( image1, image2, true, small );

	regions_list_t cand_rs2;
	vcl_vector<int> sup_rs;

	prop.propagate( &ref_rs1, &ref_rs2, &cand_rs1, cand_rs2, sup_rs, blade, radius, bias );
	vcl_cout << "DEBUG2: " << cand_rs2.size() << " " << &cand_rs2 << vcl_endl; //DEBUG

  // Save regions output
  vrf.save_regions(cand_rs2, argv[6]);

  // Save supports
  save_ints(sup_rs, argv[7]);
} // main

