#include <vcl_sstream.h>
#include <vcl_iostream.h> 
#include <vcl_fstream.h>

#include <cstdlib>

#include <vil1/vil1_image.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_load.h>
#include <vil1/vil1_save.h>
#include <misc_utility/defines.h>               // def of color_buffer_t
#include <misc_utility/gaussian_smoothing.h>    // vic_gaussian_smoothing
#include <misc_utility/region_growing.h>        // grow_uniform_region
#include <misc_utility/loaders.h>               // save_uniform_region
#include <regions/uniform_region.h>
#include <region_utility/region_draw.h>         // draw_uniform_region     
#include <misc_utility/draw.h>                  // HSVtoRGB
#include <misc_utility/count_pixels.h>          // border_color
#include <misc_utility/byte_interpolation.h>    // dist_rgb


void process_arguements(int argc, char** argv, bool& noback, int& sid) {
  for(int a = 6; a < argc; a++) {
    if(strcmp(argv[a],"-noback") == 0) {
      vcl_cout << "Background colored regions will not be outputted." << vcl_endl;
      noback = true;
    }
    if(strcmp(argv[a],"-first_sid") == 0) {
      sid = atoi(argv[a+1]);
      vcl_cout << "First synchronization id = " << sid << vcl_endl;
    }
    
  }
}


int main(int argc, char** argv)
{
  if (argc < 6) {
    vcl_cout << "Usage: " << argv[0] 
	     << " image_input.tif seeds_x seeds_y regions.unif regions_image.tif [-noback] [-first_sid sid]" << vcl_endl;
    vcl_cout << "Default: first synchronization id (sid) = 1" << vcl_endl;
    return 1;
  }

  // Parse arguements
  bool noback = false;;
  int sid = 1;
  process_arguements(argc, argv, noback, sid);

  // Load image
  vcl_cout << "Loading image..." << vcl_endl;
  vil1_image image = vil1_load(argv[1]);
  color_buffer_t* buf = new color_buffer_t(image);

  // Smooth image in the 3 color bands
  vcl_cout << "Smoothing image..." << vcl_endl;
  color_buffer_t* buf_smooth = new color_buffer_t(image);                 
  vic_gaussian_smoothing(*buf, *buf_smooth);
  color_buffer_t* buf_out = new color_buffer_t(image);              // copy
  vic_gaussian_smoothing(*buf, *buf_out);                           // smooth twice as cannot make a copy of a color_buffer_t, only of a vil1_image

  // estimate background color
  vil1_rgb<unsigned char> bcol(border_color(buf));
  vcl_cout << "Border color: " << (int)bcol.R() << " " << (int)bcol.G() << " " << (int)bcol.B() << vcl_endl;

  // Grow uniform regions
  int w = buf->width();
  int h = buf->height();
  int seeds_x = atoi(argv[2]);
  int seeds_y = atoi(argv[3]);
  vcl_cout << "Growing regions..." << vcl_endl;
  uniform_regions_list_t urs;
  bool* xflag = new bool[w*h];                                      // grown and accepted regions flag, these pixels can't be reused
  for(int y=0; y<h; y++) {
    for(int x=0; x<w; x++) {
      xflag[x + y*w] = false;
    }
  }
  for(int sx=0; sx<seeds_x; sx++) {
    for(int sy=0; sy<seeds_y; sy++) {
      int x = sx * w/seeds_x;
      int y = sy * h/seeds_y;
      vcl_cout << "Seed " << x << ", " << y << vcl_endl;
      int area = 0;
      bool* flag = grow_uniform_region(buf_out, xflag, area, x, y);
      
      if(area > 3000) {   // make threshold param, add more meaningful criteria (compactness, etc.)
	// update xflag
	for(int y=0; y<h; y++) {
	  for(int x=0; x<w; x++) {
	    if(flag[x + y*w])
	      xflag[x + y*w] = true;
	  }
	}

	//if(area < 50000) {
	uniform_region* ur = new uniform_region(flag, buf_smooth);
	bool accept = true;
	vil1_rgb<unsigned char> urcol(ur->get_color());
	if(noback && dist_rgb(urcol, bcol) < 50) {
	  vcl_cout << "  can't accept: background region." << vcl_endl;
	    accept = false;
	}

	if(accept) {
	  ur->set_sid(sid);
	  urs.push_back(ur);
	  vcl_cout << "  region accepted, area = " << area << vcl_endl;
	  sid++;
	}
	//}
      }

      delete flag;
    }
  }
  
  // Save regions to file
  save_uniform_regions(&urs, argv[4]);         
  vcl_cout << "Found " << urs.size() << " uniform regions" << vcl_endl;

  // Save drawn region (testing)
  for(int u=0; u<urs.size(); u++) {
    float r,g,b;
    float hue = (urs[u]->get_sid())*(360./16.);    // 16 different colors
    hue = hue - 360.0*((int)(hue/360.0));          // cycle to have hue in [0,360]
    vcl_cout << "sid = " << urs[u]->get_sid() << " hue = " << hue << vcl_endl;
    HSVtoRGB( &r, &g, &b, hue, 0.6, 200.0 );
    vil1_rgb<unsigned char> col((int)r, (int)g, (int)b);
    draw_uniform_region(urs[u], col, buf_out);
  }
  vil1_save(*buf_out, argv[5]);
  
  
  vcl_cout << "Done" << vcl_endl;
} // main

