/* This is the equivalent of LearningMultiViewMosaic.m */

#include <cstdlib>
#include <cstdio>

#include <vcl/vcl_string.h>
#include <vcl/vcl_fstream.h>
#include <misc_utility/defines.h>
#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_load.h>

#include <region_utility/misc.h>
#include <misc_utility/dilation.h>
#include <main_processing/imageanalyzer.h>
#include <main_processing/viewpairmatcher.h>
#include <main_processing/mosaicpropagator.h>
#include <region_utility/clusterfilter.h>

#include <region_utility/Vitto_region_file.h>


typedef struct {
	int          viewID;
	vcl_string   viewString;
	vcl_string   fileName;
	vcl_string   fileBase;
	vcl_string   fileExt;

	vil1_image   image;
	bw_buffer_t* mapBuf;

	regions_list_t             tesselation;
	vcl_vector<regions_list_t> vRegionSets;
} View;

typedef struct {
	vcl_string       objName;
	vcl_vector<View> vViews;
} Subject;

typedef struct {
	int     viewID;
	region* reg;
} TrackEntry;

bool cmpViewID( const View &v1, const View &v2 )
{
	return v1.viewID < v2.viewID;
}


vcl_string makeMapName( const vcl_string &fName )
{
	vcl_string mapFileName = fName;
	int extPos = fName.rfind('.');
	int slashPos = fName.rfind('/');
	if( slashPos == vcl_string::npos )
		slashPos = -1;
	mapFileName.replace(extPos, mapFileName.length()-extPos, ".png");
	mapFileName.insert(extPos, "-map");
	mapFileName.insert(slashPos+1, "maps/");
	
	return mapFileName;
}


int lastInteger( const vcl_string &str, vcl_string *intStr, vcl_string *prefix )
/* Return the last integer found in the string str, or -1 if none found.
 * If intStr is not NULL, the integer as it is (with trailing zeros)
 * will be returned in it. If prefix is not NULL, the part of the
 * string preceding the integer will be returned in it. */
{
	int start = -1, end = -1;
	for( int c = str.length()-1; c >= 0; c-- ) {
		if( str[c] >= '0' && str[c] <= '9' ) {
			if( end < 0 )
				end = c;
			start = c;
		}
		else if( end >= 0 ) {
			break;
		}
	}
	if( end < 0 )
		return -1;

	vcl_string iStr = str.substr(start, end+1-start);
	if(intStr)
		*intStr = iStr;
	if(prefix)
		*prefix = str.substr(0, start);
	return atoi( iStr.c_str() );
}


bool fileIsValid( const vcl_string &fName )
{
	vcl_ifstream test( fName.c_str() );
	return test.good();
}


bool isInVector( int x, const vcl_vector<int> &vec )
{
	for( unsigned int i=0; i<vec.size(); i++ ) {
		if( x == vec[i] )
			return true;
	}
	return false;
}


void deleteRegions( const vcl_vector<int> &rmv_sids,
                    regions_list_t &regions_v1, regions_list_t &regions_v2 )
/* Remove and delete regions that have sids in rmv_sids
 * from two regions_list_t's that are synchronized. */
{
	unsigned int i=0;
	while( i<regions_v1.size() ) {
		if( isInVector( regions_v1[i]->get_sid(), rmv_sids ) ) {
			delete regions_v1[i];
			regions_v1.erase( regions_v1.begin()+i );
			delete regions_v2[i];
			regions_v2.erase( regions_v2.begin()+i );
		}
		else
			i++;
	}
}


/* Structure of a .tracks file:
 * MVT [version number]
 * <object name>
 * [number of views N] [number of tracks M]
 * <space-separated list of N view IDs from filenames>
 * <Start of region listing = M sets of maximum N+1 lines, containing the following:>
 *   [int trackLength]
 *   [int viewID] [int type] [double[2] pt0] [double[2] pt1] [double[2] pt2] [double[2] maj] [double[2] min]
 *   [int viewID] [int ....
 */
void saveTracks( const vcl_string &tracksFile,
                 const vcl_vector< vcl_vector<TrackEntry> > &vTracks,
                 const Subject &obj )
{
	vcl_ofstream outFile( tracksFile.c_str() );

	outFile << "MVT 1" << vcl_endl
	        << obj.objName << vcl_endl
	        << obj.vViews.size() << " " << vTracks.size() << vcl_endl;

	for( unsigned int v=0; v<obj.vViews.size(); v++ )
		outFile << obj.vViews[v].viewString << " ";
	outFile << vcl_endl;

	for( unsigned int i=0; i<vTracks.size(); i++ ) {
		outFile << vTracks[i].size() << vcl_endl;

		for( unsigned int v=0; v<vTracks[i].size(); v++ ) {
			region *reg = vTracks[i][v].reg;
			int regType = reg->get_specific_type();
			if( reg->get_type() == region::ELL_REGIONTYPE )
				regType = 21; // for compatibility
			outFile << vTracks[i][v].viewID << " "
			        << regType << " "
			        << reg->get_pt0()->x() << " "
			        << reg->get_pt0()->y() << " "
			        << reg->get_pt1()->x() << " "
			        << reg->get_pt1()->y() << " "
			        << reg->get_pt2()->x() << " "
			        << reg->get_pt2()->y() << " "
			        << reg->get_pt1()->x() << " "
			        << reg->get_pt1()->y() << " "
			        << reg->get_pt2()->x() << " "
			        << reg->get_pt2()->y() << vcl_endl;
		}
	}
}


void clusterImages( vcl_vector<Subject> &vObjects, vcl_vector<char*> fileNames  )
/* Create objects according to file names.
 * File names should be of the form objectName##suffix, with ## an integer
 * Example:
 *   object1-01.png
 *   object1-02.png
 *   object2-01.png
 *   ...
 * will result in two objects, 'object1-' (2 views) and 'object2-'.
 *
 * All fields in each view will be set, except image, mapBuf, tesselation and vRegionSets. */
{
	for( unsigned int i=0; i<fileNames.size(); i++ ) {
		vcl_string intStr, prefix;
		vcl_string sFileName(fileNames[i]);
		int viewID = lastInteger( sFileName, &intStr, &prefix );
		if( viewID < 0 ) {
			vcl_cerr << "Error: no view ID found in filename \"" << sFileName << "\"" << vcl_endl
			         << "Skipping file." << vcl_endl;
			continue;
		}
		
		Subject *theObject = NULL;
		for( unsigned int j=0; j<vObjects.size(); j++ ) {
			if( prefix == vObjects[j].objName ) {
				theObject = &(vObjects[j]);
				break;
			}
		}
		if( theObject == NULL ) {
			vObjects.resize( vObjects.size()+1 );
			theObject = &(vObjects.back());
			theObject->objName = prefix;
		}

		theObject->vViews.resize( theObject->vViews.size()+1 );
		View &theView = theObject->vViews.back();
		theView.viewID     = viewID;
		theView.viewString = intStr;
		theView.fileName   = sFileName;

		vcl_string::size_type dotPos = sFileName.rfind(".");
		if( dotPos != vcl_string::npos ) {
			theView.fileBase = sFileName.substr( 0, dotPos );
			theView.fileExt  = sFileName.substr( dotPos+1, sFileName.length()-dotPos-1 );
		}
		else {
			theView.fileBase = sFileName;
			theView.fileExt  = "";
		}
	}
	
	for( unsigned int o=0; o<vObjects.size(); o++ )
		sort( vObjects[o].vViews.begin(), vObjects[o].vViews.end(), cmpViewID );
}


void print_usage( char *progName )
{
	vcl_cout << "Usage:" << vcl_endl
	         << progName << " [options] imageFile1 imageFile2 [...]" << vcl_endl << vcl_endl;

	vcl_cout << "Options:" << vcl_endl
					 << "  -V: turn verbose off" << vcl_endl
					 << "  -f: force recalculation, even if intermediate files are present" << vcl_endl
					 << "  -r N: only match N views on each side of each view" << vcl_endl
					 << "        (default = 2, set to 0 for all views)" << vcl_endl
					 << "  -m M: only keep the M best initial matches for each view pair" << vcl_endl
					 << "        (default = 300, set to 0 for all matches)" << vcl_endl;
}


void process_arguments( int argc, char **argv, vcl_vector<char*> &fileNames,
                        bool &verbose, bool &useReload,
                        unsigned int &maxMatches, unsigned int &vRange )
{
	// Defaults
	verbose = true;
	useReload = true;
	maxMatches = 300;
	vRange = 2;
	
	int i = 0;
	while( ++i < argc ) {
		if( argv[i][0] != '-' )
			fileNames.push_back(argv[i]);
		else {
			char *arg = argv[i];
			int len = strlen(arg);
			for( int j=1; j<len; j++ ) {
				switch( arg[j] ) {
				case 'V':
					verbose = false;
					break;
				case 'f':
					useReload = false;
					break;
				case 'r':
					if( i+1 == argc || sscanf(argv[++i], "%u", &vRange) != 1 ) {
						vcl_cerr << "Error: -r expects an unsigned integer als next argument." << vcl_endl;
						print_usage(argv[0]);
						exit(RET_WRONG_ARG);
					}
					break;
				case 'm':
					if( i+1 == argc || sscanf(argv[++i], "%u", &maxMatches) != 1 ) {
						vcl_cerr << "Error: -m expects an unsigned integer als next argument." << vcl_endl;
						print_usage(argv[0]);
						exit(RET_WRONG_ARG);
					}
					break;
				default:
					vcl_cerr << "Warning: unknown option -" << arg[j] << vcl_endl;
					break;
				}
			}
		}
	}

	if(verbose) {
		vcl_cout << "Verbose is on" << vcl_endl
		         << (useReload ? "Using" : "Not using") << " previous data if available" << vcl_endl
		         << "Using max " << maxMatches << " matches" << vcl_endl
		         << "Matching " << vRange << " views on each side" << vcl_endl
		         << "Found " << fileNames.size() << " image files" << vcl_endl;
	}
}


int main(int argc, char **argv)
{
	bool verbose, useReload; // inverse of old 'force' setting
	unsigned int maxMatches, vRange;
	vcl_vector<char*> fileNames;
	process_arguments( argc, argv, fileNames, verbose, useReload, maxMatches, vRange );

	if( ! fileNames.size() ) {
		vcl_cerr << "Error: you need to provide at least one image file." << vcl_endl;
		print_usage(argv[0]);
		exit(RET_WRONG_ARG);
	}

	vcl_vector<Subject> vObjects;
	clusterImages( vObjects, fileNames );
	if(verbose)
		vcl_cout << vObjects.size() << " objects found" << vcl_endl;

	vcl_vector<vcl_string> featureTypes;
	featureTypes.push_back("tuytelaars_ell");
	featureTypes.push_back("mser");
	featureTypes.push_back("h_affine");


	for( unsigned int o=0; o<vObjects.size(); o++ ) {
		vcl_string &objName = vObjects[o].objName;
		vcl_cout << "Modelling object '" << objName << "'" << vcl_endl
		         << "====================================" << vcl_endl;
		unsigned int numViews = vObjects[o].vViews.size();


		/*---- Process all input images: tesselate and find regions ----*/
		for( unsigned int f=0; f<numViews; f++ ) {
			View &theView = vObjects[o].vViews[f];

			vcl_string &sFileName = theView.fileName;
			vcl_string &sFileBase = theView.fileBase;
			vcl_string &sFileExt  = theView.fileExt;
			vcl_string sMapName   = makeMapName(sFileName);

			/*---- Load image ----*/
			if(verbose)
				vcl_cout << "Loading image " << sFileName << vcl_endl;
			vil1_image  &image   = theView.image;
			bw_buffer_t *&theMap = theView.mapBuf;
			image = vil1_load( sFileName.c_str() );
			theMap = NULL;
			if( fileIsValid(sMapName) ) {
				if(verbose)
					vcl_cout << "Map found in file " << sMapName << vcl_endl;
				vil1_image mapImg = vil1_image_as_byte( vil1_load( sMapName.c_str() ) );

				// segmentation map is dilated to ensure complete coverage of object
				bw_buffer_t tmpBuf(mapImg);
				theMap = new bw_buffer_t(tmpBuf);
				dilation5(tmpBuf, *theMap);
			}
			else {
				if(verbose)
					vcl_cout << "No map found." << vcl_endl;
				sMapName = "";
			}

			/*---- Analyze image or reload data ----*/
			regions_list_t &tesselation            = theView.tesselation;
			vcl_vector<regions_list_t> &regionSets = theView.vRegionSets;

			bool reload = useReload;
			if( ! fileIsValid( sFileBase + "_tess.regmV" ) )
				reload = false;
			for( unsigned int i=0; reload && i<featureTypes.size(); i++ ) {
				if( ! fileIsValid( sFileBase + "_" + featureTypes[i] + ".regmV" ) )
					reload = false;
			}

			if(!reload) {
				ImageAnalyzer imgAna( featureTypes, verbose );
				imgAna.analyzeImage( tesselation, regionSets, image, sFileName.c_str(), theMap );
			}

			Vitto_region_file vrf;
			vcl_string dataFileName = sFileBase + "_tess.regmV";
			if(reload)
				vrf.load_regions( tesselation, dataFileName.c_str(), NULL );
			else
				vrf.save_regions( tesselation, dataFileName.c_str() );
			vcl_cout << "Tesselation: " << tesselation.size() << " regions" << vcl_endl;

			for( unsigned int i=0; i<featureTypes.size(); i++ ) {
				dataFileName = sFileBase + "_" + featureTypes[i] + ".regmV";
				if(reload) {
					regionSets.resize(i+1);
					vrf.load_regions( regionSets[i], dataFileName.c_str(), NULL );
				}
				else
					vrf.save_regions( regionSets[i], dataFileName.c_str() );
				vcl_cout << "Region set " << i+1 << ": " << regionSets[i].size() << " regions\n";
			}
		}


		/*---- Only use v1 = 0 and v2 = 1 for this test! */
		for( unsigned int v1=0; v1<1; v1++ ) {

			View &view1 = vObjects[o].vViews[v1];
			vil1_image &image1 = view1.image;
			ClusterFilter cFilter1( image1, verbose );

			for( unsigned int v2=1; v2<2; v2++ ) {
				View &view2 = vObjects[o].vViews[v2];
				vil1_image &image2 = view2.image;

				/*---- Match ----*/
				vcl_cout << "View pair " << view1.viewID << " <-> " << view2.viewID << vcl_endl
				         << "----------------------------------" << vcl_endl;
				regions_list_t matchRegs1, matchRegs2;
				bool reload = useReload;
				vcl_string matchFile1 = "matches_" + objName + view1.viewString + "-" +
				                        view2.viewString + "_" + view1.viewString + ".regmV";
				vcl_string matchFile2 = "matches_" + objName + view1.viewString + "-" +
				                        view2.viewString + "_" + view2.viewString + ".regmV";

				if( !reload || ! fileIsValid(matchFile1) || ! fileIsValid(matchFile2) ) {
					reload = false;

					ViewPairMatcher vpMatcher( image1, image2, false, verbose );
					/* Assignment of sids is identical to Matlab code: tesselation gets 1~N,
					 * matching (= support) regions get N+1~M. */
					int sid = view1.tesselation.back()->get_sid()+1;
					vpMatcher.matchRegionSets( view1.vRegionSets, view2.vRegionSets,
					                           matchRegs1, matchRegs2, maxMatches, sid );
				}

				Vitto_region_file vrf;
				if(reload) {
					vrf.load_regions( matchRegs1, matchFile1.c_str(), NULL );
					vrf.load_regions( matchRegs2, matchFile2.c_str(), NULL );
				}
				else {
					vrf.save_regions( matchRegs1, matchFile1.c_str() );
					vrf.save_regions( matchRegs2, matchFile2.c_str() );
				}

				/*---- ExploDe ----*/
				unsigned int nIterTotal = 0;
				while(true) {
					unsigned int nIter;
					vcl_cout << "How many iterations? ";
					vcl_cin >> nIter;
					if(nIter == 0)
						break;

					while(nIter-- > 0) {
						//regions_list_t propRegions1;
						regions_list_t propRegions2;
						vcl_vector<int> config_sids;

						MosaicPropagator prop( image1, image2, verbose );
						prop.mosaicPropagate( matchRegs1, matchRegs2, view1.tesselation, propRegions2, config_sids );
						/*for( unsigned int i=0; i<propRegions2.size(); i++ ) {
							int s = propRegions2[i]->get_sid();
							// for tesselation, region with sid s is at position s-1 in vector!
							propRegions1.push_back( copy_region( view1.tesselation[s-1] ) );
						}*/

						/* PostProcessViewPair: filter clusters */
						/*vcl_vector<int> rmv_sids;
						ClusterFilter cFilter2( image2, verbose );
						cFilter1.filter( propRegions1, 0, 3, NULL, NULL, &rmv_sids );
						deleteRegions( rmv_sids, propRegions1, propRegions2 );
						rmv_sids.clear();
						cFilter2.filter( propRegions2, 0, 3, NULL, NULL, &rmv_sids );
						deleteRegions( rmv_sids, propRegions1, propRegions2 );*/

						//delete_regions(propRegions1);
						delete_regions(propRegions2);
						nIterTotal++;
						vcl_cout << "Performed " << nIterTotal << " iterations of total uselessness." << vcl_endl;
						vcl_cout << "Iterations remaining for this loop: " << nIter << vcl_endl;
					}
				}
				delete_regions(matchRegs1);
				delete_regions(matchRegs2);
			}
		}

		/* Clean up: delete all region vectors from this object:
		 *   tesselations, regionSets and propRegions */
		for( unsigned int v1=0; v1<numViews; v1++ ) {
			View &theView = vObjects[o].vViews[v1];
			delete_regions( theView.tesselation );
			for( unsigned int i=0; i<featureTypes.size(); i++ )
				delete_regions( theView.vRegionSets[i] );
		}
	}

	if(verbose)
		vcl_cout << "Gabba gabba gabba!" << vcl_endl;
	return 0;
}
