#include <vcl/vcl_list.h>
#include <vcl/vcl_utility.h>
#include <vcl/vcl_iostream.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_load.h>

#include <cstdlib>

#include <misc_utility/defines.h>
#include <regions/parall_region.h>
#include <regions/ell_region.h>
#include <region_utility/Vitto_region_file.h>   // saving
#include <region_utility/old_region_file.h>
#include <region_utility/VIBES_region_file.h>
#include <region_utility/misc.h>                // saving
#include <vil1/vil1_image_as.h>                 // greylevel conversion


#define SIFT_LENGTH 128

typedef struct _SiftKey {
	float x, y;                         /* Subpixel location of keypoint. */
	float cornerness;                   /* Normally unused */
	float scale, ori;                   /* Scale and orientation (range [-PI,PI]) */
	int   type;                         /* Normally unused */
	float laplacian;                    /*   "         "   */
	int   extremum_type;                /*   "         "   */
	float m11, m12, m21, m22;           /* This is... well, some kind of matrix. */
	unsigned char descrip[SIFT_LENGTH]; /* Vector of descriptor values */
} SiftKey;


int ReadSiftKeys( vcl_vector<SiftKey> & keyPoints, const char *fName )
{
	int num, len, val;           

	/* Swapped num,len for Miko's new descriptor program */
	vcl_ifstream inFile( fName );
	inFile >> len >> num;
	if( len != SIFT_LENGTH ) {
		vcl_cerr << fName << ": Keypoint descriptor length invalid (should be 128).";
		return 0;
	}

	for( int i = 0; i < num; i++ ) {
		SiftKey k;
		inFile >> k.x >> k.y >> k.cornerness >> k.scale >> k.ori;
		inFile >> k.type >> k.laplacian >> k.extremum_type;
		inFile >> k.m11 >> k.m12 >> k.m21 >> k.m22;

		for( int j = 0; j < len; j++ ) {
			inFile >> val;
			if( val < 0 || val > 255 )
				vcl_cerr << "Warning: invalid keypoint file value at " << i << "," << j
				         << " (" << val << "). Must be in [0,255].";
			k.descrip[j] = (unsigned char)val;
		}
		keyPoints.push_back(k);
	}

	return num;
}

void SaveSiftKeys( const vcl_vector<SiftKey> & keyPoints, const char *fName )
{
	vcl_ofstream outFile( fName );
	
	outFile << SIFT_LENGTH << "\n";
	outFile << keyPoints.size() << "\n";
	for( unsigned int i = 0; i < keyPoints.size(); i++ ) {
		const SiftKey & k = keyPoints[i];
		outFile << k.x << " " << k.y << " " << k.cornerness << " " << k.scale << " " << k.ori << " ";
		outFile << k.type << " " << k.laplacian << " " << k.extremum_type << " ";
		outFile << k.m11 << " " << k.m12 << " " << k.m21 << " " << k.m22 << " ";
		for( int j = 0; j < SIFT_LENGTH; j++ )
			outFile << (int)k.descrip[j] << " ";
		outFile << "\n";
	}
}


vcl_string makeMapName( const char *fName )
{
	vcl_string mapFileName = fName;
	int extPos = mapFileName.rfind('.');
	int slashPos = mapFileName.rfind('/');
	if( slashPos == vcl_string::npos )
		slashPos = -1;
	mapFileName.insert(extPos, "-map");
	mapFileName.insert(slashPos+1, "maps/");
	
	return mapFileName;
}

int fileIsValid( vcl_string & fName )
{
	vcl_ifstream test( fName.c_str() );
	return test.good();
}


int main(int argc, char** argv)
{
	if (argc < 3)
	{
		vcl_cout << "Usage: " << argv[0] << " image features1 [features2 [...]]"  << vcl_endl;
		vcl_cout << "Supported files: .regmV, .regvib, .regpnf, .rege2nfs, .sift" << vcl_endl;
		vcl_cout << "Warning: input files will be overwritten!" << vcl_endl;
		abort();
	}
	
	// load background map, if any
	vcl_string mapFileName = makeMapName( argv[1] );
	if( ! fileIsValid( mapFileName ) ) {
		// try other format
		int extPos = mapFileName.rfind('.');
		if( mapFileName.substr( extPos, 4 ) == ".png" )
			mapFileName.replace( extPos, 4, ".jpg" );
		else
			mapFileName.replace( extPos, 4, ".png" );
		
		if( ! fileIsValid( mapFileName ) ) {
			vcl_cout << "No map found, doing nothing." << vcl_endl;
			return 0;
		}
	}
	
	vcl_cout << "Map found in file " << mapFileName << vcl_endl;
	vil1_image map = vil1_image_as_byte( vil1_load( mapFileName.c_str() ) );
	bw_buffer_t mapBuff(map);
	// a dilation step could be useful here

	vil1_image* image = NULL; // dummy image
	Vitto_region_file vrf;
	old_region_file orf;
	VIBES_region_file brf;
	
	// process all feature files
	for( int f = 2; f < argc; f++ ) {
		vcl_string fName = argv[f];
		int extPos = fName.rfind('.');
		vcl_string ext = fName.substr( extPos, fName.length()-extPos+1 );
		regions_list_t rs_in, rs_out;

		if( ext.substr(0,4) == ".reg" ) {
			if( ext == ".regmV" )
				vrf.load_regions( rs_in, argv[f], image );
			else if( ext == ".regvib" )
				brf.load_regions( rs_in, argv[f], image );
			else
				orf.load_regions(rs_in, argv[f], image);
			
			for( int i = 0; i < rs_in.size(); i++ ) {
				region *r = rs_in[i];
				point_t *p0 = r->get_pt0();
				if( mapBuff((int)p0->x(), (int)p0->y()) > 127 ) {
					rs_out.push_back(r);
				}
			}

			vcl_cout << "Saving " << rs_out.size() << " of " << rs_in.size() << " regions to " << argv[f] << vcl_endl;
			if( ext == ".regmV" )
				vrf.save_regions( rs_out, argv[f] );
			else if( ext == ".regvib" )
				brf.save_regions( rs_out, argv[f] );
			else
				orf.save_regions( rs_out, argv[f] );
		}
		else if( ext == ".sift" ) {
			vcl_vector<SiftKey> keys_in, keys_out;
			ReadSiftKeys( keys_in, argv[f] );
			
			for( unsigned int i = 0; i < keys_in.size(); i++ ) {
				int x = (int)(keys_in[i].x+.5),  y = (int)(keys_in[i].y+.5);
				if( mapBuff(x,y) > 127 ) {
					keys_out.push_back( keys_in[i] );
				}
			}
			
			vcl_cout << "Saving " << keys_out.size() << " of " << keys_in.size() << " keypoints to " << argv[f] << vcl_endl;
			SaveSiftKeys( keys_out, argv[f] );
		}
		else {
			vcl_cout << "Unknown file format: " << ext << "!!! Skipping." << vcl_endl;
		}
	}
	
	return 1;
}
