
#include <cstdlib>
#include <cstring>

#include <vcl/vcl_list.h>
#include <vcl/vcl_utility.h>
#include <vcl/vcl_iostream.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_load.h>

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <regions/ell_region.h>
#include <region_utility/old_region_file.h>
#include <region_matching/reg_match_params.h>
#include <region_matching/classic_region_matcher.h>
#include <region_matching/multi_region_matcher.h>
#include <region_matching/pairwise_indexed_matches.h>
#include <region_matching/pairwise_indexed_match.h>

#include <region_utility/Vitto_region_file.h>   // saving
#include <region_utility/misc.h>                // copy_region
#include <misc_utility/loaders.h>               // saving similarity scores

#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/drgbnorm_computor.h>
#include <region_matching/ncc_drgbnorm_computor.h>
#include <region_matching/ncc_drgbsnorm_computor.h>
#include <region_matching/ncc_drgb_computor.h>

#include <region_propagation/raywalk.h>

#include <mvl/FMatrix.h>
#include <mvl/HomgPoint2D.h>
#include <mvl/FMatrixComputeRANSAC.h>

const int NCC = 1;
const int NCCDRGBSNORM = 2;

void process_arguements(int argc, char** argv, bool& multi, bool& ransac, bool& save_sims, char*& sims_file, bool& rethresh,
			int& sim, float& thresh1, float& thresh2) {

  for(int a = 8; a < argc; a++) {
    if(strcmp(argv[a],"-multi") == 0) {
      vcl_cout << "Multimatches mode activated." << vcl_endl;
      multi = true;
    }

    if(strcmp(argv[a],"-ransac") == 0) {
      vcl_cout << "Ransac filter after matching activated." << vcl_endl;
      ransac = true;
    }

    if(strcmp(argv[a],"-sim") == 0) {
      a++;
      sims_file = argv[a];
      save_sims = true;
      vcl_cout << "Similarity scores of matched regions will be saved to " << sims_file << vcl_endl;
    }

    if(strcmp(argv[a],"-norethresh") == 0) {
      rethresh = false;
      vcl_cout << "Refinement & re-thresholding deactivated." << vcl_endl;
    }

    if(strcmp(argv[a],"-ncc") == 0) {
      sim = NCC;
      vcl_cout << "Similarity measure set to ncc" << vcl_endl;
      thresh1 = atof(argv[a+1]);
      thresh2 = atof(argv[a+2]);
    }
  }
}


// applies ransac EG filter to rs1,rs2
// output are pruned rs1,rs2 (in/out params)
void ransac_filter(regions_list_t &rs1, regions_list_t &rs2) {
  // Prepare points
  vcl_vector<HomgPoint2D> points1;
  vcl_vector<HomgPoint2D> points2;
  for(int rx=0; rx<rs1.size(); rx++) {
    point_t* c1 = rs1[rx]->center();
    points1.push_back(HomgPoint2D(c1->x(), c1->y()));    // creates new points
    point_t* c2 = rs2[rx]->center();
    points2.push_back(HomgPoint2D(c2->x(), c2->y()));
  }
 
  // Perform the fit using Phil Torr's RANSAC
  FMatrixComputeRANSAC computor(true,9); // 9 -> error tolerance (in pixels ?)
  FMatrix f = computor.compute(points1, points2);

  vcl_cout << "FMatrixComputeRANSAC with rank truncation:\nF = " << f << vcl_endl;

  double d = 0;
  for(unsigned int i = 0; i < points1.size(); ++i)
     d += f.image1_epipolar_distance_squared(&points1[i], &points2[i]);
  vcl_cout << "Error = " << d/points1.size() << vcl_endl;

  // Get inliers to the epipolar geometry
  vcl_vector<bool> inliers(computor.get_inliers());
  regions_list_t prs1, prs2;
  for(int i=0; i<inliers.size(); i++) {
    if(inliers[i]) {
      prs1.push_back(rs1[i]);
      prs2.push_back(rs2[i]);
    }
  }
  rs1 = prs1;
  rs2 = prs2;
}


bool replace_region(region* r, regions_list_t& rs) {
  bool found = false;
  for(int rx=0; rx<rs.size(); rx++) {
    region* s = rs[rx];
    if(r->get_sid() == s->get_sid()) {
      found = true;
      rs[rx] = r;
      break;
    }
  }
  if(!found) {
    vcl_cout << "replace_region: Error: region " << r->get_sid() << " not in list." << vcl_endl;
    exit(1);
  }
}


int main(int argc, char** argv)
{
  if (argc < 8) {
    vcl_cout << "Usage: " << argv[0] << " image1 regs_in1.regmV image2 regs_in2.regmV regs_out1.regmV regs_out2.regmV matches_1_2.dat";
    vcl_cout << " [-norethresh] [-multi] [-ransac] [-ncc thresh1 thresh2] [-sim sims_1_2.dat]" << vcl_endl;
    vcl_cout << "Default: one-to-one matches, no ransac filtering, re-thresholding" << vcl_endl;
    vcl_cout << "sim = ncc + drgbsnorm > 0.6 > 1.0" << vcl_endl;
    abort();
  }

  // load image1
  vcl_cout << "Loading image1 " << argv[1] << " ..." << vcl_endl;
  vil1_image image1 = vil1_load(argv[1]);
  color_buffer_t* buf1 = new color_buffer_t(image1);

  // load regions1
  vcl_cout << "Loading regions1 " << argv[2] << " ..." << vcl_endl;
  Vitto_region_file vrf;
  regions_list_t regions1;
  vrf.load_regions(regions1, argv[2], &image1);
  vcl_cout << regions1.size() << " regions loaded" << vcl_endl;

  // load image2
  vcl_cout << "Loading image2 " << argv[3] << " ..." << vcl_endl;
  vil1_image image2 = vil1_load(argv[3]); 
  color_buffer_t* buf2 = new color_buffer_t(image2);

  // load regions2
  vcl_cout << "Loading regions2 " << argv[4] << " ..." << vcl_endl;
  regions_list_t regions2;
  vrf.load_regions(regions2, argv[4], &image2);
  vcl_cout << regions2.size() << " regions loaded" << vcl_endl;

  // Processing arguements
  bool multi = false;
  bool ransac = false;
  bool save_sims = false;
  bool rethresh = true;
  char* sims_file = NULL;
  int sim = NCCDRGBSNORM;
  float thresh1 = 0.6;
  float thresh2 = 1.0;
  process_arguements(argc, argv, multi, ransac, save_sims, sims_file, rethresh, sim, thresh1, thresh2);

  // set matching parameters
  reg_match_params params;
  if(rethresh)
    params.min_sim = thresh1;      // low first threshold, good for ncc_drbnormcomp
  else
    params.min_sim = thresh2;      // directly high

  // initialize similarity computor
  similarity_computor* sc = NULL;
  if(sim == NCCDRGBSNORM)  sc = new ncc_drgbsnorm_computor(buf1, buf2);                 // single scalefactor version
  if(sim == NCC)           sc = new crosscorrelation_computor(buf1, buf2);
  sc->set_resolution(params.similarity_resolution);

  // initialize region matcher
  classic_region_matcher* rm = NULL;
  if(!multi)
    rm = new classic_region_matcher(&image1, &image2, regions1, regions2, sc, &params);  // sets pairwise_indexed_match static members
  else
    rm = new multi_region_matcher(&image1, &image2, regions1, regions2, sc, &params);    // sets pairwise_indexed_match static members
  rm->compute_statistics();                                                              // gets variances of every component of feat vect for all reg types

  // match regions
  vcl_cout << "Matching regions ... " << vcl_endl;

  rm->match_regions();
  pairwise_indexed_matches mtches;
  rm->get_matches(mtches);

  // print out matches
  vcl_cout << vcl_endl << vcl_endl << mtches.size() << " matches found: " << vcl_endl;

  // convert to region list
  rotation_finder rotf(buf1, buf2);                                      // check color sim meas
  rotf.set_resolution(params.rotation_finder_resolution);
  rotf.set_stepsize(params.search_for_rotation_stepsize);
  crosscorrelation_computor ncc(buf1, buf2);
  ncc.set_resolution(params.similarity_resolution);
  drgbnorm_computor drgbnorm(drgbnorm_computor(buf1, buf2));
  drgbnorm.set_resolution(params.similarity_resolution);
  drgbsnorm_computor drgbsnorm(drgbsnorm_computor(buf1, buf2));
  drgbsnorm.set_resolution(params.similarity_resolution);
  drgb_computor drgb(drgb_computor(buf1, buf2));
  drgb.set_resolution(params.similarity_resolution);

  // prepare refiner for post-matching filtering
  raywalk refiner(&ncc, NULL, NULL, false);                     // false -> verbose
  float b0[] = {-14, 2, 14};
  float b1[] = {-14, 2, 14};
  float b2[] = {0.6, 0.1, 1.8};
  float b3[] = {0.6, 0.1, 1.8};
  float b4[] = {-3.1416/4.0, 3.1416/16, 3.1416/4.0};
  float b5[] = {-1, 0.2, 1};
  float* b[] = {b0, b1, b2, b3, b4, b5};                        // bounds matrix  (float**)
  refiner.set_bounds(b);

  // Checkout and post-filter all matches
  regions_list_t rs1, rs2;
  vcl_vector<float> sims;
  int i = 0;

  if(rethresh) vcl_cout << "Refinement and rethresholding ... " << vcl_endl;
  for (mtches.reset(); mtches.next_match(); ) {
    // Extract match <r1,r2> from mtches datastructure
    match* m = mtches.get_match();
    pairwise_indexed_match* pwi_m = m->cast_to_pairwise_indexed_match();
    region* r1 = pwi_m->get_orig_region(); region* r2 = pwi_m->get_ref_region();
    region* rr1 = NULL;

    // Output values of various sim meas
    if(r1->get_type() == region::ELL_REGIONTYPE) {
      ell_region* er1 = r1->cast_to_ellregion();
      ell_region* er2 = r2->cast_to_ellregion();
      rotf.set_region1(er1);
      rotf.set_region2(er2);
      rotf.get_crosscorrelation();
      region* rr1t = rotf.get_rotated_region1();  // could be corrupted after rerunning rotf ! Do not use this ptr further
      rr1 = copy_region(rr1t);
    } else {
      rr1 = r1;
    }

    i++;
    vcl_cout << "Match: " << i << " " << rr1->get_sid() << " <-> " << r2->get_sid() << " ncc = " << ncc.get_similarity(rr1, r2);
    vcl_cout << " , drgb = " << drgb.get_similarity(rr1, r2);
    vcl_cout << " , drgbnorm = " << drgbnorm.get_similarity(rr1, r2);
    vcl_cout << " , drgbsnorm = " << drgbsnorm.get_similarity(rr1, r2);
    float s = sc->get_similarity(rr1, r2);
    vcl_cout << " , sim = " << s;

    if(rethresh && s <= thresh2) {
      // refine if needed
      r2 = refiner.refine(rr1, r2);
      s = sc->get_similarity(rr1, r2);
      vcl_cout << " , post-ref sim = " << s;
    }

    // rethreshold
    if(s > thresh2) {
      rs1.push_back(rr1);             // save match into output lists
      rs2.push_back(r2);
      replace_region(rr1, regions1);  // alter input list with rotfinetuned region (no effect if parall)
      sims.push_back(s);
    }

    vcl_cout << vcl_endl;
  } // loop on mtches
  if(rethresh) vcl_cout << vcl_endl << vcl_endl << rs1.size() << " matches after rethresholding" << vcl_endl;

  // RANSAC
  if(ransac & rs1.size() > 11) {                  // invoking it with less than 12 matches may lead to suspend
    ransac_filter(rs1, rs2);                      // prunes rs1,rs2 of outliers
    vcl_cout << vcl_endl << rs1.size() << " matches after ransac filter" << vcl_endl;
  }
 
  // Save vitto format refined-matches
  vcl_cout << "Saving vitto formatted output regions. Including rotational aligniment for first view" << vcl_endl;
  if(rethresh) vcl_cout << ", and refinements for second view" << vcl_endl;
  vrf.save_regions(rs1, argv[5]);
  vrf.save_regions(rs2, argv[6]);

  // Save matches datafile
  vcl_cout << "Saving matches datafile" << vcl_endl;
  save_matches_file(rs1, rs2, argv[7]);

  // Save all regs of view1 with matched ones rotationally fixed
  vcl_cout << "Saving vitto formatted input regions for view1, with rotational aligniment" << vcl_endl;
  vrf.save_regions(regions1, argv[2]);
  vrf.save_regions(regions2, argv[4]);

  // Save sims if needed
  if(save_sims) {
    vcl_cout << "Saving similarity scores" << vcl_endl;
    save_floats(sims, sims_file);
  }

  delete sc;

  vcl_cout << "Done." << vcl_endl;
}
