/************************************************************************
Invariant keypoint matching.
Based on demo software by D. Lowe.
Author: David Lowe, V. Ferrari, A. Thomas

Modified by Ferrari and A. Thomas
                     
match.c:
This file contains a program to read keypoints, and output matches as
  coordinates and/or indices.
*************************************************************************/

#include <cstdlib>
#include <cstdarg>
#include <cstdio>
#include <cstring>
#include "match_sift.h"


/* Threshold for matching features. OK range: around 40000~70000 */
#define MATCH_THRESH  70000

/*----------------------------- Routines ----------------------------------*/

/* Top level routine.  Read PGM images and keypoints from files given
   in command line arguments, then call FindMatches.
*/
int main (int argc, char **argv)
{
	int arg = 0, count = 0, all = 0;
	Keypoint k1 = NULL, k2 = NULL;
	char* idfile = NULL;
	char* valfile = NULL;
	char* matchesfile = NULL;

	/* Parse command line arguments and read given files.  The command
	   line must specify two files of keypoints and optionally, 2 output files
	   using command line arguments as follows:
	   match -k1 k1.key -k2 k2.key -ids idfile -matches matchfile -a
	*/
	while (++arg < argc) {
		if (! strcmp(argv[arg], "-k1"))
			k1 = ReadKeyFileVIBES(argv[++arg]);
		else if (! strcmp(argv[arg], "-k2"))
			k2 = ReadKeyFileVIBES(argv[++arg]);
		else if (! strcmp(argv[arg], "-ids"))
			idfile = argv[++arg];
		else if (! strcmp(argv[arg], "-val"))
			valfile = argv[++arg];
		else if (! strcmp(argv[arg], "-matches"))
			matchesfile = argv[++arg];
		else if (! strcmp(argv[arg], "-a"))
			all = 1;
		else
			FatalError("Invalid command line argument: %s", argv[arg]);
	}
	if (k1 == NULL || k2 == NULL)
		FatalError("Invalid arguments\nUsage: %s -k1 k1.key -k2 k2.key [-a] [-ids idfile] [-matches matchfile]", argv[0]);

	if( k1->index == -1 || k2->index == -1 ) { /* if any of the files is empty */
		count = 0;
		if(matchesfile != NULL) {
			FILE *file = fopen(matchesfile, "w");
			fclose(file);
		}
		if(idfile != NULL) {
			FILE *file = fopen(idfile, "w");
			fclose(file);
		}
	}
	else if( all )
		count = FindSoftMatches(k1, k2, matchesfile, idfile);
	else
		count = FindSoft3Matches(k1, k2, matchesfile, idfile);
    
	// save number of matches in output file
	if(valfile) {
		FILE* file = fopen(valfile, "w");
		fprintf(file, "%d", count);
		fclose(file);
	}

	exit(0);
}


/* This function prints an error message and exits.  It takes a variable
   number of arguments that function just like those in printf.
*/
void FatalError(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "Error: ");
	vfprintf(stderr, fmt, args);
	fprintf(stderr,"\n");
	va_end(args);
	exit(1);
}

/* This reads a keypoint file from a given filename and returns the list
   of keypoints (VIBES format)
*/
Keypoint ReadKeyFileVIBES(char *filename)
{
	FILE *file;
	fprintf(stdout,"OPENING file %s\n", filename );
	file = fopen (filename, "rb");
	if (! file)
		FatalError("Could not open file: %s", filename);

	return ReadKeysVIBES(file);
}

/* Read keypoints from the given file pointer and return the list of
   keypoints.  The file format starts with 2 integers giving the total
   number of keypoints and the size of descriptor vector for each
   keypoint (currently assumed to be 128). Then each keypoint is
   specified by 4 floating point numbers giving subpixel row and
   column location, scale, and orientation (in radians from -PI to
   PI).  Then the descriptor vector for each keypoint is given as a
   list of integers in range [0,255].
   VIBES version */
Keypoint ReadKeysVIBES(FILE *fp)
{
	int i, j, num, len, val;
	Keypoint k, keys = NULL;
	float cornerness, m11, m12, m21, m22, laplacian;
	int type, extremum_type;              

	/* Swapped num,len for Miko's new descriptor program */
	if (fscanf(fp, "%d %d", &len, &num) != 2)
		FatalError("Invalid keypoint file beginning.");

	if (len != 128)
		FatalError("Keypoint descriptor length invalid (should be 128).");

	if( num == 0 ) {
		fprintf(stdout, "File contains no keys, nothing to match.\n", num);
		keys = (Keypoint) malloc(sizeof(struct KeypointSt));
		keys->index = -1;
		keys->next = NULL;
		return keys;
	}

	for (i = 0; i < num; i++) {
		/* Allocate memory for the keypoint.*/
		k = (Keypoint) malloc(sizeof(struct KeypointSt));
		k->index = i+1; /* +1 for Matlab */
		k->next = keys;
		keys = k;
		k->descrip = (unsigned char*) malloc(len);
           
		if (fscanf(fp, "%f %f %f %f %f %d %f %d %f %f %f %f",
		           &(k->col), &(k->row), &cornerness, &(k->scale), &(k->ori),
		           &type, &laplacian, &extremum_type, &m11, &m12, &m21, &m22) != 12)
			FatalError("Invalid keypoint file format. Expected 12 values.");

		for (j = 0; j < len; j++) {
			if ( fscanf(fp, "%d", &val) != 1 || val < 0 || val > 255 ) {
				FatalError("Invalid keypoint file value at %d,%d (%d). Must be in [0,255].", i, j, val );
			}
			k->descrip[j] = (unsigned char) val; /* fixed char -> unsigned char */
		}
	}

	fprintf(stdout, "Load %d keys\n", num);
	return keys;
}


/* Given a pair of images and their keypoints, build soft-matches
   Then write the result to a file.
*/
int FindSoftMatches(Keypoint keys1, Keypoint keys2, char *matchesfile, char *idfile)
{
	fprintf(stdout,"Looking for soft-matches.\n");
	Keypoint k1, k2;
	int count = 0;

	/* Match the keys in list keys1 to their best matches in keys2.
	 */
	FILE* mfile = NULL, *ifile = NULL;
	if(matchesfile != NULL)
		mfile = fopen(matchesfile, "w");
	if(idfile != NULL)
		ifile = fopen(idfile, "w");

	for (k1= keys1; k1 != NULL; k1 = k1->next) {
		for (k2= keys2; k2 != NULL; k2 = k2->next) {
			unsigned int d = DistSquared(k1, k2);
			if (d < MATCH_THRESH) {
				count++;
				if (matchesfile != NULL)
					fprintf(mfile, "%f %f %f %f\n", k1->col, k1->row, k2->col, k2->row);
				if (idfile != NULL)
					fprintf(ifile, "%d %d\n", k1->index, k2->index);
			}
		}
	}

	/* Close matches file */
	if(matchesfile != NULL)
		fclose(mfile);
	if(idfile != NULL)
		fclose(ifile);

	/* Write number of matches to standard output. */
	fprintf(stdout,"Found %d matches.\n", count);
	return count;
}

/* Soft-match keypoints: each point from keys1 is matched to at-most-3 points from keys2;
   each match must have dist < MATCH_THRESH
*/
int FindSoft3Matches(Keypoint keys1, Keypoint keys2, char *matchesfile, char *idfile)
{
    Keypoint k, k2;
    int count = 0;

    /* Match the keys in list keys1 to their best matches in keys2.
    */
    FILE *mfile = NULL, *ifile = NULL;
    if(matchesfile != NULL)
      mfile = fopen(matchesfile, "w");
    if(idfile != NULL)
      ifile = fopen(idfile, "w");

    for (k= keys1; k != NULL; k = k->next) {
      //fprintf(stdout, "Matching key %f, %f\n", k->col, k->row);

      /* find soft-3-matches for k */
      unsigned int dsq, distsq1 = 100000000, distsq2 = 100000000, distsq3 = 100000000;
      Keypoint minkey1 = NULL, minkey2 = NULL, minkey3 = NULL;

      for (k2 = keys2; k2 != NULL; k2 = k2->next) {
	dsq = DistSquared(k, k2);

	if (dsq < distsq1 && dsq < MATCH_THRESH) {
          distsq3 = distsq2;
          minkey3 = minkey2;
	  distsq2 = distsq1;
          minkey2 = minkey1;
	  distsq1 = dsq;
	  minkey1 = k2;
	} else if (dsq < distsq2 && dsq < MATCH_THRESH) {
          distsq3 = distsq2;
          minkey3 = minkey2;
	  distsq2 = dsq;
          minkey2 = k2;
	} else if (dsq < distsq3 && dsq < MATCH_THRESH) {
          distsq3 = dsq;
          minkey3 = k2;
        }
      } /* loop over keys2 */

      /* write top-3-matches */
      if (minkey1 != NULL) {
	count++;
        if (matchesfile != NULL)
	  fprintf(mfile, "%f %f %f %f\n", k->col, k->row, minkey1->col, minkey1->row);
        if (idfile != NULL)
	  fprintf(ifile, "%d %d\n", k->index, minkey1->index);
      } /* write minkey1 */
      if (minkey2 != NULL) {
	count++;
        if (matchesfile != NULL)
	  fprintf(mfile, "%f %f %f %f\n", k->col, k->row, minkey2->col, minkey2->row);
        if (idfile != NULL)
	  fprintf(ifile, "%d %d\n", k->index, minkey2->index);
        //fprintf(stdout, "2-soft-match found\n");
      } /* write minkey2 */
      if (minkey3 != NULL) {
	count++;
        if (matchesfile != NULL)
	  fprintf(mfile, "%f %f %f %f\n", k->col, k->row, minkey3->col, minkey3->row);
        if (idfile != NULL)
	  fprintf(ifile, "%d %d\n", k->index, minkey3->index);
	//fprintf(stdout, "3-soft-match found\n\n");
      } /* write minkey3 */

    } /* loop over keys1 */


    /* Close matches file */
    if(matchesfile != NULL)
      fclose(mfile);
    if(idfile != NULL)
      fclose(ifile);

    /* Write number of matches to standard output. */
    fprintf(stdout,"Found %d matches.\n", count);
    return count;
}


/* Return squared distance between two keypoint descriptors.
*/
unsigned int DistSquared(Keypoint k1, Keypoint k2)
{
    int i, dif;
    unsigned int distsq = 0;
    unsigned char *pk1, *pk2; /* fixed this from original which used signed chars. Yikes! */

    pk1 = (unsigned char*) k1->descrip;
    pk2 = (unsigned char*) k2->descrip;

    for (i = 0; i < 128; i++) {
      dif = (int) *pk1++ - (int) *pk2++;
      distsq += dif * dif;
    }
    return distsq;
}
