#include <vcl_sstream.h>
#include <vcl_iostream.h> 
#include <vcl_fstream.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_load.h>
#include <vil1/vil1_save.h>
#include <misc_utility/defines.h>                          // def of color_buffer_t
#include <region_utility/Vitto_region_file.h>
#include <region_matching/ncc_drgbsnorm_computor.h>        // similarity measure
#include <region_matching/intgr_ncc_drgbsnorm.h>
#include <misc_utility/loaders.h>                          // save_floats
#include <region_utility/misc.h>                           // compute_affine
#include <regions/parall_region.h>


vcl_vector<float>* ComputeSimilarities(region* R1, region* R2, similarity_computor& sc) {
  // in this primitive implementation, use a 4x4 grid of 50x50 regions
  // centered at R1 and oriented parallel to the coord axes

  float* A = compute_affine(R1, R2);
  vcl_vector<float>* sims = new vcl_vector<float>;
  float x0 = R1->get_pt0()->x();
  float y0 = R1->get_pt0()->y();
  regions_list_t prs1, prs2;
  float sx = 50;
  float sy = 50;
  for(int rx=-2; rx<2; rx++) {
    for(int ry=-2; ry<2; ry++) {                           // count of area positions rel to top-left corner
      // Construct area
      point_t* p0 = new point_t(x0+rx*sx, y0+ry*sy);       // top-left corner
      point_t* p1 = new point_t(x0+rx*sx+sx, y0+ry*sy);    // top-right corner
      point_t* p2 = new point_t(x0+rx*sx, y0+ry*sy+sy);    // bottom-left corner
      parall_region* pr1 = new parall_region(p0, p1, p2);  
      prs1.push_back(pr1);

      // Project to buf2
      region* pr2 = copy_region(pr1);
      project_region(pr1, pr2, A);
      prs2.push_back(pr2);
      
      // Evaluate potential for this area
      // will change to a smoother function
      float s =  sc.get_similarity(pr1, pr2);          
      sims->push_back(s);
    }
  }

  // Matlab visualiz
  Vitto_region_file vrf;
  vrf.save_regions(prs1, "areas1");
  vrf.save_regions(prs2, "areas2");

  return sims;
}


int main(int argc, char** argv)
{
  if (argc != 6) {
    vcl_cout << "Usage: " << argv[0] 
	     << " reg1.regmV image1.tif reg2.regmV image2.tif potentials.dat" << vcl_endl;
    return 1;
  }

  // Load region
  vcl_cout << "Loading region in Vitto format..." << vcl_endl;
  Vitto_region_file vrf;
  regions_list_t rs1, rs2;
  vrf.load_regions(rs1, argv[1], (vil1_image*)NULL);         // last param just needed by signature
  vrf.load_regions(rs2, argv[3], (vil1_image*)NULL);
  if(rs1.size() != rs2.size()) {
    vcl_cout << "Not the same amount of regions in the two images. Improper data format." << vcl_endl;
    return -1;
  }

  // Load images
  vcl_cout << "Loading images..." << vcl_endl;
  vil1_image image1 = vil1_load(argv[2]);
  vil1_image image2 = vil1_load(argv[4]);
  color_buffer_t* buf1 = new color_buffer_t(image1);
  color_buffer_t* buf2 = new color_buffer_t(image2);
  
  // Preparing similarity computor
  //ncc_drgbsnorm_computor sc(buf1, buf2);
  //sc.set_resolution(30);
  //sc.set_adaptive_resolution(true);
  intgr_ncc_drgbsnorm sc(buf1, buf2);
  vcl_vector<float>** sims = new vcl_vector<float>*[rs1.size()];
  for(int rx=0; rx<rs1.size(); rx++) {
    vcl_vector<float>* s = ComputeSimilarities(rs1[rx], rs2[rx], sc);
    sims[rx] = s;
  }

  // Output numerical results
  save_floats(sims, rs1.size(), argv[5]);

  // Output image with results
  // greylevel coding (white = highest-score, black = lowest-score)
  // of potential of each area -> good to check shape/pos of areas and
  // meaningfulness/helpfulness of potential-assessment function
  // (high response where high potential ?)

  vcl_cout << "Done" << vcl_endl;
} // main

