#include <vcl_sstream.h>
#include <vcl_iostream.h> 
#include <vcl_fstream.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_save.h>
#include <vil1/vil1_load.h>
#include <misc_utility/defines.h>                // def of color_buffer_t
#include <region_utility/Vitto_region_file.h>
#include <region_utility/region_draw.h>          // overlap_share, draw_region
#include <misc_utility/loaders.h>                // save_ints
#include <region_utility/misc.h>                 // bounding_box, bb_intersect


int main(int argc, char** argv) {
  if (argc != 4) {
    vcl_cout << "Usage: " << argv[0] 
	     << " regs1.regmV regs2.regmV overlap.dat" << vcl_endl;
    return 1;
  }
  // For every pair of regions (r1,r2), with r1 in regs1 and r2 in regs2,
  // overlap(r1,r2) contains the ratio of r2's area which overlaps with r1
  // (overlap.dat is the returned result)

  // Load regions 1
  vcl_cout << "Loading regions " << argv[1] << vcl_endl;
  Vitto_region_file vrf;
  regions_list_t rs1;
  vrf.load_regions(rs1, argv[1], (vil1_image*)NULL);             // last param just needed by signature
  vcl_cout << rs1.size() << " regions loaded" << vcl_endl;

  // Load regions 2
  vcl_cout << "Loading regions " << argv[2] << vcl_endl;
  regions_list_t rs2;
  vrf.load_regions(rs2, argv[2], (vil1_image*)NULL);             // last param just needed by signature
  vcl_cout << rs2.size() << " regions loaded" << vcl_endl;

  // Automatic estimates image size needed
  point_t pmin, pmax;
  regions_list_t rs12(rs1);
  //rs12.append(rs2);           // no vector append available
  for(int r_inx = 0; r_inx < rs2.size(); r_inx++) {
    rs12.push_back(rs2[r_inx]);
  }
  vcl_cout << rs12.size() << " total regions" << vcl_endl;
  bounding_box(rs12, pmin, pmax);
  vcl_cout << "Regions distributed over box: (" << pmin.x() << "," << pmin.y() << ") -> (" << pmax.x() << "," << pmax.y() << ")" << vcl_endl;
  if(pmin.x() < 0 || pmax.y() < 0) {
    vcl_cout << "WARNING: some regions have coordinates <0. Unpredictable behavior" << vcl_endl;
  }

  // Create empty buffer of the correct size
  color_buffer_t* buf = new color_buffer_t((int)pmax.x(), (int)pmax.y());

  // Compute overlaps
  vil1_rgb<unsigned char> col(255, 255, 255);                    // regs drawn in white
  vil1_rgb<unsigned char> black(0, 0, 0);     
  vcl_vector<float>* overlaps[rs1.size()];
  vcl_cout << "Computing overlaps" << vcl_endl;
  for(int r1_inx=0; r1_inx<rs1.size(); r1_inx++) {
    draw_region(rs1[r1_inx], col, buf);                         // draw r1 on currently empty buf
    overlaps[r1_inx] = new vcl_vector<float>();
    for(int r2_inx=0; r2_inx<rs2.size(); r2_inx++) {
      // Quick rejection test
      float o = 0.0;
      if(bb_intersect(rs1[r1_inx], rs2[r2_inx])) {              // compute exact overlap only if regions bounding boxes intersect (otherwise it's sure 0)
	//vcl_cout << "r1 = " << rs1[r1_inx]->get_sid() << ", r2 = " << rs2[r2_inx]->get_sid() << " intersect." << vcl_endl;
	o = overlap_share(rs2[r2_inx], col, buf);               // percentage of r2's surface intersecting with r1
      }
      overlaps[r1_inx]->push_back(o);
    }
    draw_region(rs1[r1_inx], black, buf);                       // delete r1 -> buf is now empty
  }

  // Save out result
  vcl_cout << "Saving overlaps to " << argv[3] << vcl_endl;
  save_floats(overlaps, rs1.size(), argv[3]);
 
  // Clean
  delete buf;

  vcl_cout << "Done." << vcl_endl;
} // main


