#include <vcl/vcl_string.h>
#include <vcl/vcl_fstream.h>
#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_load.h>

#include <cstring>

#include <regions/region.h>
#include <misc_utility/defines.h>
#include <misc_utility/executor.h>
#include <region_extraction/multi_region_extractor.h>
#include <region_extraction/tesselator.h>
#include <region_extraction/sift_descriptor.h>

#include "imageanalyzer.h"


ImageAnalyzer::ImageAnalyzer( const vcl_vector<vcl_string> &featureTypes,
                              bool bVerbose )
{
	verbose = bVerbose;
	fTypes = featureTypes;
}

ImageAnalyzer::~ImageAnalyzer()
{

}



int ImageAnalyzer::analyzeImage( regions_list_t &tesselation,
                                 vcl_vector<regions_list_t> &regionSets,
                                 const vil1_image &image, const char *imFile,
                                 const bw_buffer_t *theMap,
                                 float radius, bool secondLayer )
{
	vcl_string sFileName;
	
	if( imFile == NULL )
		sFileName = "";
	else {
		sFileName = imFile;
		vcl_string sFileBase, sFileExt;
		
		vcl_string::size_type dotPos = sFileName.rfind(".");
		if( dotPos != vcl_string::npos ) {
			sFileBase = sFileName.substr( 0, dotPos );
			sFileExt  = sFileName.substr( dotPos+1, sFileName.length()-dotPos-1 );
		}
		else {
			sFileBase = sFileName;
			sFileExt  = "";
		}
		
		/* Krystian's tools only accept PNG, PGM or PPM, so convert if necessary */
		if( sFileExt != "png" && sFileExt != "pgm" && sFileExt != "ppm" ) {
			vcl_string sNewName(sFileBase);
			sNewName += ".png";
			vcl_ifstream test( sNewName.c_str() );
			if( ! test.good() ) {
				Executor exec(verbose);
				vcl_string cmd = "convert " + sFileName + " " + sNewName;
				if( exec.execute( cmd, false ) != 0 )
					return RET_EXEC_FAIL;
			}
			
			sFileName = sNewName;
		}
	}


	/* Extract features & their SIFT descriptors */

	char *fName = NULL;

	if( sFileName != "" ) {
		fName = new char[sFileName.length()+1];
		strcpy( fName, sFileName.c_str() );
	}
	MultiRegionExtractor xtractor( &image, fName, theMap );
	xtractor.setVerbose(verbose);
	SiftDescriptor desc( &image, fName, verbose );

	for( unsigned int i=0; i<fTypes.size(); i++ ) {
		regions_list_t regions;
		int result = xtractor.extractRegions( regions, fTypes[i], true );
		if( result )
			return result;
		if(verbose)
			vcl_cout << "Extracted " << regions.size() << " regions of type '" << fTypes[i] << "'" << vcl_endl;

		result = desc.get_descriptors(regions);
		if( result )
			return result;
		
		regionSets.push_back(regions);
	}
	
	if( fName != NULL )
		delete [] fName;

	/* Generate tesselation */
	float radiusX = radius,  radiusY = radius,
	      distX   = radius*.5, distY = radius*.5;
	Tesselator tess( image, theMap );
	tess.setVerbose(verbose);

	int sid = tess.tesselate( tesselation, radiusX, radiusY, distX, distY, 1 );

	if(secondLayer) {
		/* Add a 2nd layer of tesselated regions. These are smaller, but equally spaced,
		 * as the first layer, but they do slighlty intersect, should produce better
		 * looking segmentations */
		radiusX *= .6; radiusY *= .6;
		sid = tess.tesselate( tesselation, radiusX, radiusY, distX, distY, sid );
	}

	/* Assign unique sids to other regions */
	for( unsigned int j=0; j<regionSets.size(); j++ ) {
		regions_list_t &regions = regionSets[j];
		for( unsigned int i=0; i<regions.size(); i++ )
			regions[i]->set_sid(sid++);
	}

	return 0;
}
