
#include <region_propagation/int_propagator.h>
#include <region_propagation/blade_propagator.h>
#include <region_matching/ncc_drgbsnorm_computor.h>

#include "rscfilter.h"
#include "mosaicpropagator.h"


#define MY_MAX(x,y) (x) > (y) ? (x) : (y)


MosaicPropagator::MosaicPropagator( const vil1_image &image1, const vil1_image &image2,
                                    bool bVerbose, bool small )
{
	verbose = bVerbose;
	buf1 = new color_buffer_t(image1);
	buf2 = new color_buffer_t(image2);

	ncc = new crosscorrelation_computor(buf1, buf2); // sim meas for refinement
	ncc->set_resolution(30);
  sc  = new ncc_drgb_computor(buf1, buf2);         // sim meas for support selection and acceptance decision
                                                   // (notice the col trafo transfered has only one scalefactor)
                                                   // (dep on propagator::prepare_color_trafo code)
  sc->set_adaptive_resolution(true);               // bigger regions get more pixels -> attempt to better assest them
	if(verbose)
		vcl_cout << "Adaptive resolution in acceptance similarity measure" << vcl_endl;
  sc->set_resolution(30);

	refiner = new raywalk(ncc, NULL, NULL, false);   // false -> verbose
	if(small) {
		float sb0[] = {-8, 1, 8};
		float sb1[] = {-8, 1, 8};
		float sb2[] = {0.8, 0.1, 1.3};
		float sb3[] = {0.8, 0.1, 1.3};
		float sb4[] = {-3.1416/8.0, 3.1416/16, 3.1416/8.0};
		float sb5[] = {-0.5, 0.1, 0.5};
		float* small_bounds[] = {sb0, sb1, sb2, sb3, sb4, sb5};  // bounds matrix  (float**)
		refiner->set_bounds(small_bounds);
	}
	else {
		float nb0[] = {-14, 2, 14};
		float nb1[] = {-14, 2, 14};
		float nb2[] = {0.6, 0.1, 1.8};
		float nb3[] = {0.6, 0.1, 1.8};
		float nb4[] = {-3.1416/4.0, 3.1416/16, 3.1416/4.0};
		float nb5[] = {-1, 0.2, 1};
		float* normal_bounds[] = {nb0, nb1, nb2, nb3, nb4, nb5}; // bounds matrix  (float**)
		refiner->set_bounds(normal_bounds);
	}
}

MosaicPropagator::~MosaicPropagator()
{
	delete refiner;
	delete sc;
	delete ncc;
	delete buf1;
	delete buf2;
}


void MosaicPropagator::propagate( regions_list_t *ref_rs1,  regions_list_t *ref_rs2,
                                  regions_list_t *cand_rs1, regions_list_t &cand_rs2,
                                  vcl_vector<int> &sup_rs,
                                  bool blade, float radius,
                                  vcl_vector<float> *bias )
{
	propagator* prop = NULL;
	if(blade) {
		prop = new blade_propagator(buf1, buf2, sc, ref_rs1, ref_rs2, refiner, false);
		((blade_propagator*)prop)->set_sections(6);

		if(verbose) {
			vcl_cout << "Blade propagation mode." << vcl_endl;
			vcl_cout << "sections = 6" << vcl_endl;
		}
	}
	else {
		prop = new int_propagator(buf1, buf2, sc, ref_rs1, ref_rs2, refiner, false); // false -> verbose
		((int_propagator*)prop)->set_max_dist(radius);
		//((int_propagator*)prop)->set_max_refs(3);     // best-3-ref mode
		if(bias)
			prop->set_bias(bias);

		if(verbose) {
			vcl_cout << "Integrated propagation mode." << vcl_endl;
			vcl_cout << "radius = " << radius << vcl_endl;
		}
	}

	prop->set_sim_t(1.0);    
	prop->set_min_var(200);
	if(verbose) {
		vcl_cout << "similarity threshold = 1.0" << vcl_endl;
		vcl_cout << "min var = 200" << vcl_endl;

		vcl_cout << vcl_endl << "Propagating regions (resolution = " << sc->get_resolution() << ")..." << vcl_endl;
	}

	cand_rs2 = *(prop->propagate(cand_rs1)); // cand_rs2 is not necessarily corresp to cand_rs1 (cand_rs2 contain only succ prop regs)
	sup_rs  = prop->get_supports();          // get sids of successful reference regions

	if(verbose) {
		vcl_cout << "Done" << vcl_endl;
		vcl_cout << "Propagated " << prop->get_prop_count() << " regions." << vcl_endl;
		vcl_cout << "Average sim = " << prop->average_sim() << vcl_endl;
	}

	delete prop;
}


void MosaicPropagator::mosaicPropagate( regions_list_t &ref_rs1,
                                        regions_list_t &ref_rs2,
                                        regions_list_t &cand_rs1,
                                        regions_list_t &prop_rs2,
                                        vcl_vector<int> &config_sids,
                                        float filter_thresh,
                                        unsigned int stop_thresh,
                                        vcl_vector< vcl_vector<int> > *config_hist,
                                        vcl_vector< vcl_vector<int> > *propag_hist,
                                        vcl_vector< vcl_vector<int> > *refern_hist,
                                        vcl_vector< vcl_vector<int> > *remove_hist )
{
	float radius = buf1->width();
	if( buf1->height() > radius )	radius = buf1->height();
	radius /= 6;      // adaptive max cand-to-sup dist to image resolution

	int max_cand_sid = getMaxSid(cand_rs1);
	int max_ref_sid = getMaxSid(ref_rs1);
	if( max_cand_sid > max_ref_sid ) {
		vcl_cerr << "ERROR in mosaicPropagate: max_cand_sid > max_ref_sid!" << vcl_endl;
		return;
	}

	ncc_drgbsnorm_computor simCo(buf1, buf2);
	RSCFilter rsc(verbose);

	// Sparse array for candidate regions cand_rs1, for easy access
	regions_list_t cand_sparse_v1( max_cand_sid+1, (region*)NULL );
	for( unsigned int i=0; i<cand_rs1.size(); i++ )
		cand_sparse_v1[ cand_rs1[i]->get_sid() ] = cand_rs1[i];

	// List of accepted regions as a sparse array, i.e. region with sid s is at vector[s]
	regions_list_t cfg_sparse_v1( max_ref_sid+1, (region*)NULL );
	// New regions will end up in cfg_sparse_v2[i] with i <= max_cand_sid
	// The rest will be copies of existing regions.
	regions_list_t cfg_sparse_v2( max_ref_sid+1, (region*)NULL );
	// Similarities between regions. Contrary to the Matlab version we will
	// only calculate new similarities, this will give a nice speed-up!
	vcl_vector<float> cfg_similarities( max_ref_sid+1, -1.0 );

	// we start out with all support regions as accepted regions
	for( unsigned int i=0; i<ref_rs1.size(); i++ ) {
		int s = ref_rs1[i]->get_sid();
		cfg_sparse_v1[s] = ref_rs1[i];
		cfg_sparse_v2[s] = ref_rs2[i];
		cfg_similarities[s] = simCo.get_similarity( ref_rs1[i], ref_rs2[i] );
	}
	
	vcl_vector< vcl_vector<int> > *vConfigSids;
	if(config_hist)
		vConfigSids = config_hist;
	else
		vConfigSids = new vcl_vector< vcl_vector<int> >;


	unsigned int iter = 0;
	bool needMore = 1;
	while( needMore && iter < 10 ) { // max 10 iters  (avoid too much processing)
		/*---- Propagation ----*/
		/* Make list of regions not yet propagated = cand_rs1 \ cfg_rs1 */
		regions_list_t cur_dest_regs_v1;
		for( unsigned int i=0; i<cand_rs1.size(); i++ ) {
			if( cfg_sparse_v1[ cand_rs1[i]->get_sid() ] )
				continue;
			cur_dest_regs_v1.push_back(cand_rs1[i]);
		}

		regions_list_t support1, support2;
		// de-sparsify for propagator
		for( int s=0; s<=max_ref_sid; s++ ) {
			if( cfg_sparse_v1[s] ) {
				support1.push_back(cfg_sparse_v1[s]);
				support2.push_back(cfg_sparse_v2[s]);
			}
		}
		if(verbose) {
			vcl_cout << "Total: " << support1.size() << " regions" << vcl_endl;
			vcl_cout << "Candidates: " << cur_dest_regs_v1.size() << " regions" << vcl_endl;
			vcl_cout << "Iter " << iter+1 << vcl_endl << "========" << vcl_endl;
		}

		regions_list_t propagated;
		vcl_vector<int> sup_rs;
		propagate( &support1, &support2, &cur_dest_regs_v1, propagated, sup_rs, false, radius );
		if(refern_hist)
			refern_hist->push_back(sup_rs);

		/* Copy propagated regions into current cfg of regions */
		if(propag_hist)
			propag_hist->resize(iter+1);
		for( unsigned int i=0; i<propagated.size(); i++ ) {
			int s = propagated[i]->get_sid();
			if(propag_hist)
				(*propag_hist)[iter].push_back(s);
			// calculate similarity for new regions and drop if below threshold
			float sim = simCo.get_similarity( cand_sparse_v1[s], propagated[i] );
			if( sim < 1-2*filter_thresh ) {
				delete propagated[i];
				continue;
			}
			assert( cfg_sparse_v2[s] == NULL );
			cfg_sparse_v1[s] = cand_sparse_v1[s];
			cfg_sparse_v2[s] = propagated[i];
			cfg_similarities[s] = sim;
		}

		/*---- Filtering ----*/
		vcl_vector<int> surviving_sids, removed_sids;
		unsigned int nFiltered = rsc.filter( surviving_sids,
		                                     cfg_sparse_v1, cfg_sparse_v2,
		                                     &removed_sids );
		// new regions: propagated ones that survived the filters
		unsigned int numNew = getNumCommon( propagated, surviving_sids );
		// Remove filtered regions from the cfg
		for( unsigned int i=0; i<removed_sids.size(); i++ ) {
			int s = removed_sids[i];
			cfg_sparse_v1[s] = NULL;
			if( removed_sids[i] <= max_cand_sid )
				delete cfg_sparse_v2[s];
			cfg_sparse_v2[s] = NULL;
		}
		if(remove_hist)
			remove_hist->push_back(removed_sids);
		if(verbose)
			vcl_cout << "Removed " << removed_sids.size() << " regions" << vcl_endl;

		/*---- Prepare next iter ----*/
		// continuation criterion: increase in new regions > stop_thresh
		needMore = ( numNew >= stop_thresh );

		// Save history
		vConfigSids->resize(iter+1);
		for( int s=0; s<=max_ref_sid; s++ ) {
			if( cfg_sparse_v1[s] )
				(*vConfigSids)[iter].push_back(s);
		}

		// Avoid infinite loop:
    // has current cfg already appeared in the past ?
		for( unsigned int i=0; i<iter; i++ ) {
			if( (*vConfigSids)[i].size() == (*vConfigSids)[iter].size() ) {
				if( (*vConfigSids)[i] == (*vConfigSids)[iter] ) {
					if(verbose)
						vcl_cout << "WARNING: Infinite loop detected: iter " << i+1 << " has identical config as current iter " << iter+1 << vcl_endl;
					needMore = 0;
				}
			}
		}

		iter++;
	}

	// copy propagated regions to prop_rs2 and create final list of sids
	for( int s=0; s<=max_cand_sid; s++ ) {
		if(cfg_sparse_v2[s]) {
			prop_rs2.push_back(cfg_sparse_v2[s]);
			config_sids.push_back(s);
		}
	}
	for( int s=max_cand_sid+1; s<=max_ref_sid; s++ ) {
		if(cfg_sparse_v2[s])
			config_sids.push_back(s);
	}

	if(config_hist == NULL)
		delete vConfigSids;
}


int MosaicPropagator::getMaxSid( const regions_list_t &regs )
{
	int s, maxSid = 0;
	for( unsigned int i=0; i<regs.size(); i++ ) {
		if( (s = regs[i]->get_sid()) > maxSid )
			maxSid = s;
	}
	return maxSid;
}


unsigned int MosaicPropagator::getNumCommon( const regions_list_t &regs,
                                             const vcl_vector<int> &sids )
{
	unsigned int nc = 0;
	for( unsigned int i=0; i<regs.size(); i++ ) {
		int sid = regs[i]->get_sid();
		for( unsigned int j=0; j<sids.size(); j++ ) {
			if( sid == sids[j] ) {
				nc++;
				break;
			}
		}
	}
	return nc;
}
