#ifndef MOSAIC_PROPAGATOR_H
#define MOSAIC_PROPAGATOR_H

/* This class is the equivalent of the MosaicPropagation.m file (duh)
 * and includes the RegionsPropagator.C */

#include <misc_utility/defines.h>
#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/ncc_drgb_computor.h>
#include <region_propagation/raywalk.h>


class MosaicPropagator
{
public:
	MosaicPropagator( const vil1_image &image1, const vil1_image &image2,
	                  bool bVerbose=false, bool small=false );
	~MosaicPropagator();

	/* This is the equivalent of the old RegionsPropagator.C */
	/* Performs a single propagation step.
	 * cand_rs2 will be overwritten (make sure it doesn't contain allocated regions)
	 * with propagated regions from cand_rs1, using ref_rs1 and ref_rs2 as support.
	 * sup_rs will be filled with the sids from the reference regions that have been
	 * successfully propagated.
	 * Set blade to true to use blade propagation instead of the default integrated
	 * propagation. Providing bias is only useful for integrated propagation. */
	void propagate( regions_list_t *ref_rs1,  regions_list_t *ref_rs2,
	                regions_list_t *cand_rs1, regions_list_t &cand_rs2,
	                vcl_vector<int> &sup_rs,
	                bool blade=false, float radius=120.0,
	                vcl_vector<float> *bias=NULL );

	/* This is the equivalent of MosaicPropagation.m */
	/* Mix of ControlledPropagation and CompositePropagationReform
	 * used in model building for POR-mosaic approach.
	 * The propagated regions from cand_rs1 will be stored in prop_rs2.
	 * The sids of the entire final configuration will be returned in config_sids.
	 * RESTRICTION: all sids in ref_rs must be larger than the largest
	 *              sid in cand_rs1!!!
	 * You can retrieve the evolution of the configuration, propagated,
	 * successful reference and filtered region sids with the respective
	 * _hist vectors. */
	void mosaicPropagate( regions_list_t &ref_rs1,  regions_list_t &ref_rs2,
	                      regions_list_t &cand_rs1, regions_list_t &prop_rs2,
												vcl_vector<int> &config_sids, /* could be changed to just ref_sids */
	                      float filter_thresh=0.0, unsigned int stop_thresh=1,
	                      vcl_vector< vcl_vector<int> > *config_hist=NULL,
	                      vcl_vector< vcl_vector<int> > *propag_hist=NULL,
	                      vcl_vector< vcl_vector<int> > *refern_hist=NULL,
	                      vcl_vector< vcl_vector<int> > *remove_hist=NULL );

	/* Return largest sid found in regs */
	int getMaxSid( const regions_list_t &regs );

	/* Set verbose output */
	void setVerbose( bool v )
	{ verbose = v; };

private:
	/* Return number of regions in regs whose sid is in sids. */
	unsigned int getNumCommon( const regions_list_t &regs, const vcl_vector<int> &sids );

	bool verbose;
	color_buffer_t *buf1;
	color_buffer_t *buf2;
	crosscorrelation_computor* ncc;
	ncc_drgb_computor* sc;
	raywalk* refiner;
};

#endif //MOSAIC_PROPAGATOR_H
