#include <vcl_iostream.h>

#include <region_utility/misc.h>
#include <region_propagation/raywalk.h>
#include <region_propagation/GDSixDim.h>

#include "regionsrefiner.h"


RegionsRefiner::RegionsRefiner( const color_buffer_t *buf1, const color_buffer_t *buf2,
                                bool bVerbose, bool small, bool use6dgd ) :
	imgBuf1(buf1),
	imgBuf2(buf2)
{
	verbose = bVerbose;
	sc = new crosscorrelation_computor(buf1, buf2);
	sc->set_resolution(30);                      // default res=50 is too slow

	if(use6dgd)
		ref = new GDSixDim(sc, NULL, NULL, false); // last param -> verbose
	else
		ref = new raywalk(sc, NULL, NULL, false);
  
	if(small) {
		float sb0[] = {-8, 1, 8};
		float sb1[] = {-8, 1, 8};
		float sb2[] = {0.8, 0.1, 1.3};
		float sb3[] = {0.8, 0.1, 1.3};
		float sb4[] = {-3.1416/8.0, 3.1416/16, 3.1416/8.0};
		float sb5[] = {-0.5, 0.1, 0.5};
		float* small_bounds[] = {sb0, sb1, sb2, sb3, sb4, sb5};  // bounds matrix  (float**)
		ref->set_bounds(small_bounds);
	}
	else {
		float nb0[] = {-14, 2, 14};
		float nb1[] = {-14, 2, 14};
		float nb2[] = {0.6, 0.1, 1.8};
		float nb3[] = {0.6, 0.1, 1.8};
		float nb4[] = {-3.1416/4.0, 3.1416/16, 3.1416/4.0};
		float nb5[] = {-1, 0.2, 1};
		float* normal_bounds[] = {nb0, nb1, nb2, nb3, nb4, nb5}; // bounds matrix  (float**)
		ref->set_bounds(normal_bounds);
	}
	aoiref = new aoi_refiner(sc, NULL, NULL, false);
	float br[] = {0, 3.1416/36.0, 2*3.1416};     // rotational search bounds
	aoiref->set_bounds(br);
}

RegionsRefiner::~RegionsRefiner()
{
	delete aoiref;
	delete ref;
	delete sc;
}


void RegionsRefiner::refine( const regions_list_t &regions1,
                             const regions_list_t &regions2,
                             regions_list_t &regions2_ref,
                             vcl_vector<float> *pre_sims,
                             vcl_vector<float> *ref_sims )
{
	if(verbose)
		vcl_cout << "Refining regions (resolution = " << sc->get_resolution() << ")..." << vcl_endl;
	region *reg1, *reg2, *reg2_aoiref;
	float avg_init = 0.0;                         // average initial similarity
	float avg_refined = 0.0;                      // average refined similarity

	if( regions2_ref.size() > 0 )
		delete_regions(regions2_ref);

	for (int rx = 0; rx < regions1.size(); rx++) {
		reg1 = regions1[rx];
		reg2 = regions2[rx];                        // same index in list
		if( reg1->get_sid() != reg2->get_sid() ) {
			vcl_cerr << vcl_endl << vcl_endl << "WARNING: not corresponding regions ids: " << reg1->get_sid() << " != " << reg2->get_sid() << vcl_endl;
		}
		if(verbose)
			vcl_cout << "Refining region " << reg1->get_sid() << " : ";
		if( reg1->get_type() == region::ELL_REGIONTYPE ) {
			reg2_aoiref = aoiref->refine(reg1, reg2);
			reg2_aoiref = copy_region(reg2);          // avoid aoiref to simulate prop ref
			if(pre_sims)
				pre_sims->push_back(aoiref->get_init_similarity());
			regions2_ref.push_back(ref->refine(reg1, reg2_aoiref));
			if(verbose) {
				avg_init += aoiref->get_init_similarity();
				vcl_cout << aoiref->get_init_similarity();
			}
			delete reg2_aoiref;
		}
		else {                                      // parall reg here
			regions2_ref.push_back(ref->refine(reg1, reg2));
			if(pre_sims)
				pre_sims->push_back(ref->get_init_similarity());
			if(verbose) {
				avg_init += ref->get_init_similarity();
				vcl_cout << ref->get_init_similarity();
			}
		}

		if(ref_sims)
			ref_sims->push_back(ref->get_max_similarity());
		if(verbose) {
			avg_refined += ref->get_max_similarity();
			vcl_cout << "  ->  " << ref->get_max_similarity() << vcl_endl;
		}
	}
	vcl_cout << vcl_endl;

	if(verbose) {
		// Print side-info on screen
		vcl_cout << "Average init sim:    " << avg_init/(float)regions1.size() << vcl_endl;
		vcl_cout << "Average refined sim: " << avg_refined/(float)regions1.size() << vcl_endl;
	}
}
