#include <algorithm>
#include <set>
#include <cstdlib>
using namespace std;

#include <vcl/vcl_iostream.h>
#include <regions/region.h>

#include "order_statistic_tree.h"
#include "rscfilter.h"


const double PI = atan(1.0) * 4.0;
// Threshold for determining whether points are on a line
// TODO: check whether this test and the treatment of such points is the same as in the Matlab filter.
const double eps = 0.0000001;

vcl_ostream &operator<<( vcl_ostream &out, pt p )
{
  out << "(" << real( p ) << "," << imag(p) << ")";
  return out;
}


RSCFilter::RSCFilter( bool bVerbose )
{
	verbose = bVerbose;
}

RSCFilter::~RSCFilter()
{
}


unsigned int RSCFilter::filter( vcl_vector<int> &surviving_sids,
                                const regions_list_t &ref_regs,
                                const regions_list_t &test_regs,
                                vcl_vector<int> *filter_history )
{
	if( ref_regs.size() != test_regs.size() ) {
		vcl_cerr << "ERROR in RSCFilter::filter: not the same number of regions in the two views." << vcl_endl;
		return 0;
	}
	vcl_vector<pt> model, img;
	vcl_vector<int> sids;
	for( unsigned int i=0; i<ref_regs.size(); i++ ) {
		if( ref_regs[i] == NULL )
			continue;

		point_t *p = ref_regs[i]->get_pt0();
		model.push_back( pt( p->x(), p->y() ) );

		p = test_regs[i]->get_pt0();
		img.push_back( pt( p->x(), p->y() ) );

		sids.push_back( ref_regs[i]->get_sid() );
	}

	return filter( surviving_sids, model, img, &sids, filter_history );
}


// TODO: change the behaviour in the case of collinear points.
// As it is implemented now, collinear points will get a lower number of satisfied
// constraints, hence a larger number of violations, hence are more likely to be
// removed even if they're in perfect agreement.
unsigned int RSCFilter::filter( vcl_vector<int> &surviving_ids,
												        vcl_vector<pt> &model,
                                vcl_vector<pt> &img,
                                vcl_vector<int> *ids,
                                vcl_vector<int> *filter_history )
{
	if( model.size() != model.size() ) {
		vcl_cerr << "ERROR in RSCFilter::filter: not the same number of regions in the two views." << vcl_endl;
		return 0;
	}

	vcl_vector<int> indices, cnt, rmHist;
	if(ids)
		indices = *ids;
	else {
		for( unsigned int i=0; i<model.size(); i++ )
			indices.push_back(i+1);
	}

	// initialize for each point number of satisfied sidedness constraints
	init( model, img, cnt );

	// remove point by point
	unsigned int n = model.size();
	while( n > 2 ) {
		// find worst point
		unsigned int k = 0;
		for( unsigned int j = 0; j < n; j++ ) {
			if( cnt[j] < cnt[k] )
				k = j;
		}

		// Stop criterion: no points are involved in more than 15% of the maximal number of mismatches
		unsigned int v = (n-1)*(n-2)/2; // number of constraints
		if( v-cnt[k] <= .15*v )
			break;

		// remove it
		remove_point( k, model, img, cnt );
		rmHist.push_back( indices[k] );
		indices.erase( indices.begin() + k );
		n--;
	}

	surviving_ids = indices;
	if( filter_history )
		*filter_history = rmHist;

	return rmHist.size();
}


double RSCFilter::normalize_angle( double a )
{
	while( a < 0 )
		a += 2*PI;
	while( a >= 2*PI )
		a -= 2*PI;
	return a;
}


/*
  i : first point of the triangle
  complexity: O(n*log(n)), where n = number of points
  returns a vector where each element corresponds to the number of points
  which are on the left side of a line - drawn between i and this point -
  for the model and the image.
  left side means, not on the line!
  
  I hate comments, so have a look at the simple algorithm
  there everything should be obvious.
*/
vcl_vector<int> RSCFilter::report_points_on_left_side( int i, vcl_vector<pt> model,
                                                       vcl_vector<pt> img )
{
	typedef pair< double, int > PDI;
	int n = model.size();
	vcl_vector< PDI > model_a;
	vcl_vector< PDI > img_a;
	vcl_vector< int > res( n, 0 );
 
	for( int j = 0; j < n; j++ ) {
		if( model[i] == model[j] || img[i] == img[j] )
			continue;
		else {
			model_a.push_back( PDI( arg( model[j] - model[i] ), j ) );
			img_a.push_back( PDI( arg( img[j] - img[i] ), j ) );
		}
	}

	sort( model_a.begin(), model_a.end() );
	sort( img_a.begin(), img_a.end() );
	vcl_vector< int > to_model_a( n );
	n = model_a.size();
	for( int j = 0; j < n; j++ )
		to_model_a[model_a[j].second] = j;

	sum_statistic< int > cnt;  
	cnt.resize( n+1 );

	vcl_vector< int > start(n), end(n);

	// prepare links
	for( int l = 0, r = 0, ll = 0; l < n; l++ ) {
		double last = normalize_angle( model_a[ll%n].first - model_a[l].first );
		while( r < l+n ) { // add further points
			double curr = normalize_angle( model_a[r%n].first - model_a[l].first );
			if( curr >= PI-eps || curr < last )
				break;
			last = curr;
			r++;
		}
		while( ll < r ) { // remove points on line
			if( normalize_angle( model_a[ll%n].first - model_a[l].first ) > eps )
				break;
			ll++;
		}
		start[l] = ll;
		end[l] = r;
	}

	for( int l = 0, r = 0, ll = 0; l < n; l++ ) { // process all the points
		// l : second point of the triangle
		double last = normalize_angle( img_a[ll%n].first - img_a[l].first );
		while( r < l+n ) { // add further points
			double curr = normalize_angle( img_a[r%n].first - img_a[l].first );
			if( curr >= PI-eps || curr < last )	
				break;
			last = curr;
			cnt.set_value( to_model_a[ img_a[r%n].second ], 1 ); // mark it
			r++;
		}
		while( ll < r ) { // remove points on line
			if( normalize_angle( img_a[ll%n].first - img_a[l].first ) > eps )
				break;
			cnt.set_value( to_model_a[ img_a[ll%n].second ], 0 );
			ll++;
		}

		int ref = to_model_a[ img_a[ l ].second ];
		int a = start[ref];
		int b = end[ref];
		if( b == a + n )
			res[ img_a[l].second ] = cnt.get_statistic( 0, n );
		else if( b <= n )
			res[ img_a[l].second ] = cnt.get_statistic( a, b );
		else
			res[ img_a[l].second ] = r-ll - cnt.get_statistic( b%n, a );
	}
	return res;
}


vcl_vector<int> RSCFilter::report_points_on_left_and_right_side( int i,
                                                                 vcl_vector<pt> model,
                                                                 vcl_vector<pt> img )
{
	vcl_vector< int > left, right;
	left = report_points_on_left_side( i, model, img );
	for( int j = 0; j < model.size(); j++ ) {
		model[j] = conj( model[j] );
		img[j] = conj( img[j] );
	}    
	right = report_points_on_left_side( i, model, img );
	for( int j = 0; j < left.size(); j++ )
		left[j] += right[j];
	return left;
}


/*
  fills cnt[i] with the number of sidedness constraints satisfied for point i
*/
void RSCFilter::init( const vcl_vector<pt> &model, const vcl_vector<pt> &img,
                      vcl_vector<int> &cnt )
{
	int n = model.size();
	cnt.resize( n, 0 );
	for( int i = 0; i < n; i++ ) {
		vcl_vector< int > tmp = report_points_on_left_side( i, model, img );
		for( int j = 0; j < n; j++ )
			cnt[i] += tmp[j];
	}
}


/*
  removes point i from all vectors (relative ordering remains)
  and updates cnt
*/
void RSCFilter::remove_point( int i, vcl_vector<pt> &model, vcl_vector<pt> &img,
                              vcl_vector<int> &cnt )
{
	vcl_vector< int > tmp = report_points_on_left_and_right_side( i, model, img );
	for( int k = 0; k < model.size(); k++ )
		cnt[k] -= tmp[k];
	img.erase( img.begin() + i );
	model.erase( model.begin() + i );
	cnt.erase( cnt.begin() + i );
}


/*-- Various methods for testing purposes, can be removed --*/

/*
  simple algorithm to compare results...
 */
vcl_vector<int> RSCFilter::report_points_on_left_side_( int i, 
                                                        vcl_vector<pt> model,
                                                        vcl_vector<pt> img )
{
	int n = model.size();
	vcl_vector< int > res( n, 0 );
	for( int j = 0; j < n; j++ ) {
		if( model[j] == model[i] || img[j] == img[i] )
			continue;
		for( int k = 0; k < n; k++ ) {
			if( model[k] == model[i] || img[k] == img[i] )
				continue;
			/*vcl_cout << j << " " << k << " "
			         << normalize_angle( arg( model[k] - model[i] ) - arg( model[j] - model[i] ) )
			         << " "
			         << normalize_angle( arg( img[k] - img[i] ) - arg( img[j] - img[i] ) ) 
			         << endl;*/
			if( normalize_angle( arg( model[k] - model[i] ) - arg( model[j] - model[i] ) ) < PI-eps && 
					normalize_angle( arg( img[k] - img[i] ) - arg( img[j] - img[i] ) ) < PI-eps &&
					normalize_angle( arg( model[k] - model[i] ) - arg( model[j] - model[i] ) ) > eps &&
					normalize_angle( arg( img[k] - img[i] ) - arg( img[j] - img[i] ) ) > eps )
				res[j]++;
		}
	}
	return res;
}


vcl_vector<pt> RSCFilter::rand_points( int n )
{
	vcl_vector< pt > res( n );
	for( int i = 0; i < n; i++ )
		res[i] = pt( rand() % 100, rand() % 100 );
	return res;
}


vcl_vector<pt> RSCFilter::normalizerand_points( int n )
{
	vcl_vector< pt > res( n );
	for( int i = 0; i < n; i++ )
		res[i] = pt( rand() % 100, rand() % 100 );
	return res;
}


template<class T>
void RSCFilter::print( vcl_vector<T> r )
{
	for( int i = 0; i < r.size(); i++ )
		vcl_cout << r[i] << " ";
	vcl_cout << endl;
}


void RSCFilter::test_report_points_on_left_side()
{
	vcl_vector< pt > model, img;
	model.push_back(pt(352,843));
	model.push_back(pt(372,243));
	model.push_back(pt(888,418));
	model.push_back(pt(548,417));
	model.push_back(pt(139,595));
	model.push_back(pt(728,711));
	model.push_back(pt(186,332));
	model.push_back(pt(970,147));
	model.push_back(pt(621,264));
	model.push_back(pt(38,330));
 
	img.push_back(pt(266,56));
	img.push_back(pt(389,685));
	img.push_back(pt(424,508));
	img.push_back(pt(350,557));
	img.push_back(pt(978,635));
	img.push_back(pt(765,855));
	img.push_back(pt(92,56));
	img.push_back(pt(209,56));
	img.push_back(pt(967,545));
	img.push_back(pt(662,375));
	while( 1 ) {
		int n = 1 + ( rand() % 100 );
		//n = 10;
		cout << n << endl;
		// create some random points
		model = rand_points( n );
		img = rand_points( n );

		for( int i = 0; i < n; i++ ) {
			vector< int > r = report_points_on_left_side( i, model, img );
			vector< int > r1 = report_points_on_left_side_( i, model, img );
			if( r != r1 ) {
				cout << "ERROR" << endl;
				print( model );
				print( img );
				print( r );
				print( r1 );
			}
		}
		//break;
	}
}


int RSCFilter::testRandom()
{
	vcl_vector< pt > model, img;
	while( 1 ) {
		int n = 1 + ( rand() % 1000 );
		cout << n << endl;

		// create some random points
		model = rand_points( n );
		img = rand_points( n );

		// initialize for each point number of satisfied sidedness constraints
		vcl_vector< int > cnt;
		init( model, img, cnt );

		// remove point by point
		for( int i = n; i > 0; i-- ) {

			// find worst point
			int k = 0;
			for( int j = 1; j < i; j++ )
				if( cnt[j] < cnt[k] ) k = j;

			// remove it
			remove_point( k, model, img, cnt );

			// check if everything is correct by calculating cnt2 from scratch
			/*vcl_vector< int > cnt2;
			init( model, img, cnt2 );
			assert( cnt == cnt2 );*/
		}
	}
	return 0;
}
