#include <vcl_iostream.h>
#include <vil1/vil1_image.h>
#include <vil1/vil1_rgb_byte.h>

#include <algorithm>

#include <region_utility/misc.h>
#include <main_processing/regionsrefiner.h>

#include "viewpairmatcher.h"


/* Parameters for the refiner */
#define SMALL_SEARCH false
#define GD6D         false


ViewPairMatcher::ViewPairMatcher( const vil1_image &image1, const vil1_image &image2,
                                  bool nccOnly, bool bVerbose )
{
	verbose = bVerbose;
	imgBuf1 = new color_buffer_t(image1);
	imgBuf2 = new color_buffer_t(image2);

	siftMatcher = new EuclideanSoftMatcher( 70000, verbose );

	if( nccOnly )
		simCo = new crosscorrelation_computor(imgBuf1, imgBuf2);
	else
		simCo = new ncc_drgbsnorm_computor(imgBuf1, imgBuf2);
  simCo->set_resolution(30);
  simCo->set_adaptive_resolution(true);
}

ViewPairMatcher::~ViewPairMatcher()
{
	delete simCo;
	delete siftMatcher;
	delete imgBuf1;
	delete imgBuf2;
}


int ViewPairMatcher::matchRegionSets( const vcl_vector<regions_list_t> &regionSets1,
                                      const vcl_vector<regions_list_t> &regionSets2,
                                      regions_list_t &finalRegions1,
                                      regions_list_t &finalRegions2,
																			unsigned int maxMatches, int first_sid,
                                      bool keepDescriptors )
{
	if( finalRegions1.size() > 0 )
		delete_regions(finalRegions1);
	if( finalRegions2.size() > 0 )
		delete_regions(finalRegions2);

	regions_list_t matchRegs1, matchRegs2;   // will hold the same regs as input
	for( unsigned int i=0; i<regionSets1.size(); i++ ) {
		vcl_vector<EuclMatch> vMatches;
		const regions_list_t &regs1 = regionSets1[i];
		const regions_list_t &regs2 = regionSets2[i];
		
		unsigned int nMatches = siftMatcher->match_soft3( vMatches, regs1, regs2 );
		if(verbose)
			vcl_cout << "Found " << nMatches << " soft 3-matches for regions type #" << i << vcl_endl;
		/* Copy matching regions to merged list */
		for( unsigned int j=0; j<vMatches.size(); j++ ) {
			EuclMatch &m = vMatches[j];
			matchRegs1.push_back(regs1[m.index1]);
			matchRegs2.push_back(regs2[m.index2]);
		}
	}
	if( matchRegs1.size() == 0 )
		return 0;
	
	/* Calculate similarities between all matches */
	vcl_vector<EuclMatch> vAllMatches;
	compute_similarities( vAllMatches, matchRegs1, matchRegs2 );

	/* Keep only the maxMatches*3 best matches */
	if( maxMatches && vAllMatches.size() > maxMatches*3 ) {
		if(verbose)
			vcl_cout << "Keeping only best " << maxMatches*3 << " of "
			         << vAllMatches.size() << " initial matches" << vcl_endl;
		sort( vAllMatches.begin(), vAllMatches.end(), cmpEuclMatchDesc );
		vAllMatches.erase( vAllMatches.begin()+maxMatches*3, vAllMatches.end() );
	}

	/* Clone regions and match sids */
	regions_list_t newRegions1, newRegions2; // will hold copies of regions
	regions_list_t tempRegions2;             // to be refined
	for( unsigned int i=0; i<vAllMatches.size(); i++ ) {
		region *copyReg = copy_region( matchRegs1[vAllMatches[i].index1] );
		int sid = copyReg->get_sid();
		/* It's not 100% efficient to first copy the invariants and then
		 * delete them, but it would be too complicated to do it otherwise */
		if( !keepDescriptors )
			copyReg->clear_invariants();
		newRegions1.push_back(copyReg);

		copyReg = copy_region( matchRegs2[vAllMatches[i].index2] );
		copyReg->set_sid(sid);
		if( !keepDescriptors )
			copyReg->clear_invariants();
		tempRegions2.push_back(copyReg);
	}
	
	/* Refine the matching regions and recalculate similarities */
	RegionsRefiner rRefiner( imgBuf1, imgBuf2, verbose, SMALL_SEARCH, GD6D );
	rRefiner.refine( newRegions1, tempRegions2, newRegions2 );
	delete_regions(tempRegions2);
	vAllMatches.clear();
	compute_similarities( vAllMatches, newRegions1, newRegions2 );

	/* Keep only matches with sim>=1.0, and eventually the maxMatches best ones */
	sort( vAllMatches.begin(), vAllMatches.end(), cmpEuclMatchAsc );
	unsigned int firstOK = 0;
	do {
		if( vAllMatches[firstOK].distance >= 1.0 )  // 'distance' is actually similarity here
			break;
		delete newRegions1[vAllMatches[firstOK].index1];
		delete newRegions2[vAllMatches[firstOK].index2];
	}	while( ++firstOK < vAllMatches.size() );
	if(verbose)
		vcl_cout << "Filter out matches with sim < 1.0 --> "
		         << vAllMatches.size()-firstOK << " matches kept" << vcl_endl;
	if( firstOK == vAllMatches.size() )
		return 0;

	if( maxMatches && vAllMatches.size()-firstOK > maxMatches ) {
		if(verbose)
			vcl_cout << "Keeping only best " << maxMatches << " refined matches" << vcl_endl;
		do {
			delete newRegions1[vAllMatches[firstOK].index1];
			delete newRegions2[vAllMatches[firstOK].index2];
		} while( ++firstOK < vAllMatches.size()-maxMatches );
	}

	/* Copy to output */
	int sid = first_sid;
	do {
		region *finalReg = newRegions1[vAllMatches[firstOK].index1];
		if( first_sid > -1 )
			finalReg->set_sid(sid);
		finalRegions1.push_back(finalReg);

		finalReg = newRegions2[vAllMatches[firstOK].index2];
		if( first_sid > -1 )
			finalReg->set_sid(sid++);
		finalRegions2.push_back(finalReg);
	} while( ++firstOK < vAllMatches.size() );

	return finalRegions1.size();
}


void ViewPairMatcher::compute_similarities( vcl_vector<EuclMatch> &vMatches,
                                            const regions_list_t &regions1,
                                            const regions_list_t &regions2,
                                            int id_offset )
{
	for( unsigned int i=0; i<regions1.size(); i++ ) {
		float sim = simCo->get_similarity( regions1[i], regions2[i] );
		EuclMatch newMatch = { i+id_offset, i+id_offset, sim };
		vMatches.push_back(newMatch);
	}
}
