#include "abl_utility.h"

#include <stdlib.h>
#include <time.h>

#include <sys/types.h> 
#include <sys/stat.h>
#include "memory.h"


#ifdef _ABLUTILITYCANUSEZLIB_
#include "zlib.h"

//  This function creates a gzipped file named ZipFileName from the uncompressed
//  file InFileName.
bool ABLZipFromFileToFile(const char *InFileName, const char *ZipFileName)
{
  struct _stat FileStat;
  int Result = _stat(InFileName, &FileStat);

  if ( Result != 0 )
  {
    return false;
  }

  int FileSize = FileStat.st_size;

  gzFile fin = gzopen(InFileName, "rb");

  if ( fin == NULL )
  {
    return false;
  }

  unsigned char *input_buffer = new unsigned char[FileSize];
  int NumRead = gzread(fin, input_buffer, FileSize);
  gzclose(fin);

  gzFile fout = gzopen(ZipFileName, "wb");

  if ( fout == NULL )
  {
    delete [] input_buffer;
    return false;
  }

  int NumWritten = gzwrite(fout, input_buffer, FileSize);
  gzclose(fout);

  if ( NumWritten != FileSize )
  {
    return false;
  }

  delete [] input_buffer;
	
  return true;
}


//  This reads an uncompressed file and gzips it to the buffer Buf.  Buf must be
//  preallocated to size BufSize.  BufSize comes back with the amount of memory that
//  was actually used.
bool ABLZipFromFileToMem(const char *FileName, unsigned char *Buf, int &BufSize)
{
  struct _stat FileStat;
  int Result = _stat(FileName, &FileStat);

  if ( Result != 0 )
  {
    return false;
  }

  int FileSize = FileStat.st_size;

  if ( BufSize < FileSize )
  {
    return false;
  }

  FILE* F = fopen(FileName, "rb");

  if ( F == NULL )
  {
    return false;
  }

  unsigned char *TempBuf = new unsigned char[FileSize];

  fread(TempBuf, sizeof(unsigned char), FileSize, F);
  fclose(F);

  unsigned long TempBSize = (unsigned long)BufSize;

  int ERR = compress(Buf, &TempBSize, TempBuf, FileSize);

  BufSize = (int)TempBSize;

  if ( ERR != 0 )
  {
    return false;
  }

  return true;
}

//  This function creates a gzipped file named ZipFileName from the uncompressed data
//  in the array Data/
bool ABLZipFromMemToFile(const unsigned char *Data, const int &DataSize, const char *ZipFileName)
{
  gzFile fout = gzopen(ZipFileName, "wb");
  if ( fout == NULL )
  {
    return false;
  }

  unsigned char *TempBuf = new unsigned char[DataSize];
  memcpy(TempBuf, Data, DataSize);

  int NumWritten = gzwrite(fout, TempBuf, (unsigned long)DataSize);
  delete [] TempBuf;
  gzclose(fout);
	
  if ( NumWritten != DataSize )
  {
    return false;
  }

  return true;
}


//  The destination buffer must be preallocated to at least (SourceLen + 0.01*SourceLen + 12).  DestLen will 
//  be returned with the size of buffer that was used.
bool ABLZipFromMemToMem(const unsigned char *Source, const int &SourceLen, unsigned char *Dest, int &DestLen)
{
  DestLen = SourceLen + (int)(0.01*(double)SourceLen) + 20;
  unsigned long TempDestLen = (unsigned long)DestLen;
  int ERR = compress(Dest, &TempDestLen, Source, SourceLen);
  DestLen = (int)TempDestLen;

  if ( ERR != 0 )
  {
    return false;
  }

  return true;
}


//  This function creates an umcompressed file called UnZippedFileName from the gzipped
//  file ZipFileName.
bool ABLUnZipFromFileToFile(const char *ZipFileName, const char *UnZippedFileName)
{

  gzFile InFile = gzopen(ZipFileName, "rb");

  if ( InFile == NULL )
  {
    return false;
  }

  int HugeNum = 10000000;
  unsigned char *HugeBuf = new unsigned char[HugeNum];
  int NumRead = gzread(InFile, HugeBuf, HugeNum);

  gzclose(InFile);

  FILE* OutFile = fopen(UnZippedFileName, "wb");

  if ( OutFile == NULL )
  {
    delete [] HugeBuf;
    return false;
  }

  fwrite(HugeBuf, sizeof(unsigned char), NumRead, OutFile);
  fclose(OutFile);
  delete [] HugeBuf;

  return true;
}


//  The data buffer must be preallocated.  OutPutSize must be the size of the buffer.  OutPutSize
//  returns as the number of bytes that were actually used in the buffer.
bool ABLUnZipFromFileToMem(const char *ZipFileName, unsigned char * data, int &OutPutSize)
{
  gzFile ZipFile = gzopen(ZipFileName, "rb");

  if ( ZipFile == NULL )
  {
    return false;
  }

  OutPutSize = gzread(ZipFile, data, OutPutSize);
  gzclose(ZipFile);

  return true;
}


//  You need to allocate your own memory for the destination buffer. DestLen needs to be the size of the allocated
//  buffer Dest, but it will come back with the amount of memory that was actually used.
bool ABLUnZipFromMemToMem(const unsigned char *Source, const int &SourceLen, unsigned char * Dest, int &DestLen)
{
  unsigned long TempDestLen = (unsigned long)DestLen;
  int ERR = uncompress(Dest, &TempDestLen, Source, SourceLen);
  DestLen = (int)TempDestLen;

  if ( ERR != 0 )
  {
    return false;
  }

  return true;
}


//  This uncompresses a gzipped memory buffer to a file.  It uses a lot memory in relation to the
//  size of the buffer, but it needs to have an upper bound on the size of the uncompressed buffer.
bool ABLUnZipFromMemToFile(const unsigned char *Source, const int &SourceLen, char *FileName)
{
  unsigned long TempLen = 20*SourceLen;
  unsigned char *TempBuf = new unsigned char[TempLen];

  int ERR = uncompress(TempBuf, &TempLen, Source, SourceLen);

  if ( ERR != 0 )
  {
    return false;
  }

  FILE *F = fopen(FileName,"wb");

  if ( F == NULL )
  {
    delete [] TempBuf;
    return false;
  }

  fwrite(TempBuf, sizeof(unsigned char), TempLen, F);
  fclose(F);

  return true;
}

#endif


double DistanceFromAPointToALineSegment(const double &p0x, const double &p0y, const double &p1x, 
                                        const double &p1y, const double &p2x, const double &p2y)
{
  static double d1,d2,u,x,y;
	
  u = ((p0x-p1x)*(p2x-p1x) + (p0y-p1y)*(p2y-p1y)) / 
    ((p2x-p1x)*(p2x-p1x) + (p2y-p1y)*(p2y-p1y));

  if ( u >= 0.0 && u <= 1 )
  {
    x = p1x + u*(p2x - p1x);
    y = p1y + u*(p2y - p1y);
    return sqrt( (p0x-x)*(p0x-x) + (p0y-y)*(p0y-y) );
  }
	
  d1 = (p0x-p1x)*(p0x-p1x) + (p0y-p1y)*(p0y-p1y);
  d2 = (p0x-p2x)*(p0x-p2x) + (p0y-p2y)*(p0y-p2y);

  return sqrt( MINTWO(d1, d2) );
}

double DistanceBetweenTwoLineSegments(const double &x1, const double &y1, const double &x2, const double &y2,
                                      const double &x3, const double &y3, const double &x4, const double &y4)
{
  static double dist1, dist2, dist3, dist4;

  dist1 = DistanceFromAPointToALineSegment(x1, y1, x3, y3, x4, y4);
  dist2 = DistanceFromAPointToALineSegment(x2, y2, x3, y3, x4, y4);
  dist3 = DistanceFromAPointToALineSegment(x3, y3, x1, y1, x2, y2);
  dist4 = DistanceFromAPointToALineSegment(x4, y4, x1, y1, x2, y2);

  return MINFOUR(dist1, dist2, dist3, dist4);
}

double DistanceFromAPointToALine(double x1, double y1, double x2, double y2, double X, double Y)
{
  double NormSquareStart = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
  double ux = X - x1;
  double uy = Y - y1;
  double u1x = (( (X - x1)*(x2 - x1) + (Y - y1)*(y2 - y1) )
                /NormSquareStart) * (x2 - x1);
  double u1y = (( (X - x1)*(x2 - x1) + (Y - y1)*(y2 - y1) )
                /NormSquareStart) * (y2 - y1);

  return sqrt( (u1x-ux)*(u1x-ux) + (u1y-uy)*(u1y-uy) );
}
