
/******************************************************************************
*
*	Title:	ABLUtility
*	Author:	Aaron Luttman
*	Start Date:	July 9, 2001
*	Notes:	This is a header and source file pair with constants, macros and some
*			C-style utility functions for generic use.  Some of the functions 
*			require #defines as they may use other libraries.  Each function 
*			(or set of functions) has its (their) own documentation here in the 
*			header.  Should be ANSI C/C++ compatible.
*
*	If you have questions or comments, feel free to email me at
*	aaron.luttman@esat.kuleuven.ac.be.
*
******************************************************************************/


#include <stdio.h>
#include <math.h>

#include <misc_utility/defines.h>

/*
	Here are some basic constants.  They probably exist elsewhere, but I don't trust people.
*/
#define ABL_BIGNUM			1e6					//  Screw machine precision
#define ABL_EPSILON			1e-10				//	Screw machine precision
#define ABL_PI				3.14159265358979
#define ABL_PI_OVER_2		1.57079632679490
#define ABL_PI_OVER_4		0.78539816339745
#define ABL_2_PI			6.28318530717959
#define ABL_e				2.71828182845905
#define ABL_e_SQUARED		7.38905609893065


/*
	Here are some macros for returning the min and max of two, three or four numbers.
	These are also defined in 'defines.h'
*/
// #define MAXTWO(a,b)			( (double)a < (double)b ) ? b : a
// #define MAXTHREE(a, b, c)	MAXTWO( MAXTWO(a,b), (double)c )
// #define MAXFOUR(a,b,c,d)	MAXTWO( MAXTWO(a,b), MAXTWO(c,d) )

// #define MINTWO(a,b)			( (double)a < (double)b ) ? a : b
// #define MINTHREE(a,b,c)		MINTWO( MINTWO(a,b), (double)c )
// #define MINFOUR(a,b,c,d)	MINTWO( MINTWO(a,b), MINTWO(c,d) )
////////////////////////////////////////////////////////////////////////////////////


/*
	Some more macros for commonly used stuff
*/
#define ABL_SQUARE(x)		x * x
#define ABL_CUBE(x)			x * x * x
#define ABL_FOURTH_POWER(x) x * x * x * x



/*	
	Here are the functions for gzipping stuff.  They require the zlib library
	and the #define below must be defined somewhere (in stdafx.h for windows programs,
	whereever crazy linux people put their #defines under crazy linux).  There are
	six functions; one to zip from file to file, one to zip from memory to file,
	one to zip from memory to memory (where the user must allocate sufficient mem),
	one to unzip from file to file, one to unzip from file to memory, and one to unzip
	from mem to mem (where the user allocates as beofre). They all
	return true if they were successful, and false if they were unsuccessful.  It is 
	assumed that zlib.lib (static library) is in your path.

	#define _ABLUTILITYCANUSEZLIB_
*/

#ifdef _ABLUTILITYCANUSEZLIB_
bool ABLZipFromFileToFile(const char *InFileName, const char *ZipFileName);
bool ABLZipFromFileToMem(const char *FileName, unsigned char *Buf, int &BufSize);
bool ABLZipFromMemToFile(const unsigned char *data, const int &DataSize, const char *ZipFileName);
bool ABLZipFromMemToMem(const unsigned char *Source, const int &SourceLen, unsigned char * Dest, int &DestLen);
bool ABLUnZipFromFileToFile(const char *ZipFileName, const char *UnZippedFileName);
bool ABLUnZipFromFileToMem(const char *ZipFileName, unsigned char * data, int &OutPutSize);
bool ABLUnZipFromMemToMem(const unsigned char *Source, const int &SourceLen, unsigned char * Dest, int &DestLen);
bool ABLUnZipFromMemToFile(const unsigned char *Source, const int &SourceLen, char *FileName);
#endif
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////



/*
	This function returns the distance between the point (p0x,p0y) and the line segment determined
	by (p1x,p1y) and (p2x,p2y).  Note: this distance is either the distance from p0 to p1, the distance
	from p0 to p2, or it is the orthogonal distance from p0 to the LINE determined by p1 and p2.
	The line segments in that Distance function are defined by (x1,y1)(x2,y2) and (x3,y3)(x4,y4).
*/
double DistanceFromAPointToALineSegment(const double &p0x, const double &p0y, const double &p1x, 
                                        const double &p1y, const double &p2x, const double &p2y);

double DistanceBetweenTwoLineSegments(const double &x1, const double &y1, const double &x2, const double &y2,
                                      const double &x3, const double &y3, const double &x4, const double &y4);

//  In this case the line is given by the points (x1,y1) and (x2,y2) and the point we care about is (X,Y)
double DistanceFromAPointToALine(double x1, double y1, double x2, double y2, double X, double Y);
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////



/*
	Some macros for basic geometry stuff.  This avoids having to use function calls for normal stuff.
*/
//  Note for DIST that only double values make sense.  If you pass, non-doubles, it converts to doubles.
#define ABL_DIST(x1, y1, x2, y2) sqrt( (((double)x1-(double)x2)*((double)x1-(double)x2)) + ( ((double)y1-(double)y2)*((double)y1-(double)y2) ) )
#define ABL_NORM_2D(x, y) sqrt( (double)x*(double)x + (double)y*(double)y )
#define ABL_NORM_3D(x, y, z) sqrt( (double)x*(double)x + (double)y*(double)y + (double)z*(double)z )



/*
	Just a macro for rounding to the nearest integer.  It only makes sense to use this with doubles and
	floats.  With ints or chars you don't need to round.
*/
#define ABLround(x) ( (int) (double)x+0.5 )
#ifndef round
#define round ABLround
#endif



/*
	Just to have byte defined
*/
#ifndef byte
#define byte unsigned char
#endif

//  This is a define so I can easily switch back and forth between doubles and floats in all my stuff
#define ABLFloat float 

//  Stuff for use in ABLImage
#define _ABLIMAGECANUSEMIHAILINE_
