#include "dilation.h"

#include <vcl_cmath.h>
#include <vcl_vector.h>

#include <vcl/vcl_list.h>
#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_resample.h>

#include <misc_utility/defines.h>

// if anyone feels like it, he can implement the general case!

// color_buffer_t*
// rgb_dilation(const color_buffer_t*buf, const double sigma)
// { 
//   int sizex = buf->width();
//   int sizey = buf->height();

//   // Create 1-D mask:
//   double cutoff = 0.01;
//   double lc = -2 * vcl_log(cutoff); // cutoff guaranteed > 0
//   int radius = (lc<=0) ? 0 : 1 + int(vcl_sqrt(lc)*sigma); // sigma guaranteed >= 0
//   int size = 2*radius + 1;
//   vcl_vector<float> mask(size);
//   double halfnorm = 0.5;
//   mask[radius] = 1.0;
//   for (int x=1; x<=radius; ++x) {
//     double v = vcl_exp(-0.5*x*x/(sigma*sigma));
//     mask[radius - x] = mask[radius + x] = v;
//     halfnorm += v;
//   }

//   // normalise mask
//   double mass_scale = 1.0/(1 + 2*halfnorm);
//   for (int x=0; x< size; ++x)
//     mask[x] *= mass_scale;

//   color_buffer_t *tmp_buf = new color_buffer_t(sizex, sizey);    

//   // create separate buffers for each colorband
//   bw_buffer_t* rTemp = 
//     new bw_buffer_t(sizex, sizey);
//   bw_buffer_t* gTemp = 
//     new bw_buffer_t(sizex, sizey);
//   bw_buffer_t* bTemp = 
//     new bw_buffer_t(sizex, sizey);
  
//   for (int i=0; i<sizex; i++)
//     for (int j=0; j<sizey; j++)
//     { 
//       vil1_rgb<unsigned char> pixel = (*buf)(i, j);
//       (*rTemp)(i, j) = pixel.R();
//       (*gTemp)(i, j) = pixel.G();
//       (*bTemp)(i, j) = pixel.B();
//     }

//   // smooth each colorband separately
//   vil1_memory_image_of<float>* rBuf = 
//     new vil1_memory_image_of<float>(vil1_smooth_gaussian((*rTemp), sigma));
//   vil1_memory_image_of<float>* gBuf = 
//     new vil1_memory_image_of<float>(vil1_smooth_gaussian((*gTemp), sigma));
//   vil1_memory_image_of<float>* bBuf = 
//     new vil1_memory_image_of<float>(vil1_smooth_gaussian((*bTemp), sigma));

//   // create new color image based on smoothed colorbands
//   for (int i=0; i<sizex; i++)
//     for (int j=0; j<sizey; j++)
//     {
//       vil1_rgb<unsigned char> pixel;
//       pixel.r = (vxl_byte)rint((*rBuf)(i, j)) ;
//       pixel.g = (vxl_byte)rint((*gBuf)(i, j)) ;
//       pixel.b = (vxl_byte)rint((*bBuf)(i, j)) ;
	
//       (*tmp_buf)(i, j)=pixel;
//     }

//   return tmp_buf;
// }



//-----------------------------------------------------------------------------
// Hardwired hyper-optimized version for 5x5 mask
//-----------------------------------------------------------------------------
void dilation5(const bw_buffer_t&in, bw_buffer_t& out) 
{
  // Fast access data
  int w = in.width();
  int h = in.height();

  // Initialization for image borders
  for (int i=0; i < 4; i++)
  {
    out(i,0) = in(i,0);
    out(i,1) = in(i,1);
    out(i,h-1) = in(i,h-1);
    out(i,h-2) = in(i,h-2);
  }

  // First part: horiz 'convolution'
  for(int y = 0; y < h; y++) 
  {
    vxl_byte dBuff[5];
    dBuff[1] = in(0, y);
    dBuff[2] = in(1, y);
    dBuff[3] = in(2, y);
    dBuff[4] = in(3, y);
		
    for(int x = 0; x < w-4; x++) // dilate (x+2, y) pixel
    {
      dBuff[x%5] = in(x+4, y);
      
      vxl_byte maxPix = dBuff[0];
      for( int p = 1; p < 5; p++ ) {
        if( dBuff[p] > maxPix )
          maxPix = dBuff[p];
      }
      out(x+2, y) = maxPix;
    }
  }
  
  // Second part: vert convolution
  for(int x = 0; x < w; x++) 
  {
    vxl_byte dBuff[5];
    dBuff[1] = out(x, 0);
    dBuff[2] = out(x, 1);
    dBuff[3] = out(x, 2);
    dBuff[4] = out(x, 3);
		
    for(int y = 0; y < h-4; y++)   // dilate (x, y+2) pixel
    {
      dBuff[y%5] = out(x, y+4);
      
      vxl_byte maxPix = dBuff[0];
      for( int p = 1; p < 5; p++ ) {
        if( dBuff[p] > maxPix )
          maxPix = dBuff[p];
      }
      out(x, y+2) = maxPix;
    }
  }
} // dilation


// Vittorio's Gaussian smoothing applied to color images,
// by first splitting the three colorbands and later merging them again
void dilation5(const color_buffer_t& in, color_buffer_t& out)
{
  int sizex = in.width();
  int sizey = in.height();

  // create separate buffers for each colorband
  bw_buffer_t rTemp(sizex, sizey);
  bw_buffer_t gTemp(sizex, sizey);
  bw_buffer_t bTemp(sizex, sizey);
  
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    { 
      vil1_rgb<unsigned char> pixel = in(i, j);
      rTemp(i, j) = pixel.R();
      gTemp(i, j) = pixel.G();
      bTemp(i, j) = pixel.B();
    }

  // dilate each colorband separately
  bw_buffer_t rBuf(sizex, sizey);
  bw_buffer_t gBuf(sizex, sizey);
  bw_buffer_t bBuf(sizex, sizey);

  dilation5(rTemp, rBuf);
  dilation5(gTemp, gBuf);
  dilation5(bTemp, bBuf);

  // create new color image based on dilated colorbands
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    {
      vil1_rgb<unsigned char> pixel;
      pixel.r = (vxl_byte)rint(rBuf(i, j)) ;
      pixel.g = (vxl_byte)rint(gBuf(i, j)) ;
      pixel.b = (vxl_byte)rint(bBuf(i, j)) ;
	
      out(i, j)=pixel;
    }
}

