
#include <cmath>
#include "draw.h"

// Draw contours of areas colored in col_i in buf_i on top of buf_o (overlay)
// with black outline (improve visibility when printed)
void draw_outlined_contours(color_buffer_t* buf_i, vil1_rgb<unsigned char>& col_i, color_buffer_t* buf_o, vil1_rgb<unsigned char>& col_o, int w) {
  // draw solid, one-color contour of width w on an intermediate buffer buf_m
  color_buffer_t* buf_m = new color_buffer_t(buf_i->width(), buf_i->height());  
  draw_contours(buf_i, col_i, buf_m, col_o, w);

  // draw solid contour in destination buffer buf_o
  draw_contours(buf_i, col_i, buf_o, col_o, w);

  // draw outline: contour of buf_m's contours, with width 1 and color black
  vil1_rgb<unsigned char> black(0,0,0);                     // try to have more contrast in b/w papers
  draw_contours(buf_m, col_o, buf_o, black, 1);
  delete buf_m;
}


// Draw contours of areas colored in col_i in buf_i on top of buf_o (overlay)
// w = contour width
void draw_contours(color_buffer_t* buf_i, vil1_rgb<unsigned char>& col_i, color_buffer_t* buf_o, vil1_rgb<unsigned char>& col_o, int w) {
  vil1_rgb<unsigned char> black(0,0,0);                     // try to have more contrast in b/w papers
  bool solid = false;                                      // latest pixel's on or off ?

  // Line-scan pass
  for(int y=0; y<buf_i->height(); y++) {
    solid = false;
    for(int x=0; x<buf_i->width(); x++) {
      if( (*buf_i)(x,y) == col_i && !solid) {
	// start of a new solid run
	(*buf_o)(x,y) = col_o;                                      // central-contour
	if(x < buf_i->width()-1 && w>1)  (*buf_o)(x+1,y) = col_o;   // clip double-contour into image
        if(x > 0 && w>2)                 (*buf_o)(x-1,y) = col_o;   // clip triple-contour into image
        if(x < buf_i->width()-2 && w>3)  (*buf_o)(x+2,y) = col_o;   // clip 4th contour into image
        if(x > 1 && w>4)                 (*buf_o)(x-2,y) = col_o;   // clip 5th contour into image
	if(x < buf_i->width()-3 && w>5)  (*buf_o)(x+3,y) = col_o;   // clip 6th contour into image
	solid = true;
      }
       
      if( ! ((*buf_i)(x,y) == col_i) && solid) {
	// end a solid run
	(*buf_o)(x-1,y) = col_o;                                    // central contour
	if(x > 1 && w>1)                 (*buf_o)(x-2,y) = col_o;   // clip double-contour into image
        if(w>2)                          (*buf_o)(x,y) = col_o;     // no need to clip triple-contour (on current pixel, obviously in image)
        if(x > 2 && w>3)                 (*buf_o)(x-3,y) = col_o;   // clip 4th contour into image
        if(x < buf_i->width()-1 && w>4)  (*buf_o)(x+1,y) = col_o;   // clip 5th contour into image
	if(x > 3 && w>5)                 (*buf_o)(x-4,y) = col_o;   // clip 6th contour into image
	solid = false;
      }
    } // loop over x
  } // loop over y


  // Column-scan pass
  for(int x=0; x<buf_i->width(); x++) {
    solid = false;
    for(int y=0; y<buf_i->height(); y++) {
      if( (*buf_i)(x,y) == col_i && !solid) {
	// start of a new solid run
	(*buf_o)(x,y) = col_o;                                      // central-contour
	if(y < buf_i->height()-1 && w>1) (*buf_o)(x,y+1) = col_o;   // clip double-contour into image
        if(y > 0 && w>2)                 (*buf_o)(x,y-1) = col_o;   // clip triple-contour into image
        if(y < buf_i->height()-2 && w>3) (*buf_o)(x,y+2) = col_o;   // clip 4th contour into image
        if(y > 1 && w>4)                 (*buf_o)(x,y-2) = col_o;   // clip 5th contour into image
	if(y < buf_i->height()-3 && w>5) (*buf_o)(x,y+3) = col_o;   // clip 6th contour into image
	solid = true;
      }
       
      if( ! ((*buf_i)(x,y) == col_i) && solid) {
	// end a solid run
	(*buf_o)(x,y-1) = col_o;                                    // central-contour
	if(y > 1 && w>1)                 (*buf_o)(x,y-2) = col_o;   // clip double-contour into image
        if(w>2)                          (*buf_o)(x,y) = col_o;     // no need to clip triple-contour (on current pixel, obviously in image)
	if(y > 2 && w>3)                 (*buf_o)(x,y-3) = col_o;   // clip 4th contour into image
        if(y < buf_i->height()-1 && w>4) (*buf_o)(x,y+1) = col_o;   // clip 5th contour into image
	if(y > 3 && w>5)                 (*buf_o)(x,y-4) = col_o;   // clip 6th contour into image
	solid = false;
      }
    } // loop over y
  } // loop over x

} // draw_contours


////////////////////////////
// Color conversion
////////////////////////////
// From www.cs.rit.edu/~ncs/color/t_convert.html
void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) {
  int i;
  float f, p, q, t;
  
  if( s == 0 ) {
    // achromatic (grey)
    *r = *g = *b = v;
    return;
  }
  
  h /= 60;                        // sector 0 to 5
  i = (int)floor( h );
  f = h - i;                      // factorial part of h
  p = v * ( 1 - s );
  q = v * ( 1 - s * f );
  t = v * ( 1 - s * ( 1 - f ) );
  
  switch( i ) {
  case 0:
    *r = v;
    *g = t;
    *b = p;
    break;
  case 1:
    *r = q;
    *g = v;
    *b = p;
    break;
  case 2:
    *r = p;
    *g = v;
    *b = t;
    break;
  case 3:
    *r = p;
    *g = q;
    *b = v;
    break;
  case 4:
    *r = t;
    *g = p;
    *b = v;
    break;
  default:                // case 5:
    *r = v;
    *g = p;
    *b = q;
    break;
  } 
} // HSVtoRGB


// From www.cs.rit.edu/~ncs/color/t_convert.html
// r,g,b values are from 0 to 1
// h = [0,360], s = [0,1], v = [0,255]
// if s == 0, then h = -1 (undefined)
void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) {
  float min_rgb, max_rgb, delta;
  min_rgb = MINTWO( MINTWO( r, g ), b );
  max_rgb = MAXTWO( MAXTWO( r, g ), b );

  *v = max_rgb;       // v
  delta = max_rgb - min_rgb;
  if( max_rgb != 0 )
    *s = delta / max_rgb; // s
  else {
    // r = g = b = 0// s = 0, v is undefined
    *s = 0;
    *h = 0; // BEW: trick, but only if case of pure black ...
    return;
  }
  if( r == max_rgb )
    *h = ( g - b ) / delta;// between yellow & magenta
  else if( g == max_rgb )
    *h = 2 + ( b - r ) / delta;// between cyan & yellow
  else
    *h = 4 + ( r - g ) / delta;// between magenta & cyan
  *h *= 60;// degrees
  if( *h < 0 )
    *h += 360;
} // RGBtoHSV
