#include "gaussian_smoothing.h"

#include <vcl_cmath.h>
#include <vcl_vector.h>

#include <vcl/vcl_list.h>
#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_load.h>
#include <vil1/vil1_save.h>

#include <vil1/vil1_smooth.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_resample.h>

#include <misc_utility/defines.h>

color_buffer_t*
rgb_gaussian_smoothing(const color_buffer_t*buf, const double sigma)
{ 
  int sizex = buf->width();
  int sizey = buf->height();

  // Create 1-D mask:
  double cutoff = 0.01;
  double lc = -2 * vcl_log(cutoff); // cutoff guaranteed > 0
  int radius = (lc<=0) ? 0 : 1 + int(vcl_sqrt(lc)*sigma); // sigma guaranteed >= 0
  int size = 2*radius + 1;
  vcl_vector<float> mask(size);
  double halfnorm = 0.5;
  mask[radius] = 1.0;
  for (int x=1; x<=radius; ++x) {
    double v = vcl_exp(-0.5*x*x/(sigma*sigma));
    mask[radius - x] = mask[radius + x] = v;
    halfnorm += v;
  }

  // normalise mask
  double mass_scale = 1.0/(1 + 2*halfnorm);
  for (int x=0; x< size; ++x)
    mask[x] *= mass_scale;

  color_buffer_t *tmp_buf = new color_buffer_t(sizex, sizey);    

  // create separate buffers for each colorband
  bw_buffer_t* rTemp = 
    new bw_buffer_t(sizex, sizey);
  bw_buffer_t* gTemp = 
    new bw_buffer_t(sizex, sizey);
  bw_buffer_t* bTemp = 
    new bw_buffer_t(sizex, sizey);
  
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    { 
      vil1_rgb<unsigned char> pixel = (*buf)(i, j);
      (*rTemp)(i, j) = pixel.R();
      (*gTemp)(i, j) = pixel.G();
      (*bTemp)(i, j) = pixel.B();
    }

  // smooth each colorband separately
  vil1_memory_image_of<float>* rBuf = 
    new vil1_memory_image_of<float>(vil1_smooth_gaussian((*rTemp), sigma));
  vil1_memory_image_of<float>* gBuf = 
    new vil1_memory_image_of<float>(vil1_smooth_gaussian((*gTemp), sigma));
  vil1_memory_image_of<float>* bBuf = 
    new vil1_memory_image_of<float>(vil1_smooth_gaussian((*bTemp), sigma));

  // create new color image based on smoothed colorbands
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    {
      vil1_rgb<unsigned char> pixel;
      pixel.r = (vxl_byte)rint((*rBuf)(i, j)) ;
      pixel.g = (vxl_byte)rint((*gBuf)(i, j)) ;
      pixel.b = (vxl_byte)rint((*bBuf)(i, j)) ;
	
      (*tmp_buf)(i, j)=pixel;
    }

  return tmp_buf;
}



//-----------------------------------------------------------------------------
// Hardwired version (much faster)
// using fixed 7x1 mask given by the Pascal triangle algo: (1 6 15 20 15 6 1)
//-----------------------------------------------------------------------------
void vic_gaussian_smoothing(const bw_buffer_t&in, bw_buffer_t& out) 
{
  // Fast access data
  int w = in.width();
  int h = in.height();

  // Initialization for image borders ->???
  /*for (int i=0; i < 6; i++)
  {
    out(i,0) = in(i,0);
    out(i,1) = in(i,1);
    out(i,2) = in(i,2);
    out(i,h-1) = in(i,h-1);
    out(i,h-2) = in(i,h-2);
    out(i,h-3) = in(i,h-3);
		}*/

  // First part: horiz convolution
  for(int y = 0; y < h; y++) 
  {
    vxl_byte next6 = in(5, y);
    vxl_byte next5 = in(4, y);
    vxl_byte next4 = in(3, y);
    vxl_byte next3 = in(2, y);
    vxl_byte next2 = in(1, y);
    vxl_byte next1 = in(0, y);
    vxl_byte next0 = 0; // cur

		// left border
		out(0,y) = (vxl_byte)round(((double)( 20*next1 + 15*next2 + 6*next3 + next4 ))/42.0);
		out(1,y) = (vxl_byte)round(((double)( 15*next1 + 20*next2 + 15*next3 + 6*next4 +
		                                      next5 ))/57.0);
		out(2,y) = (vxl_byte)round(((double)( 6*next1 + 15*next2 + 20*next3 + 15*next4 +
		                                      6*next5 + next6 ))/63.0);
		
    for(int x = 0; x < w-6; x++) // smooth (x+3, y) pixel
    {  
      next0 = next1;
      next1 = next2;
      next2 = next3;
      next3 = next4;
      next4 = next5;
      next5 = next6;
      next6 = in(x + 6, y);
			
      out(x+3, y) = (vxl_byte)round(((double)( next0 + 6*next1 + 15*next2 + 20*next3 +
					       15*next4 + 6*next5 + next6  ))/64.0);
    }
		
		// right border
		out(w-3,y) = (vxl_byte)round(((double)( 6*next6 + 15*next5 + 20*next4 + 15*next3 +
		                                        6*next2 + next1 ))/63.0);
		out(w-2,y) = (vxl_byte)round(((double)( 15*next6 + 20*next5 + 15*next4 + 6*next3 +
		                                        next2 ))/57.0);
		out(w-1,y) = (vxl_byte)round(((double)( 20*next6 + 15*next5 + 6*next4 + next3 ))/42.0);
  }
	
  // Second part: vert convolution
  for(int x = 0; x < w; x++) 
  {
    vxl_byte next6 = out(x, 5);
    vxl_byte next5 = out(x, 4);
    vxl_byte next4 = out(x, 3);
    vxl_byte next3 = out(x, 2);
    vxl_byte next2 = out(x, 1);
    vxl_byte next1 = out(x, 0);
    vxl_byte next0 = 0; // cur

		// top border
		out(x,0) = (vxl_byte)round(((double)( 20*next1 + 15*next2 + 6*next3 + next4 ))/42.0);
		out(x,1) = (vxl_byte)round(((double)( 15*next1 + 20*next2 + 15*next3 + 6*next4 +
		                                      next5 ))/57.0);
		out(x,2) = (vxl_byte)round(((double)( 6*next1 + 15*next2 + 20*next3 + 15*next4 +
		                                      6*next5 + next6 ))/63.0);
		
    for(int y = 0; y < h-6; y++)   // smooth (x, y+3) pixel
    {
      next0 = next1;
      next1 = next2;
      next2 = next3;
      next3 = next4;
      next4 = next5;
      next5 = next6;
      next6 = out(x, y+6);
			
      out(x, y+3) = (vxl_byte)round(((double)( next0 + 6*next1 + 15*next2 + 20*next3 +
					       15*next4 + 6*next5 + next6  ))/64.0);
    }

		// bottom border
		out(x,h-3) = (vxl_byte)round(((double)( 6*next6 + 15*next5 + 20*next4 + 15*next3 +
		                                        6*next2 + next1 ))/63.0);
		out(x,h-2) = (vxl_byte)round(((double)( 15*next6 + 20*next5 + 15*next4 + 6*next3 +
		                                        next2 ))/57.0);
		out(x,h-1) = (vxl_byte)round(((double)( 20*next6 + 15*next5 + 6*next4 + next3 ))/42.0);
  }
} // vic_gaussian_smoothing


// Vittorio's Gaussian smoothing applied to color images,
// by first splitting the three colorbands and later merging them again
void vic_gaussian_smoothing(const color_buffer_t& in, color_buffer_t& out)
{
  int sizex = in.width();
  int sizey = in.height();

  // create separate buffers for each colorband
  bw_buffer_t rTemp(sizex, sizey);
  bw_buffer_t gTemp(sizex, sizey);
  bw_buffer_t bTemp(sizex, sizey);
  
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    { 
      vil1_rgb<unsigned char> pixel = in(i, j);
      rTemp(i, j) = pixel.R();
      gTemp(i, j) = pixel.G();
      bTemp(i, j) = pixel.B();
    }

  // smooth each colorband separately
  bw_buffer_t rBuf(sizex, sizey);
  bw_buffer_t gBuf(sizex, sizey);
  bw_buffer_t bBuf(sizex, sizey);

  vic_gaussian_smoothing(rTemp, rBuf);
  vic_gaussian_smoothing(gTemp, gBuf);
  vic_gaussian_smoothing(bTemp, bBuf);

  // create new color image based on smoothed colorbands
  for (int i=0; i<sizex; i++)
    for (int j=0; j<sizey; j++)
    {
      vil1_rgb<unsigned char> pixel;
      pixel.r = (vxl_byte)rint(rBuf(i, j)) ;
      pixel.g = (vxl_byte)rint(gBuf(i, j)) ;
      pixel.b = (vxl_byte)rint(bBuf(i, j)) ;
	
      out(i, j)=pixel;
    }
}

