

#include <misc_utility/region_growing.h>
#include <vcl/vcl_stack.h>
#include <region_utility/misc.h>                  // in_image
#include <misc_utility/byte_interpolation.h>      // get_col_interpolation_Vic, dist_rgb


bool* grow_uniform_region(color_buffer_t* buf, bool* xflag, int& area, int seed_x, int seed_y) {
  // Grow an approximately uniformly colored region starting from pixel (seed_x, seed_y),
  // using a simple recursive flood-fill algorithm.
  // The fill criteria is relaxed: the color of the pixel under examination
  // must lie in a sphere around the seed pixel's color.
  // Mantains a flag-buffer to know which pixels are already discovered.
  // This buffer represents the region to be returned.

  // Precondition: seed in buf
  if(!(in_image(seed_x, seed_y, buf->width(), buf->height()))) {
    vcl_cout << "ERROR: seed = (" << seed_x << ", " << seed_y << ") out of image" << vcl_endl;
    area = 0;
    return NULL;
  }

  // Precondition: seed not in xflag
  if(xflag[seed_x + seed_y*buf->width()]) {
    area = 0;
    return NULL;
  }

  // Initialize flag buffer
  vcl_cout << "Initializing flag buffer" << vcl_endl;
  int w = buf->width();
  int h = buf->height();
  bool* flag = new bool[w*h];
  for(int y=0; y<h; y++) {
    for(int x=0; x<w; x++) {
      flag[x + y*w] = false;
    }
  }
  
  // Run flood-fill
  vcl_cout << "Starting flood-fill" << vcl_endl;
  vil1_rgb<unsigned char> col = (*buf)(seed_x, seed_y);                // imp: need to copy it, not only pointer-to
  flood_fill_iter(seed_x, seed_y, &col, buf, flag, xflag, area);

  return flag;
}


// Standard recursive version
// OBSOLETE: not includes recent xflag trick
void flood_fill(int x, int y, vil1_rgb<unsigned char>* col, color_buffer_t* buf, bool* flag) {
  //vcl_cout << "(" << x << ", " << y << ")" << vcl_endl;

  // Parameters
  static float thresh = 50.0;      // maximal distance of buf(x,y) from col
  static vil1_rgb<unsigned char> green(0,255,0);  // by keeping an object, could spare this redoing -> speed !
  // making it static already achieves avoid-redoing-work ?

  // Init
  int w = buf->width();
  int h = buf->height();

  // pixel (x,y) is alreadsy sure to be part of the region
  (*buf)(x,y) = green;
  flag[x + y*w] = true;
  
  // check 4-neighbors and recurse
  if(x < w-1) if(!flag[x+1 + y*w])   if(dist_rgb((*buf)(x+1,y), *col) < thresh) flood_fill(x+1, y, col, buf, flag);     // right pixel
  if(x > 0) if(!flag[x-1 + y*w])     if(dist_rgb((*buf)(x-1,y), *col) < thresh) flood_fill(x-1, y, col, buf, flag);     // left pixel
  if(y < h-1) if(!flag[x + (y+1)*w]) if(dist_rgb((*buf)(x,y+1), *col) < thresh) flood_fill(x, y+1, col, buf, flag);     // below pixel
  if(y > 0) if(!flag[x + (y-1)*w])   if(dist_rgb((*buf)(x,y-1), *col) < thresh) flood_fill(x, y-1, col, buf, flag);     // above pixel
  // change *col to current buf color in the above to have it fill through slowly changing areas (also color shades then, cool!)
}


// Iterative version
// maybe it has more success in large image due not to use implicit recursion stack in c++
void flood_fill_iter(int x, int y, vil1_rgb<unsigned char>* col, color_buffer_t* buf, bool* flag, bool* xflag, int& area) {
  // Parameters
  float thresh = 50.0;                            // maximal distance of buf(x,y) from col
  vil1_rgb<unsigned char> green(0,255,0);

  // Init
  int w = buf->width();
  int h = buf->height();
  vcl_stack<int> x_stack;
  vcl_stack<int> y_stack;
  x_stack.push(x);
  y_stack.push(y);
  
  while(!x_stack.empty()) {
    x = x_stack.top();  x_stack.pop();
    y = y_stack.top();  y_stack.pop();

    // pixel (x,y) is already sure to be part of the region
    flag[x + y*w] = true;
    area++;
  
    // check 4-neighbors
    if(x < w-1) if(!flag[x+1 + y*w]   && !xflag[x+1 + y*w])   if(dist_rgb((*buf)(x+1,y), *col) < thresh) {x_stack.push(x+1); y_stack.push(y);}     // right pixel
    if(x > 0)   if(!flag[x-1 + y*w]   && !xflag[x-1 + y*w])   if(dist_rgb((*buf)(x-1,y), *col) < thresh) {x_stack.push(x-1); y_stack.push(y);}     // left pixel
    if(y < h-1) if(!flag[x + (y+1)*w] && !xflag[x + (y+1)*w]) if(dist_rgb((*buf)(x,y+1), *col) < thresh) {x_stack.push(x);   y_stack.push(y+1);}   // below pixel
    if(y > 0)   if(!flag[x + (y-1)*w] && !xflag[x + (y-1)*w]) if(dist_rgb((*buf)(x,y-1), *col) < thresh) {x_stack.push(x);   y_stack.push(y-1);}   // above pixel
    // change *col to current buf color in the above to have it fill through slowly changing areas (also color shades then, cool!)
  }
}


