// ----------------------------------------------------------------------
// 
// curved_region_extractor.cpp: implementation of the
//                            curved_region_extractor class.
//
// ----------------------------------------------------------------------

#include "curved_region_extractor.h"

#include <vdgl/vdgl_interpolator_cubic.h>
#include <vdgl/vdgl_interpolator_linear.h>
#include <vdgl/vdgl_edgel.h>
#include <vdgl/vdgl_edgel_chain.h>
#include <vcl_algorithm.h>

#include <vnl/vnl_math.h>
#include <vnl/vnl_matrix.h>


#include <stdio.h>


#if defined (DEBUG_GL)
#include <GL/glut.h>
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

curved_region_extractor::curved_region_extractor(reg_ext_starter *starter)
  : parall_region_extractor(starter)
{
  _Splines = NULL;

  // If we have a color image or not
  _nr_of_invariants = (_starter->get_image_buffer()->components() == 3 ? 6 : 2);

  _normalized_arclength1_step = 0.01;
  _normalized_arclength2_step = 0.01;

  _max_canonic_arclength = 2000;
  _canonic_arclength_step = 5;
  
}




curved_region_extractor::~curved_region_extractor()
{

}



// ----------------------------------------------------------------------
// Find curved affinely invariant regions near anchor point pt
// ----------------------------------------------------------------------
bool curved_region_extractor::extract_regions(point_t* pt,
                                              region::region_specific_type tp)
{

  // Select edges in anchor point neighbourhood
  edge_vector_t sub_edges;
  spline_vector_t sub_splines;


  curve_buffer_t* edge_buf = _starter->get_edge_buffer();
	
  int i, j;
  double MaxDist = _starter->get_params()->MaxDistanceFromPointToEdge;
  int min_edge_length = _starter->get_params()->min_length;
  

  
  // Search radius around corner point
  // Boundary conditions: Search circle crosses image boundaries (x direction)
  int iLow = MAXTWO(0, (int)ABLround(pt->x() - MaxDist));
  int iHigh = MINTWO(edge_buf->width(), (int)ABLround(pt->x() + MaxDist));

  
  for (i=iLow; i < iHigh; i++)
  {
    // Search radius around corner point
    // Boundary conditions: Search circle crosses image boundaries (y direction)
    int jLow = MAXTWO(0, (int)ABLround(pt->y() - MaxDist));
    int jHigh = MINTWO(edge_buf->height(), (int)ABLround(pt->y() + MaxDist));


    // Store all edges within search circle in
    // sub_list. Only those edges are stored that have a minimal
    // length.
    for (j= jLow; j < jHigh; j++)
    {
      if ( (*edge_buf)(i, j) )
      {
        //////////////////////////////////////////////////
        // If the line already exists in sub_edges,
        // then it is not stored again
        //////////////////////////////////////////////////
        vcl_vector<vdgl_digital_curve*>::iterator fe;
        fe = find(sub_edges.begin(), sub_edges.end(), (*edge_buf)(i, j));
        
        if ( fe == sub_edges.end() && (*edge_buf)(i, j)->length() > min_edge_length ){
          sub_edges.push_back( (*edge_buf)(i, j) );
        }
      }
    }
  }


#if defined VERB_CURVED
  vcl_cout << pt->x() << "  " << pt->y() << "  " << pt->w() << ":  ";
  vcl_cout << sub_edges.size() << " edge(s) found" << vcl_endl;
#endif


  // No reason to proceed here if there are no edges close enough
  // to the anchor point
  if ( sub_edges.empty() )
    return false;
  



#if defined DEBUG_GL
  // Draw search rectangle around anchor point
  // to check where curved edges are searched
  glColor3f(1,0,0);
  glBegin(GL_POINTS);

  glVertex2f(pt->x(), _starter->get_image_buffer()->height() - pt->y());
  glEnd();

  int jLow = MAXTWO(0, (int)ABLround(pt->y() - MaxDist));
  int jHigh = MINTWO(edge_buf->height(), (int)ABLround(pt->y() + MaxDist));

  glColor3f(0, 1, 0);
  glBegin(GL_LINE_LOOP);
  
  glVertex2i(iLow, _starter->get_image_buffer()->height() - jLow);
  glVertex2i(iHigh, _starter->get_image_buffer()->height() - jLow);
  glVertex2i(iHigh, _starter->get_image_buffer()->height() - jHigh);
  glVertex2i(iLow, _starter->get_image_buffer()->height() - jHigh);
  glEnd();
#endif
  
  
  for (edge_vector_t::iterator it = sub_edges.begin();
       it != sub_edges.end(); ++it){

    
#if VERB_CURVED
    vcl_cout << "  Digital curve length: " << (*it)->length() << vcl_endl;
#endif
    
    // Make a new digital curve out of an interpolator
    vdgl_edgel_chain_sptr chain = (*it)->get_interpolator()->get_edgel_chain();

    int chain_length = chain->size();
    

    // Number of control points for nurbs fitting. If this number
    // equals the edge length, this might result in too long
    // computation times. Therefore we restrict the number of
    // control points to 'nurbs_ctrl_points'.
    float sampling_interval = 1.0;
    int control_points = chain_length;
    
    // First visual inspections indicate that 30 control points
    // follow the contours very precisely.
    int nurbs_ctrl_points = 30;
    

    // For rather short edges ( < nurbs_ctrl_points ) we sample
    // along the entire length with an interval of 1, for longer edges
    // we stretch that interval accordingly.
    if ( chain_length > nurbs_ctrl_points ){
      control_points = nurbs_ctrl_points;
      
      sampling_interval = (float)chain_length/(float)nurbs_ctrl_points;
    }
    
    
    
    // --------------------------------------------------
    // -- NURBS fitting
    // --------------------------------------------------
    using namespace PLib;
    Vector_Point3Df points(control_points);
    
    // Sample edge ..
    for (int s=0; s<control_points; ++s){
      int index = vnl_math_rnd(s*sampling_interval);
      
      vdgl_edgel e = (*chain)[index];

      // Note that the (x,y) vdgl_edgel coordinates are
      // actually exchanged !
      points[s] = Point3Df(e.get_y(), e.get_x(), 1);
    }

    // .. and do the NURBS fitting. We try a global
    // interpolation here, because the curve passes through
    // the control points. Alternatives are:
    // leastSquares
    PLib::NurbsCurvef curveA;
    curveA.globalInterp(points, 3);

    sub_splines.push_back(curveA);
    
    

#if defined DEBUG_GL
    // Draw the interpolated curve
    glColor4f(0,0,1,.25);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBegin(GL_POINTS);
    
    for (double t=0; t<=1; t+=0.02){

      HPoint3Df p = curveA(t);
      
      double y = p.y();
      double x = p.x();
      
      //cout << t << "  " << y << "  " << x << endl;
      
      glVertex2f(x, _starter->get_image_buffer()->height() - y);
    }
    glEnd();
#endif
  }
  
  
  vcl_vector<parall_region*> cands = span_regions(sub_splines, pt);


  if ( cands.empty() ){
    
#if defined VERB_CURVED
    vcl_cout << "      -> No candidates found" << vcl_endl;
#endif

    return false;
  }



  // If we have found some candidates, these have to be
  // verified for the geometric shape constraints and their
  // location (if they do not extend beyond the image boundaries)
  geometry_check(cands);


  if ( cands.empty() ){
    
#if defined VERB_CURVED
    vcl_cout << "Shape and location tests failed for region candidates near " << *pt << endl;
#endif

    return false;
  }
  



  // Store the remaining regions
  while ( !cands.empty() ){
    parall_region* reg = cands.back();
    cands.pop_back();

    _regions.push_back(reg);
  }

  return true;
  
}



bool curved_region_extractor::extract_all_regions()
{
  bool Ret = true;
  vcl_list<point_t *> *Corners = _starter->get_harris_corners();
	
  for(vcl_list<point_t *>::iterator it = Corners->begin();
      it != Corners->end(); ++it)
  {
    if ( !extract_regions(*it) )
      Ret = false;
  }

  return Ret;
}




bool
curved_region_extractor::compute_invariants_1d_rgb(const point_t *ctr,
                                                   const point_t *p1, 
                                                   const point_t *p2,
                                                   vnl_vector<double> &result)
{
vil1_image *fooImage = _starter->get_image_buffer();//bana

  vil1_memory_image_of<vil1_rgb_byte> *Image = new vil1_memory_image_of<vil1_rgb_byte>(*fooImage); //bana


  // Compute the 'bounding box' of the construct
  int minx, maxx, miny, maxy;

  double tempx = p1->x() + p2->x() - ctr->x();
  double tempy = p1->y() + p2->y() - ctr->y();

  minx = (int)floor(tempx); miny = (int)floor(tempy);
  maxx = (int)ceil(tempx); maxy = (int)ceil(tempy);
  
  
  minx = vnl_math_min(minx, (int)floor(ctr->x()));
  minx = vnl_math_min(minx, (int)floor(p1->x()));
  minx = vnl_math_min(minx, (int)floor(p2->x()));

  miny = vnl_math_min(miny, (int)floor(ctr->y()));
  miny = vnl_math_min(miny, (int)floor(p1->y()));
  miny = vnl_math_min(miny, (int)floor(p2->y()));


  maxx = vnl_math_max(maxx, (int)ceil(ctr->x()));
  maxx = vnl_math_max(maxx, (int)ceil(p1->x()));
  maxx = vnl_math_max(maxx, (int)ceil(p2->x()));

  maxy = vnl_math_max(maxy, (int)ceil(ctr->y()));
  maxy = vnl_math_max(maxy, (int)ceil(p1->y()));
  maxy = vnl_math_max(maxy, (int)ceil(p2->y()));


  // This assures that there is at least a difference of two
  // between minX and maxX such that the resulting invariants
  // don't become NaN
  //maxx += 1; maxy += 1;

  //assert (minx < maxx || miny <  maxy );
  

	
  if ( (minx < 0) || (miny < 0) || (maxx > Image->width()-1) || (maxy > Image->height()-1) )
  {
    return false;
  }
	
	
  int sum = 0;
  double sumR = 0.0;
  double sumG = 0.0;
  double sumB = 0.0;
  double sumRR = 0.0;
  double sumGG = 0.0;
  double sumBB = 0.0;
	
  double zwRx = 0.0;
  double zwRy = 0.0;
  double zwGx = 0.0;
  double zwGy = 0.0;
  double zwBx = 0.0;
  double zwBy = 0.0;
	
  double det = ( p1->x() - ctr->x() )*( p2->y() - ctr->y() )  - ( p2->x() - ctr->x() )*( p1->y() - ctr->y() );
	
  for (int x = minx; x <= maxx; ++x)
  {
    for (int y = miny; y <= maxy; ++y)
    {
      double u = ( (p2->y() - ctr->y() )*( x - ctr->x() ) - ( p2->x() - ctr->x() )*( y - ctr->y() ) ) / det;
      double v = ( (p1->x() - ctr->x() )*( y - ctr->y() ) - ( p1->y() - ctr->y() )*( x - ctr->x() ) ) / det;

      if ( (fabs(u) <= 1) && ( fabs(v) <= 1) )
      {
        if ( (u > 0)  && (v > 0) )
        {
          vil1_rgb_byte pixel = (*Image)(x, y);
					
          sum++;
          sumR += (double)pixel.R();
          sumG += (double)pixel.G();
          sumB += (double)pixel.B();	
          sumRR += ABL_SQUARE((double)pixel.R());
          sumGG += ABL_SQUARE((double)pixel.G());
          sumBB += ABL_SQUARE((double)pixel.B());
					
          zwRx += x * (double)pixel.R();
          zwRy += y * (double)pixel.R();
          zwGx += x * (double)pixel.G();
          zwGy += y * (double)pixel.G();
          zwBx += x * (double)pixel.B();
          zwBy += y * (double)pixel.B();
        }
      }
    }
  }

	
  zwRx /= sumR;
  zwRy /= sumR;
  zwGx /= sumG;
  zwGy /= sumG;
  zwBx /= sumB;
  zwBy /= sumB;
	
  point_t q(p1->x() + p2->x() - ctr->x(), p1->y() + p2->y() - ctr->y(), 1.0);
	
  double area = (ctr->x() - p1->x()) * (ctr->y() - p2->y() )- (ctr->y() - p1->y()) * (ctr->x() - p2->x());
  double areaR1 = ( (zwRx - p1->x())  * (zwRy - p2->y()) - (zwRy - p1->y())  * (zwRx - p2->x()) );
  double areaR2 = ( (zwRx - ctr->x()) * (zwRy - q.y())   - (zwRy - ctr->y()) * (zwRx - q.x()) );
  double areaG1 = ( (zwGx - p1->x())  * (zwGy - p2->y()) - (zwGy - p1->y())  * (zwGx - p2->x()) );
  double areaG2 = ( (zwGx - ctr->x()) * (zwGy - q.y())   - (zwGy - ctr->y()) * (zwGx - q.x()) );
  double areaB1 = ( (zwBx - p1->x())  * (zwBy - p2->y()) - (zwBy - p1->y())  * (zwBx - p2->x()) );
  double areaB2 = ( (zwBx - ctr->x()) * (zwBy - q.y())   - (zwBy - ctr->y()) * (zwBx - q.x()) );
	
  double MaxSize = _starter->get_params()->MaxSideLength;

  if ( area > ABL_SQUARE(MaxSize)*ABL_PI/2.0)
  {

    
#if defined VERB_CURVED==3
    vcl_cout << "    region gets too large" << vcl_endl;
#endif
    
    return false;
  }
	
  if ( _offset_invariants )
  {
    double N = sqrt(sumRR*sum/sumR/sumR - 1);
    areaR1 /= N;
    areaR2 /= N;
    N = sqrt(sumGG*sum/sumG/sumG - 1);
    areaG1 /= N;
    areaG2 /= N;
    N = sqrt(sumBB*sum/sumB/sumB - 1);
    areaB1 /= N;
    areaB2 /= N;
  }
	
  result.resize(6);

  result(0) = 1000*fabs(areaR1/area); 
  result(1) = 1000*fabs(areaG1/area); 
  result(2) = 1000*fabs(areaB1/area);
  result(3) = 1000*fabs(areaR2/area);
  result(4) = 1000*fabs(areaG2/area);
  result(5) = 1000*fabs(areaB2/area);
	
  return true;
}



// ----------------------------------------------------------------------
// Does anybody use this ?
// ----------------------------------------------------------------------
bool curved_region_extractor::ComputeInvariants1D_ZW(point_t *ctr, point_t *p1, 
                                                   point_t *p2, vnl_vector<double> &result)
{
  vil1_memory_image_of<unsigned char> *Image = _starter->get_BW_image_buffer();

  int minx, maxx, miny, maxy;
  miny = (int)(MINFOUR(ctr->y(), p1->y(), p2->y(), p1->y() + p2->y() - ctr->y())); 
  maxy = (int)(MAXFOUR(ctr->y(), p1->y(), p2->y(), p1->y() + p2->y() - ctr->y()) + 1); 
	
  minx = (int)(MINFOUR(ctr->x(), p1->x(), p2->x(), p1->x() + p2->x() - ctr->x())); 
  maxx = (int)(MAXFOUR(ctr->x(), p1->x(), p2->x(), p1->x() + p2->x() - ctr->x()) + 1); 
	
  if ( (minx < 0) || (miny < 0) || (maxx > Image->width() -1) || (maxy > Image->height() -1) )
  {
    return false;
  }
	
	
  double det = ( p1->x() - ctr->x() )*( p2->y() - ctr->y() ) - ( p2->x() - ctr->x() )*( p1->y() - ctr->y() );
	
  double zwx = 0.0;
  double zwy = 0.0;
  double sumI = 0;
  double sumII = 0;
  int nb = 0;
	
  for (int x = minx; x <= maxx; x++)
  {
    for (int y = miny; y <= maxy; y++)
    {
      double u = ( ( p2->y() - ctr->y() )*( x - ctr->x() ) - ( p2->x() - ctr->x() )*( y - ctr->y() ) ) / det;
      double v = ( ( p1->x() - ctr->x() )*( y - ctr->y() ) - ( p1->y() - ctr->y() )*( x - ctr->x() ) ) / det;

      if ( (fabs(u) <= 1) && ( fabs(v) <= 1) )
      {
        if ( (u > 0)  && (v > 0) )
        {
          sumI += (*Image)(x, y);
          sumII += ABL_SQUARE( (*Image)(x, y) );
          nb++;
					
          zwx += x * (*Image)(x, y);
          zwy += y * (*Image)(x, y);
        }
      }
    }
  }
	
  if (sumI == 0) 
  {
    zwx = 1; 
    zwy = 1;
  }
  else
  {
    zwx /= sumI;
    zwy /= sumI;
  }
	
  result.resize(2);

  if (nb == 0)
  {
#if defined VERB_CURVED==3
    vcl_cout << "    warning: nb = 0 in compute_1d_invariants..." << vcl_endl; 
#endif
    
    result(0) = 100;
    result(1) = 100;
    return true;
  }
	
  //    double gem = (double)sumI / (double)nb;
	
  double area = ( ctr->x() - p1->x() ) * ( ctr->y() - p2->y() ) - ( ctr->y() - p1->y() ) * ( ctr->x() - p2->x() );
  double areaI1 = ( ( zwx - p1->x() )  * ( zwy - p2->y() ) - ( zwy - p1->y() ) * ( zwx - p2->x() ) );
  double areaI2 = ( ( zwx - ctr->x() ) * ( zwy - p1->y() - p2->y() + ctr->y() ) - ( zwy - ctr->y() ) * 
                    ( zwx - p1->x() - p2->x() + ctr->x() ) );
	
  double MaxSize = _starter->get_params()->MaxSideLength;

  if ( area > ABL_SQUARE(MaxSize)*ABL_PI/2.0){
    
#if defined VERB_CURVED==3
      vcl_cout << "    region gets too large" << vcl_endl;
#endif
      
    return false;
  }
	
  if ( _offset_invariants )
  {
    double N = sqrt(sumII * nb / sumI / sumI - 1);
    areaI1 /= N;
    areaI2 /= N;
  }
	
  result(0) = 1000*fabs(areaI1/area);
  result(1) = 1000*fabs(areaI2/area);

  return true;
}




// bool curved_region_extractor::ComputeInvariants1D(point_t *ctr, point_t *p1, 
//                                                 point_t *p2, vnl_vector<double> &result)
// {
//   if ( _ColorImage )
//   {
//     return ComputeInvariants1D_RGB(ctr, p1, p2, result);
//   }
//   else
//   {
//     return ComputeInvariants1D_ZW(ctr, p1, p2, result);
//   }
// }



vcl_vector<int>
curved_region_extractor::compute_extrema_1d(vnl_vector<double> vec, int max_index) const
{
  vcl_vector<int> list;
  vcl_vector<int> sublist;
  int minmax = 0;// first extremum is minimum (-1) or maximum (+1)  
	
  // first smooth curve
  int N = 5;
  for (int k=0; k<N; ++k)
  {
    vnl_vector<double> new_vec( (int)(_max_canonic_arclength/_canonic_arclength_step) );
    
    for (int i=1; i<max_index -1; i++)
      new_vec(i) = ( vec(i-1) + 2*vec(i) + vec(i+1) )/4.0;
    
    
    new_vec(0) = (2* vec(0) + vec(1) )/3.0;
    new_vec(max_index-1) = (2*vec(max_index-1) + vec(max_index-2) )/3.0;
    vec = new_vec;
  }
	
  
  // previously, some splinefitting was used here.
  // see regions.C for details about this ...
  for (int i=5; i<max_index-2; ++i)
  {
    if ( (vec(i+1) < vec(i)) && (vec(i+2) < vec(i+1) ) &&
         (vec(i-1) < vec(i)) && (vec(i-2) < vec(i-1)) ){
      
#if defined VERB_CURVED==3
      vcl_cout << "    maximum found: " << i << "   " << vec(i)<< vcl_endl;
#endif
      
      sublist.push_back(i);
      
      if (minmax == 0) 
        minmax = +1;
    }
    
    if ( (vec(i+1) > vec(i)) && (vec(i+2) > vec(i+1) ) &&
         (vec(i-1) > vec(i)) && (vec(i-2) > vec(i-1)) ){
      
#if defined VERB_CURVED==3
      vcl_cout << "    minimum found: " << i << "   " << vec(i)<< vcl_endl;
#endif
      
      sublist.push_back(i);
      
      if (minmax == 0) 
        minmax = -1;
    }
  }
  
  
  
  // filter list
  double thresh = 5;
  bool add_if_end = true;
  vcl_vector<int>::iterator it;
  
  
  if  ( !sublist.empty() ){
    it = sublist.begin();
    
    int val = *it;
    
    it = sublist.erase(it);
    
    while ( !sublist.empty() ){
      int next_val = *it;
      
      it = sublist.erase(it);
      
      if ( fabs(vec(val) - vec(next_val) ) > thresh )
      {
        if ( (minmax == -1) && (vec(val) <= 20)) // should be a minimum smaller than 20
          list.push_back(val);
        
        
        val = next_val;
        add_if_end = true;
        minmax *= -1;
      }
      else
      {
        // two points are deleted.
        if ( !sublist.empty() ){
          
          int next2val = *it;
          it = sublist.erase(it);
          
          if ( fabs(vec(val) - vec(next2val) ) < thresh ){

            if ( minmax * vec(val) < minmax * vec(next2val) )
              val = next2val; 

            
            add_if_end = false;
          }
          else {
            val = next2val; 
            add_if_end = true;
          }
        }
        else{
          add_if_end = false;
        }
      }
    }
    
    if (add_if_end)
      if ( (minmax == -1) && (vec(val) <= 20)) // should be a minimum smaller than 20
        list.push_back(val);
    
  }
  return list;
}





vcl_vector<int>
curved_region_extractor::compute_scale_1d(const vnl_matrix<double> &M,
                                        int max_index, int NumberInvariants, 
                                        vcl_vector<int> &type_identifiers) const
{
  vcl_vector<int> candidates;
  if (max_index < 2) {return candidates;}

  vcl_vector<int>::iterator it;
  vcl_vector<int> l;
	
  if ( _ColorImage )
  {
    l = compute_extrema_1d(M.get_column(0), max_index);
    for (it = l.begin(); it != l.end(); ++it)
    {

#if defined VERB_CURVED
        vcl_cout << "  Curved region of type PARALL_CURVED_R found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED_R);
    }
    l.clear();
		
		
    l = compute_extrema_1d(M.get_column(1), max_index);
    for (it = l.begin(); it != l.end(); ++it)
    {
#if defined VERB_CURVED 
      vcl_cout << "  Curved region of type PARALL_CURVED_G found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED_G);
    }
    l.clear();
		
		
    l = compute_extrema_1d(M.get_column(2), max_index);
    for (it = l.begin(); it != l.end(); ++it){
      
#if defined VERB_CURVED 
      vcl_cout << "  Curved region of type PARALL_CURVED_B found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED_B);
    }
    l.clear();
		
    l = compute_extrema_1d(M.get_column(3), max_index);
    for (it = l.begin(); it != l.end(); ++it){
      
#if defined VERB_CURVED
      vcl_cout << "  Curved region of type PARALL_CURVED found at scale " << *it << vcl_endl;
#endif
      
      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED);
    }
    l.clear();
    
    l = compute_extrema_1d(M.get_column(4), max_index);
    for (it = l.begin(); it != l.end(); ++it){

#if defined VERB_CURVED
      vcl_cout << "  Curved region of type PARALL_CURVED found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED);
    }
    l.clear();
		
    l = compute_extrema_1d(M.get_column(5), max_index);
    for (it = l.begin(); it != l.end(); ++it){

#if defined VERB_CURVED
      vcl_cout << "  Curved region of type PARALL_CURVED found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED);
    }
    l.clear();
		
  }
  else
  {
    l = compute_extrema_1d(M.get_column(0), max_index);
    for (it = l.begin(); it != l.end(); ++it){
      
#if defined VERB_CURVED 
      vcl_cout << "  Curved region of type PARALL_CURVED found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED);
    }
    l.clear();


    
    l = compute_extrema_1d(M.get_column(1), max_index);
    for (it = l.begin(); it != l.end(); ++it)
    {
#if defined VERB_CURVED 
        vcl_cout << "  Curved region of type PARALL_CURVED found at scale " << *it << vcl_endl;
#endif

      candidates.push_back(*it);
      type_identifiers.push_back(region::PARALL_CURVED);
    }
    l.clear();
  }
	


  return candidates;
}



// ----------------------------------------------------------------------
// Simple function that returns the curve parameter t which yields
// the point on the curve closest to p
// Started: 23.5.2002
// ----------------------------------------------------------------------
using namespace PLib;
double curved_region_extractor::min_dist_param(const NurbsCurvef& curve,
                                               const point_t* p) const
{
  
  int steps = 100;

  double stepsize = 1.0/(double)steps;

  double min_dist, min_s = 0;


  // Init
  min_dist = (curve(0).x() - p->x())*(curve(0).x() - p->x()) +
    (curve(0).y() - p->y())*(curve(0).y() - p->y());
  
  

  for (double s=0; s <= 1; s += stepsize){
    HPoint3Df cp = curve(s);

    double dist = (curve(s).x() - p->x())*(curve(s).x() - p->x()) +
    (curve(s).y() - p->y())*(curve(s).y() - p->y());

    if (dist < min_dist){
      min_dist = dist;
      min_s = s;
    }
  }

  return min_s;
}



// ----------------------------------------------------------------------
// Given the anchor point pt and neighbouring curves, extract all
// curved affinely invariant regions.
// 'Splines' might be misleading since this version uses NURBS curves,
// but the naming originates from Tinnes early, spline-based version.
// ----------------------------------------------------------------------
vcl_vector<parall_region*>
curved_region_extractor::span_regions(spline_vector_t& splines,
                                      const point_t* pt)
{
  vcl_vector<parall_region*> candidates;
  
  
  while ( !splines.empty() ){

    PLib::NurbsCurvef curveA = splines.back();
    splines.pop_back();


    // The curve parameter in [0,1] that yields the
    // curve point closest to pt on curveA
    double t1 = min_dist_param(curveA, pt);


    // Here we store the point-pairs related by the affinely invariant
    // parameter l (cfr. Tinnes paper). Each point at index i in pts1
    // is associated a point at index i in pts2.
    vcl_vector<point_t > pts1;
    vcl_vector<point_t > pts2;


    
    // Slide direction on curve
    // If the normalized curve parameter t1 is larger than
    // the corresponding half-arclength, we slide
    // backwards
    _normalized_arclength1_step *= (t1 > .5) ? -1 : 1;


    for (spline_vector_t::const_iterator it = splines.begin();
         it < splines.end(); ++it){

       
      PLib::NurbsCurvef curveB = (*it);
      double t2 = min_dist_param(curveB, pt);


      // Slide direction for second curve
      _normalized_arclength2_step *= t2 > .5 ? -1 : 1;


      point_t pt01(curveA(t1).x(), curveA(t1).y(), 1.0);
      point_t pt02(curveB(t2).x(), curveB(t2).y(), 1.0);


      // This variable can be seen as a curve in a canonic frame that is
      // relatively affinely equivalent to the nurbs curves that are evaluated.
      // This canonic curve with total length _max_canonic_arclength is
      // sampled in equidistant intervals:
      //
      // |==|==|==|==|--|--|--|--|--|--|--|
      // 0   canonic_arclength       _max_canonic_arclength
      //
      // For each sample point, we travel along the nurbs curve and compute
      // the relative affine parameter l till it reaches the same value as
      // the distance arc_length on the curve in the canonic frame.
      float canonic_arclength = 0;

      double affine_length1 = 0;
      double affine_length2 = 0;

      double px, py, dpx, dpy;

      
      point_t p1, p2;
      

      // We stop when either the end of a nurbs curve is reached or
      // the end of the canonic curve is reached
      bool stop = false;

      
      // Proceed in the nurbs curves parameter spaces
      t1 += _normalized_arclength1_step;
      t2 += _normalized_arclength2_step;

      

      while ( !stop ){

        canonic_arclength += _canonic_arclength_step;


        // Have we reached the end of the canonical curve ?
        stop = ( canonic_arclength > _max_canonic_arclength );


        // Have we reached the ends of the nurbs curves ?
        stop = ( (t1 > 1) || (t1 < 0) );
        stop = ( (t2 > 1) || (t2 < 0) );
        

        // Now go along curveA until the affine_length1 equals
        // the canonic arclength
        bool canonic_arclength_reached = false;
        
        while( !stop && !canonic_arclength_reached ){
          
          px = curveA(t1).x();
          py = curveA(t1).y();


#if defined DEBUG_GL
          // Draw rectangle while sliding along the curve
          glBegin(GL_POLYGON);
          glColor4f(1,0,0,.05);
          glVertex2f(px-3, _starter->get_image_buffer()->height() - py-3);
          glVertex2f(px+3, _starter->get_image_buffer()->height() - py-3);
          glVertex2f(px+3, _starter->get_image_buffer()->height() - py+3);
          glVertex2f(px-3, _starter->get_image_buffer()->height() - py+3);
          glEnd();
#endif
        
        
          // We also need the derivative at t1
          dpx = curveA.derive(t1, 1).x();
          dpy = curveA.derive(t1, 1).y();
        
        
          // Now we can compute the determinant
          // We could certainly use the vxl determinant function
          // for matrices, but I think that this form of inlining
          // is faster.
          affine_length1 += fabs( ((dpx - (pt01.y() - py)) - (dpy - (pt01.x() - px))) *
                                  _normalized_arclength1_step);


          canonic_arclength_reached = (affine_length1 > canonic_arclength);

          // Proceed in the nurbs curves parameter spaces
          t1 += _normalized_arclength1_step;

          stop = ( (t1 > 1) || (t1 < 0) ); // ? true : false;
        }


        // Point associated to the canonic arclength interval
        if ( ! stop )
          p1.set(px, py, 1);

        

        // Now go along curveB until the affine_length2 equals
        // the canonic arclength
        canonic_arclength_reached = false;
        
        while( !stop && !canonic_arclength_reached ){
          
  
          px = curveB(t2).x();
          py = curveB(t2).y();


#if defined DEBUG_GL
          // Draw rectangle while sliding along the curve
          glBegin(GL_POLYGON);
          glColor4f(0,1,0,.05);
          glVertex2f(px-3, _starter->get_image_buffer()->height() - py-3);
          glVertex2f(px+3, _starter->get_image_buffer()->height() - py-3);
          glVertex2f(px+3, _starter->get_image_buffer()->height() - py+3);
          glVertex2f(px-3, _starter->get_image_buffer()->height() - py+3);
          glEnd();
#endif
        
        
          // We also need the derivative at t1
          dpx = curveB.derive(t2, 1).x();
          dpy = curveB.derive(t2, 1).y();
        
        
          // Now we can compute the determinant
          // We could certainly use the vxl determinant function
          // for matrices, but I think that this form of inlining
          // is faster.
          affine_length2 += fabs( ((dpx - (pt02.y() - py)) - (dpy - (pt02.x() - px))) *
                                  _normalized_arclength2_step);


          canonic_arclength_reached = (affine_length2 > canonic_arclength);

          // Proceed in the nurbs curves parameter spaces
          t2 += _normalized_arclength2_step;

          stop = ( (t2 > 1) || (t2 < 0) ); 
        }
        
        // Point associated to the canonic arclength interval
        if ( ! stop ){
          p2.set(px, py, 1);
          pts2.push_back(p2);
          pts1.push_back(p1);
        }
      }

      
#if defined DEBUG_GL
      if ( !pts1.empty() ){
        
        glColor3f(1,1,0);
        glBegin(GL_LINE_STRIP);
        glVertex2f(pts2.back().x(), _starter->get_image_buffer()->height() - pts2.back().y());
        glVertex2f(pt->x(), _starter->get_image_buffer()->height() - pt->y());
        glVertex2f(pts1.back().x(), _starter->get_image_buffer()->height() - pts1.back().y());
        glEnd();
      }
      
#endif


      // Invariants
      vnl_matrix<double> invariants(pts1.size(), 6, 0);
      int max_index = 0;
      vcl_vector<int> type_identifiers;
      


      // --------------------------------------------------
      // Compute the function A(l) over
      // all affinely equivalent point pairs on the two
      // nurbs curves. Since A(l) might not be computed
      // for all point pairs (image boundaries etc.),
      // we denote with max_index the location where the
      // computation of A(l) stopped.
      // --------------------------------------------------
      for (unsigned int i=0; i<pts1.size(); ++i){

        vnl_vector<double> res;

        
        if ( compute_invariants_1d_rgb(pt, &pts1[i], &pts2[i], res) ){
          invariants.set_row(i, res);
          
          max_index = i;
        }
      }


      // --------------------------------------------------
      // The extrema of the function A(l) are
      // affinely invariant.
      // --------------------------------------------------
      vcl_vector<int> extremata = compute_scale_1d(invariants, max_index,
                                                   _nr_of_invariants,
                                                   type_identifiers);



      // --------------------------------------------------
      // If extremata were found, then these are
      // region candidates
      // --------------------------------------------------
      for (unsigned int i=0; i< extremata.size(); ++i){
        
        int index = extremata[i];
        int type = type_identifiers[i];

        point_t* p1 = new point_t(pts1[index]);
        point_t* p2 = new point_t(pts2[index]);


        if ((p1->x() - pt->x()) * (p2->y() - pt->y()) -
            (p1->y() - pt->y()) * (p2->x() - pt->x()) < 0.0){
          
          point_t* tmp = p1;
          p1 = p2;
          p2 = tmp;
        }


        parall_region* region = new parall_region(pt, p1, p2,
                                                (region::region_specific_type)type);
        

        candidates.push_back(region);

        delete p1; p1 = NULL;
        delete p2; p2 = NULL;
      }
    }
  }


  return candidates;
  
}



// ----------------------------------------------------------------------
// All regions in regvec are checked for geometric constraints and their
// location. Candidates not fulfilling the constraints are removed.
// Started: 28.5.2002
// ----------------------------------------------------------------------
void curved_region_extractor::geometry_check(vcl_vector<parall_region*>& regvec) const
{
  vcl_vector<parall_region*>::iterator it;

  int min_size = _starter->get_params()->min_size;
  int max_size = _starter->get_params()->max_size;
  
  
  for (it = regvec.begin(); it < regvec.end(); ++it){

    parall_region* region = *it;

    bool reject = false;

    // Shape check
    reject =  !region->good_region(2.0, ABL_SQUARE(min_size)*ABL_PI*2,
                                  ABL_SQUARE(max_size)*ABL_PI/2.0);
    

#if defined VERB_CURVED
    if ( reject )
      vcl_cout << "  Shape check failed !" << vcl_endl;
#endif

    // Location check
    if ( (region->get_min_x() < 0 ) || 
         (region->get_max_x() > _starter->get_image_buffer()->width()-1 ) || 
         (region->get_min_y() < 0 ) || 
         (region->get_max_y() > _starter->get_image_buffer()->height()-1 ) ){
      
      reject = true;

#if defined VERB_CURVED
      vcl_cout << "  Location check failed (extends beyond image boundaries)" << vcl_endl;
#endif
      
    }
    

    if ( reject )
      it = regvec.erase(it);
  }
  
}



// ----------------------------------------------------------------------
// Computation of a relative affine invariant parameter l
// 
//   l = integral[ abs( | p'(t)  pt0i - p(t) | ) dt ]
// 
// The integration stops when the end of the curve is reached
// ----------------------------------------------------------------------
