#include <region_extraction/ell_region_extractor.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_save.h>
#include <vil1/vil1_image_as.h>
#include <region_utility/aoi_computor.h>

#include <conversions/geometry/osl_to_vdgl.h>

//#define VERB_ELLIPTICAL


//-------------------------------------------------------------------------------
//--- Constructors/Destructor ---------------------------------------------------
//-------------------------------------------------------------------------------

ell_region_extractor::ell_region_extractor(reg_ext_starter* starter)
  :region_extractor(starter)
{
  _anchor = NULL;
  _ec = NULL;
  _dc = NULL;
  _mombuf = NULL;
	
}


ell_region_extractor::~ell_region_extractor()
{
	
}

//-------------------------------------------------------------------------------
//--- region Extraction ---------------------------------------------------------
//-------------------------------------------------------------------------------

bool ell_region_extractor::extract_regions(point_t* pt,
                                           region::region_specific_type tp)
{
  if (!pt) 
  {
    vcl_cerr << "Error: extract_regions called with pt == NULL" << vcl_endl; 
    return false;
  }
	
  _anchor = pt;
	
  if (!find_points_on_rays())
  {
    //vcl_cout << "Find Points on Rays Failed" << vcl_endl;
    if ( _ec )
    {
      delete _ec; 
      _ec = NULL;
    }
		
    if ( _dc )
    {
      delete _dc;
      _dc = NULL;
    }
		
    if ( _mombuf )
    {
      delete _mombuf;
      _mombuf = NULL;
    }
    return false;
  }
	
  _minx = (int)floor(_dc->get_interpolator()->get_min_x());
  _miny = (int)floor(_dc->get_interpolator()->get_min_y());
  _maxx = (int)ceil(_dc->get_interpolator()->get_max_x());
  _maxy = (int)ceil(_dc->get_interpolator()->get_max_y());
	
  _sx = _maxx - _minx + 1;
  _sy = _maxy - _miny + 1;
	
  if (!link_points())
  {
    //vcl_cout << "Link Points Failed" << vcl_endl;
    if ( _ec )
    {
      delete _ec; 
      _ec = NULL;
    }
		
    if ( _dc )
    {
      delete _dc;
      _dc = NULL;
    }
		
    if ( _mombuf )
    {
      delete _mombuf;
      _mombuf = NULL;
    }
    return false;
  }
	
  if (!compute_ellipse())
  {
    //vcl_cout << "Compute Ellipse Failed" << vcl_endl;
    if ( _ec )
    {
      delete _ec; 
      _ec = NULL;
    }
		
    if ( _dc )
    {
      delete _dc;
      _dc = NULL;
    }
		
    if ( _mombuf )
    {
      delete _mombuf;
      _mombuf = NULL;
    }
    return false;
  }
	
  if ( _ec )
  {
    delete _ec; 
    _ec = NULL;
  }
	
  if ( _dc )
  {
    delete _dc;
    _dc = NULL;
  }
	
  if ( _mombuf )
  {
    delete _mombuf;
    _mombuf = NULL;
  }
	
  return true;
}

bool ell_region_extractor::extract_all_regions()
{
  bool ok = true;
  points_list_t *extrema = _starter->get_intensity_extrema();
	
  for (points_list_t::iterator it = extrema->begin(); it != extrema->end(); ++it)
  {
    if ( !extract_regions(*it) )
    {
      ok = false;
    }
  }
	
  return ok;
}

//-------------------------------------------------------------------------------
//--- Helper Functions  ---------------------------------------------------------
//-------------------------------------------------------------------------------

bool ell_region_extractor::find_points_on_rays()
{
#if defined VERB_ELLIPTICAL
    vcl_cout << "ell_region_extractor: Find points on rays..." << vcl_endl;
#endif
	
  // Local variables setup
  int NB_RAYS = _params.number_of_rays_EllRegExt;
  int MINSIZE = _params.min_size_EllRegExt;
  int MAXSIZE = _params.max_size_EllRegExt;
	
  vil1_memory_image_of<byte> *imagebuf = _starter->get_BW_image_buffer();
	
  point_t* pt = _anchor; // shortcut and readability
	
  _ec = new osl_edgel_chain(NB_RAYS + 1);
	
  int i_chain = 0; // Aaron changed this to double - I've no idea why...
  double extremum, extremum2, sum, f;
	

  // coherence with Tinne algo:
  double f0 = get_byte_interpolation_no_check_slowround(*imagebuf, _anchor->x(), _anchor->y());
	
  double cur_x = _anchor->x();  // we KNOW are integers coords (extremum computed!)
  double cur_y = _anchor->y();
  int extri = 10;
  int prev_extri = 0, extri2 = 0;
  int sizex = imagebuf->width();//byte_buf->GetSizeX();
  int sizey = imagebuf->height();//byte_buf->GetSizeY();
	
  /* check all rays */
  for (double theta = 0; theta < 2*ABL_PI; theta += 2 * ABL_PI/(double)NB_RAYS, i_chain++)
  {
    /* pt2: endpoint of ray (but inside image boundaries) */
    point_t* pt2 = new point_t(pt->x()+MAXSIZE*cos(theta), pt->y()+MAXSIZE*sin(theta), 1);
    if (pt2->x() <0)
    {
      pt2->set(0, pt->y() + (pt2->y() - pt->y())/(pt2->x() - pt->x())*(0-pt->x())); 
    }
    if (pt2->x() > sizex-1)
    {
      pt2->set(sizex-1, pt->y() + (pt2->y() - pt->y())/(pt2->x() - pt->x())*(sizex-1-pt->x())); 
    }
    if (pt2->y() <0)
    {
      pt2->set(pt->x() + (pt2->x() - pt->x())/(pt2->y() - pt->y())*(0-pt->y()), 0); 
    }
    if (pt2->y() > sizey-1)
    {
      pt2->set(pt->x() + (pt2->x() - pt->x())/(pt2->y() - pt->y())*(sizey-1-pt->y()), sizey-1); 
    }
		
		
    extremum = 2;       // global extremum
    extremum2 = 2;      // local extremum (close to extremum on previous ray)
    sum = 0;  
    prev_extri = extri; // position of extremum along previous ray
		
    cur_x = pt->x();  
    cur_y = pt->y(); 
		
    double incr_x = (pt2->x()-pt->x())/(double)MAXSIZE;
    double incr_y = (pt2->y()-pt->y())/(double)MAXSIZE;
		
    /* evaluate function along ray */
    for (int i=1; i<MINSIZE; i++)
    {
      cur_x += incr_x;
      cur_y += incr_y;
      f = get_byte_interpolation_no_check_slowround(*imagebuf/**byte_buf*/, cur_x, cur_y);
			
      sum += fabs(f0-f);
    }
		
    for (int i=MINSIZE; i<MAXSIZE; i++)
    {
      cur_x += incr_x;
      cur_y += incr_y;
      f = get_byte_interpolation_no_check_slowround(*imagebuf, cur_x, cur_y);
			
      double cur_str = fabs(f0-f) / ( MAXTWO((double)(sum/(double)i), (double)5.0) );
			
      /* search global maximum */
      if (cur_str > extremum) 
      {
        extremum = cur_str; 
        extri = i;
      }
			
      /* search local maximum (close to maximum on previous ray) */
      if (fabs(prev_extri - i) <= 5) 
      {
        if ((cur_str > extremum2) &&
            (fabs(prev_extri - i) != 5)) // make sure it is an extremum
        {
          extremum2 = cur_str;
          extri2 = i;
        }
      }
			
      sum += fabs(f0-f);
    }
		
    if (extremum == 2) // no extremum found
    {
      /* no extremum found along ray -> region extraction failed */
      i_chain--;
      delete pt2;

      
#if defined VERB_ELLIPTICAL
      vcl_cout << theta*180/ABL_PI << "   BREAK  (" << i_chain << ")" << vcl_endl;
#endif
			
      return false;      
    }
    else
    {
      if ( (i_chain != 0) && (fabs(extri - prev_extri) > 5))
      {
        /* global extremum far from extremum on previous ray
           check alternative without jumping (extremum2) */
#if defined VERB_ELLIPTICAL
          vcl_cout << "JUMP detected" << vcl_endl;
          vcl_cout << "extremum " << extremum << " at " 
                   << pt->x()*(1-extri/(double)MAXSIZE) + pt2->x()*(extri/(double)MAXSIZE) << ", " 
                   << pt->y()*(1-extri/(double)MAXSIZE) + pt2->y()*(extri/(double)MAXSIZE) << vcl_endl;
          vcl_cout << "possible alternative without jumping:" << vcl_endl;
          vcl_cout << "new extremum " << extremum2 << " at " 
                   << pt->x()*(1-extri2/(double)MAXSIZE) + pt2->x()*(extri2/(double)MAXSIZE) << ", " 
                   << pt->y()*(1-extri2/(double)MAXSIZE) + pt2->y()*(extri2/(double)MAXSIZE) << vcl_endl;
#endif
				
        if (extremum2 > (MAXTWO(extremum*0.7, 2.0)) )
        {
          extremum = extremum2;
          extri = extri2;

#if defined VERB_ELLIPTICAL
          vcl_cout << "new extremum accepted" << vcl_endl;
#endif
        } 
      }
			
      /* add new extremum to osl_edgelchain */
      _ec->SetX(pt->x()*(1-extri/(double)MAXSIZE) + pt2->x()*(extri/(double)MAXSIZE), i_chain);
      _ec->SetY(pt->y()*(1-extri/(double)MAXSIZE) + pt2->y()*(extri/(double)MAXSIZE), i_chain);

      
#if defined VERB_ELLIPTICAL
        vcl_cout << theta*180/ABL_PI << "\t" << extri << "\t";
        vcl_cout << _ec->GetX(i_chain) << ", " << _ec->GetY(i_chain) << vcl_endl;
#endif
			
      /* try to eliminate 'outliers' */
      if ( (i_chain > 1) && 
           ( pow(_ec->GetX(i_chain) - _ec->GetX(i_chain-2), 2) + 
             pow(_ec->GetY(i_chain) - _ec->GetY(i_chain-2), 2)     < 
             0.5 * MAXTWO((pow(_ec->GetX(i_chain) - _ec->GetX(i_chain-1), 2) + 
                           pow(_ec->GetY(i_chain) - _ec->GetY(i_chain-1), 2)),
                          (pow(_ec->GetX(i_chain-2) - _ec->GetX(i_chain-1), 2) + 
                           pow(_ec->GetY(i_chain-2) - _ec->GetY(i_chain-1), 2))) ) )
      {
        _ec->SetX(_ec->GetX(i_chain), i_chain-1);
        _ec->SetY(_ec->GetY(i_chain), i_chain-1);

        
#if defined VERB_ELLIPTICAL
          vcl_cout << "removed outlier " << vcl_endl;
#endif
          
      }
    }
    delete pt2;
  }
	
  /* close osl_edgel_chain */
#if defined VERB_ELLIPTICAL
  vcl_cout << "closing osl_edgelchain" << vcl_endl;
#endif
	
  while (i_chain <= NB_RAYS)
  {
    _ec->SetX(_ec->GetX(0), i_chain);
    _ec->SetY(_ec->GetY(0), i_chain);
    i_chain++;
  }
	
  /* display edgel chain */
  _dc = new vdgl_digital_curve(osl_to_vdgl(*_ec));
	
  return true;
} // find_points_on_rays


bool ell_region_extractor::link_points()
{
#if defined VERB_ELLIPTICAL
    vcl_cout << "ell_region_extractor: Link Points..." << vcl_endl;
#endif
	
  // Local vars shortcuts
  point_t* pt = _anchor;
	
  _mombuf = new vil1_memory_image_of<unsigned char>(_sx, _sy);
  memset(_mombuf->get_buffer(), 0, sizeof(unsigned char)*_sx*_sy);
	
  /* set buffervalues equal to 100 for pixel along osl_edgel_chain
     look out not to leave any gaps */
  int L = _ec->size();
  double x = _ec->GetX(0);
  double y = _ec->GetY(0);
  double prev_x, prev_y;
  int xx, yy, sgn_x, sgn_y, tx, ty;
	
  for (int i=1; i<L; i++)
  {
    prev_x = x;
    prev_y = y;
    x = _ec->GetX(i);
    y = _ec->GetY(i);
		
    if ( fabs(x-prev_x) > fabs(y-prev_y) )
    {
      sgn_x = (x > prev_x ? 1: -1);
			
      for (xx = (int)round(prev_x); xx*sgn_x <= (int)round(x)*sgn_x; xx += sgn_x)
      {
        yy = (int) (prev_y + (xx-prev_x)/(x-prev_x) * (y-prev_y));
				
        if ( ( _miny <= yy) && (yy <= _maxy) )
        {
          tx = (int)round(xx) - _minx;
          ty = (int)round(yy) - _miny;
          (*_mombuf)(tx, ty) = 100;
        }
      }
    }
    else
    {
      sgn_y =  ( y > prev_y ? 1: -1 );
      for (yy = (int)round(prev_y); yy*sgn_y <= (int)round(y)*sgn_y; yy += sgn_y)
      {
        xx = (int) (prev_x + (yy-prev_y)/(y-prev_y) * (x-prev_x));
				
        if ( ( _minx <= xx ) && ( xx <= _maxx ) )
        {
          tx = (int)round(xx) - _minx;
          ty = (int)round(yy) - _miny;
          (*_mombuf)(tx, ty) = 100;
        }
      }
    }
  }
	
  /* fill up pixels from interior of osl_edgel_chain 
     starting from the center of the buffer and adding
     points above, below, left and right until osl_edgel_chain is reached */
#if defined VERB_ELLIPTICAL
    vcl_cout << "filling interior" << vcl_endl;
#endif
	
  vcl_list<int> xlist;
  vcl_list<int> ylist;
	
  int ii = (int)round(pt->x()) - _minx;
  int jj = (int)round(pt->y()) - _miny;
	
  (*_mombuf)(ii, jj) = 100;
	
  xlist.push_front(ii);
  ylist.push_front(jj);
	
  while ( !(xlist.empty()) )
  {
    ii = *(xlist.begin());
    xlist.erase(xlist.begin());
    jj = *(ylist.begin());
    ylist.erase(ylist.begin());
		
    //check point above
    if ( jj > 0 )
    {
      if ( (*_mombuf)(ii, jj-1) == 0 )
      {
        (*_mombuf)(ii, jj-1) = 100;
        xlist.push_front(ii);
        ylist.push_front(jj-1);
      }      
    }
    //check point below
    if ( jj < _sy-1 )
    {
      if ( (*_mombuf)(ii, jj+1) == 0 )
      {
        (*_mombuf)(ii, jj+1) = 100;
        xlist.push_front(ii);
        ylist.push_front(jj+1);
      }
    }
    //check point left
    if ( ii > 0 )
    {
      if ( (*_mombuf)(ii-1, jj) == 0 )
      {
        (*_mombuf)(ii-1, jj) = 100;
        xlist.push_front(ii-1);
        ylist.push_front(jj);
      }
    }
    //check point right
    if ( ii < _sx-1 )
    {
      if ( (*_mombuf)(ii+1, jj) == 0 )
      {
        (*_mombuf)(ii+1, jj) = 100;
        xlist.push_front(ii+1);
        ylist.push_front(jj);
      }
    }
  }
	
  return true;
} // link_points


bool ell_region_extractor::compute_ellipse()
{
#if defined VERB_ELLIPTICAL
    vcl_cout << "ell_region_extractor: Compute Ellipse..." << vcl_endl;
#endif
	
  vil1_memory_image_of<byte> *imagebuf = _starter->get_BW_image_buffer();
	
  int N = 0;
  double sumx = 0.0;
  double sumy = 0.0;
  double sumxx = 0.0;
  double sumxy = 0.0;
  double sumyy = 0.0;
  double I;
  double sumI = 0.0;
  double sumIx = 0.0;
  double sumIy = 0.0;

  for (int i=0; i<_sx; i++)
  {
    for (int j=0; j<_sy; j++)
    {
      if ( (*_mombuf)(i, j ) == 100 )
      {
        N++;
        sumx += i;
        sumy += j;
        sumxx += i*i;
        sumxy += i*j;
        sumyy += j*j;
				
        I = (double)((*imagebuf)(i + _minx, j + _miny));//byteABL_PIxel(*byte_buf, i + minx, j + miny);
				
        sumI += I;
        sumIx += i*I;
        sumIy += j*I;
      }
    }
  }
	
  sumx /= N;
  sumy /= N;
	
#if defined VERB_ELLIPTICAL
    vcl_cout << "zwtepunt:" << sumx << ", " << sumy << vcl_endl;
#endif
	
  sumxx -= N*sumx*sumx;
  sumxy -= N*sumx*sumy;
  sumyy -= N*sumy*sumy;
  sumxx /= N;
  sumxy /= N;
  sumyy /= N;
	
#if defined VERB_ELLIPTICAL
    vcl_cout << "spread: " << sumxx << ", " << sumxy << ", " << sumyy << vcl_endl;
#endif
	
	
  /* compute ellipse parameters */
  double th = 0.5 * atan2(2*sumxy,(sumxx-sumyy));
	
#if defined VERB_ELLIPTICAL
    vcl_cout << "theta: " << th*180/ABL_PI << vcl_endl;
#endif
	
  double sth = sin(th);
  double cth = cos(th);
  double sumuu = cth*cth*sumxx + sth*sth*sumyy + 2*cth*sth*sumxy;
  double sumvv = sth*sth*sumxx + cth*cth*sumyy - 2*sth*cth*sumxy;
  double a = sqrt(sumuu)*2;
  double b = sqrt(sumvv)*2;
	
  /* change ellipse size */
  a *= _params.resize_factor_EllRegExt;
  b *= _params.resize_factor_EllRegExt;
	
#if defined VERB_ELLIPTICAL
    vcl_cout << "a: " << a << vcl_endl;
    vcl_cout << "b: " << b << vcl_endl;
#endif
	
  point_t* ctr = new point_t(_minx + sumx, _miny + sumy, 1.0);
  point_t* p1 = new point_t(_minx + sumx + a*cth, _miny + sumy + a*sth, 1.0);
  point_t* p2 = new point_t(_minx + sumx - b*sth, _miny + sumy + b*cth, 1.0);
	
#if defined VERB_ELLIPTICAL
    vcl_cout << "center: " << *ctr << vcl_endl;
    vcl_cout << "pt1: " << *p1 << vcl_endl;
    vcl_cout << "pt2: " << *p2 << vcl_endl;
#endif


    ell_region* region = new ell_region(ctr, p1, p2,
                                        region::ELL, _anchor,
                                        _starter->get_image_buffer()); // still not comm _image par
    
    // check whether still inside image
  int W = _starter->get_BW_image_buffer()->width()-1;
  int H = _starter->get_BW_image_buffer()->height()-1;
	
  double rmix = region->get_min_x(); double rmiy = region->get_min_y();
  double rmax = region->get_max_x(); double rmxy = region->get_max_y();
    
  if ( ( rmix < 0 ) || ( rmax > W  ) ||
       ( rmiy < 0 ) || ( rmxy > H ) )
  {
#if defined VERB_ELLIPTICAL
      vcl_cout << "region rejected: outside image " << vcl_endl; 
#endif
      
      delete region;
      delete ctr;
      delete p1;
      delete p2;
      region = NULL;
      return false;
  }

  // rotate the axes to AOI
  find_orientation_aoi(region, *(_starter->get_BW_image_buffer()));
	
  // if ( region->good_region(0.6, 100, 1500) )  {
  _regions.push_back(region);
  //}
  //else
  //{
  // #if defined VERB_ELLIPTICAL
  //     vcl_cout << " region rejected...Goodregion() failed\n" << vcl_endl;
  // #endif
  //   }
	
  delete ctr;
  delete p1;
  delete p2;
	
  return true;
}


bool ell_region_extractor::find_orientation_aoi(ell_region *Reg,
                                                vil1_image &Image)
{
#if defined VERB_ELLIPTICAL
    vcl_cout << "ell_region_extractor: Find Orientation AOI..." << vcl_endl;
#endif

  aoi_computor comp(Reg, &Image);

  if( !comp.compute() ) 
  {
    vcl_cerr << "ell_region_extractor: AOI computation failed" << vcl_endl;
    return false;
  }

#if defined VERB_ELLIPTICAL
    vcl_cout << "ell_region_extractor: rotating region over " 
	     << (comp.get_rotation_angle())*180/3.1416 << " degrees" << vcl_endl;
#endif

  Reg->rotate_axes((comp.get_rotation_angle())*180/3.1416);
	
  return true;
}
