// ----------------------------------------------------------------------
//
// StraightregionExtractor.cpp: implementation of the
//                              StraightregionExtractor class.
//
// ----------------------------------------------------------------------

#include "homogeneous_region_extractor.h"
#include <misc_utility/defines.h>
#include <region_utility/region_deskew.h>

#include <vil1/vil1_image_as.h>
#include <vil1/vil1_smooth.h>
#include <vil1/vil1_save.h>

#include <vsol/vsol_point_2d.h>

#include <vnl/vnl_matrix.h>
#include <vnl/vnl_math.h>

#include <fstream>
#include <cstdio>

#if defined (DEBUG_GL)
#include <GL/glut.h>
#endif



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
homogeneous_region_extractor::homogeneous_region_extractor(reg_ext_starter *starter)
  : parall_region_extractor(starter)
{
  _extraction_statistics.resize(4, 0);
}




homogeneous_region_extractor::~homogeneous_region_extractor()
{

}





bool homogeneous_region_extractor::extract_all_regions()
{
  points_list_t *HarrisCorners = _starter->get_harris_corners();


  // FOR DEBUGGING
  _extraction_statistics[0] = HarrisCorners->size();
  

  for(points_list_t::iterator it = HarrisCorners->begin();
      it != HarrisCorners->end(); ++it)
  {
    extract_regions(*it);
  }

  return true;
}




bool homogeneous_region_extractor::extract_regions(point_t* pt,
                                                region::region_specific_type tp)
{
  
  int MinSize = _starter->get_params()->MinSideLength;
  int MaxSize = _starter->get_params()->MaxSideLength;


  // Are there any straight lines near the
  // corner point pt ? If not, then abort !
  if ( !compute_lines(pt) ){


    // update statistics
    _extraction_statistics[1] += 1;

#if defined VERB_HOMOG
    vcl_cout << "No lines around " << *pt << vcl_endl;
#endif
    
    return false;
  }

  
  

  // Among all the straight lines found near the anchor point,
  // construct basic parallelogram shaped regions and maximize
  // them in size. The result is returned as a vector containing the
  // maximized regions.
  vcl_vector<parall_region*> maximum_regions = maximize_regions();



  // First geometric triage:
  // All those maximized regions that do not fulfill the geometric
  // constraints checked by 'good_region' are subsequently
  // removed
  for (vcl_vector<parall_region*>::iterator it = maximum_regions.begin();
       it < maximum_regions.end(); ++it){

    parall_region* reg = *it;

    if ( !reg->good_region(1000, ABL_SQUARE(MinSize)*ABL_PI*2, 1000000000) ){

      // According to the STL documentation, upon calling erase(it),
      // the element pointed to by it is destroyed, so there should
      // be no need to delete reg (however I'm not sure)
      it = maximum_regions.erase(it);
      _extraction_statistics[2] += 1;
    }
  }


  // It does certainly make no sense to continue without
  // candidates ;-)
  if ( maximum_regions.empty() )
    return false;




  vcl_vector<parall_region*> candidates;



  // Now we search within the maximized regions for possible candidates
  // for homogeneous regions
  candidates = extract_candidates(maximum_regions);
  
    
#if defined DEBUG_GL
  for (vcl_vector<parall_region*>::const_iterator it = candidates.begin();
       it < candidates.end(); ++it){
    
    parall_region* reg = *it;
  
  
    point_t* c = reg->get_pt0();
    point_t* p1 = reg->get_pt1();
    point_t* p2 = reg->get_pt2();
    point_t* p3 = reg->get_pt3();

      
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1,1,0,.25);
      
    glBegin(GL_LINE_LOOP);
    glVertex2f(c->x(), _starter->get_image_buffer()->height() - c->y());
    glVertex2f(p1->x(), _starter->get_image_buffer()->height() - p1->y());
    glVertex2f(p3->x(), _starter->get_image_buffer()->height() - p3->y());
    glVertex2f(p2->x(), _starter->get_image_buffer()->height() - p2->y());
    glEnd();
    }
#endif
        

  // Second geometric triage:
  // This time, all extracted region candidates are checked for geometric
  // constraints, and only the surviving ones can be considered as
  // real, homogeneous regions.
  for (vcl_vector<parall_region*>::iterator it = candidates.begin();
       it < candidates.end(); ++it){
    
    parall_region* reg = *it;

    
    if ( !reg->good_region(10.0, ABL_SQUARE(MinSize)*ABL_PI*2,
                             ABL_SQUARE(MaxSize)*ABL_PI/2.0) ){
      it = candidates.erase(it);
      _extraction_statistics[3] += 1;
    }
    else{
      _regions.push_back(reg);
      _extraction_statistics[1] += 1;
    }
  }

  
  return (candidates.empty() ?  false :true );
}




points_list_t
homogeneous_region_extractor::find_rgb_scales(vcl_list<int>& temp_types)
{

#if defined VERB_HOMOG
    vcl_cout << "find 2D scales -- 1Q -- RGB" << vcl_endl;
#endif

  points_list_t scales;
	
  int sx = ds_buf->width();
  int sy = ds_buf->height();


  // Buffer for color bands
  vil1_memory_image_of<vxl_byte> R_buf(sx, sy);
  vil1_memory_image_of<vxl_byte> G_buf(sx, sy);
  vil1_memory_image_of<vxl_byte> B_buf(sx, sy);


  // Buffer containing gradient functions
  vil1_memory_image_of<float> gradR_buf(sx, sy);
  vil1_memory_image_of<float> gradG_buf(sx, sy);
  vil1_memory_image_of<float> gradB_buf(sx, sy);


  // Initialization
  R_buf.fill(0); gradR_buf.fill(0);
  G_buf.fill(0); gradG_buf.fill(0);
  B_buf.fill(0); gradB_buf.fill(0);
  

	
  // Split color bands
  split_color_bands(*ds_buf, R_buf, G_buf, B_buf);



  // Compute the affinely invariant gradient function
  // for each color band separately
  compute_gradient_buffer(gradR_buf, R_buf);
  compute_gradient_buffer(gradG_buf, G_buf);
  compute_gradient_buffer(gradB_buf, B_buf);

  
  scales = compare_grad_buffers(gradR_buf, gradG_buf, gradB_buf,
                                R_buf, G_buf, B_buf);


  int s = scales.size();

  for (int i=0; i<s; ++i)
    temp_types.push_back((int)(region::PARALL_HOMOGENEOUS));
  
  

  return scales;
  
}




bool homogeneous_region_extractor::compute_gradient_buffer(vil1_memory_image_of<float>& grad_buf,
                                                         const vil1_memory_image_of<vxl_byte>& buf) const
{
  int sx = buf.width();
  int sy = buf.height();


  // Temporary buffers for storing gradients in x and y
  vil1_memory_image_of<float> xgrads(sx-1, sy);
  vil1_memory_image_of<float> ygrads(sx-1, sy);


  // Initialization
  xgrads.fill(0);
  ygrads.fill(0);
  
  
  // Better smooth the original color buffers
  //vil1_memory_image_of<vxl_byte> smoothed(vil1_smooth_gaussian(buf, 3.0));


  // --------------------------------------------------
  // Computation of affinely invariant gradient
  // buffers
  // --------------------------------------------------
  
  // Sum of x-gradients in y-direction
  for (int j=3; j<sy; ++j)
    for (int i=3; i<sx-1; ++i){

      // 	floatPixel(xgrads, i, j) = (bytePixel(buf, i, j) - bytePixel(buf, i+1, j)) + 
      // 	floatPixel(xgrads, i, j-1);
      xgrads(i, j) = ((float)buf(i, j) - (float)buf(i+1, j)) + xgrads(i, j-1);
    }


  
  // Sum of y-gradients in x-direction
  for (int j=3; j<sy-1; ++j)
    for (int i=3; i<sx; ++i){

      // 	floatPixel(ygrads, i, j) = (bytePixel(buf, i, j) - bytePixel(buf, i, j+1)) + 
      // 	floatPixel(ygrads, i-1, j);
      ygrads(i, j) = ((float)buf(i, j) - (float)buf(i, j+1)) + ygrads(i-1, j);
    }


  // The normalization
  // We are interested only in the absolute value
  float max = 0;
  
  for (int i=3; i<sx-1; ++i)
    for (int j=3; j<sy-1; ++j){
      grad_buf(i, j) = xgrads(i, j) * ygrads(i, j)/(i*j);

      if ( grad_buf(i,j) < 0 )
        grad_buf(i,j) *= (-1);

      if ( grad_buf(i,j) > max )
        max = grad_buf(i,j);
    }


  
  // Thresholding
  max /= 4.0;
  for (int i=3; i<sx-1; i++)
    for (int j=3; j<sy-1; j++){

      if ( grad_buf(i,j) < max )
        grad_buf(i,j) = 0.0;
      
    }


  // Non-maximum suppression
  int delta = 3;
  for (int j=delta ; j<sy-delta; j++)
    for (int i=delta; i<sx-delta; i++){

      if ( grad_buf(i,j) > 0 ){
        
        for (int jj = j-delta; jj<j+delta; jj++)
          for (int ii = i-delta; ii<i+delta; ii++){
            
            
            if ( grad_buf(ii, jj) > grad_buf(i, j) )
              grad_buf(i, j) = 0.0;
          }
        
      }
    }
  

      
  return true;
}




points_list_t
homogeneous_region_extractor::compare_grad_buffers(vil1_memory_image_of<float>& gradR,
                                                 vil1_memory_image_of<float>& gradG,
                                                 vil1_memory_image_of<float>& gradB,
                                                 const vil1_memory_image_of<vxl_byte>& R,
                                                 const vil1_memory_image_of<vxl_byte>& G,
                                                 const vil1_memory_image_of<vxl_byte>& B)
{

  int sx = gradR.width();
  int sy = gradR.height();

  vcl_vector<int*> scales_R;
  vcl_vector<int*> scales_G;
  vcl_vector<int*> scales_B;


  for (int j=3; j<sy; j++)
    for (int i=3; i<sx; i++){
      
      if ( gradR(i, j) != 0 ){
        int* cand = new int(2);
        cand[0] = i; cand[1] = j;
	scales_R.push_back(cand);
      }
      if ( gradG(i, j) != 0 ){
        int* cand = new int(2);
        cand[0] = i; cand[1] = j;
	scales_G.push_back(cand);
      }
      if ( gradB(i, j) != 0 ){
        int* cand = new int(2);
        cand[0] = i; cand[1] = j;
	scales_B.push_back(cand);
      }
    }


  
#if defined VERB_HOMOG
  cout << "scales_R: " << scales_R.size() << endl;
  cout << "scales_G: " << scales_G.size() << endl;
  cout << "scales_B: " << scales_B.size() << endl;
#endif
  


  vcl_vector<float> contrastR(3,-1);
  vcl_vector<float> contrastG(3,-1);
  vcl_vector<float> contrastB(3,-1);


  if ( !scales_R.empty() )
    contrastR = filter_grad_scales(scales_R, R);
  if ( !scales_G.empty() )
    contrastG = filter_grad_scales(scales_G, G);
  if ( !scales_B.empty() )
    contrastB = filter_grad_scales(scales_B, B);


#if defined VERB_HOMOG
  cout << "contrastR: " << contrastR[0] << " " << contrastR[1]<< "  " << contrastR[2] << endl;
  cout << "contrastG: " << contrastG[0] << " " << contrastG[1]<< "  " << contrastG[2] << endl;
  cout << "contrastB: " << contrastB[0] << " " << contrastB[1]<< "  " << contrastB[2] << endl;
#endif
  



  // Find maximum
  float rgb[3];
  rgb[0] = contrastR[2];
  rgb[1] = contrastG[2];
  rgb[2] = contrastB[2];
  
  float max = rgb[0];
  unsigned max_index = 0;
  for (unsigned i=1; i<3; i++)
    if (rgb[i] > max ){
      max = rgb[i];
      max_index = i;
    }

  vcl_vector<float> selection;
  switch(max_index){
  case 0:
    selection = contrastR;
    break;
  case 1:
    selection = contrastG;
    break;
  case 2:
    selection = contrastB;
    break;
  }


  points_list_t candidates;

  if ( (selection[0] != -1) && (selection[1] != -1) ){
    point_t* cand = new
      point_t(selection[0], selection[1], 1.0 );

    candidates.push_back(cand);
  }

  
  return candidates;
  
}




vcl_vector<float>
homogeneous_region_extractor::filter_grad_scales(vcl_vector<int*>& scales,
                                               const vil1_memory_image_of<vxl_byte>& buf)
{

  vcl_vector<int*> tmp_scales;
  
  
  // Find the minimum distance from the origin 
  float length_ratio = 5;

  //int sx = buf.width();
  //int sy = buf.height();

  //int size = min(sx, sy);
  //float min_distance = (float)size / 5.0;   // We favour meaningful sizes
  
  unsigned row_index = 0;

  
  // Remove all scale candidates not fulfilling the geometric
  // length-ratio constraint
  while ( !scales.empty() ){
    
    int* p = scales.back();
    scales.pop_back();
    
    
    // Check ratio
    float ratio = (float)p[0]/(float)p[1];

    
    
    if (ratio < 1)
      ratio = (float)p[1]/(float)p[0];
    

    // Are there any significant edges within the area spanned by
    // the candidate for the homogeneous region ?
    if ( ratio < length_ratio ){
    
//       bool penalty = false;

//       for (int i=0; i<p[0]; ++i)
//         for (int j=0; j<p[1]; ++j){

//           if ( (*ds_line_buf)(i,j) ){

//             penalty = true;
//             break;
//           }
//         }

     
//       if ( !penalty )
        tmp_scales.push_back(p);
    }
#ifdef VERB_HOMOG
    else{
      cout << "    " << p[0] <<" " << p[1] << " rejected (length ratio)" << endl;
    }
    
#endif
  }
  
  // If empty then abort
  if ( tmp_scales.empty() ){
    vcl_vector<float> no(3,-1);
    return no;
  }
  
  
  vnl_matrix<float> stats(tmp_scales.size(), 3, 0);
  
  
  // Calculate variance over point candidates 
  while ( !tmp_scales.empty() ){
    int* p = tmp_scales.back();
    tmp_scales.pop_back();
    
    
    int x = p[0];
    int y = p[1];
    int offset = 3;
    
    
    float sum=0; float sum_squared = 0;
    int j=1;  // Needed for computing the statistics
    
    for (int jj=offset; jj<(y-offset); ++jj)     // Remember: Border of buffer is empty (jj=offset)
      for (int ii =offset; ii<(x-offset); ++ii){ // See: compute_grad_buffers()
        // ii, jj -> x,y-offset instead of x,y (30.10.2000)
        float val = (float)buf(ii, jj);
        
        sum += val;
        sum_squared += val*val;
        j++;
      }
    
    float variance = (sum_squared - (1.0/(float)j)*sum*sum)/(float)j;
    
    
    
    
    // Save result in stats matrix
    // Note: Assert-style boundary checks can be turned off
    // by setting '#define NDEBUG'
    stats(row_index, 0) = x;
    stats(row_index, 1) = y;
    stats(row_index, 2) = variance;
    
    ++row_index;
  }

  
#if defined VERB_HOMOG
  cout << "Stats:\n " << stats << endl;
#endif
  
  
  // Find minimum variance
  float min_var = stats(0,2);
  float var = 0.0;
  unsigned min_row_index = 0;
  
  
  for (unsigned i=1; i<(unsigned)stats.rows(); ++i){
    var = stats(i,2);
    
    if ( var < min_var ){
      min_var = var;
      min_row_index = i;
    }
  }
  
  
  if ( stats.rows() > 1){
    float N = stats.rows() - 1;
    float medium = 0;
    for (unsigned i=0; i<(unsigned)stats.rows(); i++)
      if ( i != min_row_index )
        medium += stats(i,2);
    
    medium /= N;
    
    vcl_vector<float> res(3,0);
    res[0] = stats(min_row_index,0);
    res[1] = stats(min_row_index,1);
    res[2] = vnl_math_abs(min_var - medium);
    
    return res;
  }
  
  vcl_vector<float> res(3,0);
  vnl_vector<float> row = stats.get_row(0);
  
  res[0] = row(0);
  res[1] = row(1);
  res[2] = row(2);
  
  
  return res;
  
}



// ----------------------------------------------------------------------
// Make a nice, formatted printout of the extraction
// statistics.
// Started: 16.5.2002
// ----------------------------------------------------------------------
void homogeneous_region_extractor::print_statistics() const
{
  printf("\n\n");
  printf("Seedpoints                             : %i\n", _extraction_statistics[0]);
  printf("No lines near seed                     : %i\n", _extraction_statistics[1]);
  printf("Maximal search region geometry fails   : %i\n", _extraction_statistics[2]);
  printf("Extrema geometry fails                 : %i\n", _extraction_statistics[3]);
  vcl_cout << vcl_endl;
}




// ----------------------------------------------------------------------
// Among all straight lines found around the anchor point,
// construct regions and scale them to their maximum size or till
// they hit the image boundaries
// Started: 16.5.2002
// ----------------------------------------------------------------------
vcl_vector<parall_region*> homogeneous_region_extractor::maximize_regions()
{

  vcl_vector<parall_region*> maximized_regions;

  vdgl_digital_curve *l1, *l2;
  point_t* ctr;
  parall_region* reg;
  
  curves_list_t::iterator it1, it2;
  points_list_t::iterator PointIt = _centers->begin();


  it1 = _Lines1->begin();
  it2 = _Lines2->begin();


  while ( !_Lines1->empty() )
  {	
    l1 = *it1;	it1 = _Lines1->erase(it1);
    l2 = *it2;	it2 = _Lines2->erase(it2);
    ctr = *PointIt;	PointIt = _centers->erase(PointIt);
		
    // Set p1 and p2 to maximum region (2*MAXSIZE)*(2*MAXSIZE)
    maximum_region(*ctr, *l1, *l2, p1, p2);

    
    // switch if necessary
    if ((p1->x()-ctr->x()) * (p2->y()-ctr->y()) - (p1->y()-ctr->y()) * (p2->x()-ctr->x()) < 0.0)
    {
      point_t* tmp = p1;
      p1 = p2;
      p2 = tmp;
    }

    reg = new parall_region(ctr, p1, p2);

    maximized_regions.push_back(reg);
  }

  return maximized_regions;
}




// ----------------------------------------------------------------------
// Extraction of region candidates in the deskewed domain.
// For each maximized region, possible candidates for homogeneous
// regions are extracted after the maximized region has been deskewed.
// The result is converted to parall_regions (i.e. skewed again)
// and returned.
// Started: 16.5.02
// ----------------------------------------------------------------------
vcl_vector<parall_region*>
homogeneous_region_extractor::extract_candidates(const vcl_vector<parall_region*>& maximized) 
{
  vcl_vector<parall_region*> candidates;


  for (vcl_vector<parall_region*>::const_iterator it = maximized.begin();
       it < maximized.end(); ++it){

    vcl_list<int> temp_types;

    parall_region* reg = *it;

    point_t* ctr = reg->get_pt0();
    point_t* p1 = reg->get_pt1();
    point_t* p2 = reg->get_pt2();

    int dist1 = (int)ABLround(ABL_DIST(ctr->x(), ctr->y(), p1->x(), p1->y()));
    int dist2 = (int)ABLround(ABL_DIST(ctr->x(), ctr->y(), p2->x(), p2->y()));

      
    // Deskew the maximized region
    vil1_image* buf = _starter->get_image_buffer();
    vil1_pixel_format_t T = vil1_pixel_format(*buf);
    if (T == VIL1_RGB_BYTE)
      ds_buf = deskew_region(reg, (vil1_memory_image_of<vil1_rgb_byte>*)buf, dist1, dist2);
    else 
    {
      vcl_cout << "Error: straight_region_extractor doesn't know how to deal with images of type " 
	       << T << vcl_endl;
      return candidates;
    }
    // ds_line_buf = deskew_region(reg, _starter->get_line_buffer(), dist1, dist2);



    // Now search for possible canidates within the deskewed, maximized region
    // buffer
    points_list_t deskewed_cands = find_rgb_scales(temp_types);


    int type = temp_types.front();
    

    // Put result in regions
    while ( !deskewed_cands.empty() ){
      
      point_t* pt = deskewed_cands.back();
      deskewed_cands.pop_back();
      
      float cx = ctr->x();
      float cy = ctr->y();
      int sx = ds_buf->width(); // 2*MAX_SIZE+1 or MAX_SIZE
      int sy = ds_buf->height(); // 2*MAX_SIZE+1 or MAX_SIZE
      
        
      // Affine parameters needed to transform region
      // candidate points from canonical deskewed coordinate frame
      // back to image coordinate frame
      float scx1 = (p1->x() - cx) / sx;
      float scy1 = (p1->y() - cy) / sx;
      float scx2 = (p2->x() - cx) / sy;
      float scy2 = (p2->y() - cy) / sy;



      point_t* point1 = new point_t(cx + pt->x()*scx1, 
                                    cy + pt->x()*scy1, 1);
      point_t* point2 = new point_t(cx + pt->y()*scx2, 
                                    cy + pt->y()*scy2, 1);



      // Make the new, skewed region
      parall_region* region = new parall_region(ctr, point1, point2,
                                              (region::region_specific_type)type);

      candidates.push_back(region);

      
      // We no longer need the deskewed buffer
      delete ds_buf; ds_buf = NULL;

      // .. and the deskewed candidate point
      delete pt; pt = NULL;
    }
  }
  return candidates;
}
  
