// ParallregionExtractor.cpp: implementation of the ParallregionExtractor class.
//
//////////////////////////////////////////////////////////////////////

#include <region_extraction/parall_region_extractor.h>

#include <misc_utility/defines.h>
#include <vsol/vsol_point_2d.h>
#include <vdgl/vdgl_interpolator_cubic.h>
#include <vdgl/vdgl_interpolator_linear.h>
#include <vdgl/vdgl_edgel.h>
#include <vdgl/vdgl_edgel_chain.h>

#include <vgl/vgl_homg_point_2d.h>
#include <vgl/vgl_homg_line_2d.h>
#include <vgl/algo/vgl_homg_operators_2d.h>


#include <vcl_algorithm.h>


#if defined (DEBUG_GL)
#include <GL/glut.h>
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

parall_region_extractor::parall_region_extractor(reg_ext_starter *starter)
  : region_extractor(starter)
{
  _offset_invariants = false;

  vil1_image *buf = _starter->get_image_buffer();
  if (vil1_pixel_format(*buf) == VIL1_RGB_BYTE)
    _ColorImage = true;
  else
    _ColorImage = false;

  _centers = new points_list_t;
  _Lines1 = new curves_list_t;
  _Lines2 = new curves_list_t;
}





parall_region_extractor::~parall_region_extractor()
{
  for(points_list_t::iterator pit = _centers->begin();
      pit != _centers->end(); ++pit)
  {
    delete *pit;
  }
  _centers->clear();
  delete _centers;
  
  curves_list_t::iterator cit;
  for(cit = _Lines1->begin(); cit != _Lines1->end(); ++cit)
  {
    delete *cit;
  }
  _Lines1->clear();
  delete _Lines1;
  
  for(cit = _Lines2->begin(); cit != _Lines2->end(); ++cit)
  {
    delete *cit;
  }
  _Lines2->clear();
  delete _Lines2;
}




bool parall_region_extractor::compute_lines(point_t *Corner)
{
  bool straight_edges = false;

  
  _Lines1->clear();
  _Lines2->clear();
  _centers->clear();


  
  // select lines in corner neighbourhood
  curves_list_t* sub_lines = new curves_list_t;


  
  // Intersection point of lines surrounding the corner point
  point_t* pt;


  vil1_memory_image_of<vdgl_digital_curve*>* line_buf = _starter->get_line_buffer();
	
  int i, j;
  double MaxDist = _starter->get_params()->MaxDistanceFromPointToEdge;


  
  // Search radius around corner point
  // Boundary conditions: Search circle crosses image boundaries (x direction)
  int iLow = MAXTWO(0, (int)ABLround(Corner->x() - MaxDist));
  int iHigh = MINTWO(line_buf->width(), (int)ABLround(Corner->x() + MaxDist));
  double min_dist;

  
  for (i=iLow; i < iHigh; i++)
  {
    // Search radius around corner point
    // Boundary conditions: Search circle crosses image boundaries (y direction)
    int jLow = MAXTWO(0, (int)ABLround(Corner->y() - MaxDist));
    int jHigh = MINTWO(line_buf->height(), (int)ABLround(Corner->y() + MaxDist));


    // Store all straight lines within search circle in
    // sub_list
    for (j= jLow; j < jHigh; j++)
    {
      if ( (*line_buf)(i, j) )
      {

        //////////////////////////////////////////////////
        // If the line already exists in sub_lines,
        // then it is not stored again
        //////////////////////////////////////////////////
        curves_list_t::iterator fe;
        fe = find(sub_lines->begin(), sub_lines->end(), (*line_buf)(i, j));
        
        if ( fe == sub_lines->end() ){
          sub_lines->push_back( (*line_buf)(i, j) );
        }
        
      }
    }
  }


#if defined (VERB_PARALL)
  if ( _verbose ) 
  {
    vcl_cout << sub_lines->size() << " lines found";
    vcl_cout << " for cornerpoint: " << *Corner << vcl_endl;


    
#if defined (DEBUG_GL)
    for (curves_list_t::iterator TestIt = sub_lines->begin();
         TestIt != sub_lines->end(); ++TestIt)
    {
      //vcl_cout << "  " << (*(*TestIt)) << vcl_endl;
      vsol_point_2d_sptr p0 = (*TestIt)->p0();
      vsol_point_2d_sptr p1 = (*TestIt)->p1();

      glColor4f(1.0, 0, 0, 1);

      glBegin(GL_POINTS);
      glVertex2f(Corner->x(), _starter->get_image_buffer()->height() - Corner->y());
      glEnd();
      
      glColor4f(1.0, 1.0, 0, 1.0);
      
      
      glBegin(GL_LINES);
      glVertex2f(p0->y(), _starter->get_image_buffer()->height() - p0->x());
      glVertex2f(p1->y(), _starter->get_image_buffer()->height() - p1->x());
      glEnd();
    }
#endif  // DEBUG_GL

  }
#endif  // VERB_PARALL
  
  
  
  double dist;
  min_dist = MaxDist;
  vdgl_digital_curve *l1, *l2;
  pt = new point_t;
  
  
  curves_list_t::iterator iIter, jIter, TempIter;

  
  for (iIter = sub_lines->begin(); iIter != sub_lines->end(); ++iIter)
  {
    l1 = *iIter;
    TempIter = iIter;
    ++TempIter;
    
    for (jIter = TempIter; jIter != sub_lines->end(); ++jIter)
    {
      l2 = *jIter;
      
      if ( l1 != l2 )
      {
        pt->set(-10.0, -10.0);
        intersect_lines(l1, l2, pt);


        
#if defined (VERB_PARALL)
        if ( _verbose ) 
        {
          vcl_cout << "intersection point: " << *pt << vcl_endl;

#if defined (DEBUG_GL)
          glColor4f(0, 1.0, 0, 1.0);

          glBegin(GL_POINTS);
          glVertex2f(pt->x(), _starter->get_image_buffer()->height() - pt->y());
          glEnd();
#endif  // DEBUG_GL
        }
#endif  // VERB_PARALL

        
        
        dist = sqrt( (pt->x()-Corner->x())*(pt->x()-Corner->x()) + 
                     (pt->y()-Corner->y())*(pt->y()-Corner->y()) );
        
        
        // Yes, there are indeed straight edges in the neighbourhood
        // around the Harris corner point !
        if ( dist < MaxDist )
        {
          straight_edges = true;
          

          
          if (dist < min_dist)
          {
            min_dist = dist;
          }
          
          // If necessary, split lines in 2 sublines ...
          vdgl_digital_curve *l11, *l12, *l21, *l22;

          
          if ( split_edgelchain_from_point(l1, l11, l12, pt) )
          {
            
            // In case both lines l1 and l2 could be
            // splitted succesfully
            if ( split_edgelchain_from_point(l2, l21, l22, pt) )
            {
              _Lines1->push_back( new vdgl_digital_curve(*l11) );
              _Lines2->push_back( new vdgl_digital_curve(*l21) );
              _centers->push_back( new point_t(*pt) );
							
              _Lines1->push_back( new vdgl_digital_curve(*l11) );
              _Lines2->push_back( new vdgl_digital_curve(*l22) );
              _centers->push_back( new point_t(*pt) );
							
              _Lines1->push_back( new vdgl_digital_curve(*l12) );
              _Lines2->push_back( new vdgl_digital_curve(*l22) );
              _centers->push_back( new point_t(*pt) );
							
              _Lines2->push_back( new vdgl_digital_curve(*l12) );
              _Lines2->push_back( new vdgl_digital_curve(*l21) );
              _centers->push_back( new point_t(*pt) );
              
              delete l11; delete l12; delete l21; delete l22;
            }
            
            // Only l1 could be splitted, but l2 not
            else
            {
              _Lines1->push_back( new vdgl_digital_curve(*l11) );
              _Lines2->push_back( new vdgl_digital_curve(*l2) );
              _centers->push_back( new point_t(*pt) );
              
              _Lines1->push_back( new vdgl_digital_curve(*l12) );
              _Lines2->push_back( new vdgl_digital_curve(*l2) );
              _centers->push_back( new point_t(*pt) );
              
              delete l11; delete l12;
            }
          }
          else
          {
            // In case l2 can be splitted, but not l1
            if ( split_edgelchain_from_point(l2, l21, l22, pt) )
            {
              _Lines1->push_back( new vdgl_digital_curve(*l1) );
              _Lines2->push_back( new vdgl_digital_curve(*l21) );
              _centers->push_back( new point_t(*pt) );
              
              _Lines1->push_back( new vdgl_digital_curve(*l1) );
              _Lines2->push_back( new vdgl_digital_curve(*l22) );
              _centers->push_back( new point_t(*pt) );
              
              delete l21; delete l22;
            }
            
            // Neither l1 and l2 can be splitted
            else
            {
              _Lines1->push_back( new vdgl_digital_curve(*l1) );
              _Lines2->push_back( new vdgl_digital_curve(*l2) );
              _centers->push_back( new point_t(*pt) );
            }
          }
        }
        
#if defined (VERB_PARALL)
        else 
        {
          vcl_cout << "  too far away from corner (" << dist << ")"<< vcl_endl;
        }
#endif
        
      }
    }
  }
  
  delete pt; pt = NULL;
  
  // select only the points that could be the closest
  // (assuming noise on the position of the corner of 2 pixel)
  vdgl_digital_curve *line;
  
  curves_list_t::iterator Line1It, Line2It;
  
  Line1It = _Lines1->begin();
  Line2It = _Lines2->begin();


  // Temporary iterators needed, since removal of elements
  // of a list invalidates the iterators pointing to the elements
  // being removed
  points_list_t::iterator tmp_Pi;  
  curves_list_t::iterator tmp_Line1It, tmp_Line2It;
 
  for (points_list_t::iterator Pi = _centers->begin();
       Pi != _centers->end(); ){

    pt = *Pi;
    
    if ( ABL_DIST(pt->x(), pt->y(), Corner->x(), Corner->y()) > min_dist + 2)
    {
      // remove point from list and move to next point
      tmp_Pi = Pi;
      
      Pi = _centers->erase(tmp_Pi);
      delete pt;

      // remove line from list and move to next line
      tmp_Line1It = Line1It;
      line = *Line1It;
      
      Line1It = _Lines1->erase(tmp_Line1It);
      delete line;

      // remove line from list and move to next line
      tmp_Line2It = Line2It;
      line = *Line2It;
     
      Line2It = _Lines2->erase(tmp_Line2It);
      delete line;
    }   
    else
    {
      // just move to next point and lines
      ++Line1It; ++Line2It; ++Pi;
    }
  }
  
  delete sub_lines; sub_lines = NULL;
  


  
#if defined (VERB_PARALL)
  if ( _verbose && straight_edges )
  {
    vcl_cout << "_centers: " << _centers->size() << vcl_endl;
    vcl_cout << "_Lines1 : " << _Lines1->size() << vcl_endl;
    vcl_cout << "_Lines2 : " << _Lines2->size() << vcl_endl;
  }
#endif

  
  return straight_edges;
}





// ----------------------------------------------------------------------
// This function computes the intersection of the lines l1 and l2.
// It assumes that 
// they are lines, so there is no error checking done for that.
// ----------------------------------------------------------------------
bool parall_region_extractor::intersect_lines(vdgl_digital_curve *l1, vdgl_digital_curve *l2, 
                                           point_t *pt)
{

  vsol_point_2d_sptr p10 = l1->p0();
  vsol_point_2d_sptr p11 = l1->p1();
  
  vsol_point_2d_sptr p20 = l2->p0();
  vsol_point_2d_sptr p21 = l2->p1();


  point_t hp10(p10->x(), p10->y(), 1.0);
  point_t hp11(p11->x(), p11->y(), 1.0);

  point_t hp20(p20->x(), p20->y(), 1.0);
  point_t hp21(p21->x(), p21->y(), 1.0);

  vgl_homg_operators_2d<double> op;
  vgl_homg_line_2d<double> hl1 = op.join(hp10, hp11);
  vgl_homg_line_2d<double> hl2 = op.join(hp20, hp21);

  point_t intersection = op.intersection(hl1, hl2);
  
  
  pt->set(intersection.y()/intersection.w(), intersection.x()/intersection.w());

  return true;
}




// ----------------------------------------------------------------------
// Split an edgel chain closest to a point
// ----------------------------------------------------------------------
bool parall_region_extractor::split_edgelchain_from_point(vdgl_digital_curve *l2,
                                                     vdgl_digital_curve*& l21, 
                                                     vdgl_digital_curve*& l22,
                                                     point_t *pt)
{
  //  First we find the nearest point on the curve to the pt
  vdgl_edgel ClosestEdgel;
  int ClosestEdgelIndex = 0;
  
  double Dist, MinDist = 1000000000;
	
  vdgl_edgel_chain_sptr Chain = l2->get_interpolator()->get_edgel_chain();
  int Size = Chain->size();

  for(int k=0; k<Size; k++)
  {
    vdgl_edgel CurrentEdgel = Chain->edgel(k);
    Dist = sqrt( (pt->x()-CurrentEdgel.x())*(pt->x()-CurrentEdgel.x()) + 
                 (pt->y()-CurrentEdgel.y())*(pt->y()-CurrentEdgel.y()) );

    if ( Dist < MinDist )
    {
      MinDist = Dist;
      ClosestEdgel = CurrentEdgel;
      ClosestEdgelIndex = k;
    }
  }


  // Minimum edge length for splitted edges
  int MinSide = _starter->get_params()->MinSideLength;


  // In case the splitted edge is too short, forget it
  if ( ClosestEdgelIndex < MinSide || Size - ClosestEdgelIndex < MinSide )
  {
    return false;
  }


  // Otherwise, create the new edges
  vcl_vector<vdgl_edgel> NewChain;
  for(int k=0; k<ClosestEdgelIndex; k++)
  {
    NewChain.push_back(Chain->edgel(k));
  }
	
  vdgl_interpolator_linear *Interp = new vdgl_interpolator_linear(new vdgl_edgel_chain(NewChain));
  l21 = new vdgl_digital_curve(Interp);

	
  vcl_vector<vdgl_edgel> NewChain2;
  for(int k=ClosestEdgelIndex; k<Size; k++)
  {
    NewChain2.push_back(Chain->edgel(k));
  }
	
  vdgl_interpolator_linear *Interp2 = new vdgl_interpolator_linear(new vdgl_edgel_chain(NewChain2));
  l22 = new vdgl_digital_curve(Interp2);
	
  return true;
}



bool parall_region_extractor::split_color_bands(const vil1_memory_image_of<vil1_rgb_byte>& buf, 
						vil1_memory_image_of<byte>& R_buf,
						vil1_memory_image_of<byte>& G_buf, 
						vil1_memory_image_of<byte>& B_buf)
{
  if ( R_buf.width() != buf.width() || R_buf.height() != buf.height() )
  {
    return false;
  }

  if ( G_buf.width() != buf.width() || G_buf.height() != buf.height() )
  {
    return false;
  }

  if ( B_buf.width() != buf.width() || B_buf.height() != buf.height() )
  {
    return false;
  }

  for(int k=0; k<buf.width(); k++)
  {
    for(int j=0; j<buf.height(); j++)
    {
      vil1_rgb_byte Pixel = buf(k,j);
      R_buf(k,j) = Pixel.R();
      G_buf(k,j) = Pixel.G();
      B_buf(k,j) = Pixel.B();
    }
  }

  return true;
}



bool parall_region_extractor::maximum_region(const point_t &ctr,
                                             vdgl_digital_curve& l1, 
                                             vdgl_digital_curve& l2,
                                             point_t*& p1, 
                                             point_t*& p2)
{


#if defined VERB_PARALL
    vcl_cout << "make maximum region" << vcl_endl;
#endif

  
  vil1_image *buf = _starter->get_image_buffer();
  int sizex = buf->width();
  int sizey = buf->height();

  int MaxSize = _starter->get_params()->MaxSideLength;

  
  // --------------------------------------------------
  // Set Pt1
  // Depending on the Euclidean distance from straight line
  // intersection point to the line endpoints
  // Note: vdgl_digital_curve endpoint coords need to be
  //       exchanged (x <-> y)
  // --------------------------------------------------
  if (ABL_DIST(ctr.x(), ctr.y(),l1.p0()->y(),l1.p0()->x()) >
      ABL_DIST(ctr.x(),ctr.y(),l1.p1()->y(),l1.p0()->x()) )
  {
    p1 = new point_t(l1.p0()->y(), l1.p0()->x(), 1);
  }
  else
  {
    p1 = new point_t(l1.p1()->y(), l1.p1()->x(), 1);
  }



  float dist = ABL_DIST(ctr.x(), ctr.y(), p1->x(), p1->y());


  // Scale Pt1 to maximum
  p1->set(ctr.x() + (p1->x() - ctr.x()) * (float)MaxSize / dist,
          ctr.y() + (p1->y() - ctr.y()) * (float)MaxSize / dist);

  
	
  // --------------------------------------------------
  // Set Pt2
  // Dito as for Pt1
  // --------------------------------------------------
  if (ABL_DIST(ctr.x(), ctr.y(),l2.p0()->y(),l2.p0()->x()) >
      ABL_DIST(ctr.x(),ctr.y(),l2.p1()->y(),l2.p0()->x()) )
  {
    p2 = new point_t(l2.p0()->y(), l2.p0()->x(), 1);
  }
  else
  {
    p2 = new point_t(l2.p1()->y(), l2.p1()->x(), 1);
  }

  
  dist = ABL_DIST(ctr.x(), ctr.y(), p2->x(), p2->y());


  
  // Scale Pt2 to maximum
  p2->set(ctr.x() + (p2->x() - ctr.x()) * (float)MaxSize / dist,
          ctr.y() + (p2->y() - ctr.y()) * (float)MaxSize / dist);


#if defined (DEBUG_GL)
//   glEnable(GL_BLEND);
  
//   glColor4f(0, 1.0, 1.0, 0.25);
//   glPointSize(8);
  
//   glBegin(GL_POINTS);
//   glVertex2f(p2->x(), sizey-p2->y());
//   glEnd();
  
//   glDisable(GL_BLEND);
#endif 
	
	
  // shrink until first quadrant fits in the image
  while ( ( p1->x() > sizex-1) ||
	  ( p1->x() < 0)      ||
	  ( p1->y() > sizey-1) ||
	  ( p1->y() < 0)      ||
		
	  ( p2->x() > sizex-1) ||
	  ( p2->x() < 0)      ||
	  ( p2->y() > sizey-1) ||
	  ( p2->y() < 0)      ||
		
	  ( p1->x() + p2->x() - ctr.x() > sizex -1) ||
	  ( p1->x() + p2->x() - ctr.x() < 0      ) ||
	  ( p1->y() + p2->y() - ctr.y() > sizey -1) ||
	  ( p1->y() + p2->y() - ctr.y() < 0      ) )
  {
    p1->set(ctr.x() + 0.9 * (p1->x() - ctr.x()), ctr.y() + 0.9 * (p1->y() - ctr.y()));
    p2->set(ctr.x() + 0.9 * (p2->x() - ctr.x()), ctr.y() + 0.9 * (p2->y() - ctr.y()));
  }
	

#if defined (DEBUG_GL)
//   glColor4f(0, 1.0, 1.0, 1.0);

//   glBegin(GL_LINES);
//   glVertex2f(ctr.x(), sizey - ctr.y());
//   glVertex2f(p1->x(), sizey - p1->y());
//   glEnd();

//   glColor4f(1,1,1,1);
  
//   glBegin(GL_LINES);
//   glVertex2f(ctr.x(), sizey - ctr.y());
//   glVertex2f(p2->x(), sizey - p2->y());
//   glEnd();
#endif 
  
  return true;
}
