#include "reg_ext_starter.h"
#include <region_extraction/reg_ext_params.h>

#include <osl/osl_harris.h>
#include <osl/osl_canny_ox.h>
#include <osl/osl_canny_ox_params.h>
#include <osl/osl_edge.h>
#include <osl/osl_fit_lines.h>
#include <osl/osl_fit_lines_params.h>
#include <conversions/geometry/osl_to_vdgl.h>

#include <vgl/vgl_homg_point_2d.h>
#include <vdgl/vdgl_digital_curve.h>
#include <vdgl/vdgl_fit_line.h>

#include <vsol/vsol_group_2d.h>

#include <vil1/vil1_image_as.h>
#include <vil1/vil1_pixel.h>
#include <vil1/vil1_save.h>

#include <vnl/vnl_vector.h>
#include <vnl/vnl_matrix.h>
#include <vnl/algo/vnl_svd.h>

#include <misc_utility/defines.h>
#include <misc_utility/abl_utility.h>
#include <misc_utility/gaussian_smoothing.h>

#include <cstring>
#include <ctime>


//-----------------------------------------------------------------------------
//-- Constructor
//   We wanted to keep buf const, but this results in a conflict with the
//   Harris class as its constructor discards the const qualifier.
//-----------------------------------------------------------------------------
reg_ext_starter::reg_ext_starter(const vil1_image& im,
                                 reg_ext_params* p)
{
  if (vil1_pixel_format(im) == VIL1_RGB_BYTE)
    _image = new color_buffer_t(im);
  else
    _image = new bw_buffer_t(im);
  _byte_buf = new bw_buffer_t(vil1_image_as_byte(*_image));

  _corners = NULL;
  _extrema = NULL;

  _edgebuf = NULL;
  _linebuf = NULL;
	
  _edges = NULL;
  _lines = NULL;

  _verbose = false;
	
  _params = p;	//  Note that _params does not own this memory.
}

//-----------------------------------------------------------------------------
//-- Destructor
//-----------------------------------------------------------------------------
reg_ext_starter::~reg_ext_starter()
{	
  if ( _corners )
  {
    for (points_list_t::iterator it = _corners->begin();
         it != _corners->end(); ++it)
    {
      delete *it;
    }
    _corners->clear();
    delete _corners;
  }

  if ( _extrema )
  {
    for (points_list_t::iterator it1 = _extrema->begin();
         it1 != _extrema->end(); ++it1)
    {
      delete *it1;
    }
    _extrema->clear();
    delete _extrema;
  }

  if ( _edges )
  {
    for(curves_list_t::iterator it2 = _edges->begin(); it2 != _edges->end(); ++it2)
    {
      delete *it2;
    }
    _edges->clear();
    delete _edges;
  }

  if ( _lines )
  {
    for(curves_list_t::iterator LineIt = _lines->begin();
        LineIt != _lines->end(); ++LineIt)
    {
      delete *LineIt;
    }
    _lines->clear();
    delete _lines;
  }

	
  if ( _edgebuf )
  {
    delete _edgebuf;
  }

  if ( _linebuf )
  {
    delete _linebuf; 
  }

  if ( _image )
  {
    delete _image;
  }

  if ( _byte_buf )
  {
    delete _byte_buf;
  }
}




//-----------------------------------------------------------------------------
//-- Harris corner extraction 
//-----------------------------------------------------------------------------
bool reg_ext_starter::find_harris_corners()
{
  // Prepare corner list
  if ( _corners )
  {
    //  Shitty delete that must be done for vcl_vector
    for (points_list_t::iterator it = _corners->begin();
         it != _corners->end(); ++it)
    {
      delete *it;
    }
    _corners->clear();
  }
  else
  {
    _corners = new points_list_t;
  }
	
  osl_harris_params hparams = _params->params_Harris;
  osl_harris *harris = new osl_harris(hparams);
	
  // Actual corner detection
  harris->compute(*_byte_buf);
  
  // Get results
  vcl_vector<float> X,Y;
  harris->get_corners(X, Y);
  delete harris;
	
  // Convert them to list
  for(unsigned k=0; k<X.size(); k++)
  {
    _corners->push_back((new point_t(X[k], Y[k])));
  }
	
  return true;
} 




//-----------------------------------------------------------------------------
//-- Extraction of intensity extrema using Vittorio's highly optimized
//   version.
//-----------------------------------------------------------------------------
bool reg_ext_starter::find_intensity_extrema()
{
  // Prepare corner list
  if ( _extrema )
  {
    for (points_list_t::iterator it = _extrema->begin(); it != _extrema->end(); ++it)
    {
      delete *it;
    }
    _extrema->clear();
  }
  else
  {
    _extrema = new points_list_t;
  }
	
  int sizex = _byte_buf->width();
  int sizey = _byte_buf->height();
  int size = _params->size_IntExtrema; // neighborhood side length
  int k;
	
  // Smooth the buffer
  bw_buffer_t tmp_buf(sizex, sizey);
  vic_gaussian_smoothing(*_byte_buf,tmp_buf);

  int **hor_min = new int*[sizex];  // could be made 'unsigned'
  int **hor_max = new int*[sizex];

  for(k=0; k<sizex; k++)
  {
    hor_min[k] = new int[sizey];
    hor_max[k] = new int[sizey];

    memset(hor_min[k], 0, sizey*sizeof(int));
    memset(hor_max[k], 0, sizey*sizeof(int));
  }
	
  // General algo description : check every pixel for extrema
  // a pixel is an extrema if it has the max|min intensity
  // of its 21x21 neighborhood
	
  // first part of optimised algo: prepare a buffer
  // where each pixel tells the x coord of its block's
  // maximum (prepare also a minimum buffer)
#if defined VERB_STARTER
    vcl_cout << "First part: hor_min|max buffer filling" << vcl_endl;
#endif

    //clock_t start,end;
    //start = clock();
  for (int j=0; j<sizey; j++) 
  {
    for (int i=size/2; i<sizex-size/2; i++) 
    {
			
      // long version only at first pixel of each line,
      // and when the previous block minimum is exactly
      // on the block's first pixel
      bool dotest = true;
			
      if(i == size/2) 
      {
        dotest = true;
      }
      else if (hor_max[i-1][j] == i-size/2-1 || 
	       hor_min[i-1][j] == i-size/2-1)
      {
        dotest = true;
      }
      else // can do fast min/max
      {
        if( tmp_buf(hor_max[i-1][j], j) >= tmp_buf(i+size/2, j) ) 
        {
          // >= because must be biased toward the left (to match original algo)
          hor_max[i][j] = hor_max[i-1][j];
        }
        else
        {
          hor_max[i][j] = i+size/2;
        }
				
        if( tmp_buf(hor_min[i-1][j], j) <= tmp_buf(i+size/2, j))
        {
          hor_min[i][j] = hor_min[i-1][j];
        }
        else
        {
          hor_min[i][j] = i+size/2;
        }
				
        dotest = false;
      }
			
      if(dotest) 
      {
        byte maximum = 0;
        byte minimum = 255;
        int maxi = i-size/2;
        int mini = i-size/2;
				
        for (int ii=i-size/2; ii <= i+size/2; ii++) 
        {
          if ( tmp_buf(ii, j) > maximum ) 
          {
            maximum = tmp_buf(ii, j);
            maxi = ii;
          }
					
          if ( tmp_buf(ii, j) < minimum ) 
          {
            minimum = tmp_buf(ii, j);
            mini = ii;
          }
        }
				
        hor_max[i][j] = (int) maxi;
        hor_min[i][j] = (int) mini;
				
      }
			
    }
  } 

	
  // Second part of optimised algo: minimize along
  // columns
#if defined VERB_STARTER
  vcl_cout << "Second part: vertical max/min-imisation" << vcl_endl;
#endif
  
  //bool okmin, okmax;
  for (int i=size/2; i<sizex-size/2; i++) 
  {
    for (int j=size/2; j<sizey-size/2; j++) 
    {	
      byte maximum = 0;
      byte minimum = 255;
      int maxj = j-size/2;
      int minj = j-size/2;
			
      // check for max in all lines only if it's max in this line
      if(hor_max[i][j] == i) 
      {
        //maximum = tmp_buf(i, j);
        //maxj = j;
        //okmax = true;
        for (int jj=j-size/2; jj <= j+size/2; jj++) 
        {
          if( tmp_buf(hor_max[i][jj], jj) > maximum ) 
          {
            //okmax = false;
            maxj = jj;
            maximum = tmp_buf(hor_max[i][jj], jj);
            //if(maxj != j) break;
          }
        }
      }
			
      // check for min in all lines only if it's min in this line
      if(hor_min[i][j] == i) 
      {
        //minimum = tmp_buf(i, j);
        //minj = j;
        //okmin = true;
        for (int jj=j-size/2; jj <= j+size/2; jj++) 
        {
          if( tmp_buf(hor_min[i][jj], jj) < minimum ) 
          {
            //okmin = false;
            minj = jj;
            minimum = tmp_buf(hor_min[i][jj], jj);
            //if(minj != j) break;
          }
        }
      }
			
      if( (minimum != 255 && minj == j && hor_min[i][minj] == i) || 
          (maximum != 0 && maxj == j && hor_max[i][maxj] == i)) 
	//if (okmax || okmin && minimum != 255 && maximum != 0)
      {
        point_t *pt = new point_t(i, j, 1.0);
        _extrema->push_back(pt);
      }
			
    }
  }
 
  //end = clock();
  //vcl_cout << "Found " << _extrema->size() << " extremas" << vcl_endl;
  //vcl_cout << "Took " << double(end-start)/CLOCKS_PER_SEC << vcl_endl;

  for(k=0; k<sizex; k++)
  {
    delete [] hor_min[k];
    delete [] hor_max[k];
  }

  delete [] hor_min; delete [] hor_max;
	
  return true;
} 




//-----------------------------------------------------------------------------
//-- Edge detection
//-----------------------------------------------------------------------------
bool reg_ext_starter::find_edges()
{
  // Call Canny and write to the EdgelChain list edges
  osl_canny_ox_params Params;
	
  Params.border_size = _params->border_size_CannyOX;
  Params.border_value = _params->border_value_CannyOX;
  Params.edge_min = _params->edge_min_CannyOX;
  Params.follow_strategy = _params->follow_strategy_CannyOX;
  Params.gauss_tail = _params->gauss_tail_CannyOX;
  Params.high = _params->high_CannyOX;
  Params.join_flag = _params->join_flag_CannyOX;
  Params.junction_option = _params->junction_option_CannyOX;
  Params.low = _params->low_CannyOX;
  Params.max_width = _params->max_width_CannyOX;
  Params.min_length = _params->min_length_CannyOX;
  Params.scale = _params->scale_CannyOX;
  Params.sigma = _params->sigma_CannyOX;

  osl_canny_ox *filter = new osl_canny_ox(Params);
	
  // Prepare edge list
  if ( !_edges )
  {
    _edges = new curves_list_t;
  }
  else
  {
    for (curves_list_t::iterator it = _edges->begin();
         it != _edges->end(); ++it){
      
      delete *it;
    }
    
    _edges->clear();
    delete _edges;
    _edges = new curves_list_t;
    
  }


  //////////////////////////////////////////////////
  _o_edges = new vcl_list<osl_edge*>;
  //////////////////////////////////////////////////
  
  // Edge detection
  vcl_list<osl_edge*> *TempEdges = new vcl_list<osl_edge*>;
  if (_image)
    filter->detect_edges(*_image, TempEdges);
  else
    filter->detect_edges(*_byte_buf, TempEdges);
  delete filter;


  // Line fitting
  vcl_list<osl_edge*> *TempLines = new vcl_list<osl_edge*>;
  osl_fit_lines_params FitParams;
  osl_fit_lines LineFit(FitParams);
  LineFit.incremental_fit_to_list(TempEdges, TempLines);


  
  vcl_list<osl_edge*>::iterator Iter = TempLines->begin();
  while( Iter != TempLines->end() )
  {
    osl_edgel_chain Chain = *(*Iter);

    if ( Chain.size() > _params->MinSideLength )
    {
      vdgl_digital_curve *NewChain = new vdgl_digital_curve( osl_to_vdgl(Chain) );
      _edges->push_back( NewChain );

      //////////////////////////////////////////////////
      osl_edge* o_e = new osl_edge(**Iter);
      _o_edges->push_back(o_e);
      //////////////////////////////////////////////////
    }
		
    Iter = TempLines->erase(Iter);
  }

  vcl_list<osl_edge*>::iterator  DEI;
  for(DEI = TempEdges->begin(); DEI != TempEdges->end(); ++DEI)
  {
    delete *DEI;
  }
  for(DEI = TempLines->begin(); DEI != TempLines->end(); ++DEI)
  {
    delete *DEI;
  }
  TempLines->clear(); TempEdges->clear();
  delete TempLines; delete TempEdges;
	
#if defined VERB_STARTER
  vcl_cout << _edges->size() << " edges found" << vcl_endl;
#endif

  //create_edge_buffer();
	
  return true;
}




//-----------------------------------------------------------------------------
//-- Line fitting
//-----------------------------------------------------------------------------
bool reg_ext_starter::fit_lines_to_edges()
{
  // Do we have edges at all ?
  if ( !_edges )
  {
    find_edges();
  }
  
  //  First we prepare the _lines variable
  if ( _lines )
  {
    for(curves_list_t::iterator it = _lines->begin();
        it != _lines->end(); ++it)
    {
      delete *it;
    }
    _lines->clear();
    delete _lines;
    
    _lines = new curves_list_t;
  }
  else
  {
    _lines = new curves_list_t;
  }
  
  //  Now we go through the current edges and, if they are actually lines, we make lines of them
  //  It will find any lines that lie along the edge and push them onto the list.
  for(curves_list_t::iterator EdgeIter = _edges->begin();
      EdgeIter != _edges->end(); ++EdgeIter)
  {
    vdgl_digital_curve *Curve = *EdgeIter;
    find_lines_from_curve(Curve);
  }
  
  return true;
}





//-----------------------------------------------------------------------------
//-- Extraction of edge buffer
//-----------------------------------------------------------------------------
bool reg_ext_starter::create_edge_buffer()
{
  // Do we have edges at all ?
  if ( !_edges )
    find_edges();

	
  int sizex = _byte_buf->width();
  int sizey = _byte_buf->height();
	

#if defined VERB_STARTER
    vcl_cout << "Extracting edge buffer.." << vcl_endl;
#endif
	
	
  if ( _edgebuf )
    _edgebuf->fill(NULL);
  else {
    _edgebuf = new curve_buffer_t(sizex,sizey);
    _edgebuf->fill(NULL);
  }


  
	
  // Each buffer pixel that is covered by an edge points to the particular
  // edge in memory.
  for(curves_list_t::iterator it = _edges->begin();
      it != _edges->end(); ++it ){

    vdgl_digital_curve *edge = *it;
    vdgl_edgel_chain_sptr Chain = edge->get_interpolator()->get_edgel_chain();

    int length = Chain->size();
		
    for (int i=0; i<length; i++){

      vdgl_edgel Edgel = Chain->edgel(i);


      // Range check needed because the interpolator yields
      // edgel locations that are slightly outside the image
      // boundaries
      int xc = (int)Edgel.get_x();
      int yc = (int)Edgel.get_y();

      if ( xc >=0 && xc < sizey && yc >= 0 && yc < sizex)
        (*_edgebuf)(yc, xc) = edge;
    }
  }
	
	
#if defined VERB_STARTER
  vcl_cout << "done !" << vcl_endl;
#endif
	
  return true;
}





//-----------------------------------------------------------------------------
//-- Extraction of line buffer
//-----------------------------------------------------------------------------
bool reg_ext_starter::create_line_buffer()
{
  // Do we have edges at all ?
  if ( !_lines )
  {
    fit_lines_to_edges();
  }
	
  int sizex = _byte_buf->width();
  int sizey = _byte_buf->height();
	
#if defined VERB_STARTER
  vcl_cout << "Extracting line buffer..";
#endif
	
	
  if ( _linebuf )
    _linebuf->fill(NULL);
  else{
    _linebuf = new curve_buffer_t(sizex, sizey);
    _linebuf->fill(NULL);
  }

  
	
  // Each buffer pixel that is covered by a fitted line points to the particular
  // line in memory.
  for (curves_list_t::iterator it = _lines->begin();
       it != _lines->end(); ++it){
    
    vdgl_digital_curve* line = *it;
    vdgl_edgel_chain_sptr Chain = line->get_interpolator()->get_edgel_chain();
    int L = Chain->size();

    for(int k=0; k<L; k++)
    {
      vdgl_edgel Edgel = Chain->edgel(k);
      //_linebuf->SetElement((int)Edgel.get_y(), (int)Edgel.get_x(), line);

      //////////////////////////////////////////////////
      // PROBLEM HERE ??
      //////////////////////////////////////////////////
      (*_linebuf)((int)Edgel.get_y(), (int)Edgel.get_x()) = line;
    }
  }
	
	
#if defined VERB_STARTER
  vcl_cout << "done !" << vcl_endl;
#endif
	
  return true;
}


// ----------------------------------------------------------------------
//  This function will make new vdgl_digital_curves along a line from
//  the Curve that is entered . We will do a least squares test to see
//  if there is a line or not (algebraic error)
//  ----------------------------------------------------------------------
bool reg_ext_starter::find_lines_from_curve(vdgl_digital_curve *Curve)
{
  int k, j, MinLineLength = _params->MinSideLength;
  double Error, ErrorThresh = 0.5;

  vdgl_edgel_chain_sptr Chain = Curve->get_interpolator()->get_edgel_chain();

  int ChainLength = Chain->size();

  if ( ChainLength < MinLineLength )
  {
    return true;
  }

  vnl_vector<double> X(ChainLength), Y(ChainLength);
  vnl_matrix<double> A(MinLineLength, 2);
  vnl_vector<double> TempY(MinLineLength);

  for(k=0; k<ChainLength; k++)
  {
    vdgl_edgel Edgel = Chain->edgel(k);
    X(k) = Edgel.x();
    Y(k) = Edgel.y();
  }

  //int CurrentIndex = 0;

  for(k=0; k<=ChainLength-MinLineLength; k++)
  {
    //  The line starts with the current index
    int StartIndex = k;

    //  Now we need to find the end of the line.  It is at least of length 15, but we try to grow it out 
    //  further.
    bool ErrorSmallEnough = true;
    bool EndOfChain = !( StartIndex+MinLineLength < ChainLength );
		
    if ( !EndOfChain )
    {
      int EndIndex = StartIndex + MinLineLength;

      vdgl_edgel StartEdgel = Chain->edgel(StartIndex);
      vdgl_edgel EndEdgel = Chain->edgel(StartIndex + MinLineLength);
      double StartX = StartEdgel.x();
      double StartY = StartEdgel.y();
      double EndX = EndEdgel.x();
      double EndY = EndEdgel.y();
      double CurrentErrorVal, MaxErrorVal = -1.0;

      for(j=0; j<MinLineLength; j++)
      {
        vdgl_edgel TempEdgel = Chain->edgel(StartIndex+j);
        CurrentErrorVal = DistanceFromAPointToALine(StartX, StartY, EndX, EndY, 
                                                    TempEdgel.x(), TempEdgel.y());
        if ( CurrentErrorVal > MaxErrorVal )
        {
          MaxErrorVal = CurrentErrorVal;
        }
      }

      if ( MaxErrorVal < ErrorThresh && !EndOfChain )
      {
        while( ErrorSmallEnough && !EndOfChain )
        {
          int CurrentLength = EndIndex - StartIndex + 1;
          vdgl_edgel CurrentEdgel = Chain->edgel(StartIndex+CurrentLength);

          Error = DistanceFromAPointToALine(StartX, StartY, EndX, EndY, CurrentEdgel.x(), 
                                            CurrentEdgel.y());
          ErrorSmallEnough = ( Error < ErrorThresh );
					
          if ( ErrorSmallEnough )
          {
            EndIndex++;
            EndOfChain = !( EndIndex < ChainLength );
          }
        }

        vcl_vector<vdgl_edgel> edgels;
        for(j=StartIndex; j<EndIndex; j++)
        {
          vdgl_edgel Edgel = Chain->edgel(j);
          edgels.push_back(Edgel);
        }

        vdgl_digital_curve *NewChain = new vdgl_digital_curve(new vdgl_interpolator_linear(new vdgl_edgel_chain(edgels)));
        _lines->push_back(NewChain);

        k = EndIndex + 1;
      }

    }

  }

  return true;
}
