#ifndef reg_ext_starter_h
#define reg_ext_starter_h

//------------------------------------------------------------------------------
// .NAME reg_ext_starter - Class for geometric preprocessing before region extraction
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   reg_ext_starter.h
// .FILE   reg_ext_starter.C
//
// .SECTION Description
//    The class reg_ext_params takes care of some geometric preprocessing needed  
//    for region extraction (Harris Corner detection, Canny osl_edge detection, Linefitting, 
//    Creation of an osl_edge-buffer and line-buffer for the geometry-based region extractors;
//    Intensity extrema detection for the intensity-based region extractors).
//
//    reg_ext_starter is meant to be a member variable of regionExtractor,
//    but it can also be used on its own.
//
// 
// .SECTION Author:
//          Tinne Tuytelaars,  Andreas Turina, Vittorio Ferrari
//          ESAT / KULeuven            BIWI / ETH
//
// .SECTION Modifications:
//
//
//
//--------------------------------------------------------------------------------

/*******************************************************************************
*
*	Ported to vxl by Aaron Luttman
*	Notes:	This class owns the following memory: _bute_buf, _edgebuf, _linebuf, 
*	_extrema, _corners, _edges, _lines, _image.
*
*******************************************************************************/

#include <misc_utility/abl_utility.h>
#include <misc_utility/defines.h>
#include "reg_ext_params.h"

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vgl/vgl_homg_point_2d.h>
#include <vdgl/vdgl_digital_curve.h>
#include <vcl/vcl_list.h>


template class vil1_memory_image_of<vdgl_digital_curve*>;


typedef vil1_memory_image_of<vdgl_digital_curve*> curve_buffer_t;




//////////////////////////////////////////////////
class osl_edge;
//////////////////////////////////////////////////

class reg_ext_starter
{
public:
  reg_ext_starter(const vil1_image &im,
                  reg_ext_params* p);
  ~reg_ext_starter();
	
	
  // Spatial information extraction
  inline reg_ext_params *get_params() { return _params; }
  inline points_list_t* get_intensity_extrema();  // copy vcl_list or return (unsafe) pointer ?
  inline points_list_t* get_harris_corners();     // copy vcl_list or return (unsafe) pointer ?
  inline curves_list_t* get_canny_edges();           // copy vcl_list or return (unsafe) pointer ?
  inline curves_list_t* get_lines();                // copy vcl_list or return (unsafe) pointer ?
         

  inline curve_buffer_t* get_edge_buffer();
  inline curve_buffer_t* get_line_buffer();
  
        
  inline vil1_image* get_image_buffer() { return _image; }
  inline vil1_memory_image_of<byte>* get_BW_image_buffer() { return _byte_buf; }

  // FOR DEBUGGING
  inline vcl_list<osl_edge*>* return_o_edges();
        
	
protected:
  reg_ext_params* _params;
  bool find_intensity_extrema();
  bool find_harris_corners();
  bool find_edges();
  bool fit_lines_to_edges();
  bool create_edge_buffer();
  bool create_line_buffer();

  
  void vic_gaussian_smooth(const vil1_memory_image_of<unsigned char> &in,
                         vil1_memory_image_of<unsigned char> &out);
  bool find_lines_from_curve(vdgl_digital_curve *Curve);
	
private:
  vil1_image *_image;
  vil1_memory_image_of<unsigned char> *_byte_buf;
	

  vil1_memory_image_of<vdgl_digital_curve*>* _edgebuf;
  vil1_memory_image_of<vdgl_digital_curve*>* _linebuf;
        
	
  points_list_t* _extrema;
  points_list_t* _corners;
	
  curves_list_t* _edges;	// These automatically get fit to lines 
  curves_list_t* _lines;	//  These are the implicit lines that
  // actually matter for the region finding

  // FOR DEBUG PURPOSES
  vcl_list<osl_edge*>* _o_edges;
        
	
  bool _verbose;
	
};


// --------------------------------------------------
// Inline functions
// --------------------------------------------------

points_list_t* reg_ext_starter::get_intensity_extrema()
{
  if (!_extrema)
    find_intensity_extrema();
  
  return _extrema;
}




points_list_t* reg_ext_starter::get_harris_corners()
{
  if (!_corners)
    find_harris_corners();
  
  return _corners;
}




curves_list_t* reg_ext_starter::get_canny_edges()
{
  if (!_edges)
    find_edges();
	
  return _edges;
}



//////////////////////////////////////////////////
vcl_list<osl_edge*>* reg_ext_starter::return_o_edges()
{
  return _o_edges;
  
}
//////////////////////////////////////////////////



curves_list_t* reg_ext_starter::get_lines()
{
  if (!_lines)
    fit_lines_to_edges();
  
  return _lines;
}






curve_buffer_t* reg_ext_starter::get_edge_buffer()
{
  if (!_edgebuf) 
    create_edge_buffer();
  
  return _edgebuf;
}





curve_buffer_t* reg_ext_starter::get_line_buffer()
{
  if (!_linebuf)
    create_line_buffer();
  
  return _linebuf;
}



#endif
