// straight_region_extractor.cpp: implementation of the straight_region_extractor class.
//
//////////////////////////////////////////////////////////////////////

#include "straight_region_extractor.h"
#include <misc_utility/defines.h>
#include <region_utility/region_deskew.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_pixel.h>
#include <vsol/vsol_point_2d.h>


#include <fstream>


#if defined (DEBUG_GL)
#include <GL/glut.h>
#endif



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

straight_region_extractor::straight_region_extractor(reg_ext_starter *starter)
  : parall_region_extractor(starter)
{

}




straight_region_extractor::~straight_region_extractor()
{

}




bool straight_region_extractor::extract_regions(point_t* corner,
						region::region_specific_type tp)
{
  vdgl_digital_curve *l1, *l2;
  point_t* ctr;
  parall_region* reg;

  int MinSize = _starter->get_params()->MinSideLength;
  int MaxSize = _starter->get_params()->MaxSideLength;

  compute_lines(corner);

  curves_list_t::iterator it1, it2;
  points_list_t::iterator PointIt = _centers->begin();
  points_list_t::iterator ScalesIter;

  vil1_image *buf = _starter->get_image_buffer();

  it1 = _Lines1->begin();
  it2 = _Lines2->begin();

  while ( !_Lines1->empty() )
  {	
    l1 = *it1;	it1 = _Lines1->erase(it1);
    l2 = *it2;	it2 = _Lines2->erase(it2);
    ctr = *PointIt;	
    PointIt = _centers->erase(PointIt);
		
    // Set p1 and p2 to maximum region (2*MAXSIZE)*(2*MAXSIZE)
    maximum_region(*ctr, *l1, *l2, p1, p2);

    
    // switch if necessary
    if ((p1->x()-ctr->x()) * (p2->y()-ctr->y()) - (p1->y()-ctr->y()) * (p2->x()-ctr->x()) < 0.0)
    {
      point_t* tmp = p1;
      p1 = p2;
      p2 = tmp;
    }

    //deskew region
    reg = new parall_region(ctr, p1, p2);

        
    if ( reg->good_region(ABL_SQUARE(MinSize)*ABL_PI*2, 1000, 1000000000) ) // reject if too small
    {
      
      
    int dist1 = (int)ABLround(ABL_DIST(ctr->x(), ctr->y(), p1->x(), p1->y()));
    int dist2 = (int)ABLround(ABL_DIST(ctr->x(), ctr->y(), p2->x(), p2->y()));
      

    // or, equivalently: if(_ColorImage)
    //vil1_pixel_format_t T = vil1_pixel_format(*buf);
    //if (T == VIL1_RGB_BYTE) {
      //vcl_cout << "Before deskewing" << vcl_endl;
      ds_buf = deskew_region(reg, (vil1_memory_image_of<vil1_rgb_byte>*)buf, dist1, dist2);
      //exit(1);
      //}
      //else 
    /*{
      vcl_cout << "Error: straight_region_extractor doesn't know how to deal with images of type " 
	       << T << vcl_endl;
      return false;
    }*/
  
    delete reg;
			
			
    // search in 2D for apropriate regions
    vcl_list<int> temp_types;
    vcl_list<int>::iterator TempTypesIt = temp_types.begin();
    points_list_t scales = find_2d_scales(temp_types);
      

    
#if defined VERB_STRAIGHT
    vcl_cout << scales.size() << " = " << temp_types.size() << " different scales " << vcl_endl;
#endif

    
    // put result in regions
    if ( !scales.empty() )
    {
      float cx = ctr->x();
      float cy = ctr->y();
      int sx = ds_buf->width(); // 2*MAX_SIZE+1 or MAX_SIZE
      int sy = ds_buf->height(); // 2*MAX_SIZE+1 or MAX_SIZE
      
      float scx1 = (p1->x() - cx) / sx;
      float scy1 = (p1->y() - cy) / sx;
      float scx2 = (p2->x() - cx) / sy;
      float scy2 = (p2->y() - cy) / sy;
      
      TempTypesIt = temp_types.begin();
      for (ScalesIter = scales.begin(); ScalesIter != scales.end(); ++ScalesIter)
      {
        point_t* pt_s = *ScalesIter;
        int type = *TempTypesIt;


        
#if defined VERB_STRAIGHT
        vcl_cout << pt_s->x() << ", " << pt_s->y() << "         " << vcl_endl; 
#endif

          
        point_t* point1 = new point_t(cx + pt_s->x()*scx1, 
				      cy + pt_s->x()*scy1, 1);
        point_t* point2 = new point_t(cx + pt_s->y()*scx2, 
				      cy + pt_s->y()*scy2, 1);
#if defined VERB_STRAIGHT
          vcl_cout << point1->x() << ", " << point1->y() << "     " 
                   << point2->x() << ", " << point2->y() << vcl_endl;
#endif

          
        parall_region* region = 
	  new parall_region(ctr, point1, point2, 
			    (region::region_specific_type)type,
			    corner, _starter->get_image_buffer());
        
        bool reject = !region->good_region(2.0, ABL_SQUARE(MinSize)*ABL_PI*2, ABL_SQUARE(MaxSize)*ABL_PI/2.0);
        
        if ( (region->get_min_x() < 0) || (region->get_min_y() < 0) ||
             (region->get_max_x() >= buf->width()) ||
             (region->get_max_y() >= buf->height()))
        {
          reject = true;
        }
        
        if (!reject)
        {
          _regions.push_back(region);
        }
        else
        {
          delete region;


          
#if defined VERB_STRAIGHT
            vcl_cout << "region rejected " << vcl_endl;
#endif

            
        }
        
        delete pt_s;

        ++TempTypesIt;
      }
    }
    
    delete ds_buf;
    }
    else
    {
      delete reg;
    }
    
  }
  

  

#if defined VERB_STRAIGHT
    vcl_cout << "finished findregionStraightEdges" << vcl_endl;
    vcl_cout << _regions.size() << " regions found using straight edges" << vcl_endl;
#endif

    
  
  return true;
}





bool straight_region_extractor::extract_all_regions()
{
  points_list_t *HarrisCorners = _starter->get_harris_corners();

  for(points_list_t::iterator it = HarrisCorners->begin();
      it != HarrisCorners->end(); ++it)
  {
    extract_regions(*it);
  }

  return true;
}





double straight_region_extractor::get_interpolation(vil1_memory_image_of<float>& buf, double x, double y)
{
  int xl = (int)floor(x-0.000001); // +/- 0.000001 such that xl != xh, yl != yh
  int xh = (int)ceil(x+0.000001);
  int yl = (int)floor(y-0.000001);
  int yh = (int)ceil(y+0.000001);
	
  if ( ( xl >= 0) && (xh <= buf.width()-1) && (yl >= 0) && (yh <= buf.height()-1) )
  {
    return ( (xh-x) * ((yh-y)*buf(xl, yl) + (y-yl)*buf(xl, yh)) + (x-xl) * ((yh-y)*buf(xh, yl) + 
									    (y-yl)*buf(xh, yh))) /(xh-xl)/(yh-yl);
  }
	
  return 0.0; // point lies on the edge or outside the image
}



bool straight_region_extractor::compute_I_buffer(vil1_memory_image_of<float>& I_buf,
                                               vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
  int i,j;
	
  // initialize
  I_buf(0, 0) = (float)buf(0, 0);
  for (j=1; j<sx; j++)
  {
    I_buf(j, 0) = (float)buf(j, 0) + I_buf(j-1, 0);
  }
	
  for (j=1; j<sy; j++)
  {
    I_buf(0, j) = (float)buf(0, j) + I_buf(0, j-1);
  }
	
	
  // fill out
  for (i=1; i<sx; i++)
  {
    for (j=1; j<sy; j++)
    {
      I_buf(i, j) = (float)buf(i, j) + I_buf(i-1, j) + I_buf(i, j-1) - I_buf(i-1, j-1);
    }
  }

  return true;
}

bool straight_region_extractor::compute_II_buffer(vil1_memory_image_of<float>& II_buf,
                                                vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
  int i,j;
	
  // initialize
  II_buf(0, 0) = (float)buf(0, 0)*(float)buf(0,0);
  for (j=1; j<sx; j++)
  {
    II_buf(j, 0) = (float)buf(j, 0)*(float)buf(j,0) + II_buf(j-1, 0);
  }
	
  for (j=1; j<sy; j++)
  {
    II_buf(0, j) = (float)buf(0, j)*(float)buf(0,j) + II_buf(0, j-1);
  }
	
	
  // fill out
  for (i=1; i<sx; i++)
  {
    for (j=1; j<sy; j++)
    {
      II_buf(i, j) = (float)buf(i, j)*(float)buf(i,j) + II_buf(i-1, j) + II_buf(i, j-1) - II_buf(i-1, j-1);
    }
  }

  return true;
}

bool straight_region_extractor::compute_Ix_buffer(vil1_memory_image_of<float>& Ix_buf,
                                                vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
  int i,j;
	
  // initialize
  Ix_buf(0, 0) = 0.0;
  for (j=1; j<sy; j++)
  {
    Ix_buf(0, j) = 0.0;
  }
	
  for (j=1; j<sx; j++)
  {
    Ix_buf(j, 0) = (float)(buf(j, 0) * j) + Ix_buf(j-1, 0);
  }
	
  // fill out
  for (i=1; i<sx; i++)
  {
    for (j=1; j<sy; j++)
    {
      Ix_buf(i, j) = (float)(buf(i, j) * i) + Ix_buf(i-1, j) + Ix_buf(i, j-1) - Ix_buf(i-1, j-1);
    }
  }
	
  return true;
}

bool straight_region_extractor::compute_Iy_buffer(vil1_memory_image_of<float>& Iy_buf,
                                                vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
  int i,j;
	
  // initialize
  Iy_buf(0, 0) = 0.0;
	
  for (j=1; j<sy; j++)
  {
    Iy_buf(0, j) = (float)(buf(0, j) * j) + Iy_buf(0, j-1);
  }

  for (j=1; j<sx; j++)
  {
    Iy_buf(j, 0) = 0.0;
  }
	
  // fill out
  for (i=1; i<sx; i++)
  {
    for (j=1; j<sy; j++)
    {
      Iy_buf(i, j) = (float)(buf(i, j) * j) + Iy_buf(i-1, j) + Iy_buf(i, j-1) - Iy_buf(i-1, j-1);
    }
  }

  return true;
}

bool straight_region_extractor::compute_cog1_buffer(vil1_memory_image_of<float>& cog_buf, 
						vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
	
  vil1_memory_image_of<float> I_buf(sx, sy);
  vil1_memory_image_of<float> II_buf(sx, sy);
  vil1_memory_image_of<float> Ix_buf(sx, sy);
  vil1_memory_image_of<float> Iy_buf(sx, sy);
	
  compute_I_buffer(I_buf, buf);
  compute_II_buffer(II_buf, buf);
  compute_Ix_buffer(Ix_buf, buf);
  compute_Iy_buffer(Iy_buf, buf);
	
  compute_cog1_buffer(cog_buf, I_buf, II_buf, Ix_buf, Iy_buf);

  return true;
}

bool straight_region_extractor::compute_cog1_buffer(vil1_memory_image_of<float>& cog_buf, vil1_memory_image_of<float>& I_buf, 
						vil1_memory_image_of<float>& II_buf, vil1_memory_image_of<float>& Ix_buf, 
						vil1_memory_image_of<float>& Iy_buf)
{
  int sx = I_buf.width();
  int sy = I_buf.height();
  // double sc = sqrt(2);
  int i,j;
  int MinSize = _starter->get_params()->MinSideLength;
	
  // initialize
  for (i=0; i<sx; i++)
  {
    for (j=0; j<(MinSize-2); j++)
    {
      cog_buf(i, j) = 0.0;
    }
  }
	
  for (i=0; i<sy; i++)
  {
    for (j=0; j<(MinSize-2); j++)
    {
      cog_buf(j, i) = 0.0;
    }
  }
	
  // fill out
  double x,y;
  for (i=(MinSize-2); i<sx; i++)
  {
    for (j=(MinSize-2); j<sy; j++)
    {
      x = Ix_buf(i, j) / I_buf(i, j);
      y = Iy_buf(i, j) / I_buf(i, j);
      cog_buf(i, j) = fabs(x*j-y*i) / (float)((i+1)*(j+1));

      if ( _offset_invariants )
      {
	cog_buf(i, j) = cog_buf(i, j) / sqrt(II_buf(i, j) * i * j / I_buf(i,j) / I_buf(i, j) - 1);
      }
    }
  }

  return true;
}

bool straight_region_extractor::compute_cog2_buffer(vil1_memory_image_of<float>& cog_buf, 
						vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
	
  vil1_memory_image_of<float> I_buf(sx, sy);
  vil1_memory_image_of<float> II_buf(sx, sy);
  vil1_memory_image_of<float> Ix_buf(sx, sy);
  vil1_memory_image_of<float> Iy_buf(sx, sy);
	
  compute_I_buffer(I_buf, buf);
  compute_II_buffer(II_buf, buf);
  compute_Ix_buffer(Ix_buf, buf);
  compute_Iy_buffer(Iy_buf, buf);
	
  compute_cog2_buffer(cog_buf, I_buf, II_buf, Ix_buf, Iy_buf);

  return true;
}

bool straight_region_extractor::compute_cog2_buffer(vil1_memory_image_of<float>& cog_buf, vil1_memory_image_of<float>& I_buf,  
						vil1_memory_image_of<float>& II_buf, vil1_memory_image_of<float>& Ix_buf, 
						vil1_memory_image_of<float>& Iy_buf)
{
  int sx = I_buf.width();
  int sy = I_buf.height();
  double sc = sqrt(2);
  int i,j;
  int MinSize = _starter->get_params()->MinSideLength;
	
  // initialize
  for (i=0; i<sx; i++)
  {
    for (j=0; j<(int)((float)(MinSize-2)/sc); j++)
    {
      cog_buf(i, j) = 0.0;
    }
  }
	
  for (i=0; i<sy; i++)
  {
    for (j=0; j<(int)((float)(MinSize-2)/sc); j++)
    {
      cog_buf(j, i) = 0.0;
    }
  }
	
  // fill out
  double x,y;
  for (i=(int)ABLround((MinSize-2)/sc); i<sx; i++)
  {
    for (j=(int)ABLround((MinSize-2)/sc); j<sy; j++)
    {
      x = Ix_buf(i, j) / I_buf(i, j);
      y = Iy_buf(i, j) / I_buf(i, j);
      cog_buf(i, j) = fabs(x*j+y*i-i*j) / (float)((i+1)*(j+1));

      if ( _offset_invariants )
      {
	cog_buf(i, j) = cog_buf(i, j) / sqrt(II_buf(i, j) * i * j / I_buf(i,j) / I_buf(i, j) - 1);
      }
    }
  }

  return true;
}

bool straight_region_extractor::compute_cog123_buffers(vil1_memory_image_of<float>& cog1_buf, 
						   vil1_memory_image_of<float>& cog2_buf, 
						   vil1_memory_image_of<float>& cog3_buf, 
						   vil1_memory_image_of<byte>& buf)
{
  int sx = buf.width();
  int sy = buf.height();
	
  vil1_memory_image_of<float> I_buf(sx, sy);
  vil1_memory_image_of<float> II_buf(sx, sy);
  vil1_memory_image_of<float> Ix_buf(sx, sy);
  vil1_memory_image_of<float> Iy_buf(sx, sy);
	
  compute_I_buffer(I_buf, buf);
  compute_II_buffer(II_buf, buf);
  compute_Ix_buffer(Ix_buf, buf);
  compute_Iy_buffer(Iy_buf, buf);
	
  compute_cog123_buffers(cog1_buf, cog2_buf, cog3_buf, I_buf, II_buf, Ix_buf, Iy_buf);

  return true;
}

bool straight_region_extractor::compute_cog123_buffers(vil1_memory_image_of<float>& cog1_buf, 
						   vil1_memory_image_of<float>& cog2_buf, 
						   vil1_memory_image_of<float>& cog3_buf,
                                                   vil1_memory_image_of<float>& I_buf, 
						   vil1_memory_image_of<float>& II_buf,
                                                   vil1_memory_image_of<float>& Ix_buf, 
						   vil1_memory_image_of<float>& Iy_buf)
{
  int sx = I_buf.width();
  int sy = I_buf.height();
  double sc = sqrt(2);
  int i,j;
  int MinSize = _starter->get_params()->MinSideLength;
	
  // initialize
  for (i=0; i<sx; i++)
  {
    for (j=0; j<(int)ABLround((MinSize-2)/sc); j++)
    {
      cog1_buf(i, j) = 0.0;
      cog2_buf(i, j) = 0.0;
    }
  }
	
  for (i=0; i<sy; i++)
  {
    for (j=0; j<(int)ABLround((MinSize-2)/sc); j++)
    {
      cog1_buf(j, i) = 0.0;
      cog2_buf(j, i) = 0.0;
    }
  }
	
  // fill out
  double x,y, N;
  for (i=(int)ABLround((MinSize-2)/sc); i<sx; i++)
  {
    for (j=(int)ABLround((MinSize-2)/sc); j<sy; j++)
    {
      x = Ix_buf(i, j) / I_buf(i, j);
      y = Iy_buf(i, j) / I_buf(i, j);
      cog1_buf(i, j) = fabs(x*j-y*i) / (float)((i+1)*(j+1));
      cog2_buf(i, j) = fabs(x*j+y*i-i*j) / (float)((i+1)*(j+1));
			
      if ( _offset_invariants )
      {
	N = sqrt(II_buf(i, j) * (i+1) * (j+1) / I_buf(i,j) / I_buf(i, j) - 1);
	cog1_buf(i, j) = cog1_buf(i, j)/N;
	cog2_buf(i, j) = cog2_buf(i, j)/N;
      }
    }
  }
				
  // cog3_buf...
  for (i=0; i<sx; i++)
  {
    for (j=0; j<MinSize; j++)
    {
      cog3_buf(i, j) = 0.0;
    }
  }

  for (i=0; i<sy; i++)
  {
    for (j=0; j<MinSize; j++)
    {
      cog3_buf(j, i) = 0;
    }
  }
				
  for (i=MinSize; i<sx; i++)
  {
    for (j=MinSize; j<sy; j++)
    {
      cog3_buf(i, j) = get_interpolation(cog2_buf, i/sc, j/sc);
    }
  }
	
  return true;
}

//--- take intersection of cog-buffers --------------------------------------------

// ! precondition: 2 <= col < sizex-2
inline static bool vertical_edge(int row, int col, vil1_memory_image_of<float>& buffer, float thresh)
{
  return ( ( buffer(row, col) <= thresh) && ( buffer(row, col) <= buffer(row, col-1))    &&
	   ( buffer(row, col) <= buffer(row, col+1)) && ( buffer(row, col) / buffer(row, col-2) <= 0.5) && 
	   ( buffer(row, col) / buffer(row, col+2) <= 0.5) );
}

// ! precondition: 2 <= row < sizex-2
inline static bool horizontal_edge(int row, int col, vil1_memory_image_of<float>& buffer, float thresh)
{
  return ( ( buffer(row, col) <= thresh) && ( buffer(row, col) <= buffer(row-1, col)) &&
	   ( buffer(row, col) <= buffer(row+1, col)) && ( buffer(row, col) / buffer(row-2, col) <= 0.5) && 
	   ( buffer(row, col) / buffer(row+2, col) <= 0.5));
} 




// ! precondition: 2 <= row, col < sizex-2
inline static bool horver_edge(int row, int col, vil1_memory_image_of<float>& buffer, float thresh)
{
  return (vertical_edge(row, col, buffer, thresh) || horizontal_edge(row, col, buffer, thresh));
} 





points_list_t
straight_region_extractor::compare_cog_buffers(vil1_memory_image_of<float>& cog1_buf, 
                                           vil1_memory_image_of<float>& cog2_buf, 
                                           double thresh)
{
  int sx = cog1_buf.width();
  int sy = cog1_buf.height();
	
  int MinSize = _starter->get_params()->MinSideLength;
	
  points_list_t scales;


    
	
  for (int i=MinSize+2; i<sx-3; i++)      // sx-3 and sy-3 because of implementation of horver_edge
  {
    for (int j=MinSize+2; j<sy-3; j++)
    {
      if (horver_edge(i, j, cog1_buf, thresh))
      {
	if (horver_edge(i, j, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i, j, 1);
	  scales.push_back(pt);
	}
	else
	{
	  if (horver_edge(i, j-1, cog2_buf, thresh))
	  {
	    point_t* pt = new point_t(i, j-0.5, 1);
	    scales.push_back(pt);
	  }
					
	  if (horver_edge(i, j+1, cog2_buf, thresh))
	  {
	    point_t* pt = new point_t(i, j+0.5, 1);
	    scales.push_back(pt);
	  }
					
	  if  (horver_edge(i+1, j, cog2_buf, thresh))
	  {
	    point_t* pt = new point_t(i+0.5, j, 1);
	    scales.push_back(pt);
	  }
					
	  if  (horver_edge(i-1, j, cog2_buf, thresh))
	  {
	    point_t* pt = new point_t(i-0.5, j, 1);
	    scales.push_back(pt);
	  }
	}
      }
    }
  }
	
  // first and last column
  for (int i=MinSize; i<sx-3; i++)
  {
    if (horver_edge(i, MinSize, cog1_buf, thresh))
    {
      if (horver_edge(i, MinSize, cog2_buf, thresh))
      {
	point_t* pt = new point_t(i, MinSize, 1);
	scales.push_back(pt);
      }
      else
      {
	if (horver_edge(i, MinSize+1, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i, MinSize+0.5, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(i+1, MinSize, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i+0.5, MinSize, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(i-1, MinSize, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i-0.5, MinSize, 1);
	  scales.push_back(pt);
	}
      }
    }
		
    if (horver_edge(i, sy-3, cog1_buf, thresh))
    {
      if (horver_edge(i, sy-3, cog2_buf, thresh))
      {
	point_t* pt = new point_t(i, sy-3, 1);
	scales.push_back(pt);
      }
      else
      {
	if (horver_edge(i, sy-4, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i, sy-3.5, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(i+1, sy-3, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i+0.5, sy-3, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(i-1, sy-3, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(i-0.5, sy-3, 1);
	  scales.push_back(pt);
	}
      }
    }
		
  }
	
  // first and last row
  for (int j=MinSize; j<sy-3; j++)
  {
    if (horver_edge(MinSize, j, cog1_buf, thresh))
    {
      if (horver_edge(MinSize, j, cog2_buf, thresh))
      {
	point_t* pt = new point_t(MinSize, j, 1);
	scales.push_back(pt);
      }
      else
      {
	if (horver_edge(MinSize, j-1, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(MinSize, j-0.5, 1);
	  scales.push_back(pt);
	}
				
	if (horver_edge(MinSize, j+1, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(MinSize, j+0.5, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(MinSize+1, j, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(MinSize+0.5, j, 1);
	  scales.push_back(pt);
	}
      }
    }
		
    if (horver_edge(sx-3, j, cog1_buf, thresh))
    {
      if (horver_edge(sx-3, j, cog2_buf, thresh))
      {
	point_t* pt = new point_t(sx-3, j, 1);
	scales.push_back(pt);
      }
      else
      {
	if (horver_edge(sx-3, j-1, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(sx-3, j-0.5, 1);
	  scales.push_back(pt);
	}
				
	if (horver_edge(sx-3, j+1, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(sx-3, j+0.5, 1);
	  scales.push_back(pt);
	}
				
	if  (horver_edge(sx-4, j, cog2_buf, thresh))
	{
	  point_t* pt = new point_t(sx-3.5, j, 1);
	  scales.push_back(pt);
	}
      }
    }
  }
	
  filter_2d_scales(scales);


  return scales;
}





points_list_t
straight_region_extractor::find_2d_rgb_scales(vcl_list<int>& temp_types, bool SHOW)
{

#if defined VERB_STRAIGHT
    vcl_cout << "find 2D scales -- 1Q -- RGB" << vcl_endl;
#endif

  points_list_t scales;
	
  int sx = ds_buf->width();
  int sy = ds_buf->height();
	
  vil1_memory_image_of<byte> R_buf(sx, sy);
  vil1_memory_image_of<byte> G_buf(sx, sy);
  vil1_memory_image_of<byte> B_buf(sx, sy);
	
  vil1_memory_image_of<float> cogR1_buf(sx, sy);
  vil1_memory_image_of<float> cogR2_buf(sx, sy);
  vil1_memory_image_of<float> cogR3_buf(sx, sy);
  vil1_memory_image_of<float> cogG1_buf(sx, sy);
  vil1_memory_image_of<float> cogG2_buf(sx, sy);
  vil1_memory_image_of<float> cogG3_buf(sx, sy);
  vil1_memory_image_of<float> cogB1_buf(sx, sy);  
  vil1_memory_image_of<float> cogB2_buf(sx, sy);
  vil1_memory_image_of<float> cogB3_buf(sx, sy);
	
  // split color_buffer
  split_color_bands(*ds_buf, R_buf, G_buf, B_buf);
	
  // compute invariants
  compute_cog123_buffers(cogR1_buf, cogR2_buf, cogR3_buf, R_buf);
  compute_cog123_buffers(cogG1_buf, cogG2_buf, cogG3_buf, G_buf);
  compute_cog123_buffers(cogB1_buf, cogB2_buf, cogB3_buf, B_buf);
	
	
  // combine invariants to find scales
  float thresh = 0.5;


  // --------------------------------------------------
  // Red
  // --------------------------------------------------
#if defined VERB_STRAIGHT
  vcl_cout << "compare R1, R3..." << vcl_flush;
#endif

  
  points_list_t temp_scales = compare_cog_buffers(cogR1_buf, cogR3_buf, thresh);


#if defined VERB_STRAIGHT
  vcl_cout << "done" << vcl_endl;
#endif
  
  
  while ( !temp_scales.empty() )
  {
    scales.push_back(temp_scales.back());
    temp_scales.pop_back();
    
    temp_types.push_back((int)(region::PARALL_STRAIGHT_R));
  }


  
  // --------------------------------------------------
  // Green
  // --------------------------------------------------
#if defined VERB_STRAIGHT
  vcl_cout << "compare G1, G3..." << vcl_flush;
#endif


  temp_scales = compare_cog_buffers(cogG1_buf, cogG3_buf, thresh);

  
#if defined VERB_STRAIGHT
  vcl_cout << "done" << vcl_endl;
#endif

  
  while (!temp_scales.empty())
  {
    scales.push_back(temp_scales.back());
    temp_scales.pop_back();
    temp_types.push_back((int)(region::PARALL_STRAIGHT_G));
  }


  // --------------------------------------------------
  // Blue
  // --------------------------------------------------
#if defined VERB_STRAIGHT
  vcl_cout << "compare B1, B3..." << vcl_flush;
#endif

  
  temp_scales = compare_cog_buffers(cogB1_buf, cogB3_buf, thresh);


  
#if defined VERB_STRAIGHT
  vcl_cout << "done" << vcl_endl;
#endif

  
  
  while (!temp_scales.empty())
  {
    scales.push_back(temp_scales.back());
    temp_scales.pop_back();
    temp_types.push_back((int)(region::PARALL_STRAIGHT_B));
  }
	
  return scales;
}






points_list_t
straight_region_extractor::find_2d_zw_scales(vcl_list<int>& temp_types, bool SHOW)
{

  
#if defined VERB_STRAIGHT
  vcl_cout << "find 2D scales -- 1Q -- ZW" << vcl_endl;
#endif

  
  points_list_t scales;
  int sx = ds_buf->width();
  int sy = ds_buf->height();
	
  // compute invariants
  vil1_memory_image_of<float> cog1_buf(sx, sy);
  vil1_memory_image_of<float> cog2_buf(sx, sy);
  vil1_memory_image_of<float> cog3_buf(sx, sy);

  vil1_memory_image_of<byte> TempImage(vil1_image_as_byte(*ds_buf));     // greylevel conversion
  compute_cog123_buffers(cog1_buf, cog2_buf, cog3_buf, TempImage);

	
  // combine invariants to find scales
  float thresh = 0.2;
	
  points_list_t temp_scales = compare_cog_buffers(cog1_buf, cog3_buf, thresh);

  points_list_t::iterator it = temp_scales.begin();
  while ( !temp_scales.empty() )
  {
    scales.push_back(*it);
    it = temp_scales.erase(it);
    temp_types.push_back((int)(region::PARALL_STRAIGHT));
  }
	
  return scales;
}




// ----------------------------------------------------------------------
// Clean-up of all intersection points (they are 'blurred').
// For each point in the list, the function looks for all other points
// that are close, removes them from the list, computes the average location
// of the removed points and adds this average to a new list.
// ----------------------------------------------------------------------
void straight_region_extractor::filter_2d_scales(points_list_t& scales)
{


#if defined VERB_STRAIGHT
    vcl_cout << "filter 2Dscales ..." << vcl_endl;
#endif

  float sumx, sumy;
  float sumx2, sumy2;
  int nb;
	
  points_list_t new_scales;
  points_list_t temp_scales;
  points_list_t new_temp_scales;

  point_t* pt, *pit;
  

  while ( !scales.empty() ){

    
#if defined VERB_STRAIGHT
    vcl_cout << scales.size() << " elements in scales" << vcl_endl;
#endif

    
    // First we remove the first candidate in the scales list
    pt = scales.back();
    scales.pop_back();


    // For average location computation
    sumx = pt->x();
    sumy = pt->y();
    sumx2 = pt->x() * pt->x();
    sumy2 = pt->y() * pt->y();
    nb = 1;
    

    // We look if there are candidates close enough to pt in the
    // scales list. Any such candidates are removed and the average
    // location is updated.
    for (points_list_t::iterator it = scales.begin();
         it != scales.end(); ){
      pit = (*it);

      double d = ABL_DIST(pt->x(), pt->y(), pit->x(), pit->y());

      // This point is close enough
      if ( d <= 5 ){

        // Update the average
        sumx += pit->x();
        sumy += pit->y();
        sumx2 += pit->x() * pit->x();
        sumy2 += pit->y() * pit->y();
        nb++;


        // Remove from scales list and move to next point
        it = scales.erase(it);
        delete pit;        
      }
      else
      {
	// just move to next point
	++it;
      }
    }


    // Now that we have examined the entire candidates list,
    // we create a new point with the average position
    point_t* average = new
      point_t(sumx/nb, sumy/nb, 1.0);
    

    new_scales.push_back(average);
    delete pt;
    
  }
  
	
  // Transfer from new_scales to scales
  while ( !new_scales.empty() ){
    point_t* p = new_scales.back();
    new_scales.pop_back();

    scales.push_back(p);
  }
}



