#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>                   // greylevel conversion
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <vil1/vil1_load.h>
#include <vcl/vcl_list.h>
#include <vcl/vcl_iostream.h>

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <regions/ell_region.h>
#include <region_extraction/reg_ext_params.h>
#include <region_extraction/reg_ext_starter.h>
#include <region_extraction/ell_region_extractor.h>
#include <region_utility/old_region_file.h>
#include <region_utility/Vitto_region_file.h>
#include <region_matching/invariant_computor.h>

#ifdef PARALL_REGIONS
#include <region_extraction/straight_region_extractor.h>
#include <region_extraction/curved_region_extractor.h>
#endif

#include <region_utility/region_draw.h>         // color_share
#include <region_utility/misc.h>                // copy_region
#include <misc_utility/count_pixels.h>          // border_color
#include <misc_utility/gaussian_smoothing.h>    // vic_gaussian_smoothing
#include <misc_utility/rescaling.h>

#include <cstdlib>
#include "tuytelaars_region_extractor.h"

// Invariant computation parameters
bool COLOR = true;
bool SWITCH = false;
bool ROT_ELL = true;  // ell regions need rot invs, so they are less
bool ROT_PAR = false; // parallelogram regions have more invariants, as they are all non-rotation invariant (not needed)


TuyteRegionExtractor::TuyteRegionExtractor( const vil1_image *image, bool bNoback, bool bVerbose ) :
	theImage(image)
{
	verbose = bVerbose;
	noback = bNoback;
	org_buf = color_buffer_t(*image);
	bcol = vil1_rgb<unsigned char>(0,0,0);

	if(noback) {
		// Smooth image in the 3 color bands
		if(verbose)
			vcl_cout << "Smoothing image..." << vcl_endl;
		color_buffer_t buf_smooth(*theImage);
		vic_gaussian_smoothing(org_buf, buf_smooth);

		bcol = vil1_rgb<unsigned char>(border_color(&buf_smooth));
		if(verbose)
			vcl_cout << "Border color: " << (int)bcol.R() << " " << (int)bcol.G() << " " << (int)bcol.B() << vcl_endl;
		if(abs(bcol.R() - bcol.G()) > 15 || abs(bcol.B() - bcol.G()) > 15) {
			vcl_cout << "WARNING: Colored background, segmentation assumes it black." << vcl_endl;
			bcol = vil1_rgb<unsigned char>(0, 0, 0);
		}
		else {
			if(bcol.R() <= 100) {
				if(verbose)
					vcl_cout << "Black background detected" << vcl_endl;
				bcol = vil1_rgb<unsigned char>(0, 0, 0);
			}
			else {
				if(verbose)
					vcl_cout << "White background detected" << vcl_endl;
				bcol = vil1_rgb<unsigned char>(70, 70, 70);
			}
		}
	}
}


TuyteRegionExtractor::~TuyteRegionExtractor()
{
	
}


void TuyteRegionExtractor::extractEllRegions( regions_list_t &regions )
{
	int sizex = org_buf.cols();
	int sizey = org_buf.rows();
	float scale = 1.0;
	unsigned int iniRegions = regions.size();

	invariant_computor ic(&org_buf, COLOR, SWITCH, ROT_ELL);   // all invariants are computed on the original buffer, with the rgion scaled appropriately
	int NB_INV = ic.get_max_nb_inv();
	if(verbose) {
		vcl_cout << "Max number of invariants = " << NB_INV << vcl_endl;
		vcl_cout << "Number of invariants for ell regions: " << ic.get_nb_inv(COLOR, ROT_ELL) << vcl_endl;
	}

	// MAIN LOOP : ITERATE OVER ALL SCALES
	color_buffer_t* buf = new color_buffer_t(org_buf);     // current scale buffer, need a copy to avoid org_buf to be rescaled as well
	bool firsttime = true;
	while ((sizex > 200) && (sizey > 200) || firsttime) {
		firsttime = false;
		vcl_cout << "Current image size : " << sizex << "x" << sizey << vcl_endl;
		// extract regions
		vcl_cout << "Extracting regions..." << vcl_endl;
		reg_ext_params params; // use default parameters for the time being
		reg_ext_starter starter(*buf, &params);
		ell_region_extractor extractor(&starter);

		extractor.extract_all_regions();
		regions_list_t tmp_regions;
		extractor.get_regions(tmp_regions);

		// rescale regions to size of original image
		ell_region* reg;
		for (regions_list_t::iterator i=tmp_regions.begin(); i!=tmp_regions.end(); ++i) {
			reg = (ell_region*)(*i);

			point_t* npt0 = new point_t(reg->get_pt0()->x()/scale, 
			                            reg->get_pt0()->y()/scale);
			point_t* npt1 = new point_t(reg->get_pt1()->x()/scale, 
			                            reg->get_pt1()->y()/scale);
			point_t* npt2 = new point_t(reg->get_pt2()->x()/scale, 
			                            reg->get_pt2()->y()/scale);
			point_t* nanchor = new point_t(reg->get_anchor_point()->x()/scale,
			                               reg->get_anchor_point()->y()/scale);

			ell_region* nreg = new ell_region(npt0, npt1, npt2, 
			                                  reg->get_specific_type(), nanchor);

			if(!noback || color_share(nreg, bcol, &org_buf) < 0.4) {  // do not add background regions
				// compute invariants
				vnl_vector<double>* inv = new vnl_vector<double>(NB_INV, 0.0);
				ic.set_region(nreg);
				ic.get_invariants(*inv);
				nreg->set_invariants(inv); // makes no copy, so: do not 'delete inv' !

				// add nreg to list of extracted regions
				regions.push_back(nreg);      
			}
			else
				delete nreg;

			delete reg;
		}

		if(verbose) {
			vcl_cout << "*************************************************" << vcl_endl;
			vcl_cout << regions.size()-iniRegions << " regions extracted up to scale " << 1/scale << vcl_endl;
			vcl_cout << "*************************************************" << vcl_endl;
		}

		// rescale image for next run
		if ( (sizex*0.71 > 200) && (sizey*0.71 > 200) ) {
			color_buffer_t* nbuf = rescale_buffer(buf, 0.71);
			delete buf;
			buf = nbuf;
			scale *= 0.71;
			sizex = buf->cols();
			sizey = buf->rows();
		}
		else {
			sizex *= (int)0.71;
			sizey *= (int)0.71;
		}
	}

	delete buf;
}


#ifdef PARALL_REGIONS
void TuyteRegionExtractor::extractParRegions( regions_list_t &regions )
{
	// set parameters
	reg_ext_params params;                           // use default parameters for the time being

	// prepare to compute invariants
	invariant_computor ic(&org_buf, COLOR, SWITCH, ROT_PAR); // invariants computed always and anyway from the color buffer
	int NB_INV = ic.get_max_nb_inv();
	if(verbose) {
		vcl_cout << "Max number of invariants = " << NB_INV << vcl_endl;
		vcl_cout << "Number of invariants for parall regions: " << ic.get_nb_inv(COLOR, ROT_PAR) << vcl_endl;
	}

	// extract straight regions
	if(verbose)
		vcl_cout << "Extracting straight regions" << vcl_endl;
	reg_ext_starter starter(org_buf, &params);   
	straight_region_extractor s_extractor(&starter);
	s_extractor.extract_all_regions();
	regions_list_t newRegions;
	s_extractor.get_regions(newRegions);
	if(verbose)
		vcl_cout << newRegions.size() << " regions extracted" << vcl_endl;

	// extract curved regions
	if(verbose)
		vcl_cout << vcl_endl << "Extracting curved regions" << vcl_endl;
	curved_region_extractor c_extractor(&starter);
	c_extractor.extract_all_regions();
	c_extractor.get_regions(newRegions);
	if(verbose)
		vcl_cout << newRegions.size() << " total regions extracted" << vcl_endl;

	// Filter background regions
	if(verbose) {
		vcl_cout << "Computing invariants";
		if(noback) vcl_cout << " and filtering background regions.";
	}
	vcl_cout << vcl_endl;
	for (regions_list_t::iterator i=newRegions.begin(); i!=newRegions.end(); ++i) {
		parall_region* reg = (parall_region*)(*i);

		if(!noback || color_share(reg, bcol, &org_buf) < 0.4) { // do not add background regions
			point_t* npt0 = new point_t( *(reg->get_pt0()) );
			point_t* npt1 = new point_t( *(reg->get_pt1()) );
			point_t* npt2 = new point_t( *(reg->get_pt2()) );
			point_t* nanchor = new point_t( *(reg->get_anchor_point()) );
			parall_region* nreg = new parall_region(npt0, npt1, npt2,
			                                        reg->get_specific_type(), nanchor);

			vnl_vector<double>* inv = new vnl_vector<double>(NB_INV, 0.0);
			ic.set_region(nreg);
			ic.get_invariants(*inv);
			nreg->set_invariants(inv);                            // makes no copy, so: do not 'delete inv' !
			regions.push_back(nreg);
		}

		delete reg;
	} // loop over extracted regions
}

#else // ! PARALL_REGIONS
void TuyteRegionExtractor::extractParRegions( regions_list_t &regions )
{
	vcl_cerr << "Warning: parallel regions have not been enabled." << vcl_endl;
}

#endif // PARALL_REGIONS
