#include "classic_region_matcher.h"

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <regions/ell_region.h>
#include <region_matching/reg_match_params.h>
#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/rotation_finder.h>
#include <region_matching/matches.h>
#include <region_matching/match.h>
#include <region_matching/pairwise_indexed_matches.h>
#include <region_matching/pairwise_indexed_match.h>

#include <vcl/vcl_cmath.h>

#include <vil1/vil1_image_as.h>

#include <misc_utility/vic_math.h>


// -- Constructors / Destructors -----------------------------------

classic_region_matcher::classic_region_matcher(const vil1_image* im1,
					       const vil1_image* im2,
					       regions_list_t& reglist1, 
					       regions_list_t& reglist2,
					       similarity_computor* sc,
					       const reg_match_params* p)
  : region_matcher(im1, im2, reglist1, reglist2, p)
{
  if (vil1_pixel_format(*im1) == VIL1_RGB_BYTE)
  {
    _rgb = true;
    _grey1 = new bw_buffer_t(vil1_image_as_byte(*im1));
    _grey2 = new bw_buffer_t(vil1_image_as_byte(*im2));
  }
  else
  {
    _rgb = false;
    _grey1 = (bw_buffer_t*)im1;
    _grey2 = (bw_buffer_t*)im2;
  }

  // Set class-wide (static) members.
  pairwise_indexed_match::set_orig_regions_list(&_reglist1);
  pairwise_indexed_match::set_ref_regions_list(&_reglist2);

  _mtches = new pairwise_indexed_matches;
  _sc = sc;
  _sc->set_resolution(_params->similarity_resolution);

  _rotf = new rotation_finder(_grey1, _grey2);
  _rotf->set_resolution(_params->rotation_finder_resolution);
  _rotf->set_stepsize(_params->search_for_rotation_stepsize);

  // Find largest region type present
  int max_type = find_max_type();
  
  // Find the amount of invariants present
  int tot_invs = find_tot_invs();

  // Instantiate variance matrix
  V = new  vnl_matrix<float>(max_type+1, tot_invs, 0.0); // max_type rows x tot_invs column matrix full of 0s
}


int classic_region_matcher::find_max_type() {
  int max_type = 0;
  for(int rx=0; rx < _reglist1.size(); rx++) {
    int t = (int)(*(_reglist1[rx]->get_invariants_ptr()))[0];
    if(t > max_type) {
      max_type = t;
    }
  }
  for(int rx=0; rx < _reglist2.size(); rx++) {
    int t = (int)(*(_reglist2[rx]->get_invariants_ptr()))[0];
    if(t > max_type) {
      max_type = t;
    }
  }

  vcl_cout << "Maximal region type: " << max_type << vcl_endl;
  return max_type;
}


// General test, useful to know dimensionality of variance matrix,
// and to check out consistency of the input data
// This test is unexpensive, compared to the total running time of matching
// (it's even only a fraction of the time needed to load the regions from the disk)
int classic_region_matcher::find_tot_invs() {
  int tot_invs = 0;
  for(int rx=0; rx < _reglist1.size(); rx++) {
    int cur_invs = _reglist1[rx]->get_invariants_ptr()->size();

    if(rx==0) { 
      vcl_cout << "Number of invariants: " << cur_invs << vcl_endl;
      tot_invs = cur_invs; 
    }
    
    if(cur_invs != tot_invs) {
      vcl_cout << "WARNING: region id " << _reglist1[rx]->get_sid() << " has " << cur_invs
	       << " invariants, against " << tot_invs << " of the first one on list." << vcl_endl;
    }
  }
  for(int rx=0; rx < _reglist2.size(); rx++) {
    int cur_invs = _reglist2[rx]->get_invariants_ptr()->size();

    if(cur_invs != tot_invs) {
      vcl_cout << "WARNING: region id " << _reglist2[rx]->get_sid() << " has " << cur_invs
	       << " invariants, against " << tot_invs << " of the first one on list." << vcl_endl;
    }
  }

  return tot_invs;
}


classic_region_matcher::~classic_region_matcher() {
  if (_rgb)
  {
    delete _grey1;
    delete _grey2;
  }

  delete _mtches;
  delete _rotf;
  delete V;
}


bool classic_region_matcher::match_region1(region* reg1, region*& reg2, int& indx2) {
  return match_region_algo(reg1, reg2, _reglist2, indx2, false);
}


bool classic_region_matcher::match_region2(region* reg2, region*& reg1, int& indx1) {
  return match_region_algo(reg2, reg1, _reglist1, indx1, true);
}


// match region from reglist1 (general - two way)
// input: reg1
// output: reg2, indx
// fills in mtches member
bool classic_region_matcher::match_region(region* reg1, region*& reg2, int& indx2) {
  bool match_found = match_region_algo(reg1, reg2, _reglist2, indx2, false);
  if (!match_found) return false;
  
  int indx1 = 0;
  region* dummy = NULL;
  vcl_cout << "    ";
  match_found = match_region_algo(reg2, dummy, _reglist1, indx1, true);
  if (!match_found) return false;
  
  if (dummy == reg1) {
    pairwise_indexed_match* mtch = new pairwise_indexed_match(indx1, indx2);
    _mtches->add(mtch);
  }

  return true;
}


// Finds top few tentative matches of candiadate region reg1 in set reglist
// based soley on feature-vectors distances
// fills in tentative_* members
void classic_region_matcher::find_tentative_matches(region* reg1, regions_list_t& reglist) {
  float dist_thr = _params->max_dist;
  int nb_of_tentative_matches = _params->nb_of_tentative_matches;

  tentative_regions.clear();
  tentative_distances.clear();
  tentative_indices.clear();
  for (int i=0; i<nb_of_tentative_matches; i++) {
    tentative_regions.push_front(NULL);
    tentative_distances.push_front(dist_thr);
    tentative_indices.push_front(-1);
  }

  // run over all regions in reglist
  region* reg;
  region::region_specific_type reg1_type = reg1->get_specific_type();   // type of the candidate region
  const vnl_vector<double>* inv1 = reg1->get_invariants_ptr();          // invariants of the candidate region

  if(reg1->get_type() != region::ELL_REGIONTYPE) {
    vcl_cout << "P" << reg1_type << " ";
  } else {
    vcl_cout << "E21 ";
  }

  float dist;
  int nb_regions = reglist.size();
  for (int indx=0; indx < nb_regions; ++indx) {
    reg = reglist[indx]; // next region to be investigated

    if (reg->get_specific_type() == reg1_type) {
      const vnl_vector<double>* inv = reg->get_invariants_ptr();
      dist = mahl_distance(inv1, inv);                     
    
      if (dist < dist_thr) {
	// add region to list of tentative regions
	vcl_list<region* >::iterator regions_it = tentative_regions.begin();
	vcl_list<float >::iterator distances_it = tentative_distances.begin();
	vcl_list<int >::iterator indices_it = tentative_indices.begin();

	// (could be done more efficient with binary search, or other indexing strategy)
	// however, the slowness of pixel-similarity later on renders this stage
	// computationally-neglectable
	while ( (regions_it != tentative_regions.end()) ) {
	  if (dist < *(distances_it) ) {
	    tentative_regions.insert(regions_it, reg);
	    tentative_distances.insert(distances_it, dist);
	    tentative_indices.insert(indices_it, indx);
	    break;
	  }

	  ++regions_it;
	  ++distances_it;
	  ++indices_it;
	}

	// remove last element
	tentative_regions.pop_back();
	tentative_distances.pop_back();
	tentative_indices.pop_back();

	// adapt dist_thr to largest distance currently in tentative_distances
	dist_thr = tentative_distances.back();
      } // dist < dist_thr
    } // same type
  } // loop over reglist

} // find_tentative_matches


// Find top first most similar (pixel-based) region to reg1 in tentative_*, with sim > _params->min_sim
// Save it to reg2, and index to indx2
// two_to_one -> reg1 comes from image2, so need to swap calls to sim measures
void classic_region_matcher::find_best_sim_match(region* reg1, region*& reg2, int& indx2, bool two_to_one) {
  float sim_thr = _params->min_sim;
  reg2 = NULL;

  if ( (reg1->get_type() == region::ELL_REGIONTYPE &&
	_params->search_for_rotation) ) {
    ell_region *ereg1, *ereg2, *ereg;
    if(!two_to_one) {
      ereg1 = reg1->cast_to_ellregion();
      _rotf->set_region1(ereg1);
    } else {
      ereg2 = reg1->cast_to_ellregion();
      _rotf->set_region2(ereg2);
    }
    
    vcl_list<int >::iterator indices_it = tentative_indices.begin(); 
    for (vcl_list<region* >::iterator regions_it = tentative_regions.begin();
	 regions_it != tentative_regions.end(); ++regions_it, ++indices_it) {
      region* reg = *(regions_it);
      if (!reg) break; // no more tentative regions below dist_thr
      vcl_cout << "."; // evry time I see a dot, then a reg was below dist threshold, a actually entered the sim test
      
      ell_region* ereg = reg->cast_to_ellregion();
      if(!two_to_one) {
	_rotf->set_region2(ereg);
      } else {
	_rotf->set_region1(ereg);
      }
      _rotf->get_crosscorrelation();                                 // causes rotation of _region1 member of _rotf. 'Lowcost' as greyvalue ncc, potentially lowres
      float sim = 0.0;
      if(!two_to_one) {
	sim = _sc->get_similarity(_rotf->get_rotated_region1(), ereg); //higher cost (eg _sc of type drgbnorm), but only once (rotf applied about 72)
      } else {
	sim = _sc->get_similarity(_rotf->get_rotated_region1(), ereg2);
      }
      // not 100% efficient as if _sc is of type drgbnorm, it computes the means once again for region1, even thought they didn't change ...

      if (sim > sim_thr) {
	reg2 = reg;
	sim_thr = sim;
	indx2 = *(indices_it);
	vcl_cout << "!";
      }
   } // loop over tentative_*
 } else { // parall region or elliptical one without need to search for rotation
   vcl_list<int >::iterator indices_it = tentative_indices.begin(); 
   for (vcl_list<region* >::iterator regions_it = tentative_regions.begin();
	regions_it != tentative_regions.end(); ++regions_it, ++indices_it) {
     region* reg = *(regions_it);
     if (!reg) break; // no more tentative regions below dist_thr
     vcl_cout << "."; // eveyr time I see a dot, then a reg was below dist threshold, a actually entered the sim test
     
     float sim = 0.0;
     if(!two_to_one)
       sim = _sc->get_similarity(reg1, reg);
     else
       sim = _sc->get_similarity(reg, reg1);

     if (sim > sim_thr) {
       reg2 = reg;
       sim_thr = sim;
       indx2 = *(indices_it);
       vcl_cout << "!";
     }
   } // loop over tentative_*
 } // reg type check

  if (reg2)
    vcl_cout << reg2->get_sid() << "   (" << sim_thr << ")";
  vcl_cout << vcl_endl;
} // find_best_sim_match


// top first best match to reg1 is saved to reg2
// reg2 is searched for in reglist
// _reglist2[indx2] = reg2
bool classic_region_matcher::match_region_algo(region* reg1, region*& reg2, regions_list_t& reglist, int& indx2, bool two_to_one) {
  vcl_cout << reg1->get_sid() << " ";

  // STEP 1: select the N best matches based on Mahalanobis distance only.
  // These are stored in tentative_* members
  find_tentative_matches(reg1, reglist);

  // STEP 2: select the region with largest sim value from
  // the list of tentative regions
  find_best_sim_match(reg1, reg2, indx2, two_to_one);
       
  return (reg2 != NULL);
}


// get variances of every component of feat vect for all reg types
// consider only the _reglist2 (to mimic TrJ original version)
bool classic_region_matcher::compute_statistics() {
  // Instantiate average matrix (rows = region types; cols = invariant type)
  vnl_matrix<float>* avg = new vnl_matrix<float>(V->rows(), V->cols(), 0.0);

  // Number of regions per type (integer, but made float to avoid warnings)
  vnl_vector<float>* N = new vnl_vector<float>(V->rows(), 0.0);
  
  // Collect sums of features for each type
  for(int rx=0; rx < _reglist2.size(); rx++) {
    const vnl_vector<double>* v = _reglist2[rx]->get_invariants_ptr();
    //vnl_vector<float>* invs =  cast_to_vector_float(v);        // COMP ERR: onvalid const conv ??
    vnl_vector<float>* invs = new vnl_vector<float>(v->size());  // waste a copy, but need to cast all elements to float from double
    for(int i=0; i < v->size(); i++) {
      (*invs)[i] = (*v)[i];
    }
    int type = (int)(*invs)[0];
    
    avg->set_row(type, avg->get_row(type) + *invs);                   
    V->set_row(type, V->get_row(type) + element_product(*invs, *invs));   // not the fastest due to reinstantiations+copying due to operator overloading ?
                                                                          // could probably be faster 'by hand'
    (*N)[type] = (*N)[type] + 1.0;

    delete invs;
  }

  // Complete variance matrix V
  for (int type = 0; type < V->rows(); type++) {
    avg->set_row(type, avg->get_row(type) / (*N)[type]);
    V->set_row(type, (V->get_row(type) / (*N)[type]) - element_product(avg->get_row(type), avg->get_row(type)));
    for (int i=0; i<V->cols(); i++) {
      if ((*V)(type, i) < 0.01)                                            // an invariant with such a small covariance is useless
	(*V)(type,i) = 0.01;
    }
  }

  // Debug output
  vcl_cout << "Variance matrix: " << vcl_endl << *V << vcl_endl;

  // Notice how V(:,0) always is 0.01, as the first invariant is the region type and it is of course constant within the class of regions with a certain type
  // V(u,:) = NaN for all types u not present in the data
  delete avg;
  delete N;
}
