#ifndef _classic_region_matcherh_
#define _classic_region_matcherh_


//------------------------------------------------------------------------------
// .NAME classic_region_matcher - Class for matching 
//       two lists of affinely invariant regions based on 
//       Mahalanobis distance between invariants and 
//       a pixel-wise similarity measure
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   classic_region_matcher.h
// .FILE   classic_region_matcher.C
//
// .SECTION Description
//    Find matches in two lists of affinely invariant regions based on 
//    Mahalanobis distance between invariants and 
//    a pixel-wise similarity measure
//     
//  
// .SECTION Author:
//          Tinne Tuytelaars,   Vittorio Ferrari
//          ESAT / KULeuven           ETHZ
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/reg_match_params.h>
#include <region_matching/region_matcher.h>
#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/rotation_finder.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_memory_image_of.h>

#include <vnl/vnl_matrix.h>



class classic_region_matcher : public region_matcher
{
  // PUBLIC INTERFACE----------------------------------------------------------
public:
  // Constructors/Destructors--------------------------------------------------
  classic_region_matcher(const vil1_image* im1,
			 const vil1_image* im2,
			 regions_list_t& reglist1, 
			 regions_list_t& reglist2,
			 similarity_computor* sc,
			 const reg_match_params* p);  
  ~classic_region_matcher();  
  
  // Utility Methods-----------------------------------------------------------
  virtual bool match_region(region* reg1, region*& reg2, int& indx2); // match region from reglist1 (general-two way)
  bool match_region1(region* reg1, region*& reg2, int& indx2); // match region from reglist1
  bool match_region2(region* reg2, region*& reg1, int& indx1); // match region from reglist2 

  // Mahlanobis distance related ----------------------------------------------
  bool compute_statistics();
  int find_max_type();
  int find_tot_invs();
  inline float mahl_distance(const vnl_vector<double>* inv1, const vnl_vector<double>* inv2);

protected:
  // Helpers ------------------------------------------------------------------
  void find_tentative_matches(region* reg1, regions_list_t& reglist);
  void find_best_sim_match(region* reg1, region*& reg2, int& indx2, bool two_to_one);
  virtual bool match_region_algo(region* reg1, region*& reg2, regions_list_t& reglist, int& indx2, bool two_to_one);


  bool _rgb; // needed to free memory in destructors
  bw_buffer_t* _grey1;
  bw_buffer_t* _grey2;

  similarity_computor* _sc;
  rotation_finder* _rotf;

  vnl_matrix<float>* V;                     // features variances ('invariants variances' ;-)

  vcl_list<region*> tentative_regions;      // temp, working variables
  vcl_list<float> tentative_distances;
  vcl_list<int> tentative_indices;
};


inline float classic_region_matcher::mahl_distance(const vnl_vector<double>* inv1, const vnl_vector<double>* inv2) {
  float dist = 0;
  int size = inv1->size();

  for(int i=0; i < size; ++i)
  {
    dist += ((*inv1)[i] - (*inv2)[i])*((*inv1)[i] - (*inv2)[i])/((*V)((int)(*inv1)[0],i));
  }

  return dist;
}

#endif // _classic_region_matcherh_
