#include "crosscorrelation_computor.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <region_utility/misc.h>               // in_image



// -- Constructors / Destructors ------------------------------
crosscorrelation_computor::crosscorrelation_computor(const color_buffer_t* buf1,
						     const color_buffer_t* buf2)
  : similarity_computor(buf1, buf2)
{

}


crosscorrelation_computor::crosscorrelation_computor(const bw_buffer_t* buf1,
						     const bw_buffer_t* buf2)
{
  _buffer1 = buf1;
  _buffer2 = buf2;
  _region1 = NULL;
  _region2 = NULL;
  _buf1 = NULL;            // no color buffers
  _buf2 = NULL; 
  _rgb = false;
  
  _sim_valid = false;
  _sim_sw_valid = false;
  _switched_valid = false;

  _res = 50;               // default value - can be changed by user
}

crosscorrelation_computor::crosscorrelation_computor(const color_buffer_t* buf1, 
						     const color_buffer_t* buf2, 
						     const region* r1, const region* r2)
  : similarity_computor(buf1, buf2, r1, r2)
{
  _rgb = true;
}

crosscorrelation_computor::crosscorrelation_computor(const bw_buffer_t* buf1, 
						     const bw_buffer_t* buf2, 
						     const region* r1, const region* r2)
{
  _buffer1 = buf1;
  _buffer2 = buf2;
  _buf1 = NULL;            // no color buffers
  _buf2 = NULL; 
  _region1 = r1;
  _region2 = r2;
  _rgb = false;            // we do not want the destructor of similarity_computor to delete the bw buffers
  
  _sim_valid = false;
  _sim_sw_valid = false;
  _switched_valid = false;

  _res = 50;              // default value - can be changed by user
}
 


crosscorrelation_computor::~crosscorrelation_computor() {
  // auto invoke destructor of similarity_computor,
  // which deletes bw buffers
}


// -- Computations -------------------------------------------------------

bool crosscorrelation_computor::compute_similarity()
{
  if(!in_image(_region1, _buffer1->width(), _buffer1->height()) || !in_image(_region2, _buffer2->width(), _buffer2->height())) {
    _sim = -1;
    _sim_valid = false;
    return false;
  }

  int nb_ = 0;
  float avg1_ = 0;
  float avg2_ = 0;
  float cor1_ = 0;
  float cor2_ = 0;
  float cor12_ = 0; 

  float cx1 = _region1->get_pt0()->x();
  float cy1 = _region1->get_pt0()->y();
  float cx2 = _region2->get_pt0()->x();
  float cy2 = _region2->get_pt0()->y();
  
  float x11 = _region1->get_pt1()->x();
  float y11 = _region1->get_pt1()->y();
  float x12 = _region1->get_pt2()->x();
  float y12 = _region1->get_pt2()->y();

  float x21 = _region2->get_pt1()->x();
  float y21 = _region2->get_pt1()->y();
  float x22 = _region2->get_pt2()->x();
  float y22 = _region2->get_pt2()->y();

  float incr1_x_a1, incr1_y_a1, incr1_x_a2, incr1_y_a2, xi1, yi1;
  float incr2_x_a1, incr2_y_a1, incr2_x_a2, incr2_y_a2, xi2, yi2;
  int resl1 = _res -1;

  if (_region1->get_type() != region::ELL_REGIONTYPE) // NB_QUAD == 1
  {
    // Region 1
    xi1 = cx1; yi1 = cy1;
    incr1_x_a1 = (float)(x11 - cx1) / resl1; // incrementals for axe 1
    incr1_y_a1 = (float)(y11 - cy1) / resl1;

    incr1_x_a2 = (float)(x12 - cx1) / resl1; // incrementals for axe 2
    incr1_y_a2 = (float)(y12 - cy1) / resl1;

    // Region 2
    xi2 = cx2; yi2 = cy2;
    incr2_x_a1 = (float)(x21 - cx2) / resl1; // incrementals for axe 1
    incr2_y_a1 = (float)(y21 - cy2) / resl1;

    incr2_x_a2 = (float)(x22 - cx2) / resl1; // incrementals for axe 2
    incr2_y_a2 = (float)(y22 - cy2) / resl1;


    float pix1, pix2;
    float cur1_x, cur1_y;                   // current position inside region 1
    float cur2_x, cur2_y;                   // current position inside region 2
    float xia1, yia1;                       // x,y init axis for region 1
    
    for (int i=0; i<_res; i++) {
      xia1 = xi1 + i*incr1_x_a1;            // cur_x|y now positioned to start of other axis
      yia1 = yi1 + i*incr1_y_a1;
      cur1_x = xia1;
      cur1_y = yia1;
      
      cur2_x = xi2 + i*incr2_x_a1;
      cur2_y = yi2 + i*incr2_y_a1;
      
      for (int j=0; j<_res; j++) {
	pix1 = get_interpolation_Vic(_buffer1, cur1_x, cur1_y);
	pix2 = get_interpolation_Vic(_buffer2, cur2_x, cur2_y);
	
	nb_++;
	avg1_ += pix1;
	avg2_ += pix2;
	cor1_ += pix1*pix1;
	cor2_ += pix2*pix2;
	cor12_ += pix1*pix2;
	
	// Position in region 1
	cur1_x += incr1_x_a2;
	cur1_y += incr1_y_a2;
	// Position in region 2
	cur2_x += incr2_x_a2;
	cur2_y += incr2_y_a2;
      }
    }
  }
  else { // NB_QUAD == 4
    // Region 1
    xi1 = cx1 - (x11 - cx1) - (x12 - cx1); // initial pt at the start of 4 quadrans
    yi1 = cy1 - (y11 - cy1) - (y12 - cy1);

    incr1_x_a1 = (float)(x11 - cx1)*2.0 / resl1; // incrementals for axe 1
    incr1_y_a1 = (float)(y11 - cy1)*2.0 / resl1;

    incr1_x_a2 = (float)(x12 - cx1)*2.0 / resl1; // incrementals for axe 2
    incr1_y_a2 = (float)(y12 - cy1)*2.0 / resl1;

    // Region 2
    xi2 = cx2 - (x21 - cx2) - (x22 - cx2); // initial pt at the start of 4 quadrans
    yi2 = cy2 - (y21 - cy2) - (y22 - cy2);

    incr2_x_a1 = (float)(x21 - cx2)*2.0 / resl1; // incrementals for axe 1
    incr2_y_a1 = (float)(y21 - cy2)*2.0 / resl1;

    incr2_x_a2 = (float)(x22 - cx2)*2.0 / resl1; // incrementals for axe 2
    incr2_y_a2 = (float)(y22 - cy2)*2.0 / resl1;


    //float i1, i2, j1, j2;
    float pix1, pix2;
    float cur1_x, cur1_y; // current position inside region 1
    float cur2_x, cur2_y; // current position inside region 2
    float xia1, yia1;                           // x,y init axis for region 1
    
    for (int i=0; i<_res; i++) {
      xia1 = xi1 + i*incr1_x_a1; // cur_x|y now positioned to start of other axe
      yia1 = yi1 + i*incr1_y_a1;
      cur1_x = xia1;
      cur1_y = yia1;
      
      cur2_x = xi2 + i*incr2_x_a1;
      cur2_y = yi2 + i*incr2_y_a1;

      for (int j=0; j<_res; j++) {
	if (_region2->inside_region(cur2_x, cur2_y)) { // sufficient to test one ellipse
	  pix1 = get_interpolation_Vic(_buffer1, cur1_x, cur1_y);  
	  pix2 = get_interpolation_Vic(_buffer2, cur2_x, cur2_y); 
	  
	  nb_++;
	  avg1_ += pix1;
	  avg2_ += pix2;
	  cor1_ += pix1*pix1;
	  cor2_ += pix2*pix2;
	  cor12_ += pix1*pix2;
	} 
	// Position in region 1
	cur1_x += incr1_x_a2;
	cur1_y += incr1_y_a2;
	// Position in region 2
	cur2_x += incr2_x_a2;
	cur2_y += incr2_y_a2;
      }
    }
  }

  if (nb_ == 0)
  {vcl_cout << "error in crosscorcompute_ns: nb_ = 0 " << vcl_endl;}
  

  avg1_ /= nb_;
  avg2_ /= nb_;

  cor1_ = cor1_ - nb_*avg1_*avg1_;
  cor2_ = cor2_ - nb_*avg2_*avg2_;
  cor12_ = cor12_ - nb_*avg1_*avg2_;
  cor12_ /= sqrt(cor1_ * cor2_);

  // If reg2 is 100% homogeneous -> cor2 = 0 -> _sim = inf|nan
  if(!(cor12_ >= -1 && cor12_ <= 1)) {
    cor12_ = -1;
  }

  _sim = cor12_;
  _sim_valid = true;

  return true;
}




bool crosscorrelation_computor::compute_similarity_switched()
{
  if(!in_image(_region1, _buffer1->width(), _buffer1->height()) || !in_image(_region2, _buffer2->width(), _buffer2->height())) {
    _sim = -1;
    _sim_valid = false;
    return false;
  }

  int nb_ = 0;
  float avg1_ = 0;
  float avg2_ = 0;
  float cor1_ = 0;
  float cor2_ = 0;
  float cor12_ = 0; 

  float cx1 = _region1->get_pt0()->x();
  float cy1 = _region1->get_pt0()->y();
  float cx2 = _region2->get_pt0()->x();
  float cy2 = _region2->get_pt0()->y();
  
  float x11 = _region1->get_pt1()->x();
  float y11 = _region1->get_pt1()->y();
  float x12 = _region1->get_pt2()->x();
  float y12 = _region1->get_pt2()->y();

  float x21 = _region2->get_pt1()->x();
  float y21 = _region2->get_pt1()->y();
  float x22 = _region2->get_pt2()->x();
  float y22 = _region2->get_pt2()->y();

  float incr1_x_a1, incr1_y_a1, incr1_x_a2, incr1_y_a2, xi1, yi1;
  float incr2_x_a1, incr2_y_a1, incr2_x_a2, incr2_y_a2, xi2, yi2;
  // float resol;
  int resl1 = _res -1;

  if (_region1->get_type() != region::ELL_REGIONTYPE) //NB_QUAD == 1
  {
    // Region 1
    xi1 = cx1; yi1 = cy1;
    incr1_x_a1 = (float)(x11 - cx1) / resl1; // incrementals for axe 1
    incr1_y_a1 = (float)(y11 - cy1) / resl1;

    incr1_x_a2 = (float)(x12 - cx1) / resl1; // incrementals for axe 2
    incr1_y_a2 = (float)(y12 - cy1) / resl1;

    // Region 2 (SWITCHED !!)
    xi2 = cx2; yi2 = cy2;
    incr2_x_a2 = (float)(x21 - cx2) / resl1; // incrementals for axe 1
    incr2_y_a2 = (float)(y21 - cy2) / resl1;

    incr2_x_a1 = (float)(x22 - cx2) / resl1; // incrementals for axe 2
    incr2_y_a1 = (float)(y22 - cy2) / resl1;

    float pix1, pix2;
    float cur1_x, cur1_y; // current position inside region 1
    float cur2_x, cur2_y; // current position inside region 2
    float xia1, yia1; // x,y init axis for region 1
    
    for (int i=0; i<_res; i++) {
      xia1 = xi1 + i*incr1_x_a1; // cur_x|y now positioned to start of other axe
      yia1 = yi1 + i*incr1_y_a1;
      cur1_x = xia1;
      cur1_y = yia1;
      
      cur2_x = xi2 + i*incr2_x_a1;
      cur2_y = yi2 + i*incr2_y_a1;
      
      for (int j=0; j<_res; j++) {
	pix1 = get_interpolation_Vic(_buffer1, cur1_x, cur1_y); 
	pix2 = get_interpolation_Vic(_buffer2, cur2_x, cur2_y); 
	
	nb_++;
	avg1_ += pix1;
	avg2_ += pix2;
	cor1_ += pix1*pix1;
	cor2_ += pix2*pix2;
	cor12_ += pix1*pix2;
	
	// Position in region 1
	cur1_x += incr1_x_a2;
	cur1_y += incr1_y_a2;
	// Position in region 2
	cur2_x += incr2_x_a2;
	cur2_y += incr2_y_a2;
      }
    }
  }
  else { //NB_QUAD == 4
    // Region 1
    xi1 = cx1 - (x11 - cx1) - (x12 - cx1); // initial pt at the start of 4 quadrans
    yi1 = cy1 - (y11 - cy1) - (y12 - cy1);

    incr1_x_a1 = (float)(x11 - cx1)*2.0 / resl1; // incrementals for axe 1
    incr1_y_a1 = (float)(y11 - cy1)*2.0 / resl1;

    incr1_x_a2 = (float)(x12 - cx1)*2.0 / resl1; // incrementals for axe 2
    incr1_y_a2 = (float)(y12 - cy1)*2.0 / resl1;

    // Region 2 (SWITCHED !!)
    xi2 = cx2 - (x21 - cx2) - (x22 - cx2); // initial pt at the start of 4 quadrans
    yi2 = cy2 - (y21 - cy2) - (y22 - cy2);

    incr2_x_a2 = (float)(x21 - cx2)*2.0 / resl1; // incrementals for axe 1
    incr2_y_a2 = (float)(y21 - cy2)*2.0 / resl1;

    incr2_x_a1 = (float)(x22 - cx2)*2.0 / resl1; // incrementals for axe 2
    incr2_y_a1 = (float)(y22 - cy2)*2.0 / resl1;

    float pix1, pix2;
    float cur1_x, cur1_y; // current position inside region 1
    float cur2_x, cur2_y; // current position inside region 2
    float xia1, yia1; // x,y init axis for region 1
    
    for (int i=0; i<_res; i++) {
      xia1 = xi1 + i*incr1_x_a1; // cur_x|y now positioned to start of other axe
      yia1 = yi1 + i*incr1_y_a1;
      cur1_x = xia1;
      cur1_y = yia1;
      
      cur2_x = xi2 + i*incr2_x_a1;
      cur2_y = yi2 + i*incr2_y_a1;
      
      for (int j=0; j<_res; j++) {
	if (_region2->inside_region(cur2_x, cur2_y)) { // sufficient to test one ellipse
	  pix1 = get_interpolation_Vic(_buffer1, cur1_x, cur1_y); 
	  pix2 = get_interpolation_Vic(_buffer2, cur2_x, cur2_y); 
	  
	  nb_++;
	  avg1_ += pix1;
	  avg2_ += pix2;
	  cor1_ += pix1*pix1;
	  cor2_ += pix2*pix2;
	  cor12_ += pix1*pix2;
	}
	  // Position in region 1
	  cur1_x += incr1_x_a2;
	  cur1_y += incr1_y_a2;
	  // Position in region 2
	  cur2_x += incr2_x_a2;
	  cur2_y += incr2_y_a2;
      }
    }
  }
    
  if (nb_ == 0)
  {
    vcl_cout << "error in crosscorcompute_sw: nb_ = 0 " << vcl_endl;
    return false;
  }
  

  avg1_ /= nb_;
  avg2_ /= nb_;
  
  cor1_ = cor1_ - nb_*avg1_*avg1_;
  cor2_ = cor2_ - nb_*avg2_*avg2_;
  cor12_ = cor12_ - nb_*avg1_*avg2_;
  cor12_ /= sqrt(cor1_ * cor2_);

  _sim_sw = cor12_;
  _sim_sw_valid = true;
  
  return true;
}

