#ifndef _drgbnorm_computer2h
#define _drgbnorm_computer2h

//------------------------------------------------------------------------------
// .NAME  drgbnorm_computor - Class for computing the average euclidean distance in
//                            normalized RGB space between two affinely invariant regions.
//                           
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   drgbnorm_computor.h
// .FILE   drgbnorm__computor.cpp
//
// .SECTION Description
//   drgbnorm_computor computes the average euclidean distance in normalized RGB space 
//   between two affinely invariant regions, after aligning them by the
//   appropriate affine transformation (assuming pt0, pt1 and pt2 correspond).
//   The mean of each colorband is set to half the range (128), via scalefactors
//   at pixel-reading timee, for both regions (normalization).
//   set_region{1,2} automatically update the scalefactors via 128/mean when any region changes
//   (transparent interface, little code and max speed, especially when used with refiner)
//
// .SECTION Author:
//          Vittorio Ferrari
//               ETHZ
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/drgb_computor.h>

class drgbnorm_computor : public drgb_computor {

 public:
  drgbnorm_computor() {}
  drgbnorm_computor(const color_buffer_t* buf1,
		    const color_buffer_t* buf2);
  drgbnorm_computor(const color_buffer_t* buf1, 
		    const color_buffer_t* buf2, 
		    const region* reg1, const region* reg2);
  ~drgbnorm_computor();

 public:
  inline float get_drgbnorm()           {return get_similarity();}           // syntactic sugar
  inline float get_drgbnorm_switched()  {return get_similarity_switched();}  // syntactic sugar

  inline virtual void set_region1(const region* r1);
  inline virtual void set_region2(const region* r2);


private:

  //--- helper functions -----------------------------------------------
  inline float dist_rgb_sf(vil1_rgb_byte& pix1, vil1_rgb_byte& pix2);

};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline void drgbnorm_computor::set_region1(const region* r1) {
  // DO NOT try to optimize here by computing means iff
  // r1!=_region1 -> pointer difference does not imply coordinates
  // differnce ! Source of bug that costed hours !
  float* M = compute_means(r1, _buf1);
  M[0] = 128.0 / M[0];
  M[1] = 128.0 / M[1];
  M[2] = 128.0 / M[2];
  set_scale_factors1(M);
  delete M;
  similarity_computor::set_region1(r1);
}


inline void drgbnorm_computor::set_region2(const region* r2) {
  float* M = compute_means(r2, _buf2);
  M[0] = 128.0 / M[0];
  M[1] = 128.0 / M[1];
  M[2] = 128.0 / M[2];
  set_scale_factors2(M);
  delete M;
  similarity_computor::set_region2(r2);
}


// now with scalefactors on both regions
inline float drgbnorm_computor::dist_rgb_sf(vil1_rgb_byte& pix1, vil1_rgb_byte& pix2) {
  float dr = pix1.R()*_sr1 - pix2.R()*_sr2;
  float dg = pix1.G()*_sg1 - pix2.G()*_sg2;
  float db = pix1.B()*_sb1 - pix2.B()*_sb2;

  return sqrt( dr*dr + dg*dg + db*db );
}

#endif
