#ifndef _drgbsnorm_computer2h
#define _drgbsnorm_computer2h

//------------------------------------------------------------------------------
// .NAME  drgbsnorm_computor - Class for computing the average euclidean distance in
//                             intensity-normalized RGB space between two affinely invariant regions.
//                           
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   drgbsnorm_computor.h
// .FILE   drgbsnorm__computor.cpp
//
// .SECTION Description
//   drgbsnorm_computor computes the average euclidean distance in intensity-normalized RGB space 
//   between two affinely invariant regions, after aligning them by the
//   appropriate affine transformation (assuming pt0, pt1 and pt2 correspond).
//   The mean intensity is set to half the range (128), via scalefactors
//   at pixel-reading time, for both regions (intensity-normalization).
//   set_region{1,2} automatically update the scalefactors via 128/mean when any region changes
//   (transparent interface, little code and max speed, especially when used with refiner)
//
// .SECTION Author:
//          Vittorio Ferrari
//               ETHZ
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/drgb_computor.h>

class drgbsnorm_computor : public drgb_computor {

 public:
  drgbsnorm_computor() {}
  drgbsnorm_computor(const color_buffer_t* buf1,
		     const color_buffer_t* buf2);
  drgbsnorm_computor(const color_buffer_t* buf1, 
		     const color_buffer_t* buf2, 
		     const region* reg1, const region* reg2);
  ~drgbsnorm_computor();

 public:
  inline float get_drgbsnorm()           {return get_similarity();}           // syntactic sugar
  inline float get_drgbsnorm_switched()  {return get_similarity_switched();}  // syntactic sugar

  inline virtual void set_region1(const region* r1);
  inline virtual void set_region2(const region* r2);


private:

  //--- helper functions -----------------------------------------------
  inline float dist_rgb_sf(vil1_rgb_byte& pix1, vil1_rgb_byte& pix2);

  //--- scalefactors ---------------------------------------------------
  float _sf1;
  float _sf2;
};


//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline void drgbsnorm_computor::set_region1(const region* r1) {
  // DO NOT try to optimize here by computing means iff
  // r1!=_region1 -> pointer difference does not imply coordinates
  // difference ! Source of bug that costed hours !
  _sf1 = 128.0 / compute_mean(r1, _buffer1);          // _buffer1 is greylevel
  similarity_computor::set_region1(r1);
}


inline void drgbsnorm_computor::set_region2(const region* r2) {
  _sf2 = 128.0 / compute_mean(r2, _buffer2);
  similarity_computor::set_region2(r2);               // _buffer2 is greylevel
}


// now with scalefactor on both regions
inline float drgbsnorm_computor::dist_rgb_sf(vil1_rgb_byte& pix1, vil1_rgb_byte& pix2) {
  float dr = pix1.R()*_sf1 - pix2.R()*_sf2;
  float dg = pix1.G()*_sf1 - pix2.G()*_sf2;
  float db = pix1.B()*_sf1 - pix2.B()*_sf2;

  return sqrt( dr*dr + dg*dg + db*db );
}

#endif
