#include <vcl/vcl_iostream.h>

#include <regions/region.h>

#include "euclideansoftmatcher.h"


EuclideanSoftMatcher::EuclideanSoftMatcher( double mThresh, bool bVerbose )
{
	matchThresh = mThresh;
	verbose = bVerbose;
}

EuclideanSoftMatcher::~EuclideanSoftMatcher()
{

}


unsigned int EuclideanSoftMatcher::match_soft( vcl_vector<EuclMatch> &matches,
                                               const regions_list_t &regions1,
                                               const regions_list_t &regions2 )
{
	unsigned int nReg1 = regions1.size(),
	             nReg2 = regions2.size(),
	             nMatches = 0;

	for( unsigned int i=0; i<nReg1; i++ ) {
		/* Find all matches for this region */
		for( unsigned int j=0; j<nReg2; j++ ) {
			double d = distSquared( regions1[i]->get_invariants_ptr(),
			                        regions2[j]->get_invariants_ptr() );
			if( d < matchThresh ) {
				nMatches++;
				EuclMatch newMatch = { i, j, d };
				matches.push_back(newMatch);
			}
		}
	}

	if(verbose)
		vcl_cout << nMatches << " matches found" << vcl_endl;
	return nMatches;
}


unsigned int EuclideanSoftMatcher::match_soft3( vcl_vector<EuclMatch> &matches,
                                                const regions_list_t &regions1,
                                                const regions_list_t &regions2 )
{
	unsigned int nReg1 = regions1.size(),
	             nReg2 = regions2.size(),
	             nMatches = 0;

	for( unsigned int i=0; i<nReg1; i++ ) {
		double dBest1, dBest2, dBest3;
		unsigned int nBest1, nBest2, nBest3;
		dBest3 = dBest2 = dBest1 = matchThresh*2;

		/* Find best 3 matches for this region */
		for( unsigned int j=0; j<nReg2; j++ ) {
			double d = distSquared( regions1[i]->get_invariants_ptr(),
			                        regions2[j]->get_invariants_ptr() );
			if( d < matchThresh ) {
				if( d < dBest1 ) {
					dBest3 = dBest2;
					nBest3 = nBest2;
					dBest2 = dBest1;
					nBest2 = nBest1;
					dBest1 = d;
					nBest1 = j;
				}
				else if( d < dBest2 ) {
					dBest3 = dBest2;
					nBest3 = nBest2;
					dBest2 = d;
					nBest2 = j;
				}
				else if( d < dBest3 ) {
					dBest3 = d;
					nBest3 = j;
				}
			}
		}

		/* Store 3 best matches */
		if( dBest1 < matchThresh ) {
			nMatches++;
			EuclMatch newMatch = { i, nBest1, dBest1 };
			matches.push_back(newMatch);

			if( dBest2 < matchThresh ) {
				nMatches++;
				EuclMatch newMatch = { i, nBest2, dBest2 };
				matches.push_back(newMatch);

				if( dBest3 < matchThresh ) {
					nMatches++;
					EuclMatch newMatch = { i, nBest3, dBest3 };
					matches.push_back(newMatch);
				}
			}
		}
	}

	if(verbose)
		vcl_cout << nMatches << " matches found" << vcl_endl;
	return nMatches;
}


unsigned int EuclideanSoftMatcher::match_hard( vcl_vector<EuclMatch> &matches,
                                               const regions_list_t &regions1,
                                               const regions_list_t &regions2)
{
	unsigned int nReg1 = regions1.size(),
	             nReg2 = regions2.size(),
	             nMatches = 0;

	for( unsigned int i=0; i<nReg1; i++ ) {
		double dBest = matchThresh*2;
		unsigned int nBest;

		/* Find best match for this region */
		for( unsigned int j=0; j<nReg2; j++ ) {
			double d = distSquared( regions1[i]->get_invariants_ptr(),
			                        regions2[j]->get_invariants_ptr() );
			if( d < matchThresh && d < dBest ) {
				dBest = d;
				nBest = j;
			}
		}

		/* Store best match */
		if( dBest < matchThresh ) {
			nMatches++;
			EuclMatch newMatch = { i, nBest, dBest };
			matches.push_back(newMatch);
		}
	}

	if(verbose)
		vcl_cout << nMatches << " matches found" << vcl_endl;
	return nMatches;
}


inline double EuclideanSoftMatcher::distSquared( const vnl_vector<double> *desc1,
                                                 const vnl_vector<double> *desc2 )
{
	double distsq = 0.0;
	unsigned int dLen = desc1->size();

	for( unsigned int i=0; i<dLen; i++ ) {
		double dif = (*desc1)[i] - (*desc2)[i];
		distsq += dif*dif;
	}
	
	return distsq;
}
