#include "integral_computor.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <region_utility/misc.h>                 // in_image, compute_affine
#include <misc_utility/byte_interpolation.h>     // get_interpolation_Vic


// -- Constructors / Destructors ------------------------------
integral_computor::integral_computor(const color_buffer_t* buf1,
				     const color_buffer_t* buf2)
  : similarity_computor(buf1, buf2)
{
  _res = 4;      // number of zones per axis on which to integrate information
  _integral1 = NULL;
  _integral2 = NULL;
}


integral_computor::integral_computor(const color_buffer_t* buf1, 
				     const color_buffer_t* buf2, 
				     const region* r1, const region* r2)
  : similarity_computor(buf1, buf2, r1, r2)
{
  _res = 4;
  _integral1 = NULL;
  _integral2 = NULL;
}


integral_computor::~integral_computor() {

}


vcl_vector<vil1_rgb_byte>* integral_computor::compute_color_integral(const region* r, const color_buffer_t* buf) {
  // Preconditions
  if(!in_image(r, buf->width(), buf->height())) {
    return NULL;
  }
  if(_region1->get_type() == region::ELL_REGIONTYPE) {
    vcl_cout << "integral_computor::compute_color_integral: elliptical regions not supported." << vcl_endl;
    return NULL;
  }
  
  // Coordinate tranformation between r and an
  // hypothetical upright square of coords
  // (0,0), (q,0), (0,q), with q the number
  // of zones to average in.
  int q = (int)_res;
  point_t* q0 = new point_t(0,0);
  point_t* q1 = new point_t(q,0);
  point_t* q2 = new point_t(0,q);
  float* A = compute_affine(r->get_pt0(), r->get_pt1(), r->get_pt2(), q0, q1, q2);

  // Initialize zones counters and color accumulators
  int* zones_nb = new int[q*q];
  long* zones_r = new long[q*q];
  long* zones_g = new long[q*q];
  long* zones_b = new long[q*q];
  for(int z=0; z<q*q; z++)  {
    zones_nb[z] = 0;   // speedup with a memset ...
     zones_r[z] = 0;
     zones_g[z] = 0;
     zones_b[z] = 0;
  }
 
  // Compute BB
  int minx = (int) r->get_min_x();
  int miny = (int) r->get_min_y();
  int maxx = (int) (r->get_max_x()+1);
  int maxy = (int) (r->get_max_y()+1);
  //vcl_cout << " BB = " << minx << ", " << miny << " -> " << maxx << ", " << maxy << vcl_endl;

  // Scan BB
  vil1_rgb_byte pix;
  for(int x=minx; x<=maxx; x++) {
    for(int y=miny; y<=maxy; y++) {

      // Is the current pixel inside r ?
      if(r->inside_region(x, y)) {
	// Zone of (x,y)
	int zx = (int) (A[0]*x + A[1]*y + A[2]);
	int zy = (int) (A[3]*x + A[4]*y + A[5]);                             
	if(zx == q) zx = q-1;    // sometimes can happen last pixel off
	if(zy == q) zy = q-1;
	
	// Color of (x,y)
	pix = (*buf)(x, y);
	
	// Put pix in (zx,zy) zone
	int z = zx + zy*q;
	zones_r[z] += pix.R();
	zones_g[z] += pix.G();
	zones_b[z] += pix.B();
	zones_nb[z]++;

	// Test
	//vcl_cout << " (" << x << ", " << y << ") in zone " << zx << ", " << zy << " color " << pix << vcl_endl;
      }
    }
  }

  // Output means of zones
  vcl_vector<vil1_rgb_byte>* integral = new vcl_vector<vil1_rgb_byte>;
  //vcl_cout << "integral: ";
  for(int z=0; z<q*q; z++)  {
    integral->push_back(vil1_rgb_byte( vxl_byte(zones_r[z] / zones_nb[z]), vxl_byte(zones_g[z] / zones_nb[z]), vxl_byte(zones_b[z] / zones_nb[z]) ));
    //vcl_cout << (*integral)[z] << " ";
  }
  //vcl_cout << vcl_endl;
  //         different-area-zones (ala shape-contexts)
  //         visualiz (for vibes, for papers, for guests): zones with pixels form the org image, but brightness prop to potential (cool!)
  // 

  return integral;
}


// Instantiates a new color_buffer_t
color_buffer_t* integral_computor::draw(vcl_vector<vil1_rgb_byte>* integral) {
  // Create empty buffer the same size as the integral
  int q = _res;
  color_buffer_t* buf = new color_buffer_t(q, q);  
  
  // Draw integral
  for(int x=0; x<_res; x++) {
    for(int y=0; y<_res; y++) {
      int z = x + y*q;
      (*buf)(x, y) = (*integral)[z];
    }
  }

  return buf;
}


