#ifndef _integral_computor
#define _integral_computor

//------------------------------------------------------------------------------
// .NAME  integral_computor - Base Class for computing integrated similarity
//                            between two affinely invariant regions.
//                            This differes from other similarity computors in
//                            that instead of each pixel considered being sampled,
//                            it is the integration of a block of nearby pixels.
//                            This should achive a important smoothing effect which
//                            should strongly reduce the sensitivity to misaligniments.
//                            
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   integral_computor.h
// .FILE   integral_computor.cpp
//
// .SECTION Description
//
// .SECTION Author:
//          Vittorio Ferrari
//                ETHZ      
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>
#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/similarity_computor.h>

class integral_computor : public similarity_computor {

 public:
  integral_computor() {}
  integral_computor(const color_buffer_t* buf1,
		    const color_buffer_t* buf2);
  integral_computor(const color_buffer_t* buf1, 
		    const color_buffer_t* buf2, 
		    const region* reg1, const region* reg2);
  virtual ~integral_computor();

 public:
  inline virtual void set_region1(const region* r1);
  inline virtual void set_region2(const region* r2);

  inline float get_similarity();
  inline virtual float get_similarity(const region* r1, const region* r2);

  inline color_buffer_t* draw1() {draw(_integral1);}
  inline color_buffer_t* draw2() {draw(_integral2);}
  color_buffer_t* draw(vcl_vector<vil1_rgb_byte>* integral);

protected:
  // Computations
  vcl_vector<vil1_rgb_byte>* compute_color_integral(const region* r, const color_buffer_t* buf);

  // Data
  vcl_vector<vil1_rgb_byte>* _integral1;             // color integral for region1
  vcl_vector<vil1_rgb_byte>* _integral2;             // color integral for region2
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline void integral_computor::set_region1(const region* r1) {
  _region1 = r1;
  _sim_valid = false; 
  _sim_sw_valid = false; 
  _switched_valid = false;
  delete _integral1;
  _integral1 = compute_color_integral(r1, _buf1);
}


inline void integral_computor::set_region2(const region* r2) {
  _region2 = r2; 
  _sim_valid = false; 
  _sim_sw_valid = false; 
  _switched_valid = false;
  delete _integral2;
  _integral2 = compute_color_integral(r2, _buf2);
}


inline float integral_computor::get_similarity(const region* r1, const region* r2) {
  set_region1(r1);
  set_region2(r2);
  compute_similarity();
  return _sim;
}


inline float integral_computor::get_similarity()
{
  if (_sim_valid)
    return _sim;
  
  if (!_region1 || !_region2) 
  {
    vcl_cout << "error integral_computor: cannot compute before regions are loaded" << vcl_endl;
    return -1.0;
  }

  compute_similarity();
  return _sim;
}


#endif
