#ifndef invariant_computer2h
#define invariant_computer2h

//------------------------------------------------------------------------------
// .NAME  invariant_computor - Class for describing affinely invariant regions in an image
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   invariant_computor.h
// .FILE   invariant_computor.cpp
//
// .SECTION Description
//   invariant_computor computes a featurevector of moment invariants    
//   for a given (affinely invariant) region.
//   The moment invariants themselves are not affinely invariant.
//   but this has been taken care of by first normalizing the region 
//   for affine and photometric changes
//
// .SECTION Author:
//          Tinne Tuytelaars
//          ESAT / KULeuven
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <vnl/vnl_vector.h>
#include <vnl/vnl_matrix.h>

#include <regions/region.h>

typedef vil1_memory_image_of<vil1_rgb_byte> color_buffer_t;
typedef vil1_memory_image_of<vxl_byte> bw_buffer_t;

class invariant_computor{

 public:
  invariant_computor() {}
  invariant_computor(const color_buffer_t* buf, 
		     bool color = true, bool switch_inv = false, bool rot_inv = false);
  invariant_computor(const color_buffer_t* buf, 
		     const region* reg, 
		     bool color = true, bool switch_inv = false, bool rot_inv = false);
  ~invariant_computor();

 public:
  void set_region(const region* r) {_region = r; _invariants_computed = false;}

  inline void get_invariants(vnl_vector<double>& inv);
  static int get_nb_inv(bool color, bool rot_inv = false);
  static int get_max_nb_inv() {return 18+1;}


private:
  //--- computations ---------------------------------------------------
  void compute_invariants();

 protected:
  void compute_invariants_RGB();
  void compute_invariants_RGB_SW();  
  void compute_invariants_RGB_ROT();
  void compute_invariants_ZW();
  void compute_invariants_ZW_SW();
  void compute_invariants_ZW_ROT();

  //--- helper functions -----------------------------------------------
  void normalize_photometric_RGB(color_buffer_t* buf, 
				 bw_buffer_t* const mask);
  void normalize_photometric(bw_buffer_t* buf, 
			     bw_buffer_t* const mask);

 private:
  const region* _region;
  double* _invariants;
  const color_buffer_t* _buffer;
  color_buffer_t* _ds_buffer;
  bool _COLOR;
  bool _SWITCH_INVARIANT;
  bool _ROTATION_INVARIANT;
  
  bool _invariants_computed;

  vnl_matrix<double>* _TNP; // matrix for coordinate transformation of Fischer-analysis
  vnl_matrix<double>* _TNE; // matrix for coordinate transformation of Fischer-analysis
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
void invariant_computor::get_invariants(vnl_vector<double>& inv)
{
  if (!_invariants_computed)
    compute_invariants();

  int nb_inv = get_nb_inv(_COLOR, _ROTATION_INVARIANT);
  for (int i = 0; i < nb_inv; i++)
    inv[i] = _invariants[i];
  return;
}

#endif
