#ifndef _indexed_matchesh_
#define _indexed_matchesh_


//------------------------------------------------------------------------------
// .NAME indexed_matches - Abstract base class for all datastructures containing
//    matches between affinely invariant regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   indexed_matches.h
// .FILE   indexed_matches.cpp
//
// .SECTION Description
//    The class matches is an abstract class for storing matches. 
//    Different derived classes are foreseen, e.g. implementations 
//    based on a list of pairs of regions, based on Kd-trees (for indexing), etc. 
//    This should be sufficiently versatile to deal with both pairwise matches
//    as well as clusters, one-view, two-view and multiview matches, etc.
//  
// .SECTION Author:
//          Tinne Tuytelaars
//          ESAT / KULeuven
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <region_matching/matches.h>
#include <region_matching/match.h>
#include <regions/region.h>
#include <vil1/vil1_image.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_utility.h>

class indexed_matches : public matches
{
  // PUBLIC INTERFACE----------------------------------------------------------
public:
  // Constructors/Destructors--------------------------------------------------
  indexed_matches();  
  virtual ~indexed_matches();  
  
  // Data Access -----------------------------------------------------------
  virtual bool add(match mtch) {_matchlist->push_back(mtch);}
  virtual bool remove(match mtch) {_matchlist->remove(mtch);}

  // conversion to / from index to region
  virtual static bool valid_orig_index(int indx) const;
  virtual static bool valid_ref_index(int indx) const;
  virtual static region* get_orig_region(int orig_indx) const;
  virtual static region* get_ref_region(int ref_indx) const;
  virtual static int get_orig_index(region* orig_reg) const;
  virtual static int get_ref_index(region* ref_reg) const;

  // iterator-like operation to run over all matches / clusters
  virtual bool reset() {_it = _matchlist.reset();}

  virtual match next_match() {_it.next();}
  virtual match next_match_between(vil1_image* im1, vil1_image* im2); // pairwise matches
  virtual match next_match_with(region* orig_reg);
  virtual match next_match_with(int orig_indx);
  virtual match next_inverse_match_with(region* ref_reg);
  virtual match next_inverse_match_with(int ref_indx);

  virtual match get_match() const; // get current match
  virtual bool remove_match(); // remove the current match

  // get (first) matching region (if any)
  virtual region* get_matching_region_with(region* orig_reg);


protected:
  static regions_list_t* _orig_regions;
  static regions_list_t* _ref_regions;
  
  vcl_list<indexed_match> _matchlist;
  vcl_list<indexed_match>::iterator _it;
};


#endif // _matchesh_
