#include "moment_computor.h"

#include <misc_utility/defines.h>
#include <vcl/vcl_iostream.h>
#include <vcl/vcl_fstream.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>




//--- constructors / destructors ---------------------------------------------------------------
moment_computor::moment_computor(const bw_buffer_t *buffer, 
				 const bw_buffer_t *maskbuffer)
{
  int lin, col;
  sizex = buffer->width();
  sizey = buffer->height();
 
  frame = new int[sizex*sizey];
  mask = new int[sizex*sizey];

  green = frame;
  blue = frame;

  for(lin = 0 ; lin < sizex ; lin++)
    for(col = 0 ; col < sizey ; col++)
      {
	frame[col * sizex + lin] = (*buffer)(lin,col);
	mask[col * sizex + lin] = (*maskbuffer)(lin,col);
      }

  
  for(lin = 0 ; lin < 6 ; lin++)
    {
      i_moments_set[lin] = 0;
      s_moments_set[lin] = 0;
    }
 
 for(lin = 0 ; lin < 9 ; lin++)
    {
      rgb_moments_set[lin] = 0;
    }

}

moment_computor::moment_computor(const color_buffer_t *buffer, 
				 const bw_buffer_t *maskbuffer)
{
  int lin, col;
  sizex = buffer->width();
  sizey = buffer->height();
 
  frame = new int[sizex*sizey];
  green = new int[sizex*sizey];
  blue = new int[sizex*sizey];
  mask = new int[sizex*sizey];

  for(lin = 0 ; lin < sizex ; lin++)
    for(col = 0 ; col < sizey ; col++)
      {
	frame[col * sizex + lin] = ((*buffer)(lin,col)).R();
	green[col * sizex + lin] = ((*buffer)(lin,col)).G();
	blue[col * sizex + lin] = ((*buffer)(lin,col)).B();
	mask[col * sizex + lin] = (*maskbuffer)(lin,col);
      }
  
  for(lin = 0 ; lin < 6 ; lin++)
    {
      i_moments_set[lin] = 0;
      s_moments_set[lin] = 0;
    }
 
 for(lin = 0 ; lin < 9 ; lin++)
    {
      rgb_moments_set[lin] = 0;
    }

}


moment_computor::moment_computor(const color_buffer_t* buf, 
				 const region* reg)
{
  int lin, col;
  bw_buffer_t maskbuffer(buf->width(),buf->height());
 
  sizex = buf->width();
  sizey = buf->height();

  _region = reg;

  frame = new int[sizex*sizey];
  green = new int[sizex*sizey];
  blue = new int[sizex*sizey];
  mask = new int[sizex*sizey];
 

  for(lin = 0 ; lin < sizex ; lin++)
    for(col = 0 ; col < sizey ; col++)
      {
	frame[col * sizex + lin] = (*buf)(lin,col).R();
	green[col * sizex + lin] = (*buf)(lin,col).G();
	blue[col * sizex + lin] = (*buf)(lin,col).B();
	mask[col * sizex + lin] = _region->inside_region(lin,col);
      }

  
  for(lin = 0 ; lin < 6 ; lin++)
    {
      i_moments_set[lin] = 0;
      s_moments_set[lin] = 0;
    }
 for(lin = 0 ; lin < 9 ; lin++)
    {
      rgb_moments_set[lin] = 0;
    }
}


moment_computor::~moment_computor()
{
  if(green!= frame)
    delete[] green;

  if(blue!= frame)
    delete[] blue;

  if(frame)
    delete[] frame;

  if(mask)
    delete[] mask;
}

//-----------------------------------------------------------------------------
void moment_computor::set_region(const region* r) 
{
  int lin, col;
  _region = r;
  for(lin = 0 ; lin < sizex ; lin++)
    for(col = 0 ; col < sizey ; col++)
      {
	mask[col * sizex + lin] = _region->inside_region(lin,col);
      }
}

//--- computations ---------------------------------------------------------------
//-----------------------------------------------------------------------------


// All the shape and intensity moments  are computed at the same parsing of the buffer.
// It is useful to call this method before the invarians computation methods
// when it is likely that the invariants  need all the moments.
 void moment_computor::all_moments()
{
  int i, lin, col;
  for(i = 0 ; i < 6 ; i++)
    {
      s_moments[i] = 0;
      i_moments[i] = 0;
    }

  
  for(lin=0 ; lin < sizex ; lin++)
    for(col=0 ; col < sizey ; col++)
      
      if( mask[col * sizex + lin] != 0 )
	{
	  i_moments[0] += frame[col * sizex + lin];
	  i_moments[1] += frame[col * sizex + lin] * col;
	  i_moments[2] += frame[col * sizex + lin] * lin;
	  i_moments[3] += frame[col * sizex + lin] * lin * col;
	  i_moments[4] += frame[col * sizex + lin] * col * col;
	  i_moments[5] += frame[col * sizex + lin] * lin * lin;
	  s_moments[0] += 1;
	  s_moments[1] += col;
	  s_moments[2] += lin;
	  s_moments[3] += lin * col;
	  s_moments[4] += col * col;
	  s_moments[5] += lin * lin;
	}

  for(i = 0 ; i < 6 ; i++)
    {
      s_moments[i] = s_moments[i] *10e-4;
      i_moments[i] = i_moments[i] *10e-4;
      s_moments_set[i] = 1;
      i_moments_set[i] = 1;
    }

  return;
}


//-----------------------------------------------------------------------------
// the method computes the (p+q)-order moment; r is the power of intensity
 double moment_computor::get_moment_pq(int p, int q, int r) const 
{
  double sum=0.0;
  int col,lin;

  if(r!=0)
    {
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0 && frame[col * sizex + lin] != 0  )
	    sum += pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) lin, (double) p) * pow((double) col, (double) q);
    }
  else  
    {
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if(mask[col * sizex + lin] != 0 )
	    sum += pow((double) lin, (double) p) * pow((double) col, (double) q) ;         
    }    

  return sum*10e-4;  	
}


//-----------------------------------------------------------------------------
// the method computes the  moments with powers of intensities; r is the power of R band intensity, g for the green band, b for the blue band

 double moment_computor::get_moment_rgb(int r, int g, int b) const 
{
  double sum=0.0;
  int col,lin;
 
  if(r==0)
    {
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0   )
	    sum += pow((double) green[col * sizex + lin] , (double) g) *
	      pow((double) blue[col * sizex + lin], (double) b) ;
    }
  else  
    {
      if(g==0)
	{
	  for(lin=0 ; lin < sizex ; lin++)
	    for(col=0 ; col < sizey ; col++)
	      if( mask[col * sizex + lin] != 0   )
		sum += pow((double) frame[col * sizex + lin] , (double) r) *
		  pow((double) blue[col * sizex + lin], (double) b) ;
	}
      else  
	{ // b==0
	  for(lin=0 ; lin < sizex ; lin++)
	    for(col=0 ; col < sizey ; col++)
	      if( mask[col * sizex + lin] != 0   )
		sum += pow((double) green[col * sizex + lin] , (double) g) *
		  pow((double) frame[col * sizex + lin], (double) r) ;
	}
    }  
  return sum*10e-4;  	
}

//-----------------------------------------------------------------------------
// the method computes the first order  moments with powers of intensities; r is the power of R band intensity, g for the green band, b for the blue band;


 double moment_computor::get_moment_pqrgb(int p, int q, int r, int g, int b) const 
{
  double sum=0.0;
  int col,lin;
  int flag=0;
 
 
  if(p==0 && q==0)
    {
      flag=1;
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0   )
	    sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	      pow((double) blue[col * sizex + lin], (double) b) ;
      return sum*10e-4;  	
    }

  if(p==0 && q==1)
    {
      flag=1;
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0   )
	    sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	      pow((double) blue[col * sizex + lin], (double) b)* col ;
      return sum*10e-4;  	
    }
  if(p==1 && q==0)
    {
       flag=1;
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0   )
	    sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	      pow((double) blue[col * sizex + lin], (double) b)* lin ;
      return sum*10e-4;  	
    }
  
  if(p==1 && q==1)
    {
      flag=1;
      for(lin=0 ; lin < sizex ; lin++)
	for(col=0 ; col < sizey ; col++)
      
	  if( mask[col * sizex + lin] != 0   )
	    sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	      pow((double) blue[col * sizex + lin], (double) b) *
	      lin * col ;

      return sum*10e-4;  	
    }

  if (p==2 && q==0)
  {
    flag = 1;
    for(lin=0 ; lin < sizex ; lin++)
      for(col=0 ; col < sizey ; col++)
	
	if( mask[col * sizex + lin] != 0   )
	  sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	    pow((double) blue[col * sizex + lin], (double) b) *
	    lin * lin ;
    
    return sum*10e-4;  	
  }
  
  if (p==0 && q==2)
  {
    flag = 1;
    for(lin=0 ; lin < sizex ; lin++)
      for(col=0 ; col < sizey ; col++)
	
	if( mask[col * sizex + lin] != 0   )
	  sum +=  pow((double) frame[col * sizex + lin] , (double) r) *
	      pow((double) green[col * sizex + lin] , (double) g) *
	    pow((double) blue[col * sizex + lin], (double) b) *
	    col * col ;
    
    return sum*10e-4;  	
  }
  

  if(flag==0) vcl_cerr<<"\n Problems with moment_computor::Moment_pqrgb ->  "<<p<<"  "<<q;

  return sum*10e-4;  	
}

