#ifndef moment_computer2h
#define moment_computer2h

//------------------------------------------------------------------------------
// .NAME  moment_computor - Class for computing moments over affinely invariant regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   moment_computor.h
// .FILE   moment_computor.cpp
//
// .SECTION Description
//   moment_computor computes a generalized color moments
//   for a given (affinely invariant) region.
//   
//   M^{pq}_{abc} = int ( x^p y^q R^a G^b B^c dx dy)
//
// .SECTION Author:
//          Tinne Tuytelaars
//          ESAT / KULeuven
//
// .SECTION Modifications:
//          Bart Nabbe (conversion TJ -> VXL)
//
//----------------------------------------------------------------------------

#include <misc_utility/defines.h>
#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <regions/region.h>

class moment_computor{

 public:
  moment_computor() {}
  moment_computor(const bw_buffer_t *buffer, 
		  const bw_buffer_t *maskbuffer);
  moment_computor(const color_buffer_t *buffer, 
		  const bw_buffer_t *maskbuffer);
  moment_computor(const color_buffer_t* buf, 
		  const region* reg);
  
  ~moment_computor();

 public:
  void set_region(const region* r);

  // -- the method which computes the (p+q)-order moment; r is the power of intensity
  double get_moment_pq(int p, int q, int r) const;
// -- the method which computes the moments with powers of RGB intensities 
  double get_moment_rgb(int r, int g, int b) const;
// -- the method which computes the (p+q)-order moment; r,g,b is the power of intensity for the 3 bands
  double get_moment_pqrgb(int p, int q, int r, int g, int b) const;

  // -- Intensity moments on the pattern's image
  inline double momentI_00();
  inline double momentI_01();
  inline double momentI_10();
  inline double momentI_11();
  inline double momentI_02();
  inline double momentI_20();

  // -- Shape moments on the pattern's image
  inline double momentS_00();
  inline double momentS_01();
  inline double momentS_10();
  inline double momentS_11();
  inline double momentS_02();
  inline double momentS_20();

  // -- Moments with powers of RGB intensities
  inline double momentRGB_000();
  inline double momentRGB_001();
  inline double momentRGB_010();
  inline double momentRGB_011();
  inline double momentRGB_100();
  inline double momentRGB_101();
  inline double momentRGB_110();
  inline double momentRGB_002();
  inline double momentRGB_020();
  inline double momentRGB_200();

  // -- Common procedure call for all moments computation
  inline double compute_momentI(int p, int q);
  inline double compute_momentS(int p, int q);
  inline double compute_RGB(int r, int g, int b);

  void all_moments();

  // -- Member variables
 private:
  const region* _region;
 
  int* frame; // holds red for colour images
  int* green;
  int* blue;
  int* mask; 

  int air, aig, aib; //average intensities for the bands.
 
  int sizex, sizey;
 
  double i_moments[6];
  bool i_moments_set[6];
  double s_moments[6];
  bool s_moments_set[6];
  double rgb_moments[9]; //necessary moments when working with colour inv.
  bool rgb_moments_set[9];


};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Intensity moments on the pattern's image
 inline double moment_computor::momentI_00()
{
  if(i_moments_set[0] == 0)
    {
      i_moments[0] = compute_momentI(0,0);
      i_moments_set[0] = 1;
    }
  return i_moments[0];  
}

 inline double moment_computor::momentI_01()
{
  if(i_moments_set[1] == 0)
    {    
      i_moments[1] = compute_momentI(0,1);
      i_moments_set[1] = 1;
    }

  return i_moments[1];  
}

 inline double moment_computor::momentI_10()
{
  if(i_moments_set[2] == 0)
    {    
      i_moments[2] = compute_momentI(1, 0);
      i_moments_set[2] = 1;
    }
  return i_moments[2];  
}

 inline double moment_computor::momentI_11()
{
  if(i_moments_set[3] == 0)
    {    
      i_moments[3] = compute_momentI(1, 1);
      i_moments_set[3] = 1;
    }
  return i_moments[3];  
}

 inline double moment_computor::momentI_02()
{
  if(i_moments_set[4] == 0)
    {
      i_moments[4] = compute_momentI(0, 2);
      i_moments_set[4] = 1;
    }
  return i_moments[4];  
}

 inline double moment_computor::momentI_20()
{
  if(i_moments_set[5] == 0)
    {
      i_moments[5] = compute_momentI(2, 0);
      i_moments_set[5] = 1;
    }
  return i_moments[5];  
}



//-----------------------------------------------------------------------------
// Shape moments on the pattern's image

 inline double moment_computor::momentS_00()
{
  if(s_moments_set[0] == 0)
    {
      s_moments[0] = compute_momentS(0, 0);
      s_moments_set[0] = 1;
    }

  return s_moments[0];  
}

 inline double moment_computor::momentS_01()
{
  if(s_moments_set[1] == 0)
    {
    s_moments[1] = compute_momentS(0, 1);
    s_moments_set[1] = 1;
    }

  return s_moments[1];  
}

 inline double moment_computor::momentS_10()
{
  if(s_moments_set[2] == 0)
    {
      s_moments[2] = compute_momentS(1, 0);
      s_moments_set[2] = 1;
    }

  return s_moments[2];  
}

 inline double moment_computor::momentS_11()
{
  if(s_moments_set[3] == 0)
    {
      s_moments[3] = compute_momentS(1, 1);
      s_moments_set[3] = 1;
    }

  return s_moments[3];  
}

 inline double moment_computor::momentS_02()
{
  if(s_moments_set[4] == 0)
    {
      s_moments[4] = compute_momentS(0, 2);
      s_moments_set[4] = 1;
    }

  return s_moments[4];  
}

 inline double moment_computor::momentS_20()
{
  if(s_moments_set[5] == 0)
    {
      s_moments[5] = compute_momentS(2, 0);
      s_moments_set[5] = 1;
    }

  return s_moments[5];  
}

//-----------------------------------------------------------------------------
 inline double moment_computor::compute_momentI(int p, int q)
{
  return get_moment_pq(p, q, 1);
}

 inline double moment_computor::compute_momentS(int p, int q)
{
  return get_moment_pq(p, q, 0);
}


 inline double moment_computor::compute_RGB(int r, int g, int b)
{
  return get_moment_rgb(r, g, b);
}


//-----------------------------------------------------------------------------
// Moments with powers of RGB intensities

 inline double moment_computor::momentRGB_000()
{
  return  momentS_00();
}

 inline double moment_computor::momentRGB_001()
{
  if(rgb_moments_set[0] == 0)
    {
      rgb_moments[0] = compute_RGB(0, 0, 1);
      rgb_moments_set[0] = 1;
    }

  return rgb_moments[0];  
}

 inline double moment_computor::momentRGB_010()
{
  if(rgb_moments_set[1] == 0)
    {
      rgb_moments[1] = compute_RGB(0, 1, 0);
      rgb_moments_set[1] = 1;
    }

  return rgb_moments[1];  
}

 inline double moment_computor::momentRGB_011()
{
  if(rgb_moments_set[2] == 0)
    {
      rgb_moments[2] = compute_RGB(0, 1, 1);
      rgb_moments_set[2] = 1;
    }

  return rgb_moments[2];  
}

inline double moment_computor::momentRGB_100()
{
  if(rgb_moments_set[3] == 0)
    {
      rgb_moments[3] = compute_RGB(1, 0, 0);
      rgb_moments_set[3] = 1;
    }

  return rgb_moments[3];  
}

inline double moment_computor::momentRGB_101()
{
  if(rgb_moments_set[4] == 0)
    {
      rgb_moments[4] = compute_RGB(1, 0, 1);
      rgb_moments_set[4] = 1;
    }

  return rgb_moments[4];  
}

inline double moment_computor::momentRGB_110()
{
  if(rgb_moments_set[5] == 0)
    {
      rgb_moments[5] = compute_RGB(1, 1, 0);
      rgb_moments_set[5] = 1;
    }

  return rgb_moments[5];  
}

inline double moment_computor::momentRGB_002()
{
  if(rgb_moments_set[6] == 0)
    {
      rgb_moments[6] = compute_RGB(0, 0, 2);
      rgb_moments_set[6] = 1;
    }

  return rgb_moments[6];  
}


inline double moment_computor::momentRGB_020()
{
  if(rgb_moments_set[7] == 0)
    {
      rgb_moments[7] = compute_RGB(0, 2, 0);
      rgb_moments_set[7] = 1;
    }

  return rgb_moments[7];  
}


inline double moment_computor::momentRGB_200()
{
  if(rgb_moments_set[8] == 0)
    {
      rgb_moments[8] = compute_RGB(2, 0, 0);
      rgb_moments_set[8] = 1;
    }

  return rgb_moments[8];  
}



#endif
