#include "multi_region_matcher.h"

#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/rotation_finder.h>
#include <region_matching/matches.h>
#include <region_matching/match.h>
#include <region_matching/pairwise_indexed_matches.h>
#include <region_matching/pairwise_indexed_match.h>

#include <vcl/vcl_cmath.h>

#include <vil1/vil1_image_as.h>

#include <misc_utility/vic_math.h>


// -- Constructors / Destructors -----------------------------------

multi_region_matcher::multi_region_matcher(const vil1_image* im1,
					   const vil1_image* im2,
					   regions_list_t& reglist1, 
					   regions_list_t& reglist2,
					   similarity_computor* sc,
					   const reg_match_params* p)
  : classic_region_matcher(im1, im2, reglist1, reglist2, sc, p)
{
  
}


multi_region_matcher::~multi_region_matcher()
{
  if (_rgb)
  {
    delete _grey1;
    delete _grey2;
  }

  delete _mtches;
  delete _rotf;
  delete V;
}


// match region from reglist1
// input: reg1
// output: reg2, indx = top first most similar region
//         Fills in mtches member with top _params->max_multi_matches
//         matches
bool multi_region_matcher::match_region(region* reg1, region*& reg2, int& indx2) {
  bool match_found = match_region_algo(reg1, _reglist2);
  if (!match_found) return false;

  for(int m = 0; m < multi_regions.size(); m++) {
    if(m == 0) {
      reg2 = multi_regions[0];
      indx2 = multi_indices[0];
    }
    pairwise_indexed_match* mtch = new pairwise_indexed_match(reg1, multi_regions[m]);
    _mtches->add(mtch);
  }

  return true;
}


// top _params->max_multi_matches best matches to reg1. Saved to multi_* memebers
// matching regions are searched for in reglist
bool multi_region_matcher::match_region_algo(region* reg1, regions_list_t& reglist) {
  vcl_cout << reg1->get_sid() << " ";

  // STEP 1: select the N best matches based on Mahalanobis distance only.
  // These are stored in tentative_* members
  find_tentative_matches(reg1, reglist);

  // STEP 2: select the regions with largest sim values from
  // the list of tentative regions
  find_best_sim_matches(reg1);
       
  return (multi_regions.size() > 0);
}


void multi_region_matcher::insert_multi_match(region* r, int indx, float sim) {
  // add region to list of multi_matches
  bool inserted = false;
  vcl_vector<region*>::iterator regions_it = multi_regions.begin();
  vcl_vector<float>::iterator sims_it = multi_sims.begin();
  vcl_vector<int>::iterator indices_it = multi_indices.begin();
  for(int m = 0; m < multi_regions.size(); m++) {
    if(sim > multi_sims[m]) {
      multi_regions.insert(regions_it, r);
      multi_sims.insert(sims_it, sim);
      multi_indices.insert(indices_it, indx);
      inserted = true;
      break;
    }
    ++regions_it;
    ++sims_it;
    ++indices_it;
  }

  if(!inserted && multi_regions.size() < _params->max_multi_matches) {
    multi_regions.push_back(r);
    multi_sims.push_back(sim);
    multi_indices.push_back(indx);
    inserted = true;
  }

  if(multi_regions.size() > _params->max_multi_matches) {
    // remove last element
    multi_regions.pop_back();
    multi_sims.pop_back();
    multi_indices.pop_back();
  }
}


// Find top first _params->max_multi_matches most similar (pixel-based) regions to reg1 in tentative_*,
// with sim > _params->min_sim.
// Save them to multi_* members
void multi_region_matcher::find_best_sim_matches(region* reg1) {
  float sim_thr = _params->min_sim;
  region* reg2 = NULL;

  multi_regions.clear();
  multi_indices.clear();
  multi_sims.clear();

  if ( (reg1->get_type() == region::ELL_REGIONTYPE &&
	_params->search_for_rotation) ) {
    ell_region *ereg1, *ereg2, *ereg;
    ereg1 = reg1->cast_to_ellregion();
    _rotf->set_region1(ereg1);
        
    vcl_list<int >::iterator indices_it = tentative_indices.begin(); 
    for (vcl_list<region* >::iterator regions_it = tentative_regions.begin();
	 regions_it != tentative_regions.end(); ++regions_it, ++indices_it) {
      region* reg = *(regions_it);
      int indx = *(indices_it);
      if (!reg) break;                                               // no more tentative regions below dist_thr
      vcl_cout << ".";                                  // reg below dist threshold, and enters the sim test
      
      ell_region* ereg = reg->cast_to_ellregion();
      _rotf->set_region2(ereg);
      _rotf->get_crosscorrelation();                                 // causes rotation of _region1 member of _rotf. 'Lowcost' as greyvalue ncc, potentially lowres
      float sim = 0.0;
      sim = _sc->get_similarity(_rotf->get_rotated_region1(), ereg); // higher cost (eg _sc of type drgbnorm), but only once (rotf applied about 72)
      // not 100% efficient as if _sc is of type drgbnorm, it computes the means once again for region1, even thought they didn't change ...

      if (sim > sim_thr) {
	insert_multi_match(reg, indx, sim);
	vcl_cout << " " << reg->get_sid() << " " << " (" << sim << ") ";
      }
   } // loop over tentative_*
 } else { // parall region or elliptical one without need to search for rotation
   vcl_list<int >::iterator indices_it = tentative_indices.begin(); 
   for (vcl_list<region* >::iterator regions_it = tentative_regions.begin();
	regions_it != tentative_regions.end(); ++regions_it, ++indices_it) {
     region* reg = *(regions_it);
     int indx = *(indices_it);
     if (!reg) break; // no more tentative regions below dist_thr
     vcl_cout << "."; // eveyr time I see a dot, then a reg was below dist threshold, a actually entered the sim test
     
     float sim = 0.0;
     sim = _sc->get_similarity(reg1, reg);
     
     if (sim > sim_thr) {
       insert_multi_match(reg, indx, sim);
       vcl_cout << " " << reg->get_sid() << " (" << sim << ") ";
     }
   } // loop over tentative_*
 } // reg type check

  vcl_cout << vcl_endl;
} // find_best_sim_matches

