#include "ncc_region_matcher.h"

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/reg_match_params.h>
#include <region_matching/crosscorrelation_computor.h>
#include <region_matching/rotation_finder.h>
#include <region_matching/matches.h>
#include <region_matching/match.h>
#include <region_matching/pairwise_indexed_matches.h>
#include <region_matching/pairwise_indexed_match.h>

#include <vcl/vcl_cmath.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>


// -- Constructors / Destructors -----------------------------------

ncc_region_matcher::ncc_region_matcher(const vil1_image* im1,
				       const vil1_image* im2,
				       regions_list_t& reglist1, 
				       regions_list_t& reglist2,
				       const reg_match_params* p)
  : region_matcher(im1, im2, reglist1, reglist2, p)
{
  if (vil1_pixel_format(*im1) == VIL1_RGB_BYTE)
  {
    _rgb = true;
    _grey1 = new bw_buffer_t(vil1_image_as_byte(*im1));
    _grey2 = new bw_buffer_t(vil1_image_as_byte(*im2));
  }
  else
  {
    _rgb = false;
    _grey1 = (bw_buffer_t*)im1;
    _grey2 = (bw_buffer_t*)im2;
  }

  _mtches = new pairwise_indexed_matches;
}


ncc_region_matcher::~ncc_region_matcher()
{
  if (_rgb)
  {
    delete _grey1;
    delete _grey2;
  }

  delete _mtches;
}
  
// -- Utility Methods -----------------------------------------------------------

// match region from reglist1 (general - two way)
// input: reg1
// output: reg2, indx
bool ncc_region_matcher::match_region(region* reg1, region*& reg2, int& indx2)
{
  bool match_found = match_region1(reg1, reg2, indx2);
  if (!match_found) return false;

  int indx1;
  region* dummy;
  match_found = match_region2(reg2, dummy, indx1);
  if (!match_found) return false;

  if (dummy == reg1)
  {
    vcl_cout << "match found: " 
	     << reg1->get_pt0()->x() << ", " << reg1->get_pt0()->y() << " -> " 
	     << reg2->get_pt0()->x() << ", " << reg2->get_pt0()->y() << vcl_endl;

    pairwise_indexed_match* mtch = new pairwise_indexed_match(indx1, indx2);
    _mtches->add(mtch);
  }

  return true;
}

// match region from reglist1
bool ncc_region_matcher::match_region1(region* reg1, region*& reg2, int& indx2)
{
  if (!reg1)
    return false;

  // check type of reg1
  if (reg1->get_type() == region::ELL_REGIONTYPE)
    return match_region1_ell(reg1, reg2, indx2);
  else
    return match_region1_parall(reg1, reg2, indx2);
}

// match region from reglist2
bool ncc_region_matcher::match_region2(region* reg2, region*& reg1, int& indx1)
{
  if (!reg2)
    return false;

  // check type of reg1
  if (reg2->get_type() == region::ELL_REGIONTYPE)
    return match_region2_ell(reg2, reg1, indx1);
  else
    return match_region2_parall(reg2, reg1, indx1);
}


// match parall. region from reglist1
bool ncc_region_matcher::match_region1_parall(region* reg1, region*& reg2, int& indx2)
{
  float max_corr = _params->min_sim;
  crosscorrelation_computor ccc(_grey1, _grey2);
  ccc.set_resolution(30);
  ccc.set_region1(reg1);
  for (int indx=0; indx < _reglist2.size(); ++indx)
  {
    region* reg = _reglist2[indx];
    if (reg->get_type() == region::PARALL_REGIONTYPE)
    {
      ccc.set_region2(reg);
      float ncc = ccc.get_crosscorrelation();
      if (ncc > max_corr)
      {
	max_corr = ncc;
	reg2 = reg;
	indx2 = indx;
      }
    }
  }

  if (max_corr > _params->min_sim)
  {
    vcl_cout << reg1->get_pt0()->x() << ", " << reg1->get_pt0()->y() << " -> " 
	     << reg2->get_pt0()->x() << ", " << reg2->get_pt0()->y() 
	     << "( " << max_corr << " ) " << vcl_endl;
    return true;
  }

  return false; // no match found
}

// match parall. region from reglist1
bool ncc_region_matcher::match_region1_ell(region* reg1, region*& reg2, int& indx2)
{
  float max_corr = _params->min_sim;
  rotation_finder rf(_grey1, _grey2);
  rf.set_resolution(_params->rotation_finder_resolution);
  rf.set_stepsize(_params->rotation_finder_stepsize);

  rf.set_region1((ell_region*)reg1);
  for (int indx = 0; indx < _reglist2.size(); ++indx)
  {
    region* reg = _reglist2[indx];
    if (reg->get_type() == region::ELL_REGIONTYPE)
    {
      rf.set_region2((ell_region*)reg);
      float ncc = rf.get_crosscorrelation();
      if (ncc > max_corr)
      {
	max_corr = ncc;
	reg2 = reg;
	indx2 = indx;
      }
    }
  }

  if (max_corr > _params->min_sim)
  {
    vcl_cout << reg1->get_pt0()->x() << ", " << reg1->get_pt0()->y() << " -> " 
	     << reg2->get_pt0()->x() << ", " << reg2->get_pt0()->y() 
	     << "( " << max_corr << " ) " << vcl_endl;
    return true;
  }

  return false; // no match found
}

// match region from reglist2
bool ncc_region_matcher::match_region2_parall(region* reg2, region*& reg1, int& indx1)
{
  float max_corr = _params->min_sim;
  crosscorrelation_computor ccc(_grey1, _grey2);
  ccc.set_resolution(_params->similarity_resolution);

  ccc.set_region2(reg2);
  for (int indx = 0; indx < _reglist1.size(); ++indx)
  {
    region* reg = _reglist1[indx];
    if (reg->get_type() == region::PARALL_REGIONTYPE)
    {
      ccc.set_region1(reg);
      float ncc = ccc.get_crosscorrelation();
      if (ncc > max_corr)
      {
	max_corr = ncc;
	reg1 = reg;
	indx1 = indx;
      }
    }
  }

  if (max_corr > _params->min_sim)
  { 
    vcl_cout << reg1->get_pt0()->x() << ", " << reg1->get_pt0()->y() << " <- " 
	     << reg2->get_pt0()->x() << ", " << reg2->get_pt0()->y() 
	     << "( " << max_corr << " ) " << vcl_endl; 
    return true;
  }

  return false; // no match found
}

// match region from reglist2
bool ncc_region_matcher::match_region2_ell(region* reg2, region*& reg1, int& indx1)
{
  float max_corr = _params->min_sim;
  rotation_finder rf(_grey1, _grey2);
  rf.set_resolution(_params->rotation_finder_resolution);
  rf.set_stepsize(_params->rotation_finder_stepsize);

  rf.set_region2((ell_region*)reg2);
  for (int indx = 0; indx < _reglist1.size(); ++indx)
  {
    region* reg = _reglist1[indx];
    if (reg->get_type() == region::ELL_REGIONTYPE)
    {
      rf.set_region1((ell_region*)reg);
      float ncc = rf.get_crosscorrelation();
      if (ncc > max_corr)
      {
	max_corr = ncc;
	reg1 = reg;
	indx1 = indx;
      }
    }
  }

  if (max_corr > _params->min_sim)
  {
    vcl_cout << reg1->get_pt0()->x() << ", " << reg1->get_pt0()->y() << " <- " 
	     << reg2->get_pt0()->x() << ", " << reg2->get_pt0()->y() 
	     << "( " << max_corr << " ) " << vcl_endl;
    return true;
  }
  
  return false; // no match found
}

