#ifndef _region_matcherh_
#define _region_matcherh_


//------------------------------------------------------------------------------
// .NAME region_matcher - Abstract base class for matching 
//       two sets of affinely invariant regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   region_matcher.h
// .FILE   region_matcher.cpp
//
// .SECTION Description
//    The class region_matcher finds matches in two sets of 
//    affinely invariant regions. 
//     
//  
// .SECTION Author:
//          Tinne Tuytelaars
//          ESAT / KULeuven
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <misc_utility/defines.h>
#include <regions/region.h>
#include <region_matching/reg_match_params.h>
#include <region_matching/matches.h>
#include <region_matching/match.h>

#include <vcl/vcl_utility.h>
#include <vil1/vil1_memory_image_of.h>

class region_matcher
{
  // PUBLIC INTERFACE----------------------------------------------------------
public:
  // Constructors/Destructors--------------------------------------------------
  region_matcher(const vil1_image* im1,
		 const vil1_image* im2,
		 regions_list_t& reglist1,  // TT: wanted to make these const as well,
		 regions_list_t& reglist2,  // but that seems to give problems
		 const reg_match_params* p);  
  virtual ~region_matcher();  
  
  // Utility Methods-----------------------------------------------------------
  virtual inline bool match_regions();
  virtual bool match_region(region* reg1, region*& reg2, int& indx) = 0;  // match region from reglist1 (general-two way)
  virtual bool match_region1(region* reg1, region*& reg2, int& indx) = 0; // match region from reglist1 (one-way)
  virtual bool match_region2(region* reg2, region*& reg1, int& indx) = 0; // match region from reglist2

  // Data Access---------------------------------------------------------------
  virtual inline bool get_matches (matches& mtches); // TT: wanted to make this const,
                                                     // but that seems to give problems 
  virtual inline bool clear_matches();

protected:
  // Helper Functions----------------------------------------------------------
  virtual inline double distance(const vnl_vector<double>* inv1, const vnl_vector<double>* inv2);
  virtual inline double distance(const vnl_vector<double>& inv1, const vnl_vector<double>& inv2);


protected:
  regions_list_t _reglist1; // list of regions   (TT: wanted to make these const as well,
  regions_list_t _reglist2; // list of regions   but that seems to give problems)

  matches* _mtches; 

  const vil1_image* _im1;
  const vil1_image* _im2;

  const reg_match_params* _params; 
};


bool region_matcher::match_regions()
{
  region* dummy;
  int dummy_indx;
  for (int indx = 0; indx < _reglist1.size(); ++indx) {
    region* reg1 = _reglist1[indx];
    match_region(reg1, dummy, dummy_indx);
  }

  return true;
}


bool region_matcher::get_matches(matches& mtches) // TT: wanted to make this const,
                                                  // but that seems to give problems 
{
  for (_mtches->reset(); _mtches->next_match(); )
  {
    match* mtch = _mtches->get_match();
    mtches.add(mtch);
  }
        
  return true;
}

bool region_matcher::clear_matches()
{
  _mtches->clear();
  
  return true;
}

double region_matcher::distance(const vnl_vector<double>* inv1, const vnl_vector<double>* inv2)
{
  float dist = 0;
  int size = inv1->size();

  for(int i=0; i < size; ++i)
  {
    dist += ((*inv1)[i] - (*inv2)[i])*((*inv1)[i] - (*inv2)[i]);
  }

  return dist;
}

double region_matcher::distance(const vnl_vector<double>& inv1, const vnl_vector<double>& inv2)
{
  float dist = 0;
  int size = inv1.size();

  for(int i=0; i < size; ++i)
  {
    dist += (inv1[i] - inv2[i])*(inv1[i] - inv2[i]);
  }

  return dist;
}

#endif // _region_matcherh_
