#include "rotation_finder.h"

#include <vcl/vcl_cmath.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>

#include <misc_utility/defines.h>
#include <regions/ell_region.h>


// -- Constructors / Destructors ------------------------------
rotation_finder::rotation_finder(const color_buffer_t* buf1,
				 const color_buffer_t* buf2)
{
 _buffer1 = new bw_buffer_t(vil1_image_as_byte(*buf1));
 _buffer2 = new bw_buffer_t(vil1_image_as_byte(*buf2));
 _rgb= true; // needed to free memory in destructor

 _region1 = NULL;
 _region2 = NULL;

 _ncc_valid = false;
 _ncc_sw_valid = false;
 _switched_valid = false;

 _step = 5; // default value - can be changed by user
 _res = 30; // default value - can be changed by user

 ccc = new crosscorrelation_computor(_buffer1, _buffer2);
 ccc->set_resolution(_res);
}


rotation_finder::rotation_finder(const bw_buffer_t* buf1,
				 const bw_buffer_t* buf2)
{
  _buffer1 = buf1;
  _buffer2 = buf2;
  _rgb = false;

  _region1 = NULL;
  _region2 = NULL;
  
  _ncc_valid = false;
  _ncc_sw_valid = false;
  _switched_valid = false;

  _step = 5; // default value - can be changed by user
  _res = 30; // default value - can be changed by user

  ccc = new crosscorrelation_computor(_buffer1, _buffer2);
  ccc->set_resolution(_res);
}

rotation_finder::rotation_finder(const color_buffer_t* buf1, 
				 const color_buffer_t* buf2, 
				 const ell_region* r1, const ell_region* r2)
{
 _buffer1 = new bw_buffer_t(vil1_image_as_byte(*buf1));
 _buffer2 = new bw_buffer_t(vil1_image_as_byte(*buf2));
 _rgb= true; // needed to free memory in destructor

 // _region1 is local copy, so we can freely rotate it without affecting r1
 _region1 = new ell_region(*r1);

 // _region2 points to r2, won't be changed at any time
 _region2 = r2;

 _ncc_valid = false;
 _ncc_sw_valid = false;
 _switched_valid = false;

  _step = 5; // default value - can be changed by user
  _res = 30; // default value - can be changed by user

  ccc = new crosscorrelation_computor(_buffer1, _buffer2);
  ccc->set_resolution(_res);
}

rotation_finder::rotation_finder(const bw_buffer_t* buf1, 
				 const bw_buffer_t* buf2, 
				 const ell_region* r1, const ell_region* r2)
{
  _buffer1 = buf1;
  _buffer2 = buf2;
  _rgb = false;

 // _region1 is local copy, so we can freely rotate it without affecting r1
  _region1 = new ell_region(*r1);
  
 // _region2 points to r2, won't be changed at any time
 _region2 = r2;
  
  _ncc_valid = false;
  _ncc_sw_valid = false;
  _switched_valid = false;

  _step = 5; // default value - can be changed by user
  _res = 30; // default value - can be changed by user

  ccc = new crosscorrelation_computor(_buffer1, _buffer2);
  ccc->set_resolution(_res);
}

rotation_finder::~rotation_finder() {
  // free memory of greyscale buffers
  if (_rgb){
    delete _buffer1; 
    delete _buffer2;
  }

  // free memory of _region1, which was a local copy
  if (_region1)
    delete _region1; 

  delete ccc;
}

// -- Computations -------------------------------------------------------
void rotation_finder::set_region1(const ell_region* r1)
{
  // free memory of _region1
  if (_region1)
    delete _region1;

  // _region1 is local copy, so we can freely rotate it without affecting r1
  _region1 = new ell_region(*r1);

  _ncc_valid = false; 
  _ncc_sw_valid = false; 
  _switched_valid = false;
}

void rotation_finder::set_region2(const ell_region* r2)
{
  _region2 = r2; 

  _ncc_valid = false; 
  _ncc_sw_valid = false; 
  _switched_valid = false;
}



bool rotation_finder::compute_crosscorrelation()
{
  _ncc = -1;
  _rotation = 0;

  int rollback_steps = 0;

  for (int theta = 0; theta <360; theta += _step)
  {
    float sim = ccc->get_similarity(_region1, _region2);
    if (sim > _ncc)
    {
      _ncc = sim; 
      _rotation = theta;
      rollback_steps = 0;
    }

    _region1->rotate_axes(_step);
    rollback_steps++;
  }
  
  _region1->rotate_axes(-rollback_steps*_step);

  return true;
}


bool rotation_finder::compute_crosscorrelation_switched()
{
  _ncc_sw = -1;
  _rotation_sw = 0;


  for (int theta = 0; theta <360; theta += _step)
  {
    float sim = ccc->get_similarity_switched(_region1, _region2);
    if (sim > _ncc_sw)
    {
      _ncc_sw = sim;
      _rotation_sw = theta;
    }

    _region1->rotate_axes(_step);
  }
  
  //vcl_cout << "max. ncc of " << _ncc_sw << " at rotation " << _rotation_sw << vcl_endl;

  return true;
}

