#ifndef _rotation_finder_h__
#define _rotation_finder_h__

//------------------------------------------------------------------------------
// .NAME  rotation_finder - Class for finding the optimal rotation
//        of an elliptical region (maximizing the crosscorrelation
//        with a matching region)
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   rotation_finder.h
// .FILE   rotation_finder.cpp
//
// .SECTION Description
//   rotation_finder computes the maximum crosscorrelation between two
//   elliptical affinely invariant regions, after aligning them by the appropriate 
//   affine transformation (assuming pt0, pt1 and pt2 correspond) and rotating 
//   one of them.
//   
// .SECTION Author:
//          Tinne Tuytelaars
//          ESAT / KULeuven
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/defines.h>
#include <regions/ell_region.h>
#include <region_matching/crosscorrelation_computor.h>

class rotation_finder{

 public:
  rotation_finder() {}
  rotation_finder(const color_buffer_t* buf1,
		  const color_buffer_t* buf2);
  rotation_finder(const bw_buffer_t* buf1,
		  const bw_buffer_t* buf2);
  rotation_finder(const color_buffer_t* buf1, 
		  const color_buffer_t* buf2, 
		  const ell_region* reg1, const ell_region* reg2);
  rotation_finder(const bw_buffer_t* buf1, 
		  const bw_buffer_t* buf2, 
		  const ell_region* reg1, const ell_region* reg2);
  ~rotation_finder();

 public:
  void set_region1(const ell_region* r1);
  void set_region2(const ell_region* r2);

  inline ell_region* get_rotated_region1() {return _region1;}

  inline void set_resolution(int res) {_res = res; ccc->set_resolution(res); _ncc_valid = false;}
  inline int get_resolution() const {return _res;}
  inline void set_stepsize(int step) {_step = step;}
  inline int get_stepsize() const {return _step;}

  inline float get_crosscorrelation();
  inline float get_crosscorrelation_switched();

  inline bool is_switched();
  inline int get_rotation();
  inline int get_rotation_switched();

private:
  //--- computations ---------------------------------------------------
  bool compute_crosscorrelation();
  bool compute_crosscorrelation_switched();

  //--- helper functions -----------------------------------------------

  //--- Member variables -----------------------------------------------
 private:
  ell_region* _region1;
  const ell_region* _region2;
  float _ncc; // max. normalized crosscorrelation value
  float _ncc_sw; // max. normalized crosscorrelation value with _region1 switched
  int _rotation; // rotation angle (in degrees) for max. ncc value
  int _rotation_sw; // rotation angle (in degrees) for max. ncc value with _region2 switched
  bool _switched; // true if _ncc_sw > _ncc
  bool _rgb; // needed for freeing memory in destructor

  bool _ncc_valid; // _ncc and _rotation have valid value
  bool _ncc_sw_valid; // _ncc_sw and _rotation_sw have valid value
  bool _switched_valid; // _switched has valid value

  int _step; // stepsize for trying out different rotations (in degrees) (default 5)
  int _res; // resolution used for ncc computations (default 30)

  const bw_buffer_t* _buffer1;
  const bw_buffer_t* _buffer2;

  crosscorrelation_computor* ccc;
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline float rotation_finder::get_crosscorrelation()
{
  if (_ncc_valid)
    return _ncc;
  
  if (!_region1 || !_region2) 
  {
    vcl_cerr << "error rotation_finder: cannot compute before regions are loaded" << vcl_endl;
    return -1.0;
  }

  compute_crosscorrelation();

  return _ncc;
}

inline float rotation_finder::get_crosscorrelation_switched()
{
  if (_ncc_sw_valid)
    return _ncc_sw;

  if (!_region1 || !_region2) 
  {
    vcl_cout << "error rotation_finder: cannot compute before regions are loaded" << vcl_endl;
    return -1.0;
  }

  compute_crosscorrelation_switched();

  return _ncc_sw;
}

inline bool rotation_finder::is_switched()
{
  if (_switched_valid)
    return _switched;

  if (! _ncc_valid)
    _ncc = compute_crosscorrelation();

  if (! _ncc_sw_valid)
    _ncc_sw = compute_crosscorrelation_switched();

  _switched = (_ncc_sw > _ncc ? true : false);
  _switched_valid = true;

  return _switched;
}

inline int rotation_finder::get_rotation()
{
  if (_ncc_valid)
    return _rotation;
  
  if (!_region1 || !_region2) 
  {
    vcl_cerr << "error rotation_finder: cannot compute before regions are loaded" << vcl_endl;
    return -1;
  }

  compute_crosscorrelation();

  return _rotation;
}

inline int rotation_finder::get_rotation_switched()
{
  if (_ncc_sw_valid)
    return _rotation_sw;
  
  if (!_region1 || !_region2) 
  {
    vcl_cerr << "error rotation_finder: cannot compute before regions are loaded" << vcl_endl;
    return -1;
  }

  compute_crosscorrelation_switched();

  return _rotation_sw;
}

#endif
