#include "similarity_computor.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <cstdlib>

#include <region_utility/misc.h>                 // in_image

#include <misc_utility/byte_interpolation.h>     // get_interpolation_Vic

#include <misc_utility/vic_math.h>               // pdist


// -- Constructors / Destructors ------------------------------
similarity_computor::similarity_computor(const color_buffer_t* buf1,
					 const color_buffer_t* buf2)
{
  _buf1 = buf1;
  _buf2 = buf2;

  // create greylevel buffers
  _buffer1 = new bw_buffer_t(vil1_image_as_byte(*buf1));
  _buffer2 = new bw_buffer_t(vil1_image_as_byte(*buf2));
  _rgb = true;

  _region1 = NULL;
  _region2 = NULL;

  _sim_valid = false;
  _sim_sw_valid = false;
  _switched_valid = false;

  _res = 50; // default value - can be changed by user

  _sr1 = 1.0;   _sr2 = 1.0;
  _sg1 = 1.0;   _sg2 = 1.0;
  _sb1 = 1.0;   _sb2 = 1.0;

  _adaptive_res = false;
  _min_res = _res;
}


similarity_computor::similarity_computor(const color_buffer_t* buf1, 
					 const color_buffer_t* buf2, 
				  	 const region* r1, const region* r2)
{
  _buf1 = buf1;
  _buf2 = buf2;

  // create greylevel buffers
  _buffer1 = new bw_buffer_t(vil1_image_as_byte(*buf1));
  _buffer2 = new bw_buffer_t(vil1_image_as_byte(*buf2));
  _rgb = true;

 set_region1(r1);
 set_region2(r2);

 _sim_valid = false;
 _sim_sw_valid = false;
 _switched_valid = false;

 _res = 50; // default value - can be changed by user

  _sr1 = 1.0;   _sr2 = 1.0;
  _sg1 = 1.0;   _sg2 = 1.0;
  _sb1 = 1.0;   _sb2 = 1.0;

  _adaptive_res = false;
  _min_res = _res;
}


similarity_computor::~similarity_computor() {
  if(_rgb) {
    delete _buffer1;
    delete _buffer2;
  }
}


int similarity_computor::region_resolution(const region* r) {
  point_t* pt0 = r->get_pt0();
  point_t* pt1 = r->get_pt1();
  point_t* pt2 = r->get_pt2();
  float l1 = pdist(pt0,pt1);
  float l2 = pdist(pt0,pt2);

  if(r->get_type() == region::ELL_REGIONTYPE) {
    l1 *= 2.0;
    l2 *= 2.0;
  }

  return (int)( 0.3333 * MAXTWO(l1, l2) );
}


// Assumes both _region1 and _region1 are non-null and point to existing regions
void similarity_computor::adapt_resolution() {
  _res = _min_res;
  _res = MAXTWO(_res, region_resolution(_region1));
  _res = MAXTWO(_res, region_resolution(_region2));
}


void similarity_computor::set_scale_factors1(float* C) {
  _sr1 = C[0];
  _sg1 = C[1];
  _sb1 = C[2];
}


void similarity_computor::set_scale_factors2(float* C) {
  _sr2 = C[0];
  _sg2 = C[1];
  _sb2 = C[2];
}


// instantiates a new float[3] array.
float* similarity_computor::compute_means(const region * r, const color_buffer_t* buf) {
  float* M = new float[3];

  if(!in_image(r, buf->width(), buf->height())) {
    M[0] = 128.0;
    M[1] = 128.0;
    M[2] = 128.0;
    return M;
  }

  int nb_ = 0;
  float avgr_ = 0.0;
  float avgg_ = 0.0;
  float avgb_ = 0.0;

  float cx = r->get_pt0()->x();
  float cy = r->get_pt0()->y();
  float x1 = r->get_pt1()->x();
  float y1 = r->get_pt1()->y();
  float x2 = r->get_pt2()->x();
  float y2 = r->get_pt2()->y();

  float incr_x_a1, incr_y_a1, incr_x_a2, incr_y_a2, xi, yi;

  if(_adaptive_res) _res = MAXTWO(_min_res, region_resolution(r));
  int resl1 = _res - 1;

  if (r->get_type() != region::ELL_REGIONTYPE) // NB_QUAD == 1
  {
    xi = cx; yi = cy;
    incr_x_a1 = (float)(x1 - cx) / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy) / resl1;

    incr_x_a2 = (float)(x2 - cx) / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy) / resl1;

    vil1_rgb_byte pix;
    float cur_x, cur_y;                   // current position inside region
    float xia, yia;                       // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1;             // cur_x|y now positioned to start of other axis
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	pix = get_col_interpolation_Vic(buf, cur_x, cur_y);
	
	nb_++;
	avgr_ += pix.R(); avgg_ += pix.G(); avgb_ += pix.B(); 
	
	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }
  else { // NB_QUAD == 4
    xi = cx - (x1 - cx) - (x2 - cx); // initial pt at the start of 4 quadrans
    yi = cy - (y1 - cy) - (y2 - cy);

    incr_x_a1 = (float)(x1 - cx)*2.0 / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy)*2.0 / resl1;

    incr_x_a2 = (float)(x2 - cx)*2.0 / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy)*2.0 / resl1;

    vil1_rgb_byte pix;
    float cur_x, cur_y; // current position inside region
    float xia, yia;                           // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1; // cur_x|y now positioned to start of other axe
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	if (r->inside_region(cur_x, cur_y)) {
	  pix = get_col_interpolation_Vic(buf, cur_x, cur_y);
	  
	  nb_++;
	  avgr_ += pix.R(); avgg_ += pix.G(); avgb_ += pix.B(); 
	}  // compare with Matlab

	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }

  if (nb_ == 0)
    {vcl_cout << "error in compute_means: nb_ = 0 " << vcl_endl;}
  

  avgr_ /= nb_; avgg_ /= nb_; avgb_ /= nb_;

  M[0] = avgr_;
  M[1] = avgg_;
  M[2] = avgb_;
  
  _res = _min_res;

  return M;
} // compute_means


// instantiates a new float[3] array.
float* similarity_computor::compute_color_trafo(region* r1, region* r2) {
  // Needs color buffers, not sure to have by now, as this could be
  // an instance of crosscorrelation_computor
  if(!_buf1 || !_buf2) {
    vcl_cout << "compute_color_trafo: needs color buffers. " << vcl_endl;
    exit(1);
  }

  float* M1 = compute_means(r1, _buf1);
  float* M2 = compute_means(r2, _buf2);
  float* C = new float[3];
  C[0] = M2[0] / M1[0];
  C[1] = M2[1] / M1[1];
  C[2] = M2[2] / M2[2];

  delete M1;
  delete M2;
  return C;
}


// instantiates a new float[3] array.
// each element has the same value.
// multiplying each colorband of each pixel by it causes an increase
// of the greylevel mean from r1 to r2
float* similarity_computor::compute_intensity_trafo(region* r1, region* r2) {
  float m1 = compute_mean(r1, _buffer1);
  float m2 = compute_mean(r2, _buffer2);
  float* C = new float[3];
  float sf = m2 / m1;
  C[0] = sf;
  C[1] = sf;
  C[2] = sf;

  return C;
}



// greylevel mean of r
float similarity_computor::compute_mean(const region * r, const bw_buffer_t* buf) {
  if(!in_image(r, buf->width(), buf->height())) {
    return 128.0;
  }

  int nb_ = 0;
  float avg_ = 0.0;

  float cx = r->get_pt0()->x();
  float cy = r->get_pt0()->y();
  float x1 = r->get_pt1()->x();
  float y1 = r->get_pt1()->y();
  float x2 = r->get_pt2()->x();
  float y2 = r->get_pt2()->y();

  float incr_x_a1, incr_y_a1, incr_x_a2, incr_y_a2, xi, yi;

  if(_adaptive_res) _res = MAXTWO(_min_res, region_resolution(r));
  int resl1 = _res -1;

  if (r->get_type() != region::ELL_REGIONTYPE) // NB_QUAD == 1
  {
    xi = cx; yi = cy;
    incr_x_a1 = (float)(x1 - cx) / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy) / resl1;

    incr_x_a2 = (float)(x2 - cx) / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy) / resl1;

    float pix;
    float cur_x, cur_y;                   // current position inside region
    float xia, yia;                       // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1;             // cur_x|y now positioned to start of other axis
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	pix = get_interpolation_Vic(buf, cur_x, cur_y);
	
	nb_++;
	avg_ += pix;
	
	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }
  else { // NB_QUAD == 4
    xi = cx - (x1 - cx) - (x2 - cx); // initial pt at the start of 4 quadrans
    yi = cy - (y1 - cy) - (y2 - cy);

    incr_x_a1 = (float)(x1 - cx)*2.0 / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy)*2.0 / resl1;

    incr_x_a2 = (float)(x2 - cx)*2.0 / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy)*2.0 / resl1;

    float pix;
    float cur_x, cur_y; // current position inside region
    float xia, yia;                           // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1; // cur_x|y now positioned to start of other axe
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	if (r->inside_region(cur_x, cur_y)) {
	  pix = get_interpolation_Vic(buf, cur_x, cur_y);
	  
	  nb_++;
	  avg_ += pix;
	}

	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }

  if (nb_ == 0)
    {vcl_cout << "error in compute_mean: nb_ = 0 " << vcl_endl;}
  

  avg_ /= nb_;

  _res = _min_res;          // restore res !!
  return avg_;
} // compute_mean
