#ifndef _similarity_computor
#define _similarity_computor

//------------------------------------------------------------------------------
// .NAME  similarity_computor - Base Class for computing the similarity
//                              between two affinely invariant regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   similarity_computor.h
// .FILE   similarity_computor.cpp
//
// .SECTION Description
//
// .SECTION Author:
//          Vittorio Ferrari - Tinne Tuytelaars
//                ETHZ             ESAT/KUL
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/defines.h>
#include <regions/region.h>

class similarity_computor {

 public:
  similarity_computor() {}
  similarity_computor(const color_buffer_t* buf1,
		      const color_buffer_t* buf2);
  similarity_computor(const color_buffer_t* buf1, 
		      const color_buffer_t* buf2, 
		      const region* reg1, const region* reg2);
  virtual ~similarity_computor();

 public:
  inline virtual void set_region1(const region* r1);
  inline virtual void set_region2(const region* r2);

  inline float get_similarity();
  inline float get_similarity_switched();
  inline virtual float get_similarity(const region* r1, const region* r2);
  inline float get_similarity_switched(const region* r1, const region* r2);

  inline bool is_switched();

  inline void set_adaptive_resolution(bool ar) {_adaptive_res = ar;}

  //--- parameter setting ---------------------------------------------
  virtual void set_scale_factors1(float* C);
  virtual void set_scale_factors2(float* C);
  inline virtual void set_resolution(int res); 
  inline int get_resolution() const {return _res;}

  // --- utility functions ------------------------------------------
  float* compute_color_trafo(region* r1, region* r2);
  float* compute_intensity_trafo(region* r1, region* r2);
  float* compute_means(const region* r, const color_buffer_t* buf);
  float  compute_mean(const region* r, const bw_buffer_t* buf);

  inline vil1_rgb_byte get_col_interpolation_Vic(const color_buffer_t* buf, float x, float y);


protected:
  //--- computations ---------------------------------------------------
  virtual bool compute_similarity() = 0;
  virtual bool compute_similarity_switched() = 0;
  int region_resolution(const region* r);
  void adapt_resolution();

 protected:
  const region* _region1;
  const region* _region2;
  float _sim;       
  float _sim_sw;
  bool _sim_valid;
  bool _sim_sw_valid;
  bool _switched_valid;
  bool _switched;

  int _res;                                         // resolution used for sim computations (default 50)
  bool _adaptive_res;                               // adapt the resolution to the size of the regions ? 
  int _min_res;                                     // minimal resolution (= _res if not adaptive)
  
  const color_buffer_t* _buf1;
  const color_buffer_t* _buf2;
  const bw_buffer_t* _buffer1;                      // greylevel versions, for fast computation of greylevel means
  const bw_buffer_t* _buffer2;                      // (used by compute_intensity_trafo)

  bool _rgb;                                        // need to free bw buffers ?
  float _sr1;                                       // color scale factors for region 1
  float _sg1;
  float _sb1;
  float _sr2;                                       // color scale factors for region 2
  float _sg2;
  float _sb2;
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline void similarity_computor::set_region1(const region* r1) {
  _region1 = r1;
  _sim_valid = false; 
  _sim_sw_valid = false; 
  _switched_valid = false;
}

inline void similarity_computor::set_region2(const region* r2) {
  _region2 = r2; 
  _sim_valid = false; 
  _sim_sw_valid = false; 
  _switched_valid = false;
}


inline void similarity_computor::set_resolution(int res)
{
  _res = res; 
  _min_res = res;
  _sim_valid = false; 
  _sim_sw_valid = false; 
  _switched_valid = false;
}

inline float similarity_computor::get_similarity(const region* r1, const region* r2) {
  set_region1(r1);
  set_region2(r2);
  if(_adaptive_res) adapt_resolution();
  compute_similarity();
  _sim_valid = true;
  return _sim;
}

inline float similarity_computor::get_similarity()
{
  if (_sim_valid)
    return _sim;
  
  if (!_region1 || !_region2) 
  {
    vcl_cout << "error similarity_computor: cannot compute before regions are loaded" << vcl_endl;
    return -1.0;
  }

  if(_adaptive_res) adapt_resolution();
  compute_similarity();

  return _sim;
}

inline float similarity_computor::get_similarity_switched(const region* r1, const region* r2) {
  set_region1(r1);
  set_region2(r2);
  if(_adaptive_res) adapt_resolution();
  compute_similarity_switched();
  _sim_sw_valid = true;
  return _sim_sw;
}

inline float similarity_computor::get_similarity_switched()
{
  if (_sim_sw_valid)
    return _sim_sw;

  if (!_region1 || !_region2) 
  {
    vcl_cout << "error similarity_computor: cannot compute before regions are loaded" << vcl_endl;
    return -1.0;
  }

  if(_adaptive_res) adapt_resolution();
  compute_similarity_switched();

  return _sim_sw;
}

inline bool similarity_computor::is_switched()
{
  if (_switched_valid)
    return _switched;

  if (! _sim_valid)
    _sim = compute_similarity();

  if (! _sim_sw_valid)
    _sim_sw = compute_similarity_switched();

  _switched = (_sim_sw > _sim ? true : false);
  _switched_valid = true;
  return _switched;
}


inline vil1_rgb_byte similarity_computor::get_col_interpolation_Vic(const color_buffer_t* buf,
								   float x, float y) {
  // Fast rounding, wrong if x,y exact integers
  int xl = (int)x;
  int xh = xl+1;
  int yl = (int)y;
  int yh = yl+1;

  if(xl <= 0 || yl <= 0 || xh >= buf->width() || yh >= buf->height()) return 0;

  // it will work as vil1_rgb<byte> has operators for '+' and '*',
  // but they everytime instantiate a new object.
  // Will be faster if do it by 'hand' (below)
  /*return   ( (xh-x) * ((yh-y)*((*buf)(xl,yl)) + (y-yl)*((*buf)(xl,yh))) +
	     (x-xl) * ((yh-y)*((*buf)(xh,yl)) + (y-yl)*((*buf)(xh,yh))) )/(xh-xl)/(yh-yl);*/

  // 'by hand' version: faster because avoids creation of many vil1_rgb_byte objects,
  vil1_rgb_byte ll = (*buf)(xl,yl);
  vil1_rgb_byte lh = (*buf)(xl,yh);
  vil1_rgb_byte hl = (*buf)(xh,yl);
  vil1_rgb_byte hh = (*buf)(xh,yh);
  vxl_byte r = (vxl_byte) ( (xh-x) * ((yh-y)*ll.R() + (y-yl)*lh.R()) +
			    (x-xl) * ((yh-y)*hl.R() + (y-yl)*hh.R()) )/(xh-xl)/(yh-yl);
  vxl_byte g = (vxl_byte) ( (xh-x) * ((yh-y)*ll.G() + (y-yl)*lh.G()) +
			   (x-xl) * ((yh-y)*hl.G() + (y-yl)*hh.G()) )/(xh-xl)/(yh-yl);
  vxl_byte b = (vxl_byte) ( (xh-x) * ((yh-y)*ll.B() + (y-yl)*lh.B()) +
			    (x-xl) * ((yh-y)*hl.B() + (y-yl)*hh.B()) )/(xh-xl)/(yh-yl);

  return vil1_rgb_byte(r,g,b);
}

#endif
