#include "blade_propagator.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <region_utility/misc.h>                  // compute_affine
#include <misc_utility/vic_math.h>                // angle, pdist

// -- Constructors / Destructors ------------------------------
blade_propagator::blade_propagator(const color_buffer_t* buf1,
				   const color_buffer_t* buf2,
				   similarity_computor* sc,
				   regions_list_t* ref_rs1, regions_list_t* ref_rs2,
				   refiner* rf,
				   bool verbose)
  : propagator(buf1, buf2, sc, ref_rs1, ref_rs2, verbose)
{
 _rf = rf;
 _sections = 4;
}


blade_propagator::~blade_propagator() {
	for(int rx=0; rx<refs.size(); rx++)
		delete refs[rx];
}


// in which circular section is p, if the origin is o ?
// (amount of sections is in member '_sections')
// result in [0..(_sections-1)]
int blade_propagator::find_section(point_t* o, point_t* p) {
  float alpha = angle(o, p);                   // angle of vector o->p with left-to-right oriz axis
  float sect_size = 2*3.141592/_sections;      // angular size of a section
  int s = (int)(alpha/sect_size);              // section of alpha
  if(s == _sections) {                         // if alpha==2*pi -> s=_sections, but correct result is s=0;
    s = 0;
  }
  return s;
}


// closest region from rs to c. Only regions in indxs are considered.
// Point pt0 is used as region location.
// returns -1 if indxs is empty
int blade_propagator::closest_region(point_t* c, regions_list_t* rs, vcl_vector<int>* indxs) {
  float best_d = 1000000000.0;
  int best_indx = -1;
  for(int i=0; i<indxs->size(); i++) {
    int rx = (*indxs)[i];
    point_t* p = (*rs)[rx]->get_pt0();
    if(pdist(c, p) < best_d) {
      best_d = pdist(c, p);
      best_indx = rx;
    }
  }
  return best_indx;
}


// finds candidate regions that suppoort region srx will attempt to propagate
// fills refs structure
void blade_propagator::find_cand_regs(int srx, regions_list_t* cand_rs1) {
  if(_verbose) vcl_cout << "Finding cand regs for support reg " << ((*_ref_rs1)[srx])->get_sid() << vcl_endl;
  point_t* c = ((*_ref_rs1)[srx])->get_pt0();  // origin of 'rose' (or 'blade') considered
  
  // Binning: puts all cand regs into corresponding section
  vcl_vector<int>* sect[_sections];            // indx of regs per section
  for(int s=0; s<_sections; s++) {
    sect[s] = new vcl_vector<int>();
  }
  for(int rx=0; rx<cand_rs1->size(); rx++) {
    int s = find_section(c, ((*cand_rs1)[rx])->get_pt0());
    sect[s]->push_back(rx);                    // adds index of current cand reg into corresponding section
  }

  // Distance: find closest cand reg for each section
  for(int s=0; s<_sections; s++) {
    int cand_indx = closest_region(c, cand_rs1, sect[s]);
    if(cand_indx > -1) {                       // if there is at least one cadn reg in section s
      if(_verbose) vcl_cout << "  section " << s << " : region " << (*cand_rs1)[cand_indx]->get_sid() << vcl_endl;
      refs[cand_indx]->push_back(srx);
    }
  }

  if(_verbose) vcl_cout << vcl_endl;
  for(int s=0; s<_sections; s++) {
    delete sect[s];
  }
}


// attempts to propagate all regions in cand_rs1
regions_list_t* blade_propagator::propagate(regions_list_t* cand_rs1) {
  // Init: format refs structure
	for(int rx=0; rx<refs.size(); rx++)
		delete refs[rx];
	refs.resize(cand_rs1->size());
  for(int rx=0; rx<cand_rs1->size(); rx++)
    refs[rx]= new vcl_vector<int>();
  _supports.clear();

  // Phase1: determine which ref regs will attempt to prop each cand reg
  if(_verbose) vcl_cout << "Phase1: finding which support regions will attempt to propagate each cand reg..." << vcl_endl;
  for(int rx=0; rx<_ref_rs1->size(); rx++) {
    find_cand_regs(rx, cand_rs1);                       // updates 'refs' structure
  }

  // Phase2: propagation: for each cand, prop via all its assigned refs (refs[cand_inx]) and _always_ refine, keep the best post-ref sim
  if(_verbose) vcl_cout << "Phase2: attempting to propagate each cand reg..." << vcl_endl;
  regions_list_t* cand_rs2 = new regions_list_t;
  for (int rx=0; rx<cand_rs1->size(); rx++) {
    region* cand_r2 = propagate(rx, cand_rs1);
    if(cand_r2)
      cand_rs2->push_back(cand_r2);
  }

  return cand_rs2;
}


// attempts to propagate region rs[rx].
// uses as supports regions in refs[rx]
region* blade_propagator::propagate(int crx, regions_list_t* rs) {
	region* cand_r1 = (*rs)[crx];
	region* cand_r2 = NULL;                                       // output region
	vcl_cout << "Region " << cand_r1->get_sid() << vcl_flush;
 
	vcl_vector<int>* rrs = refs[crx];                             // support regions for candidate crx
	if(rrs->size() == 0) {
		vcl_cout << " -> No support region." << vcl_endl;
		return NULL;
	}
	float best_sim = -1000.0;
	int best_ref = -1;
	for(int i=0; i<rrs->size(); i++) {
		int rx = (*rrs)[i];
		region* ref_r1 = (*_ref_rs1)[rx];
		if(_verbose) vcl_cout << vcl_endl << "  " << "  ref = " << ref_r1->get_sid() << "." << vcl_flush;

		// Project cand_r1 to prop_r via aff trafo of ref_r1
		region* prop_r = copy_region(cand_r1);
		project_region(cand_r1, prop_r, _ref_affs[rx]);

		// Is it a uniform region ?
		if(greylevel_variance(prop_r, _buffer2) >= _min_var) {

			if(_verbose) vcl_cout << " res=" << _rf->get_similarity_computor()->get_resolution() << vcl_flush;
			_rf->get_similarity_computor()->set_scale_factors1(_ref_cols[rx]);
			region* cand_r2_ref = _rf->refine(cand_r1, prop_r);
			delete prop_r;
			_sc->set_scale_factors1(_ref_cols[rx]);                   // similarity for acceptance decision with correct color factors. 
			float sim = _sc->get_similarity(cand_r1, cand_r2_ref);    // need be recomputed because _rf might not use the same sim computor as _sc
			if(_verbose) vcl_cout << " sim = " << sim << vcl_flush;
			if(sim >= best_sim) {
				cand_r2 = cand_r2_ref;
				best_sim = sim;
				best_ref = ref_r1->get_sid();
			}
			else
				delete cand_r2_ref;
		}
		else {
			if(_verbose) vcl_cout << " Variance < " << _min_var << vcl_flush;
			delete prop_r;
		}
	} // loop over support regions regs[crx]

	if(_verbose) vcl_cout << vcl_endl;
	if(best_sim >= _sim_t) {                                      // if best sim > abs min acceptance thresh (_sim_t)
		vcl_cout << "  -> Successfully propagated. Sim = " << best_sim << ". Ref = " << best_ref << vcl_flush;
		_prop_count++;
		_tot_sim += best_sim;
		_supports.push_back(best_ref);
	}
	else {
		delete cand_r2;
		cand_r2 = NULL;
		vcl_cout << "  -> No propagation. Best sim = " << best_sim << ". Ref = " << best_ref << vcl_flush;
	}

	vcl_cout << vcl_endl;
	return cand_r2;
}


