#include "propagator.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <cstdlib>

#include <region_utility/misc.h>                  // compute_affine, greylevel_variance
#include <misc_utility/vic_math.h>                // pdist

#include <misc_utility/byte_interpolation.h>      // get_interpolation_Vic

#include <vil1/vil1_image_as.h>                     // greylevel conversion
#include <vil1/vil1_memory_image_of.h>


// -- Constructors / Destructors ------------------------------
propagator::propagator(const color_buffer_t* buf1,
		       const color_buffer_t* buf2,
		       similarity_computor* sc,
                       regions_list_t* ref_rs1, regions_list_t* ref_rs2,
		       bool verbose)
{
  _buf1 = buf1;
  _buf2 = buf2;

  // create greylevel buffers
 _buffer1 = new bw_buffer_t(vil1_image_as_byte(*buf1));
 _buffer2 = new bw_buffer_t(vil1_image_as_byte(*buf2));

 _sc = sc;
 _ref_rs1 = ref_rs1;
 _ref_rs2 = ref_rs2;
 _verbose = verbose;
 _max_dist = 1000000;   // default = no limit
 _sim_t = 0.7;          // similarity threshold (default meaningful for crosscorrelation_computor)
 _min_var = 200.0;    
 _max_refs = 1;         // max amount of possible props memorized

 _prop_count = 0;
 _tot_sim = 0.0;

 _best_rx = -1;

 _bias = NULL;
 prepare_affines();
 prepare_color_trafos();
 initialize_bias();
}


propagator::~propagator() {
	delete _buffer1;
	delete _buffer2;
	delete _bias;
	for( unsigned int i=0; i<_ref_affs.size(); i++ )
		delete [] _ref_affs[i];
	for( unsigned int i=0; i<_ref_cols.size(); i++ )
		delete [] _ref_cols[i];
}


float propagator::average_sim() {
  return _tot_sim/(float)_prop_count;
}


// computes aff transfomations for all _reference_rs
void propagator::prepare_affines() {
  vcl_cout << "Computing affine transfomations of reference regions" << vcl_endl;

	for( unsigned int i=0; i<_ref_affs.size(); i++ )
		delete [] _ref_affs[i];
  _ref_affs.resize(_ref_rs1->size(), NULL);

  region  *reg1, *reg2;
  for(int rx = 0; rx < _ref_rs1->size(); rx++) {
    reg1 = (*_ref_rs1)[rx];
    reg2 = (*_ref_rs2)[rx];                                       // same index in list
    _ref_affs[rx] = compute_affine(reg1, reg2);
  }
}


// computes color scale factors for all _reference_rs
void propagator::prepare_color_trafos() {
  //vcl_cout << "Computing color transfomations of reference regions" << vcl_endl;
  vcl_cout << "Computing intensity transfomations of reference regions" << vcl_endl;

	for( unsigned int i=0; i<_ref_cols.size(); i++ )
		delete [] _ref_cols[i];
  _ref_cols.resize(_ref_rs1->size(), NULL);

  region  *reg1, *reg2;
  for(int rx = 0; rx < _ref_rs1->size(); rx++) {
    reg1 = (*_ref_rs1)[rx];
    reg2 = (*_ref_rs2)[rx];                                       // same index in list
    //_ref_cols[rx] = _sc->compute_color_trafo(reg1, reg2);
    _ref_cols[rx] = _sc->compute_intensity_trafo(reg1, reg2);     // 3 identical color trafos
  }
}


// Initialize bias for all supports to 1 (ie: no impact)
void propagator::initialize_bias() {
  delete _bias;
  _bias = new vcl_vector<float>;
	_bias->assign( _ref_rs1->size(), 1.0 );
}




void propagator::insert_prop_region(region* r, float sim, int indx) {
  // add region to prop_* lists
  bool inserted = false;
  for(int m = 0; m < prop_regs.size(); m++) {
    if(sim > prop_sims[m]) {
      prop_regs.insert( prop_regs.begin()+m, r );
      prop_sims.insert( prop_sims.begin()+m, sim );
      ref_indx.insert ( ref_indx.begin()+m,  indx );
      inserted = true;
      break;
    }
  }

  if(!inserted) {
		if( prop_regs.size() < _max_refs) {
			prop_regs.push_back(r);
			prop_sims.push_back(sim);
			ref_indx.push_back(indx);
		}
		else
			delete r;
	}
  else if(prop_regs.size() > _max_refs) {
    // remove last element
    delete prop_regs.back();
    prop_regs.pop_back();
    prop_sims.pop_back();
    ref_indx.pop_back();
  }
}


region* propagator::propagate(region* cand_r1) {
	// Inits
	point_t* cand_c1 = cand_r1->center();
	region* cand_r2 = NULL;                    // candidate region in view2, used as container of result
	prop_regs.clear();                         // regions inside prop_regs should be deleted at this point!
	prop_sims.clear();
	ref_indx.clear();

	// Loop over all reference regions
	_best_rx = -1;                             // index of best ref reg
	for (int rx = 0; rx < _ref_rs1->size(); rx++) {
		region* ref_r1 = (*_ref_rs1)[rx];
		if(_verbose) vcl_cout << "  ref = " << ref_r1->get_sid() << ".  " << vcl_flush;
		point_t* ref_c1 = ref_r1->center();

		// Consider ref_r if closer than _max_dist to cand_r
		if(pdist(ref_c1, cand_c1) <= _max_dist) {
			if(_verbose) vcl_cout << "Close enough.  " << vcl_flush;
			// Project cand_r1 to prop_r via aff trafo of _ref_rs[rx]
			region* prop_r = copy_region(cand_r1);
			project_region(cand_r1, prop_r, _ref_affs[rx]);

			// Is it a uniform region ?
			float var = greylevel_variance(prop_r, _buffer2);
			if(var >= _min_var) {
				// Compute similarity
				_sc->set_scale_factors1(_ref_cols[rx]);                    // actually used only if sc is of type *drgb_computor, but harmless otherwise
				float s = _sc->get_similarity(cand_r1, prop_r);
				float b = (*_bias)[rx];
				if(_verbose) {
					vcl_cout << "Sim = " << s << ".   " << vcl_flush;
					if(b != 1.0) vcl_cout << "Bias = " << b << " --> " << s*b << ".  " << vcl_flush;
				}

				insert_prop_region(prop_r, s*b, rx);
			}
			else {
				if(_verbose) vcl_cout << "Variance = " << var << " < " << _min_var << vcl_flush;
				delete prop_r;
			}
		}
		else {
			if(_verbose) vcl_cout << "Distance > " << _max_dist << vcl_flush;
		}
		delete ref_c1;
		if(_verbose)  vcl_cout << vcl_endl;
	}
	delete cand_c1;

	if(_verbose) {
		vcl_cout << "  Top " << _max_refs << " referrers. " << vcl_endl;
		for(int i=0; i<prop_regs.size(); i++) {
			vcl_cout << "    ref = " << ((*_ref_rs1)[ref_indx[i]])->get_sid() << "  sim = " << prop_sims[i] << vcl_endl;
		}
	}

	if(prop_regs.size() > 0) {
		cand_r2 = prop_regs[0]; //Should make a copy???
		if(_verbose) {
			vcl_cout << "  Best ref = " << ((*_ref_rs1)[ref_indx[0]])->get_sid() << vcl_endl;
		}   
	}
	else {
		cand_r2 = NULL;
	}

	return cand_r2;
}


regions_list_t* propagator::propagate(regions_list_t* cand_rs1) {
  vcl_cout << "ERROR: executing propagator::propagate(regions_list_t* cand_rs1)" << vcl_endl;
  vcl_cout << "Should never happen. Heir classes should implement it if they need it" << vcl_endl;
  exit(1);
}
