#ifndef _propagator_h
#define _propagator_h

//------------------------------------------------------------------------------
// .NAME  propagator - Base class for propagating regions on the base of a set of
//                     reference support regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   propagator.h
// .FILE   propagator.cpp
//
// .SECTION Description
//   
// .SECTION Author:
//          Vittorio Ferrari
//               ETHZ
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <region_matching/similarity_computor.h>

class propagator {

public:
  propagator() {}
  propagator(const color_buffer_t* buf1,
	     const color_buffer_t* buf2,
	     similarity_computor* sc,
             regions_list_t* ref_rs1,  regions_list_t* ref_rs2,
	     bool verbose = false);
  virtual ~propagator();

  //--- parameter setting ----------------------------------------------
  inline void set_max_dist(float max_dist) {_max_dist = max_dist;}
  inline void set_sim_t(float sim_t) {_sim_t = sim_t;}
  inline void set_min_var(float min_var) { _min_var = min_var;}
  inline void set_max_refs(int max_refs) { _max_refs = max_refs;}

  //--- settings
  inline void set_bias(vcl_vector<float>* bias) {delete _bias; _bias = bias;}

  //--- computations ---------------------------------------------------
  virtual region* propagate(region* cand_r1);                       // returns new region (NULL if fail)
  virtual regions_list_t* propagate(regions_list_t* cand_rs1);      // same for a list of regions
  float average_sim();                                              // average similarity of propagated regions
  inline int get_prop_count() {return _prop_count;}
  inline vcl_vector<int> get_supports() {return _supports;}         // sids of successful support regions
  

protected:
  //--- helper functions -----------------------------------------------
  void prepare_affines();                                           // computes aff transfomations for all _reference_rs
  void prepare_color_trafos();                                      // prepares col trafos for all _reference_rs
  void initialize_bias();                                           // set bias of each support to 1.0
  void insert_prop_region(region* r, float sim, int indx);

  //--- members --------------------------------------------------------
  const color_buffer_t* _buf1;                      // first image, where the candidate regions are
  const color_buffer_t* _buf2;                      // second image, where they should be propagated
  const bw_buffer_t* _buffer1;                      // greylevel versions, for fast computation of greylevel variance
  const bw_buffer_t* _buffer2;                      // (used in propagate fct)

  similarity_computor* _sc;                         // used to evaluate similarity function to be optimized     
  regions_list_t* _ref_rs1;                         // support regions in the first view
  regions_list_t* _ref_rs2;                         // support regions in the second view

  vcl_vector<float*> _ref_affs;                     // reference affine transformations
  vcl_vector<float*> _ref_cols;                     // reference color tranformations

  int _prop_count;                                  // how many regions have been propagated
  float _tot_sim;                                   // total similarity of propagated regions (used to compute average)

  int _best_rx;                                     // index of best ref reg for the current cand reg

  bool _verbose;                                    // display infos on screen or not

  vcl_vector<region*> prop_regs;                    // top _max_refs props
  vcl_vector<float> prop_sims;                      // top sims
  vcl_vector<int> ref_indx;                         // indexes of top referrers

  vcl_vector<int> _supports;                        // sids of successful support regions

  vcl_vector<float>* _bias;                         // bias for each support
 
  // parameters
  float _max_dist;                                  // max distance of support region to attempt prop
  float _sim_t;                                     // similarity threshold
  float _min_var;                                   // minimal variance of prop reg
  int _max_refs;                                    // max amount of reference regions considered

};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------




#endif
