#include "raywalk.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/vic_math.h> // max6
#include <region_utility/misc.h>   // copy_region, project_region



// -- Constructors / Destructors ------------------------------
raywalk::raywalk(similarity_computor* sc,
		 region* r1, region* r2,
		 bool verbose)
  : bounded_refiner(sc, r1, r2, verbose)
{
}


raywalk::~raywalk() {
}



// should not be needed (def in refiner class)
// but otherwise compiler complains that this fct does not exists ???
region* raywalk::refine(region* r1, region* r2) {
  //vcl_cout << "In raywalk::refine(r1,r2)." << vcl_endl; // strange that I can't call refiner::refine(r1,r2) ...
  set_region1(r1);
  set_region2(r2);
  refine();
}


// Main computation
// BEW: still no color-trafo in this version
// (change when impl also propagation)
// creates a new region to be returned
region* raywalk::refine() { 
  // Center _region2
  _cen = _region2->center();
  _region2_cen = copy_region(_region2);
  _region2_cen->set_pt0(_region2_cen->get_pt0()->x() - _cen->x(), _region2_cen->get_pt0()->y() - _cen->y());
  _region2_cen->set_pt1(_region2_cen->get_pt1()->x() - _cen->x(), _region2_cen->get_pt1()->y() - _cen->y());
  _region2_cen->set_pt2(_region2_cen->get_pt2()->x() - _cen->x(), _region2_cen->get_pt2()->y() - _cen->y());

  // Main loop
  _tot_sims = 0;
  float oldv = -2; 
  _region2_tmp = copy_region(_region2);
  float maxv = _sc->get_similarity(_region1, _region2_tmp);
  _init_sim = maxv;
  _cur_best_sim = maxv;
  float* gp = _cur_best_p; // alias for short notation
  gp[0] = 0; gp[1] = 0; gp[2] = 1; gp[3] = 1; gp[4] = 0; gp[5] = 0; // initial best pars is the identity matrix
  float maxvs[6];   // record partial (1-dim) max sims ...
  float maxps[6];   // ... and 1-dim-par generating them.
  
  if(_verbose)
    vcl_cout << "Sim      T  S   r  s" << vcl_endl;

  while(maxv > oldv) {
    if(_verbose) {
      vcl_cout << maxv << " "; print_pars(gp); vcl_cout << vcl_endl; // will gp be pretty printed ?
    }

    // Collect maxima in each dimension
    oldv = maxv;
    for(int d = 0; d < 6; d++) {
      maxps[d] = refine_1d(d, maxvs[d]);
    }

    // Go for the best maxima
    int best_p_inx = max6(maxvs);
    maxv = maxvs[best_p_inx];
    _cur_best_sim = maxv;
    gp[best_p_inx] = maxps[best_p_inx];  // affects contents of _cur_best_p
  }

  // Prepares region to be returned
  float* A = construct_affine(gp);
  project_region(_region2_cen, _region2_tmp, A);
  _region2_tmp->set_pt0(_region2_tmp->get_pt0()->x() + _cen->x(), _region2_tmp->get_pt0()->y() + _cen->y());
  _region2_tmp->set_pt1(_region2_tmp->get_pt1()->x() + _cen->x(), _region2_tmp->get_pt1()->y() + _cen->y());
  _region2_tmp->set_pt2(_region2_tmp->get_pt2()->x() + _cen->x(), _region2_tmp->get_pt2()->y() + _cen->y());
  _region2_tmp->update_region();
  _max_sim = maxv;

  delete A;
  delete _region2_cen;
  return _region2_tmp;
}



float raywalk::refine_1d(int p_inx, float& max_sim) {
  float cand_s = _cur_best_sim;
  float* gp = _cur_best_p;  // alias for short notation
  float pars[6];
  pars[0] = gp[0]; pars[1] = gp[1]; pars[2] = gp[2];
  pars[3] = gp[3]; pars[4] = gp[4]; pars[5] = gp[5];

  float opt_p = pars[p_inx];       // initial best parameter
  for(float p = _bounds[p_inx][0]; p <= _bounds[p_inx][2]; p += _bounds[p_inx][1]) {
    // Transforms _region2_cen via the current A
    pars[p_inx] = p;
    float* A = construct_affine(pars);
    project_region(_region2_cen, _region2_tmp, A);

    // Recenter _region2_tmp in the image
    _region2_tmp->set_pt0(_region2_tmp->get_pt0()->x() + _cen->x(), _region2_tmp->get_pt0()->y() + _cen->y());
    _region2_tmp->set_pt1(_region2_tmp->get_pt1()->x() + _cen->x(), _region2_tmp->get_pt1()->y() + _cen->y());
    _region2_tmp->set_pt2(_region2_tmp->get_pt2()->x() + _cen->x(), _region2_tmp->get_pt2()->y() + _cen->y());
    _region2_tmp->update_region(); // must be done once, to ensure proper functioning of inside_region

    // Compute similarity
    float s = _sc->get_similarity(_region1, _region2_tmp);
    _tot_sims++;

    if(s > cand_s) {
      cand_s = s;
      opt_p = p;
    }
    delete A;
  }

  max_sim = cand_s;
  return opt_p;
}


