#include "uniform_propagator.h"
#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>
#include <region_utility/misc.h>                  // compute_affine, in_image
#include <misc_utility/byte_interpolation.h>      // get_col_interpolation_Vic, dist_rgb
#include <vil1/vil1_memory_image_of.h>

/* WARNING: contains memory leaks. Compare with other propagator code to fix this. */

// -- Constructors / Destructors ------------------------------
uniform_propagator::uniform_propagator(const color_buffer_t* buf1,
				       const color_buffer_t* buf2,
				       similarity_computor* sc,
				       regions_list_t* ref_rs1, regions_list_t* ref_rs2,
				       bool verbose)
  : propagator(buf1, buf2, sc, ref_rs1, ref_rs2, verbose)
{
  // Prepare inverse affine transformations
  _ref_affs21 = NULL;
  prepare_affines2to1();

  // Prepare width and height lookups
  w1 = _buf1->width();
  h1 = _buf1->height();
  w2 = _buf2->width();
  h2 = _buf2->height();
}


uniform_propagator::~uniform_propagator() {
}


// computes aff transfomations for all _reference_rs from view 2 to view 1
void uniform_propagator::prepare_affines2to1() {
  vcl_cout << "Computing affine transfomations of reference regions from view2 to view1" << vcl_endl;

  delete _ref_affs21;  // more involved delete ? With recurse down and delete every single float* in the cells ? Anyway used only once ...
  _ref_affs21 = new float*[_ref_rs1->size()];

  region  *reg1, *reg2;
  for(int rx = 0; rx < _ref_rs1->size(); rx++) {
    reg1 = (*_ref_rs1)[rx];
    reg2 = (*_ref_rs2)[rx];                                       // same index in list
    _ref_affs21[rx] = compute_affine(reg2, reg1);
  }
}


// Draw a uniform region in solid mode
bool* uniform_propagator::uniform_region_to_flag(uniform_region* ur, const color_buffer_t* buf) {
  // Init
  int w = buf->width();
  int h = buf->height();

  bool* flag = new bool[w*h];
  for(int y=0; y<h; y++) {
    for(int x=0; x<w; x++) {
      flag[x + y*w] = false;
    }
  }

  // Segment by segment
  for(int s=0; s < ur->number_of_segments(); s++) {
    point_t* start = ur->get_segment_start(s); 
    point_t* end = ur->get_segment_end(s);

    // Scan horizontal segment, point by point
    int y = (int)start->y();                           // segments are horizontal -> y constant
    for(int x=(int)start->x(); x<=end->x(); x++) {
      flag[x + y*w] = true;
    }
  }

  return flag;
}


bool* uniform_propagator::project_uniform_region(uniform_region* ur1, bool* flag1, int ref_rx, int& area1, int& area2) {
  // Parameters
  float thresh = 50.0;                                 // maximal distance of image2 color from expected one

  // Init
  vil1_rgb<unsigned char> colt(ur1->get_color());       // expected region color. Color trafo of support region makes it work with illum changes
  float* sf = _ref_cols[ref_rx];
  thresh *= sf[0];                                     // attempt to disadvantage getting-darker-mismatches in competition
  if(_verbose) vcl_cout << "Scalefactor = " << sf[0] << ".  " << vcl_flush;
  vil1_rgb<unsigned char> colu(int(colt.R() * sf[0]), int(colt.G() * sf[1]), int(colt.B() * sf[2]));
  bool* flag2 = new bool[w2*h2];
  for(int y2=0; y2<h2; y2++) {                             // CHG: clean the flag buffer in one instruction, like 'memset' or something !
    for(int x2=0; x2<w2; x2++) {
      flag2[x2 + y2*w2] = false;
    }
  }

  // Phase 1
  // Project ur1 from buf1 to buf2
  // first marking of the flag, so that these marked pts do not need to be processed
  // again by phase 2
  point_t tl2, br2;                                                          // bounding box of projected region in buf2
  area1 = project12(ur1, _ref_affs[ref_rx], colu, thresh, tl2, br2, flag2);  // returns area1 and bb in view2

  // Phase 2
  // Project back bb2 from buf2 to buf1 -> avoids holes and computes area2
  // here actually mark flag buffer
  area2 = project21(tl2, br2, flag1, _ref_affs21[ref_rx], colu, thresh, flag2);

  return flag2;
}


void uniform_propagator::expand_bb(int x, int y, point_t& tl, point_t& br) {
  // init condition, happens exactly once
  if(tl.x() == -1) {
    tl.set(x,y);
    br.set(x,y);
  } else {
    tl.set(MINTWO(tl.x(), x), MINTWO(tl.y(), y));
    br.set(MAXTWO(br.x(), x), MAXTWO(br.y(), y));
  }
}


int uniform_propagator::project12(uniform_region* ur1, float* A, vil1_rgb<unsigned char>& colu, float thresh, point_t& tl, point_t& br, bool* flag2) {
  // Propagate region, segment by segment
  int area1 = 0;                                       // how many pixels of ur1 get propagated ?
  tl.set(-1,-1);
  br.set(-2,-2);                                       // reverse-crossed BB: if no proj pt, avoid running phase2
  for(int s=0; s < ur1->number_of_segments(); s++) {
    point_t* start = ur1->get_segment_start(s); 
    point_t* end = ur1->get_segment_end(s);

    // Scan horizontal segment, point by point
    int y = (int)start->y();                           // segments are horizontal -> y constant
    for(int x=(int)start->x(); x<=end->x(); x++) {
      int x2 = int(A[0]*x + A[1]*y + A[2] + 0.5);      // point in second image
      int y2 = int(A[3]*x + A[4]*y + A[5] + 0.5);

      if(in_image(x2, y2, w2, h2)) {
	vil1_rgb_byte col = (*_buf2)(x2, y2);           // reading a pixel without subpix interp is good enough
	if(dist_rgb(colu, col) < thresh) {             // as image is already pre-blurred and we look for uniform region
	  flag2[x2 + y2*w2] = true;
	  expand_bb(x2, y2, tl, br);
	  area1++;
	}
      }
    } // loop over points on a segment
  } // loop over segments of the uniform region

  return area1;
}


int uniform_propagator::project21(point_t& tl2, point_t& br2, bool* flag1, float* A, vil1_rgb<unsigned char>& colu, float thresh, bool* flag2) {
  int area2 = 0;
  for(int x2=(int)tl2.x(); x2<=(int)br2.x(); x2++) {
    for(int y2=(int)tl2.y(); y2<=(int)br2.y(); y2++) {

      if(flag2[x2 + y2*w2]) {                              // if point already marked
	area2++;                                           // just count it
      } else {
	vil1_rgb_byte col = (*_buf2)(x2, y2);
	if(dist_rgb(colu, col) < thresh) {                 // condition 1: point has correct color
	  int x1 = int(A[0]*x2 + A[1]*y2 + A[2] + 0.5);    // point in buf1
	  int y1 = int(A[3]*x2 + A[4]*y2 + A[5] + 0.5);
	  // should not need to check for in_image in buf1, as bb comes from previous proj ... but maybe rounding errors ...
	  if(in_image(x1, y1, w1, h1)) {
	    if(flag1[x1 + y1*w1]) {                        // condition 2: corr pt in buf1 is on uniform region to be prop
	      flag2[x2 + y2*w2] = true;
	      area2++;
	    }
	  }
	} // correct color in buf2
      } // pt not yet marked

    } // for over y2
  } // for over x2

  return area2;
}



uniform_region* uniform_propagator::propagate(uniform_region* cand_ur1) {
  // Inits
  bool* cand_flag2 = NULL;                               // candidate region in view2 (in flag format). used as temp container
  int max_a = 0;
  int best_ref = 0;
  bool* flag1 = uniform_region_to_flag(cand_ur1, _buf1); // flag repr of cand_ur1, useful later on

  // Competition among supports
  if(_verbose) vcl_cout << "Propagating uniform region " << cand_ur1->get_sid() << vcl_endl;
  for (int rx = 0; rx < _ref_rs1->size(); rx++) {
    region* ref_r1 = (*_ref_rs1)[rx];
    if(_verbose) vcl_cout << "  ref = " << ref_r1->get_sid() << ".  " << vcl_flush;

    // Project cand_ur1 to prop_ur via aff trafo of _ref_rs[rx]
    int a1 = 0;                               // how many pixels of cand_ur1 get propagated ?
    int a2 = 0;                               // how many pixels are in the prop_flag ?
    bool* prop_flag = project_uniform_region(cand_ur1, flag1, rx, a1, a2);

    // Compute score, as area of prop_r
    if(_verbose) vcl_cout << "Prop area1 = " << a1 << ", area2 = " << a2 << vcl_flush;
    if(_verbose) {
      if(a2 < 500) vcl_cout << "   --> insufficient" << vcl_flush;
      vcl_cout << vcl_endl;
    }
   
    //a2 = a1; // try out
    if(a2 > max_a) {
      max_a = a2;
      best_ref = ref_r1->get_sid();
      delete cand_flag2;
      cand_flag2 = prop_flag;
    } else {
      delete prop_flag;
    }
  }

  delete flag1;
  _supports.push_back(best_ref);
  if(_verbose) {
      vcl_cout << "  Best ref = " << best_ref << vcl_endl;
  }   
    
  if(cand_flag2 && max_a > 500) {               // if there's a winner
    // Instantiate new uniform_region object,
    // includes conversion of flag buffer to
    // horizontal-segments representation
    if(_verbose) vcl_cout << "  --> Successfully propagated" << vcl_endl;
    uniform_region* ur2 = new uniform_region(cand_flag2, _buf2);
    ur2->set_sid(cand_ur1->get_sid());
    delete cand_flag2;
    return ur2;
  } else {
    if(_verbose) vcl_cout << "  --> Not propagated" << vcl_endl;
    delete cand_flag2;
    return NULL;
  }
}


uniform_regions_list_t* uniform_propagator::propagate(uniform_regions_list_t* cand_urs1) {
  uniform_regions_list_t* _cand_urs2 = new uniform_regions_list_t;
  _supports.clear();

  for (int rx = 0; rx < cand_urs1->size(); rx++) {
    uniform_region* cand_ur2 = propagate((*cand_urs1)[rx]);
    if(cand_ur2)
      _cand_urs2->push_back(cand_ur2);
  }
  return _cand_urs2;
}
