#include <region_utility/aoi_computor.h>
#include <misc_utility/byte_interpolation.h>

#include <vil1/vil1_image_as.h>

//-------------------------------------------------
//--- Constructors/Destructors 
//-------------------------------------------------

aoi_computor::aoi_computor(ell_region* region, vil1_image* buf)
{
  _region = region;
  _buf = buf;

  //  Get Black and White image buffer
  BlackAndWhite = new vil1_memory_image_of<unsigned char>( vil1_image_as_byte(*_buf) );
}

aoi_computor::aoi_computor(vil1_image* buf)
{
  _region = NULL;
  _buf = buf;

  //  Get Black and White image buffer
  BlackAndWhite = new vil1_memory_image_of<unsigned char>( vil1_image_as_byte(*_buf) );
}

aoi_computor::~aoi_computor()
{
  delete BlackAndWhite;
}


//--------------------------------------------------
//--- Computations
//--------------------------------------------------

bool aoi_computor::compute(ell_region* reg) {
  _region = reg;
  return compute();
}

bool aoi_computor::compute()
{
  // Shortcuts
  ell_region* r = _region;
  double cx = r->get_pt0()->x();
  double cy = r->get_pt0()->y();
	
  double x1 = r->get_pt1()->x();
  double y1 = r->get_pt1()->y();
	
  double x2 = r->get_pt2()->x();
  double y2 = r->get_pt2()->y();
	
  // Local computation variables
  double sumxx = 0;
  double sumxy = 0;
  double sumyy = 0;
  double I;
  int minx, maxx, miny, maxy;
	
  // BEW: keep this low !! Previos = 100 -> makes thing slooow !
  double res = 50; // number of pts to be computed  (per axis)
  double resol;
	
  // Bounds of virtual circular region
  minx = (int) (-res/2+0.5);
  miny = (int) (-res/2+0.5);
  maxx = (int) (res/2+0.5); // fast rounding trick ! Thanx to CVI students !
  maxy = (int) (res/2+0.5);
  resol = (int) (res/2+0.5);
	
  // Local indexes
  double i1, j1;                 // i, j coords in elliptical region
	
  double inorm, jnorm;           // i,j normalised coords in circular region
  double cgx = 0; double cgy = 0; // Photometric invariant version, center of gravity is in geometric center of region
  double cgx2 = cgx*cgx;         
  double cgy2 = cgy*cgy;
  double cgxcgy = cgx*cgy;
	
	
  // compute m20, m02, m11
#if defined VERB_AOI
    vcl_cout << "aoi_computor: computing moments" << vcl_endl;
#endif
	
  for (int i=minx; i<=maxx; ++i)
  {
    for (int j=miny; j<=maxy; ++j) 
    {
      // Normalised coords are i/resol, j/resol. 
      inorm = i/resol;
      jnorm = j/resol;
			
      // Now map back to original region
      i1 = cx + inorm * (double)(x1 - cx) + jnorm * (double)(x2 - cx);
      j1 = cy + inorm * (double)(y1 - cy) + jnorm * (double)(y2 - cy);
			
      // Compute moment needed integrals
      if( r->inside_region(i1, j1) )  // CHG: could be areason for slowness !!
      {
        //vcl_cout << "i1 = " << i1 << "\nj1 = " << j1 << vcl_endl;
        I = get_byte_interpolation_no_check(*BlackAndWhite, i1, j1);
        //vcl_cout << "I = " << I << "\n\n" << vcl_endl;
				
        sumxx += (inorm*inorm + cgx2 - 2*inorm*cgx) * I/100.0;
        sumyy += (jnorm*jnorm + cgy2 - 2*jnorm*cgy) * I/100.0;
        sumxy += (inorm*jnorm - inorm*cgy - cgx*jnorm + cgxcgy) * I/100.0;
				
      }
    }
  }
	
  // Axes of inertia orientations
#if defined VERB_AOI
    vcl_cout << "aoi_computor: computing AOIs" << vcl_endl;
#endif

  double th_min, th_max;
  axes_of_inertia(th_max, th_min, sumxx, sumyy, sumxy);
	
  // MIN Axe of intertia two points (start, end) in normalised circle
  double sx = (double)(cgx + cos(th_min));
  double sy = (double)(cgy + sin(th_min));
  double ex = (double)(cgx - cos(th_min));
  double ey = (double)(cgy - sin(th_min));
	
  // MIN Axe of intertia two points (start, end) in original region
  vgl_homg_point_2d<double>* s = new vgl_homg_point_2d<double>(cx + sx * (double)(x1 - cx) + sy * (double)(x2 - cx),
                                                               cy + sx * (double)(y1 - cy) + sy * (double)(y2 - cy), 1.0);
	
  vgl_homg_point_2d<double>* e = new vgl_homg_point_2d<double>(cx + ex * (double)(x1 - cx) + ey * (double)(x2 - cx),
                                                               cy + ex * (double)(y1 - cy) + ey * (double)(y2 - cy), 1.0); 
	
  // Fix axis into region
  r->set_min_inertia_points(s, e);
	
#if defined VERB_AOI
    vcl_cout << "aoi_computor: minor AOI points: " << s << e << vcl_endl;
#endif
	
  // MAJ Axe of intertia two points (start, end) in normalised circle
  sx = (double)(cgx + cos(th_max));
  sy = (double)(cgy + sin(th_max));
  ex = (double)(cgx - cos(th_max));
  ey = (double)(cgy - sin(th_max));
	
  // MAJ Axe of intertia two points (start, end) in original region
  s = new vgl_homg_point_2d<double>(cx + sx * (double)(x1 - cx) + sy * (double)(x2 - cx),
                                    cy + sx * (double)(y1 - cy) + sy * (double)(y2 - cy), 1.0);
	
  e = new vgl_homg_point_2d<double>(cx + ex * (double)(x1 - cx) + ey * (double)(x2 - cx),
                                    cy + ex * (double)(y1 - cy) + ey * (double)(y2 - cy), 1.0); 
	
  // Fix axis into region
  r->set_max_inertia_points(s, e);
  _theta = th_max;

#if defined VERB_AOI
    vcl_cout << "aoi_computor: major AOI points: " << s << e << vcl_endl;
#endif
    
  return true;
}


void aoi_computor::axes_of_inertia(double& th_min, double& th_max, double sumxx, double sumyy, double sumxy) 
{
  th_min = 0.5 * atan2(2*sumxy, (sumxx - sumyy));
  th_max = th_min + 3.1459/2.0;                 // solutions are always 90 degrees apart
	
  double in1 = inertia_moment(th_max, sumxx, sumyy, sumxy);
  double in2 = inertia_moment(th_min, sumxx, sumyy, sumxy);
	
  if(in1 < in2) 
  {
    // need to switch values
    double tmp_th = th_min;
    th_min = th_max;
    th_max = tmp_th;
  }
}


double aoi_computor::inertia_moment(double th, double sumxx, double sumyy, double sumxy) 
{
  double sinth = sin(th);
  double costh = cos(th);
	
  return sumxx * sinth*sinth + sumyy * costh*costh - 2 * sinth*costh*sumxy;
}
