#include "hue_entropy_computor.h"

#include <misc_utility/defines.h>
#include <regions/region.h>

#include <vcl/vcl_cmath.h>
#include <vcl/vcl_vector.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_image.h>
#include <vil1/vil1_image_as.h>
#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <region_utility/misc.h>               // in_image

#include <misc_utility/draw.h>                 // RGBtoHSV


// -- Constructors / Destructors ------------------------------
hue_entropy_computor::hue_entropy_computor(const color_buffer_t* buf)
  : saliency_computor(buf)
{
  _hist = NULL;
  _hue_bins = 12;
  _grey_bins = 4;
  _sat_thresh = 0.2;      
  _value_thresh = 0.1;  
  _tot_pixels = 0;
}



hue_entropy_computor::~hue_entropy_computor() {

}


void hue_entropy_computor::print_histogram() {
  vcl_cout << "tot pixels = " << _tot_pixels << vcl_endl;

  vcl_cout << "Hue bins: " << vcl_endl;
  float bin_size_hue = 360.0 / (float)(_hue_bins);
  for(int b=0; b < _hue_bins; b++) {
    vcl_cout << "bin " << b << " (Hue=[" << b*bin_size_hue << ", " << (b+1)*bin_size_hue << "]) : " << _hist[b] << vcl_endl;
  }

  vcl_cout << "Grey bins: " << vcl_endl;
  float bin_size_value = 1.0 / (float)(_grey_bins);
  for(int b=0; b < _grey_bins; b++) {
    vcl_cout << "bin " << b+_hue_bins << " (Value=[" << b*bin_size_value << ", " << (b+1)*bin_size_value << "]) : " << _hist[b+_hue_bins] << vcl_endl;
  }

  vcl_cout << vcl_endl;
}


float hue_entropy_computor::compute_saliency() {
  compute_hue_histogram();
  print_histogram();
  return histogram_entropy();
}


float hue_entropy_computor::histogram_entropy() {
  float e = 0.0;
  for(int b=0; b < _hue_bins + _grey_bins; b++) {
    if(_hist[b] > 0) {
      float hb = _hist[b];
      float pb = hb/(float)_tot_pixels;
      e += pb * log(1.0/pb);
    }
  }
  return e;
}



// Inserts pixel pix intro _hist
void hue_entropy_computor::put_in_histogram(vil1_rgb_byte pix) {
  float h,s,v;           // do not copy-paste in compute_hue_histogram -> would conflict with h = height of buf
  RGBtoHSV( pix.R(), pix.G(), pix.B(), &h, &s, &v );
  //vcl_cout << "hsv = [" << h << ", " << s << ", " << v << "]" << vcl_endl;

  // Is it a color ?
  if(s >= _sat_thresh && v >= _value_thresh) {
    float bin_size_hue = 360.0 / (float)(_hue_bins);
    if(h == 360.0) {
      h = 359.0;      // stay in last bin, not go outside the histogram
    }
    //vcl_cout << "hue = " << h << vcl_endl;
    _hist[ (int)(h / bin_size_hue) ]++;
  }

  // is is a greylevel ?
  if(s < _sat_thresh || v < _value_thresh) {
    float bin_size_value = 255.0 / (float)(_grey_bins);
    if(v == 255.0) {
      v = 254.0;      // stay in last bin, not go outside histogram
    }
    //vcl_cout << "value = " << v << vcl_endl;
    _hist[ _hue_bins + (int)(v / bin_size_value) ]++;
  }

  _tot_pixels++;
}



// Computes hue histogram of the _region in
// the image _buf
// affects member _hist
void hue_entropy_computor::compute_hue_histogram() {
  delete _hist;
  _hist = new float[_hue_bins + _grey_bins];
  for(int i=0; i<_hue_bins+_grey_bins; i++) {
    _hist[i] = 0;
  }
  
  _tot_pixels = 0;
  const region* r = _region;                    // syntactic shortcut
  float cx = r->get_pt0()->x();
  float cy = r->get_pt0()->y();
  float x1 = r->get_pt1()->x();
  float y1 = r->get_pt1()->y();
  float x2 = r->get_pt2()->x();
  float y2 = r->get_pt2()->y();
  int w = _buf->width();                   // used by in_image
  int h = _buf->height();

  float incr_x_a1, incr_y_a1, incr_x_a2, incr_y_a2, xi, yi;
  int resl1 = _res - 1;
  //vcl_cout << "starting to compute histogram" << vcl_endl;
  if (r->get_type() != region::ELL_REGIONTYPE) { // NB_QUAD == 1
    xi = cx; yi = cy;
    incr_x_a1 = (float)(x1 - cx) / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy) / resl1;

    incr_x_a2 = (float)(x2 - cx) / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy) / resl1;

    vil1_rgb_byte pix;
    float cur_x, cur_y;                   // current position inside region
    float xia, yia;                       // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1;             // cur_x|y now positioned to start of other axis
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	if(in_image(cur_x, cur_y, w, h)) {
	  pix = get_col_interpolation_Vic(_buf, cur_x, cur_y);
	  //vcl_cout << "putting pixel " << cur_x << " " << cur_y << vcl_endl;
          put_in_histogram(pix);
	}
	
	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }
  else { // NB_QUAD == 4
    float xi = cx - (x1 - cx) - (x2 - cx); // initial pt at the start of 4 quadrans
    float yi = cy - (y1 - cy) - (y2 - cy);

    incr_x_a1 = (float)(x1 - cx)*2.0 / resl1; // incrementals for axe 1
    incr_y_a1 = (float)(y1 - cy)*2.0 / resl1;

    incr_x_a2 = (float)(x2 - cx)*2.0 / resl1; // incrementals for axe 2
    incr_y_a2 = (float)(y2 - cy)*2.0 / resl1;

    vil1_rgb_byte pix;
    float cur_x, cur_y; // current position inside region
    float xia, yia;                           // x,y init axis
    
    for (int i=0; i<_res; i++) {
      xia = xi + i*incr_x_a1; // cur_x|y now positioned to start of other axe
      yia = yi + i*incr_y_a1;
      cur_x = xia;
      cur_y = yia;
      
      for (int j=0; j<_res; j++) {
	if (r->inside_region(cur_x, cur_y) && in_image(cur_x, cur_y, w, h)) {
	  //vcl_cout << "putting pixel " << cur_x << " " << cur_y << vcl_endl;
	  pix = get_col_interpolation_Vic(_buf, cur_x, cur_y);
	  put_in_histogram(pix);
	}

	cur_x += incr_x_a2;
	cur_y += incr_y_a2;
      }
    }
  }


} // compute_hue_histogram
