#include <vcl/vcl_iostream.h>
#include <vcl/vcl_fstream.h>

#include <misc_utility/defines.h>
#include <regions/region.h> 
#include <regions/ell_region.h>
#include <regions/parall_region.h>
#include <vgl/vgl_homg_point_2d.h>
#include <vsol/vsol_conic_2d.h>
#include <vcl/vcl_list.h>
#include <vil1/vil1_image.h>

#include "miko1_region_file.h"


Miko1_region_file::Miko1_region_file(bool verbose)
  :region_file(verbose)
{
}


Miko1_region_file::~Miko1_region_file()
{
}

bool Miko1_region_file::write_header(regions_list_t& regs, 
				     vcl_ofstream& regfile)
{
  // descriptor length
	const vnl_vector<double> *invs = regs.at(0)->get_invariants_ptr();
	if( invs == NULL )
		_descript_length = 0;
	else
		_descript_length = invs->size();
  regfile << _descript_length << vcl_endl;
  
  // number of elliptical regions
  _nb_ell_regions = 0;
  for (regions_list_t::iterator it = regs.begin(); it != regs.end(); ++it)
  {
    region* reg = *it;
    if (reg->get_type() == region::ELL_REGIONTYPE)
      _nb_ell_regions++;
  }
  regfile << _nb_ell_regions << vcl_endl;

  if (_verbose) vcl_cout << _nb_ell_regions << " ell. regions to be saved" << vcl_endl;

  return true;
}



bool Miko1_region_file::load_header(vcl_ifstream& regfile)
{
  regfile >> _descript_length;
  if (_descript_length < 0) 
  {
    vcl_cout << "Error in header - cannot load regions" << vcl_endl;
    return false;
  }
  regfile >> _nb_ell_regions;
  if (_verbose) 
    vcl_cout << _descript_length << "   " << _nb_ell_regions << vcl_endl;

  return true;
}

bool Miko1_region_file::write_region(region* reg, 
				     vcl_ofstream& regfile)
{
  if (reg->get_type() == region::ELL_REGIONTYPE)
  {
    ell_region* ellreg = reg->cast_to_ellregion();
    vsol_conic_2d* ell = ellreg->get_ellipse();
    double a, b, c, f;
    a = ell->a();
    b = ell->b();
    c = ell->c();
    f = ell->f();
    double x0 = reg->get_pt0()->x();
    double y0 = reg->get_pt0()->y();
    double scalefactor = 1.0 / (a*x0*x0 + b*x0*y0 + c*y0*y0 - f);
    double A, B, C;
    A = scalefactor * a;
    B = scalefactor * b/2.0;
    C = scalefactor * c;
    regfile << x0 << " "
	    << y0 << " "
	    << A << " " 
	    << B << " " 
      << C;
		
		// save invariants/descriptor, if any
		if( _descript_length > 0 ) {
			const vnl_vector<double>* invs = reg->get_invariants_ptr();
			for( unsigned int i=0; i<_descript_length; i++ )
				regfile << " " << (*invs)[i];
		}
		regfile << vcl_endl;

    if (_verbose)
    {
      vcl_cout << reg->get_pt0()->x() << ", "
	       << reg->get_pt0()->y() << "  "
	       << reg->get_pt1()->x() << ", "
	       << reg->get_pt1()->y() << "  "
	       << reg->get_pt2()->x() << ", "
	       << reg->get_pt2()->y() << vcl_endl;
      vcl_cout << A << "  " << B << "  " << C << "   " << vcl_endl;
    }
  }
  
  return true;
}

bool Miko1_region_file::load_regions_impl(regions_list_t& regs, 
					  vcl_ifstream& regfile,
					  vil1_image* image, 
					  points_list_t* anchors)
{
  for (int i=0; i < _nb_ell_regions; i++)
  {
    if (!load_region(regs, regfile, image, anchors))
      return false;
  }

  return true;
}

bool Miko1_region_file::load_region(regions_list_t& regs, 
				    vcl_ifstream& regfile,
				    vil1_image* image, 
				    points_list_t* anchors)
{
  double x0, y0;
  double A, B, C;
  regfile >> x0;
  regfile >> y0;
  regfile >> A;
  regfile >> B;
  regfile >> C;
  if (_verbose) vcl_cout << x0 << "  " << y0 << "  " 
			 << A << "  " << B << "  " << C << vcl_endl;

  double theta = 0.5 * atan2(2*B,C-A);
  if (_verbose) vcl_cout << "theta: " << theta*180/3.14159 << vcl_endl;
  double cost = cos(theta);
  double sint = sin(theta);
  double N = sqrt(A * cost * cost - 2 * B * cost * sint + C * sint * sint);
  double x1 = x0 + cost / N;
  double y1 = y0 - sint / N;
  if (_verbose) vcl_cout << "pt1: " << x1 << ", " << y1 << vcl_endl;
  N = sqrt(A * sint * sint + 2 * B * cost * sint + C * cost * cost);
  double x2 = x0 + sint / N;
  double y2 = y0 + cost / N;
  if (_verbose) vcl_cout << "pt2: " << x2 << ", " << y2 << vcl_endl;

  point_t* pt0 = new point_t(x0, y0);
  point_t* pt1 = new point_t(x1, y1);
  point_t* pt2 = new point_t(x2, y2);
  ell_region* reg = new ell_region(pt0, pt1, pt2, 
				   region::ELL, 
				   pt0, image);

	// load invariants/descriptor, if any
	if( _descript_length > 0 ) {
		double val;
		vnl_vector<double> *invs = new vnl_vector<double>(_descript_length);
		for( unsigned int i=0; i<_descript_length; i++ ) {
			regfile >> val;
			(*invs)[i] = val;
		}
		reg->set_invariants(invs);
	}
	
  // Set synchro id
  reg->set_sid(reg->get_ID()+1);       // set synchronization id to 1 more than the absolute, process-wide region_id. So never sid==0,
  // and also generate unique sids when loading multiple files

  regs.push_back(reg);

  return true;
}
