#include <regions/region.h>
#include <region_utility/region_deskew.h>
#include <misc_utility/byte_interpolation.h>

#define ROUND(x) (int)( (double) x + 0.5 )


// ----------------------------------------------------------------------
// Make a buffer of size sx, sy containing a 'deskewed' version
// of the region content (i.e. transformed to a circular or
// square reference region ) in a canonical coordinate frame
// ----------------------------------------------------------------------
template<class T>
vil1_memory_image_of<T>* deskew_region(const region* region, const vil1_memory_image_of<T>* buf, 
                                      int sx, int sy, bool mask)
{
  if (region->is_centered()) // is pt0 center of region or corner ?
  {
    return deskew_region_centered(region, buf, sx, sy, mask);
  }
  else
  {
    return deskew_region_not_centered(region, buf, sx, sy, mask);
  }
}


// We need this macro, otherwise the templates are not
// instantiated
#define DESKEW_REGION_INSTANTIATE(T) \
template vil1_memory_image_of<T>* \
deskew_region(const region* , const vil1_memory_image_of<T>*, int , int , bool);




// ----------------------------------------------------------------------
// Pt0 is at the center of the region, which is generally
// the case for elliptical regions
// ----------------------------------------------------------------------
template<class T> vil1_memory_image_of<T>*
deskew_region_centered(const region* region, const vil1_memory_image_of<T>* buf, 
                      int sx, int sy, bool mask)
{
  int sizex = buf->width();
  int sizey = buf->height();
  int i,j;

  vil1_memory_image_of<T>* ds_buf = new vil1_memory_image_of<T>(sx, sy);

  double x0 = region->get_pt0()->x();
  double y0 = region->get_pt0()->y();
  double x1 = region->get_pt1()->x();
  double y1 = region->get_pt1()->y();
  double x2 = region->get_pt2()->x();
  double y2 = region->get_pt2()->y();

 
  for(i=0; i <sx ; i++)
  {
    for(j=0; j <sy ; j++)
    {
      double x = x0 + (x1 - x0)*(double)(i-(sx-1)/2.0)/(double)((sx-1)/2.0) 
		    + (x2 - x0)*(double)(j-(sy-1)/2.0)/(double)((sy-1)/2.0);
      double y = y0 + (y1 - y0)*(double)(i-(sx-1)/2.0)/(double)((sx-1)/2.0) 
		    + (y2 - y0)*(double)(j-(sy-1)/2.0)/(double)((sy-1)/2.0);
      
      if ( ( ROUND(x) >= 0) && (ROUND(x) <= sizex-1) && 
	   ( ROUND(y) >= 0) && (ROUND(y) <= sizey-1) && 
           ( region->inside_region(x,y) || !mask) )
      {
        (*ds_buf)(i,j) = (*buf)(ROUND(x),ROUND(y));
      }
      else
      {
        (*ds_buf)(i,j) = T(0);
      }
    }
  }
  
  return ds_buf;
}

// ----------------------------------------------------------------------
// Non-centered version
// ----------------------------------------------------------------------
template<class T> vil1_memory_image_of<T>*
deskew_region_not_centered(const region* region, const vil1_memory_image_of<T>* buf, 
                          int sx, int sy, bool mask)
{
  int sizex = buf->width();
  int sizey = buf->height();
  int i,j;

  vil1_memory_image_of<T>* ds_buf = new vil1_memory_image_of<T>(sx, sy);

  double x0 = region->get_pt0()->x();
  double y0 = region->get_pt0()->y();
  double x1 = region->get_pt1()->x();
  double y1 = region->get_pt1()->y();
  double x2 = region->get_pt2()->x();
  double y2 = region->get_pt2()->y();

  
  for(i=0; i <sx ; i++)
  {
    for(j=0; j <sy ; j++)
    {
      double x = x0 + (x1 - x0) * ((double)i/(double)(sx-1)) 
	            + (x2 - x0) * ((double)j/(double)(sy-1));
      double y = y0 + (y1 - y0) * ((double)i/(double)(sx-1)) 
	            + (y2 - y0) * ((double)j/(double)(sy-1));
      
      if ( ( ROUND(x) >= 0) && (ROUND(x) <= sizex-1) && 
	   ( ROUND(y) >= 0) && (ROUND(y) <= sizey-1) && 
           ( region->inside_region(x,y) || !mask) )
      {
        (*ds_buf)(i,j) = (*buf)(ROUND(x), ROUND(y));
      }
      else
      {
        (*ds_buf)(i,j) = T(0);
      }
    }
  }
  
  return ds_buf;
}
