
#include <region_utility/region_draw.h>
#include <region_utility/misc.h>

// Draw regions regs in solid mode
void draw_regions(regions_list_t& regs, vil1_rgb<unsigned char>& col, color_buffer_t* buf) {
  // Draw all regions in regs one by one
  region* reg;
  for (regions_list_t::iterator it = regs.begin(); it != regs.end(); ++it) {
    reg = *it;
    draw_region(reg, col, buf);
  }
} // draw_regions


// Draw a region in solid mode (ie: all internal pixels set to color 'col')
void draw_region(region* reg, vil1_rgb<unsigned char>& col, color_buffer_t* buf) {
  // loop on all pixels in bounding box
  int w = buf->width();
  int h = buf->height();
  for(int x = (int)reg->get_min_x(); x < (int)(reg->get_max_x()+1); x++) {
    for(int y = (int)reg->get_min_y(); y < (int)(reg->get_max_y()+1); y++) {

      // draw pixel if inside region
      if (reg->inside_region(x, y) && in_image(x, y, w, h)) {
	(*buf)(x, y) = col;
      }

    }
  }
} // draw_region


// Draw a uniform region in solid mode
void draw_uniform_region(uniform_region* ur, vil1_rgb<unsigned char>& col, color_buffer_t* buf) {
  // Segment by segment
  for(int s=0; s < ur->number_of_segments(); s++) {
    point_t* start = ur->get_segment_start(s); 
    point_t* end = ur->get_segment_end(s);

    // Scan horizontal segment, point by point
    int y = (int)start->y();                           // segments are horizontal -> y constant
    for(int x=(int)start->x(); x<=end->x(); x++) {
      (*buf)(x, y) = col;
    }
  }
}


// Percentage (share) of reg's surface overlapping with
// col's parts of buf.
// output in [0,1]
float overlap_share(region* reg, vil1_rgb<unsigned char>& col, color_buffer_t* buf) {
  int tot_pixels = 0;
  int overlap_pixels = 0;
  int w = buf->width();
  int h = buf->height();

  // loop on all pixels in bounding box
  for(int x = (int)reg->get_min_x(); x < (int)(reg->get_max_x()+1); x++) {
    for(int y = (int)reg->get_min_y(); y < (int)(reg->get_max_y()+1); y++) {

      // draw pixel if inside region
      if (reg->inside_region(x, y) && in_image(x, y, w, h)) {
	tot_pixels++;
	vil1_rgb<unsigned char> pixel = (*buf)(x, y);
	if(pixel == col) {
	  overlap_pixels++;
	}
      } // inside region ?

    } // loop over x
  } // loop over y

  if(tot_pixels > 0)
    return (float)overlap_pixels/(float)tot_pixels;
  else
    return 0.0;
}


// Percentage of reg's surface with color close to col
// with col in {black, white}
float color_share(region* reg, vil1_rgb<unsigned char>& col, color_buffer_t* buf) {
  int tot_pixels = 0;
  int col_pixels = 0;
  int w = buf->width();
  int h = buf->height();

  // loop on all pixels in bounding box
  for(int x = (int)reg->get_min_x(); x < (int)(reg->get_max_x()+1); x++) {
    for(int y = (int)reg->get_min_y(); y < (int)(reg->get_max_y()+1); y++) {

      // is this pixel close to col ?
      if (reg->inside_region(x, y) && in_image(x, y, w, h)) {
	tot_pixels++;
	vil1_rgb<unsigned char> pixel = (*buf)(x, y);

	// col = black case
	if(col.R() == 0 && pixel.R() < 70 && pixel.G() < 70 && pixel.B() < 70)
	  col_pixels++;
	
	// col = white case
	if(col.R() > 0 &&
	   pixel.R() > col.R() &&
	   pixel.G() > col.G() &&
	   pixel.B() > col.B() &&
	   fabs(pixel.G() - pixel.R()) < 30 &&
	   fabs(pixel.G() - pixel.B()) < 30 &&
	   fabs(pixel.B() - pixel.R()) < 30)
	  col_pixels++;

      } // inside region ?
    } // loop over y
  } // loop over x

  return (float)col_pixels/(float)tot_pixels;
}
