#ifndef _saliency_computor
#define _saliency_computor

//------------------------------------------------------------------------------
// .NAME  saliency_computor - Base Class for computing the saliency
//                            of an affinely invariant region
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   saliency_computor.h
// .FILE   saliency_computor.cpp
//
// .SECTION Description
//
// .SECTION Author:
//          Vittorio Ferrari
//                ETHZ     
//
// .SECTION Modifications:
//
//
//----------------------------------------------------------------------------

#include <vcl_cmath.h>
#include <vcl_iostream.h>

#include <vil1/vil1_memory_image_of.h>
#include <vil1/vil1_rgb_byte.h>

#include <misc_utility/defines.h>
#include <regions/region.h>

class saliency_computor{

 public:
  saliency_computor() {}
  saliency_computor(const color_buffer_t* buf);
  virtual ~saliency_computor();

 public:
  inline float get_saliency(const region* r);
  inline void set_adaptive_resolution(bool ar) {_adaptive_res = ar;}

  //--- parameter setting ---------------------------------------------
  inline void set_resolution(int res) {_res = res; _min_res = res;}
  inline int get_resolution() const {return _res;}

  // --- utility functions ------------------------------------------
  inline vil1_rgb_byte get_col_interpolation_Vic(const color_buffer_t* buf, float x, float y);

protected:
  //--- computations ---------------------------------------------------
  virtual float compute_saliency() = 0;
  int region_resolution(const region* r);
  void adapt_resolution();

 protected:
  const region* _region;

  int _res;                                         // resolution used for computations (default 50)
  bool _adaptive_res;                               // adapt the resolution to the size of the regions ? 
  int _min_res;                                     // minimal resolution (= _res if not adaptive)
  
  const color_buffer_t* _buf;
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

inline float saliency_computor::get_saliency(const region* r) {
  _region = r;
  if(_adaptive_res) adapt_resolution();
  return compute_saliency();
}


inline vil1_rgb_byte saliency_computor::get_col_interpolation_Vic(const color_buffer_t* buf,
								 float x, float y) {
  // Fast rounding, wrong if x,y exact integers
  int xl = (int)x;
  int xh = xl+1;
  int yl = (int)y;
  int yh = yl+1;

  if(xl <= 0 || yl <= 0 || xh >= buf->width() || yh >= buf->height()) return 0;

  vil1_rgb_byte ll = (*buf)(xl,yl);
  vil1_rgb_byte lh = (*buf)(xl,yh);
  vil1_rgb_byte hl = (*buf)(xh,yl);
  vil1_rgb_byte hh = (*buf)(xh,yh);
  vxl_byte r = (vxl_byte) ( (xh-x) * ((yh-y)*ll.R() + (y-yl)*lh.R()) +
			    (x-xl) * ((yh-y)*hl.R() + (y-yl)*hh.R()) )/(xh-xl)/(yh-yl);
  vxl_byte g = (vxl_byte) ( (xh-x) * ((yh-y)*ll.G() + (y-yl)*lh.G()) +
			   (x-xl) * ((yh-y)*hl.G() + (y-yl)*hh.G()) )/(xh-xl)/(yh-yl);
  vxl_byte b = (vxl_byte) ( (xh-x) * ((yh-y)*ll.B() + (y-yl)*lh.B()) +
			    (x-xl) * ((yh-y)*hl.B() + (y-yl)*hh.B()) )/(xh-xl)/(yh-yl);

  return vil1_rgb_byte(r,g,b);
}

#endif
