#include <regions/ell_region.h>
#include <regions/ellipse_dist_func.h>

#include <vcl/vcl_iostream.h>
#include <vcl/vcl_list.h>
#include <vnl/vnl_vector.h>
#include <vnl/algo/vnl_amoeba.h>
#include <vnl/algo/vnl_rpoly_roots.h>
#include <vgl/algo/vgl_homg_operators_2d.h>
#include <vgl/vgl_box_2d.h>
#include <vsol/vsol_point_2d.h>

//-------------------------------------------------------------------
//--- CONSTRUCTORS / DESTRUCTORS ------------------------------------
//-------------------------------------------------------------------

ell_region::ell_region()
{
  _ell = new vsol_conic_2d(0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
  _pmax = new point_t;
  _pmin = new point_t;
	
  _MaxX = 0;
  _MinX = 0;
  _MaxY = 0;
  _MinY = 0;
	
  _type = region::ELL_REGIONTYPE;

  _ID = _next_ID;
  _next_ID++;
}

ell_region::ell_region(point_t* p0, point_t* p1, point_t* p2,
                       region_specific_type stype, 
		       const point_t* anchor, vil1_image* image)
  : region(p0, p1, p2, region::ELL_REGIONTYPE, stype, anchor, image)
{
  _type = region::ELL_REGIONTYPE;
  _ell = NULL;
  _pmin = new point_t;
  _pmax = new point_t;

  _MaxX = 0;
  _MinX = 0;
  _MaxY = 0;
  _MinY = 0;
	
  update_region();
}

ell_region::ell_region(const ell_region& reg) : region(reg)
{
  _ell = new vsol_conic_2d(0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
  _pmin = new point_t;
  _pmax = new point_t;

  _MaxX = 0;
  _MinX = 0;
  _MaxY = 0;
  _MinY = 0;
	
  update_region();
}

ell_region::~ell_region()
{
  if ( _ell )
  {
    delete _ell;
  }

  if ( _pmin )
  {
    delete _pmin;
  }

  if ( _pmax )
  {
    delete _pmax;
  }
}

//-------------------------------------------------------------------
//--- Data Access ---------------------------------------------------  
//-------------------------------------------------------------------


// Appends the list {_pt0, _pt1, _pt2} to pts
void ell_region::push_points(vcl_list<point_t* >& pts) const
{
  pts.push_back(get_pt0());
  pts.push_back(get_pt1());
  pts.push_back(get_pt2());
}


//-------------------------------------------------------------------
//--- Computations ------------------------------------------  
//-------------------------------------------------------------------

bool ell_region::inside_region(double x, double y) const
{
  return ( (_ell->a()*x*x + _ell->b()*x*y + _ell->c()*y*y + _ell->d()*x + _ell->e()*y + _ell->f()) * _inside <= 0 );
}

double ell_region::distance_from(const region& reg) const
{
  //  This function just splits up the distances, so it can compute the distance
  //  to an elliptic region differently than a parallelogram region.
  if ( reg.get_type() == PARALL_REGIONTYPE )
  {
    return distance_from_para(*(reg.cast_to_parallregion()));
  }
  else if ( reg.get_type() == ELL_REGIONTYPE )
  {
    return distance_from_ellip(*(reg.cast_to_ellregion()));
  }
	
  return 0.0;
}

/* keep the same elliptical region, but change the position of pt1 and pt2,
   corresponding to a rotation over theta (degrees) in the deskewed (circular) region */
bool ell_region::rotate_axes(double theta)
{
  // compute new coordinates for pt1 and pt2
  theta *= ABL_PI/180.0;
  double cosinus = cos(theta);
  double sinus = sin(theta);
	
  double a, b, x, y, p;
  _ell->ellipse_parameters(x, y, p, a, b);
	
  double cost = cos(p);
  double sint = sin(p);
	
  double x1 = _pt0->x()
    + (cost*cosinus + sint*sinus*b/a) * ( cost*(_pt1->x() - _pt0->x()) + sint*(_pt1->y() - _pt0->y()))
    + (cost*sinus*a/b - sint*cosinus) * (-sint*(_pt1->x() - _pt0->x()) + cost*(_pt1->y() - _pt0->y()));
  double x2 = _pt0->x()
    + (cost*cosinus + sint*sinus*b/a) * ( cost*(_pt2->x() - _pt0->x()) + sint*(_pt2->y() - _pt0->y()))
    + (cost*sinus*a/b - sint*cosinus) * (-sint*(_pt2->x() - _pt0->x()) + cost*(_pt2->y() - _pt0->y()));
	
  double y1 = _pt0->y()
    + (sint*cosinus - cost*sinus*b/a) * ( cost*(_pt1->x() - _pt0->x()) + sint*(_pt1->y() - _pt0->y()))
    + (sint*sinus*a/b + cost*cosinus) * (-sint*(_pt1->x() - _pt0->x()) + cost*(_pt1->y() - _pt0->y()));
  double y2 = _pt0->y()
    + (sint*cosinus - cost*sinus*b/a) * ( cost*(_pt2->x() - _pt0->x()) + sint*(_pt2->y() - _pt0->y()))
    + (sint*sinus*a/b + cost*cosinus) * (-sint*(_pt2->x() - _pt0->x()) + cost*(_pt2->y() - _pt0->y()));
	
  // change pt1 and pt2
  _pt1->set(x1, y1);
  _pt2->set(x2, y2);
	
  // update ellipse, inside variables, etc.
  update_region();
	
  return true;
}

/* keep the same elliptical region, but change the position of pt1 and pt2,
   corresponding to a rotation over theta (radians) in the deskewed (circular) region
   do not update the region, used only for refinement purposes */
bool ell_region::rotate_axes_fast(double theta)
{
  // p1 coords in the unit, centered circle
  double cost = cos(theta);
  double sint = sin(theta);
	
  // move pt1
  double x1 = _pt0->x() + cost*(_pt1->x()-_pt0->x()) + sint*(_pt2->x()-_pt0->x());
  double y1 = _pt0->y() + cost*(_pt1->y()-_pt0->y()) + sint*(_pt2->y()-_pt0->y());
  
  // p2 coords in the unit, centered circle
  cost = cos(theta+3.1416/2.0);
  sint = sin(theta+3.1416/2.0);
  double x2 = _pt0->x() + cost*(_pt1->x()-_pt0->x()) + sint*(_pt2->x()-_pt0->x());
  double y2 = _pt0->y() + cost*(_pt1->y()-_pt0->y()) + sint*(_pt2->y()-_pt0->y());

  // change pt1 and pt2
  _pt1->set(x1, y1);
  _pt2->set(x2, y2);
	
  return true;
}


//-------------------------------------------------------------------
//--- Protected functions -------------------------------------------
//-------------------------------------------------------------------

bool ell_region::update_region()
{
  bool ok = true;
	
  // update ellipse
  if (!compute_ellipse())
  {
    ok = false;
  }

  if ( !compute_boundingbox() )
  {
    ok = false;
  }
	
  static double a, b, x, y, p;
  _ell->ellipse_parameters(x, y, p, a, b);
	
  // update pmax, pmin
  _pmax->set(_pt0->x() + a * cos(p), _pt0->y() + a * sin(p));
  _pmin->set(_pt0->x() - b * sin(p), _pt0->y() + b * cos(p));
	
  // update 'inside' variables
  if (!set_inside_variables())
  {
    ok = false;
  }
	
  // update regiondisplayer, if necessary
  if (_display)
  {
    if (!_display->update())
    {
      ok = false;
    }
  }
	
  return ok;
}  

/* compute the ellipse based on pt0, pt1 and pt2 */
bool ell_region::compute_ellipse()
{
  if ( (_pt0 == NULL) || (_pt1 == NULL) || (_pt2 == NULL) )
  {
    vcl_cerr << "cannot compute ellipse if center, pt1 and pt2 are NULL" << vcl_endl; 
    return false;
  }

  static double x1, y1, x2, y2, A, B, C, N, theta, tg, a, b, tmp;
	
  x1 = _pt1->x() - _pt0->x();
  y1 = _pt1->y() - _pt0->y();
  x2 = _pt2->x() - _pt0->x();
  y2 = _pt2->y() - _pt0->y();
	
  A = y1*y1+y2*y2;
  B = -2*(x1*y1 + x2*y2);
  C = x1*x1 + x2*x2;
	
  N = fabs(x1*y2 - x2*y1);
	
  if (A != C)
  {
    theta = atan2( B, (A - C) );
  }
  else 
  {
    theta = ABL_PI/2;
  }

  theta /= 2;
	
  tg = tan(theta);
	
  if (A + B*tg + C*tg*tg < 0)
  {
    vcl_cerr << "error in compute ellipse: A + B * tg + C * tg*tg < 0 " << vcl_endl; 
    return false;
  }

  tmp = sqrt(A + B*tg + C*tg*tg );
	
  if (tmp != 0)
  {
    a = N / tmp * sqrt(1.0+tg*tg);
  }
  else 
  {
    a = 10; 
    vcl_cerr << "ERROR in Ellipse: a = inf" << vcl_endl; 
    return false;
  }
	
	
  if (tg != 0)
  {
    if (A - B/tg + C/tg/tg < 0)
    {
      vcl_cerr << "error in compute ellipse: A - B * tg + C * tg*tg <0 " << vcl_endl; 
      return false;
    }

    tmp = sqrt(A - B/tg + C/tg/tg );
		
    if (tmp != 0)
    {
      b = N / tmp * sqrt(1.0+1.0/tg/tg); 
    }
    else 
    {
      b = 10; 
      vcl_cerr << "ERROR in Ellipse: b = inf" << vcl_endl; 
      return false;
    }
  }
  else 
  {
    if (C != 0)
    {
      b = N / sqrt(C);
    }
    else
    {
      vcl_cerr << " ERROR in Ellipse: B = C = 0 -> no ellipse" << vcl_endl; 
      return false;
    }
  }
	
  if (_ell)
  {
    delete _ell;
  }
	
  vsol_point_2d P(_pt0->x(), _pt0->y());
  _ell = new vsol_conic_2d(P, a, b, theta);	
	
  // set conic matrix
  //_M = _ell->matrix();
	
  return true;
}

bool ell_region::set_inside_variables()
{
  _inside = 1;
  if ( !inside_region(*(_pt0)) )
  {
    _inside *= -1;
    //_M *= -1.0;
  }
	
  return true;
}


//  This function Computes the distance from a parallelogram to this elliptical
//  region.  It is either 0, or it is the min distance from one of the four
//  line segments which makes up the parallelogram.  It is 0 if one region is
//  contained in the other, or if the boundaries of the two regions overlap.
double ell_region::distance_from_para(const parall_region &reg) const
{
  if ( inside_region(*(reg.get_pt0())) )
  {
    return 0.0;
  }
	
  if ( reg.inside_region(_pt0->x(), _pt0->y()) )
  {
    return 0.0;
  }
	
  static double Dist[4];
	
  Dist[0] = distance_to_linesegment(reg.get_pt0(), reg.get_pt1());
  Dist[1] = distance_to_linesegment(reg.get_pt3(), reg.get_pt1());
  Dist[2] = distance_to_linesegment(reg.get_pt0(), reg.get_pt2());
  Dist[3] = distance_to_linesegment(reg.get_pt3(), reg.get_pt2());
	
  return MINFOUR(Dist[0], Dist[1], Dist[2], Dist[3]);
}

//  This function returns the distance from the interior of this ellipse to the line segment p0p1.
//  The min line segment connecting this ellipse to p0p1 is normal to this ellipse and
//  either through p0, through p1, or perpendicular to p0p1.  First we will find
//  the segment that is normal and orthogonal.  If it is not inside the segment, then
//  we will calculate the distance to the endpoints.  First we check if they overlap.
//  The line segment is parameterized as y=mx+t or x=ym+t.
double ell_region::distance_to_linesegment(point_t *p0, point_t *p1) const
{
  if ( line_segment_intersect(p0, p1) || inside_region(*p0) )
  {
    return 0.0;
  }
	
  static double a, b, c, d, e, f, X0, X1, Y0, Y1, x1, y1, x2, y2, m, t;
  static double A, B, C, u1, u2, Dist[2], delta, t1, t2;
  bool orient; 
	
  X0 = p0->x();	Y0 = p0->y();
  X1 = p1->x();	Y1 = p1->y();
	
  // get parameters of ellipse
  a = _ell->a();
  b = _ell->b();
  c = _ell->c();
  d = _ell->d();
  e = _ell->e();
  f = _ell->f();
	
  // get parametrisation of line
  if ( (fabs( X0 - X1 ) >= fabs( Y0 - Y1 )) )
  {
    if ( Y0 == Y1 )
    {
      orient = false;
      m = (-b)/(2*a);
      t = (-d)/(2*a);
    }
    else
    {
      orient = true;
      delta = -(Y0 - Y1)/(X0 - X1);
      m = (2*a-delta*b)/(2*c*delta-b);
      t = (d-delta*e)/(2*c*delta-b);
    }
  }
  else
  {
    orient = true;
    if ( X0 == X1 )
    {
      m = (-b)/(2*c);
      t = (-e)/(2*c);
    }
    else
    {
      delta = -( X0 - X1 )/( Y0 - Y1 );
      m = (b-2*delta*a)/(delta*b-2*c);
      t = (e-delta*d)/(delta*b-2*c);
    }
  }
	
  if ( orient )
  {
    A = a + b*m + c*m*m;
    B = b*t + 2*c*m*t + d + e*m;
    C = c*t*t + e*t + f;
		
    if ( B*B - 4*A*C < 0.0 )
    {
      return 0.0;
    }
		
    x1 = (-B + sqrt(B*B-4*A*C))/(2*A);
    y1 = m*x1 + t;
    x2 = (-B - sqrt(B*B-4*A*C))/(2*A);
    y2 = m*x2 + t;
  }
  else
  {
    A = a*m*m + b*m + c;
    B = 2*a*m*t + b*t + d*m + e;
    C = a*t*t + d*t + f;
		
    if ( B*B - 4*A*C < 0.0 )
    {
      return 0.0;
    }
		
    y1 = (-B + sqrt(B*B-4*A*C))/(2*A);
    x1 = m*y1 + t;
    y2 = (-B - sqrt(B*B-4*A*C))/(2*A);
    x2 = m*y2 + t;
  }
	
  // check whether normal intersects line segment or not
  u1 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
    ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
	
  u2 = ( ( x2 - X0 )*( X1 - X0 ) + ( y2 - Y0 )*( Y1 - Y0 ) ) / 
    ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
	
  // compute distance (two possibilities)
  if ( (u1<1.0 && u1>0.0) || (u2<1.0 && u2>0.0) )
  {
    if ( u1 >= 1.0 || u1 <= 0.0 )
    {
      Dist[0] = ABL_BIGNUM;
    }
    else
    {
      a = X0 + u1*(X1 - X0);
      b = Y0 + u1*(Y1 - Y0);
      Dist[0] = (a-x1)*(a-x1) + (b-y1)*(b-y1);
    }
		
    if ( u2 >= 1.0 || u2 <= 0.0 )
    {
      Dist[1] = ABL_BIGNUM;
    }
    else
    {
      a = X0 + u2*(X1 - X0);
      b = Y0 + u2*(Y1 - Y0);
      Dist[1] = (a-x2)*(a-x2) + (b-y2)*(b-y2);
    }
		
    return sqrt(MINTWO(Dist[0], Dist[1]));
  }
	
	
  //  Since the min distance is not orthogonal to the line segment, it must go through one of the
  //  endpoints.  First we will compute the distance to the first endpoint, then to the second.  We
  //  do this by computing the point on the ellipse that is nearest to each endpoint.
  t1 = vgl_homg_operators_2d<double>::distance_squared(*_ell,*p0);
  t2 = vgl_homg_operators_2d<double>::distance_squared(*_ell,*p1);
	
  return sqrt(MINTWO(t1, t2));
}


//  This function returns true if the line segment defined by p0p1 intersects the ellipse.
//  Here the line is parameterized by y=mx+t or x=ym+t
//  Look out: function returns false if line segment lies entirely inside ellipse
bool ell_region::line_segment_intersect(point_t *p0, point_t *p1) const
{
  static double X0, X1, Y0, Y1, m, t, x1, y1, gamma, sigma, delta;
  static double a, b, c, d, e, f, x0, y0, u1, u2;
  static bool orient;
	
  X0 = p0->x();	Y0 = p0->y();
  X1 = p1->x();	Y1 = p1->y();
	
  // get parameters of ellipse
  a = _ell->a();
  b = _ell->b();
  c = _ell->c();
  d = _ell->d();
  e = _ell->e();
  f = _ell->f();
	
  // get parametrization of line
  // and parameters of quadratic equation
  if ( fabs( X0 - X1 ) >= fabs( Y0 - Y1 ) )
  {
    orient = true;
    m = ( Y0 - Y1 ) / ( X0 - X1 );
    t = Y0 - m*X0;
		
    delta = a + b*m + c*m*m;
    gamma = b*t + 2*c*m*t + d + e*m;
    sigma = c*t*t + e*t + f;
  }
  else
  {
    orient = false;
    m = ( X0 - X1 ) / ( Y0 - Y1 );
    t = X0 - m*Y0;
		
    delta = a*m*m + b*m + c;
    gamma = 2*a*m*t + b*t +d*m + e;
    sigma = a*t*t + d*t + f;
  }
	
	
  // check if line intersects ellipse
  if ( gamma*gamma - 4*delta*sigma < 0.0 )
  {
    return false;
  }
	
  // check if lineSEGMENT intersects ellipse
  if ( orient )
  {
    x0 = (-gamma + sqrt(gamma*gamma - 4*delta*sigma))/(2*delta);
    y0 = m*x0 + t;
		
    x1 = (-gamma - sqrt(gamma*gamma - 4*delta*sigma))/(2*delta);
    y1 = m*x1 + t;
		
    u1 = ( ( x0 - X0 )*( X1 - X0 ) + ( y0 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
		
    u2 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
  }
  else
  {
    y0 = (-gamma + sqrt(gamma*gamma - 4*delta*sigma))/(2*delta);
    x0 = m*y0 + t;
		
    y1 = (-gamma - sqrt(gamma*gamma - 4*delta*sigma))/(2*delta);
    x1 = m*y1 + t;
		
    u1 = ( ( x0 - X0 )*( X1 - X0 ) + ( y0 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
		
    u2 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
  } 
	
  if ( ( u1 <= 1.0 && u1 >= 0.0 ) || ( u2 <= 1.0 && u2 >= 0.0 ) )
  {
    return true;
  }
	
  return false;
}


//  This computes the distance from the interior of the elliptical region reg to
//  the interior of this elliptical region.  The distance is 0 if the two regions
//  overlap or if one is contained in the other.  Otherwise it is
//  min||x-y|| where x is on reg and y is on this ellipse.
double ell_region::distance_from_ellip(const ell_region &reg) const
{
  //  First we check if one is inside the other.
  if ( reg.inside_region(_pt0->x(), _pt0->y()) )
  {
    return 0.0;
  }
	
  if ( inside_region(*(reg.get_pt0()) ))
  {
    return 0.0;
  }
	
  //  Neither ellipse is contained in the other, so we will have a true distance.  
  //  Thus we will call our friend Amoeba.
  vnl_vector<double> ParamVec(2);
	
  vsol_conic_2d* TempEll = reg.get_ellipse();
  point_t *reg_center = reg.get_pt0();
	
  ParamVec(0) = _pt0->x();
  ParamVec(1) = reg_center->x();
	
  EllipseDistFunc ErrorFunc(_ell, TempEll);
  vnl_amoeba am(ErrorFunc);
  am.minimize(ParamVec);
	
  return ErrorFunc.f(ParamVec);
}


// This function returns true if reg elliptic region intersects this elliptic region
bool ell_region::ellip_overlap(const ell_region &reg) const
{
	
  vsol_conic_2d *TempEll = reg.get_ellipse();
	
  static double a1, a2, b1, b2, c1, c2, d1, d2, e1, e2, f1, f2, a, b, c, d, e, f;
	
  // get parameters of ellipse
  a = _ell->a();
  b = _ell->b();
  c = _ell->c();
  d = _ell->d();
  e = _ell->e();
  f = _ell->f();
	
  a2 = TempEll->a();
  b2 = TempEll->b();
  c2 = TempEll->c();
  d2 = TempEll->d();
  e2 = TempEll->e();
  f2 = TempEll->f(); 
	
  a = a1 - a2;
  b = b1 - b2;
  c = c1 - c2;
  d = d1 - d2;
  e = e1 - e2;
  f = f1 - f2;
	
  if ( (-d*b*a*e + a*a*e*e + b*b*a*f + a*c*d*d - 4*a*a*c*f) < 0 )
  {
    return false;
  }
	
  return true;
}


bool ell_region::compute_boundingbox()
{
  vgl_box_2d<double> BoundingBox = vgl_homg_operators_2d<double>::compute_bounding_box(*_ell);
  _MinX = (int)BoundingBox.min_x();
  _MinY = (int)BoundingBox.min_y();
  _MaxX = (int)BoundingBox.max_x() + 1;
  _MaxY = (int)BoundingBox.max_y() + 1;

  // Workaround bug in compute_bounding_box of VxL
  float abs_min_x1 = MINTWO(_pt1->x(), _pt2->x());
  float abs_min_x2 = MINTWO(2*_pt0->x()  - _pt1->x(), 2*_pt0->x() - _pt2->x());
  float abs_min_x  = MINTWO(abs_min_x1, abs_min_x2);
  
  float abs_min_y1 = MINTWO(_pt1->y(), _pt2->y());
  float abs_min_y2 = MINTWO(2*_pt0->y()  - _pt1->y(), 2*_pt0->y() - _pt2->y());
  float abs_min_y  = MINTWO(abs_min_y1, abs_min_y2);
  if(_MinX < 2*abs_min_x - _pt0->x() - 2) {     // RHS seems to be real practical limit
    vcl_cout << "WARNING: Bug in vgl_homg_operators_2d<double>::compute_bounding_box(*_ell). _MinX == " << _MinX << " ";
    _MinX = (int)abs_min_x;
    vcl_cout << "Fixing it to : " << _MinX << vcl_endl;
    vcl_cout << "REG." << *(_pt0) << *(_pt1) << *(_pt2) << vcl_endl;
  }
  if(_MinY < 2*abs_min_y - _pt0->y() - 2) {
    vcl_cout << "WARNING: Bug in vgl_homg_operators_2d<double>::compute_bounding_box(*_ell). _MinY == " << _MinY << " ";
    _MinY = (int)abs_min_y;
    vcl_cout << "Fixing it to : " << _MinY << vcl_endl;
    vcl_cout << "REG." << *(_pt0) << *(_pt1) << *(_pt2) << vcl_endl;
  }
  
  return true;
}

//  This function computes the parametrization of the two lines that go through point anchor
//  and are tangent to the ellipse.  It works well, but is not currently used for anything.
//  (I thought I would need it, but then I didn't.)
bool ell_region::get_tangentlines_through_point(const point_t &anchor, vsol_conic_2d *Ell, 
                                                double &m1, double &b1, double &m2, double &b2)
{
  /*
    This is somewhat complicated.  We have to solve two quadratics in two unknowns.
    ax^2 + bxy + cy^2 + dx + ey + f = 0		// equation of the ellipse
    (y-y0)/(x-x0) = -(2ax+by+d)/(bx+2cy+e)	// equation which links (x,y) on the ellipse
    // to 'anchor' and makes sure the line connecting them
    // is tangent
    // to the ellipse.  This actually reduces to a
    // quadratic (not a quartic).  Each solution in y
    // though gives two solutions in x, so we have a 
    // total of four solutions.
    we use the parametrization y=mx+b for both lines, so if they are vertical, then there can
    be problems.  
  */
  static double a, b, c, d, e, f, y1, y2, x11, x21, x12, x22, y0, x0, A, B, C;
  int k;

  x0 = anchor.x();
  y0 = anchor.y();

  a = Ell->a(); b = Ell->b();
  c = Ell->c(); d = Ell->d();
  e = Ell->e(); f = Ell->f();

  //  The coeeficients of the quadractic that we solve Ay^2 + By + C = 0
  A = (-e/a*b*d/4.0-b*b*b*y0/a*x0/4.0-x0*x0*b*b/4.0+e*e/4.0+b*y0*x0*c+y0*y0*c*c+d*d/a*c/4.0 -
       e/a*y0*b*b/4.0-d/a*x0*b*b/4.0+x0*x0*a*c+d*x0*c+e*y0*c-b*b*y0*y0/a*c/4.0);
  B = (e*e*y0/2.0-1/a*y0*b*b*f/2.0+b*y0*x0*e+e*f-1/a*b*d*f/2.0-d*d/a*y0*c/2.0+2.0*y0*c*f +
       d*x0*e+y0*y0*c*e-x0*x0*b*d/2.0-d*d/a*x0*b/2.0+x0*x0*a*e-d/a*x0*y0*b*b/2.0-d/a*b*y0*y0*c/2.0);
  C = f*a*x0*x0-d*d*d/a*x0/4.0-f/a*d*d/4.0+f*y0*e+f/a*y0*y0*b*b/4.0-d*d/a*y0*b*x0/4.0+f*d*x0 +
    f*x0*y0*b+y0*y0*e*e/4.0-d*d/a*y0*e/4.0-d*d*x0*x0/4.0+f*f-d/a*b*y0*y0*e/4.0;

  y1 = (-B+sqrt(B*B-4*A*C))/(2*A);
  y2 = (-B-sqrt(B*B-4*A*C))/(2*A);

  x11 = 1/a*(-d-b*y1+sqrt(d*d+2.0*d*b*y1+b*b*y1*y1-4.0*a*f-4.0*a*c*y1*y1-4.0*a*e*y1))/2.0;
  x12 = 1/a*(-d-b*y1-sqrt(d*d+2.0*d*b*y1+b*b*y1*y1-4.0*a*f-4.0*a*c*y1*y1-4.0*a*e*y1))/2.0;

  x21 = 1/a*(-d-b*y2+sqrt(d*d+2.0*d*b*y2+b*b*y2*y2-4.0*a*f-4.0*a*c*y2*y2-4.0*a*e*y2))/2.0;
  x22 = 1/a*(-d-b*y2-sqrt(d*d+2.0*d*b*y2+b*b*y2*y2-4.0*a*f-4.0*a*c*y2*y2-4.0*a*e*y2))/2.0;

  vnl_vector<double> Val(4);
  Val(0) = a*x11*x11 + b*x11*y1 + c*y1*y1 + d*x11 + e*y1 + f;
  Val(1) = a*x12*x12 + b*x12*y1 + c*y1*y1 + d*x12 + e*y1 + f;
  Val(2) = a*x21*x21 + b*x21*y2 + c*y2*y2 + d*x21 + e*y2 + f;
  Val(3) = a*x22*x22 + b*x22*y2 + c*y2*y2 + d*x22 + e*y2 + f;

  //  Make sure the points we get are on the ellipse (all four should be).
  for(k=0;k<4;k++)
  {
    if ( Val(k) > 1e-10 )
    {
      return false;
    }
  }

  //  Find which two points are correctly tangent to the ellipse (only two should be).
  static double Slope1, Slope2, Slope3, Slope4, Tan1, Tan2, Tan3, Tan4;
  Slope1 = (y1 - y0)/(x11 - x0); Slope2 = (y1 - y0)/(x12 - x0);
  Slope3 = (y2 - y0)/(x21 - x0); Slope4 = (y2 - y0)/(x22 - x0);

  Tan1 = -(2*a*x11 + b*y1 + d)/(b*x11 + 2*c*y1 + e);
  Tan2 = -(2*a*x12 + b*y1 + d)/(b*x12 + 2*c*y1 + e);
  Tan3 = -(2*a*x21 + b*y2 + d)/(b*x21 + 2*c*y2 + e);
  Tan4 = -(2*a*x22 + b*y2 + d)/(b*x22 + 2*c*y2 + e);

  if ( fabs(Tan1-Slope1) < 1e-10 )
  {
    m1 = Slope1;
    b1 = y1 - Slope1*x11;
  }
  else if ( fabs(Tan2-Slope2) < 1e-10 )
  {
    m1 = Slope2;
    b1 = y1 - Slope2*x12;
  }
  else
  {
    return false;
  }

  if ( fabs(Tan3-Slope3) < 1e-10 )
  {
    m2 = Slope3;
    b2 = y2 - Slope3*x21;
  }
  else if ( fabs(Tan4-Slope4) < 1e-10 )
  {
    m2 = Slope4;
    b2 = y2 - Slope4*x22;
  }
  else
  {
    return false;
  }

  return true;
}

void ell_region::set_min_inertia_points(point_t* s, point_t* e) 
{
  _min_inertia_pts = new vcl_list<point_t* >;
  _min_inertia_pts->push_back(s);
  _min_inertia_pts->push_back(e);
}


void ell_region::set_max_inertia_points(point_t* s, point_t* e) 
{
  _max_inertia_pts = new vcl_list<point_t* >;
  _max_inertia_pts->push_back(s);
  _max_inertia_pts->push_back(e);
}
