#ifndef ell_region_h
#define ell_region_h

//----------------------------------------------------------------------------
// .NAME ell_region - Class for elliptical (affinely invariant) regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   ell_region.h
// .FILE   ell_region.C
//
// .SECTION Description
//    The class ell_region is a class for
//  elliptical (affinely invariant) regions.
//  The internal representation consists of three points,
//  coined 'pt0' (= center of the ellipse), 'pt1' and 'pt2'
//  (= two specific points on the ellipse)
//  Also the ellipse itself is stored and can be accessed.
// 
//  This class was written to be used with the class ell_regionFinder,
//  which allows to find regions in an affinely invariant way,
//
// .SECTION Author:
//          Tinne Tuytelaars - Andrew Turina - Vittorio Ferrari
//          ESAT / KULeuven           BIWI / ETH Zurich
//
// .SECTION Modifications:
//     07/01 Restructured and resynchronized code      
//
//----------------------------------------------------------------------------

/*******************************************************************************
*
*	Ported to vxl by Aaron Luttman
*	Notes:	This class owns the following memory: _ell, _pmin, _pmax.  They should
*	be newed and deleted by this class.
*
*******************************************************************************/

#include <vcl/vcl_iostream.h>
#include <regions/region.h>
#include <regions/parall_region.h>
#include <regions/ell_region_displayer.h>
#include <vgl/vgl_homg_point_2d.h>
#include <vsol/vsol_conic_2d.h>
#include <vnl/vnl_matrix.h>
#include <misc_utility/abl_utility.h>


class vil1_image;

class ell_region: public region
{
  public:
  ell_region();
  ell_region(point_t* p0, point_t* p1, point_t* p2,
             region_specific_type stype = region::REGION_NO_SPECIFIC_TYPE,
             const point_t* anchor = NULL, vil1_image* image = NULL);
  ell_region(const ell_region& reg);
  ~ell_region();
  
 public:
  //--- Data Access -------------------------------------------
  vsol_conic_2d* get_ellipse() const { return _ell; }
  const point_t* get_major_axis_intersection() const { return _pmax; } // does anyone ever use this!!??!
  const point_t* get_minor_axis_intersection() const { return _pmin; } // !!??!

  inline virtual double get_min_x() const { return _MinX; }
  inline virtual double get_min_y() const { return _MinY; }
  inline virtual double get_max_x() const { return _MaxX; }
  inline virtual double get_max_y() const { return _MaxY; }
  
  inline virtual double get_area() const 
    {
      double cx, cy, phi, width, height;
      _ell->ellipse_parameters(cx, cy, phi, width, height);
      return ABL_PI*width*height;
  }

  inline virtual double get_eccentricity() const 
  { 
    double x,y,p,a,b;
    _ell->ellipse_parameters(x,y,p,a,b);
    
    a /= 2.0;
    b /= 2.0;
    
    //return ( a / b > 1.0 ) ? sqrt(a*a-b*b) : sqrt(b*b-a*a); 
    return ( a > b ) ? (1.0 - (b / a)) : (1.0 - (a / b));
  } 

  inline virtual double get_elongatedness() const 
    { 
      return get_eccentricity(); 
  }

  inline virtual bool is_centered() const { return true; }

  /*
  inline virtual double GetTheta() const 
  { 
	  return _theta; 
  }
  */
  

  virtual void push_points(vcl_list<point_t*>& pts) const;

  //--- Computations ------------------------------------------  
  virtual bool inside_region(double x, double y) const;
  virtual inline bool inside_region(const point_t& pt) const { return inside_region(pt.x(), pt.y()); }
  virtual double distance_from(const region& reg) const;
  bool compute_boundingbox();
  inline point_t* center() const {return new point_t(*_pt0);};
  
  //--- region Modification -----------------------------------
  bool rotate_axes(double theta);            // change pt1_ and pt2_ by 'rotating' inside ellipse: crashes !
  bool rotate_axes_fast(double theta);       //  simpler formulas, input=radians and not invoking update()
  virtual inline bool switch_region();
  virtual bool update_region();

  //--- region I/O --------------------------------------------
  virtual inline void print(vcl_ostream& str) const;

  //--- Safe Down Casting -------------------------------------
  ell_region* cast_to_ellregion(){ return this; }
  const ell_region* cast_to_ellregion() const { return this; }


  // Added by aturina 26.3.02
  // Were these methods forgotten to declare ??
  void set_min_inertia_points(point_t* s, point_t* e);
  void set_max_inertia_points(point_t* s, point_t* e);
 
protected:
  virtual inline void create_region_displayer() const; // mutable const
  
  bool compute_ellipse();
  bool set_inside_variables();

  double distance_from_para(const parall_region &reg) const;
  double distance_from_ellip(const ell_region &reg) const;
  bool line_segment_intersect(point_t *p0, point_t *p1) const;
  bool ellip_overlap(const ell_region &reg) const;
  bool para_overlap(const parall_region &reg) const;
  double distance_to_linesegment(point_t *p0, point_t *p1) const;
  bool get_tangentlines_through_point(const point_t &anchor, vsol_conic_2d *Ell, 
                                      double &m1, double &b1, double &m2, double &b2);

  
 
  //--- Data members ----------------------------------------
protected:
  vsol_conic_2d *_ell;
  //vnl_matrix_fixed<double,3,3> _M; // conic matrix (no longer used)
  int _inside;
  point_t *_pmax, *_pmin; // points of ellipse on minor axis
  int _MinX, _MaxX, _MinY, _MaxY;

  vcl_list<point_t* >* _min_inertia_pts;
  vcl_list<point_t* >* _max_inertia_pts;

};

//===========================================================
//=== INLINE FUNCTION DEFINITIONS                         
//===========================================================

//--- region I/O --------------------------------------------
void ell_region::print(vcl_ostream& str) const
{
	/*
  str << "Ell. region " << vcl_endl;
  str << "   Pt0: " << *_pt0 << vcl_endl;
  str << "   Pt1: " << *_pt1 << vcl_endl;
  str << "   Pt2: " << *_pt2 << vcl_endl;
  if (_anchor)
    str << "   anchor: " << *_anchor << vcl_endl;
  if (_image) 
    str << "   in image " << _image->GetName() << vcl_endl;  
  if (_inv) 
    str << "   Inv: " << *_inv << vcl_endl;
	*/
}




//--- Protected Functions -----------------------------------

//mutable const
void ell_region::create_region_displayer() const
{
  _display = new ell_region_displayer(this);
}

bool ell_region::switch_region()
{
  _pt2->set(2 * _pt0->x() - _pt2->x(), 2 * _pt0->y() - _pt2->y());

  return (update_region());
}



#endif // end of class ell_region
