#include <regions/parall_region.h>
#include <regions/ell_region.h>

#include <vcl/vcl_iostream.h>
#include <vcl/vcl_list.h>

#include <vil1/vil1_rgb_byte.h>

//-------------------------------------------------------------------
//--- CONSTRUCTORS / DESTRUCTORS ------------------------------------
//-------------------------------------------------------------------
  
parall_region::parall_region()
{
  _pt3 = new point_t;
  _type = region::PARALL_REGIONTYPE; 

  _ID = _next_ID;
  _next_ID++;
}



parall_region::parall_region(const point_t* p0, const point_t* p1, 
                             const point_t* p2, region_specific_type stype,
                             const point_t* anchor, vil1_image* image)
  : region(p0, p1, p2, region::PARALL_REGIONTYPE, stype, anchor, image)
{
  _pt3 = NULL;
  update_region();
}



parall_region::parall_region(const parall_region& reg) : region(reg)
{
  _pt3 = NULL;
  update_region();
}

parall_region::~parall_region()
{
  if ( _pt3 != NULL )
  {
    delete _pt3;
  }
}


//-------------------------------------------------------------------
//--- Data Access ---------------------------------------------------
//-------------------------------------------------------------------

// ratio of axes of inner ellipse
double parall_region::get_elongatedness() const
{
  static type_t x1, y1, x2, y2, A, B, C, theta, tg, e;

  x1 = _pt1->x() - _pt0->x();
  y1 = _pt1->y() - _pt0->y();
  x2 = _pt2->x() - _pt0->x();
  y2 = _pt2->y() - _pt0->y();
	
  A = y1*y1+y2*y2;
  B = -2*(x1*y1 + x2*y2);
  C = x1*x1 + x2*x2;
	
  if (A != C)
  {
    theta = atan2( B, (A - C) );
  }
  else 
  {
    theta = ABL_PI/2;
  }

  theta /= 2;
	
  tg = tan(theta);
  e = sqrt( (A*tg*tg - B*tg + C)/(A + B*tg + C*tg*tg) );

  if (e<1) 
  {
    e = 1/e;
  }
	
  return e;
}


// Appends the list {_pt0, _pt1, _pt2} to pts
void parall_region::push_points(vcl_list<point_t*>& pts) const
{
  pts.push_back(get_pt0());
  pts.push_back(get_pt1());
  pts.push_back(get_pt2());
}

  
//-------------------------------------------------------------------
//--- Computations --------------------------------------------------  
//-------------------------------------------------------------------
bool parall_region::inside_region(type_t x, type_t y) const
{
  // check line through pt0 and pt1
  if (_inside01) 
  {
    if ( (_pt1->y() - _pt0->y())*(x - _pt0->x()) -
         (_pt1->x() - _pt0->x())*(y - _pt0->y()) < 0) 
    {
      return false;
    }
  }
  else 
  {
    if ( (_pt1->y() - _pt0->y())*(x - _pt0->x()) -
         (_pt1->x() - _pt0->x())*(y - _pt0->y()) > 0) 
    {
      return false;
    }
  }
	
  // check line through pt0 and pt2
  if (_inside02) 
  {
    if ( (_pt2->y() - _pt0->y())*(x - _pt0->x()) -
         (_pt2->x() - _pt0->x())*(y - _pt0->y()) < 0)
    {
      return false;
    }
  }
  else 
  {
    if ( (_pt2->y() - _pt0->y())*(x - _pt0->x()) -
         (_pt2->x() - _pt0->x())*(y - _pt0->y()) > 0)
    {
      return false;
    }
  }
	
  // check line through pt1 and pt3
  if (_inside13) 
  {
    if ( (_pt3->y() - _pt1->y())*(x - _pt1->x()) -
         (_pt3->x() - _pt1->x())*(y - _pt1->y()) < 0)
    {
      return false;
    }
  }
  else 
  {
    if ( (_pt3->y() - _pt1->y())*(x - _pt1->x()) -
         (_pt3->x() - _pt1->x())*(y - _pt1->y()) > 0)
    {
      return false;
    }
  }
    
  // check line through pt2 and pt3
  if (_inside23) 
  {
    if ( (_pt3->y() - _pt2->y())*(x - _pt2->x()) -
         (_pt3->x() - _pt2->x())*(y - _pt2->y()) < 0)
    {
      return false;
    }
  }
  else 
  {
    if ( (_pt3->y() - _pt2->y())*(x - _pt2->x()) -
         (_pt3->x() - _pt2->x())*(y - _pt2->y()) > 0)
    {
      return false;
    }
  }
	
  return true;
}


type_t parall_region::distance_from(const region& reg) const
{
  //  This function just splits up the distances, so it can compute the distance
  //  to an elliptic region differently than a parallelogram region.
  if ( reg.get_type() == PARALL_REGIONTYPE )
  {
    return distance_from_para(*(reg.cast_to_parallregion()));
  }
  else if ( reg.get_type() == ELL_REGIONTYPE )
  {
    return distance_from_ellip(*(reg.cast_to_ellregion()));
  }
	
  return 0.0;
}

//-------------------------------------------------------------------
//--- Protected Functions--------------------------------------------  
//-------------------------------------------------------------------
bool parall_region::update_region()
{
  bool ok = true;
	
  // update pt3
  if ( _pt3 != NULL )
  {
    delete _pt3;
  }
	
  _pt3 = new point_t( _pt1->x() + _pt2->x() - _pt0->x(),
                      _pt1->y() + _pt2->y() - _pt0->y() );
	
	
  // update 'inside' variables
  if (!set_inside_variables())
  {
    ok = false;
  }
	
  // update regiondisplayer, if necessary
  if ( _display )
  {
    if ( !_display->update() )
    {
      ok = false;
    }
  }
	
  return ok;
}

bool parall_region::set_inside_variables()
{
  _inside01 = ((_pt1->y() - _pt0->y())*(_pt2->x() - _pt0->x()) 
               - (_pt1->x() - _pt0->x())*(_pt2->y() - _pt0->y()) > 0);
  _inside02 = ((_pt2->y() - _pt0->y())*(_pt1->x() - _pt0->x()) 
               - (_pt2->x() - _pt0->x())*(_pt1->y() - _pt0->y()) > 0);
  _inside13 = ((_pt3->y() - _pt1->y())*(_pt2->x() - _pt1->x()) 
               - (_pt3->x() - _pt1->x())*(_pt2->y() - _pt1->y()) > 0);
  _inside23 = ((_pt3->y() - _pt2->y())*(_pt1->x() - _pt2->x()) 
               - (_pt3->x() - _pt2->x())*(_pt1->y() - _pt2->y()) > 0);
	
  return true;
}



// ----------------------------------------------------------------------
//  This computes the distance from this parallelogram region to
//  the Reg parallelogram region.  It is the min distance from a line segment
//  of one of the parallelgrams to a line segment of the other.  If they 
//  overlap, then we say the distance is 0.0.
// ----------------------------------------------------------------------
type_t parall_region::distance_from_para(const parall_region &reg) const
{
  static double Dist[16];
  point_t *P0, *P1, *P2, *P3;
	
  P0 = reg.get_pt0();
  P1 = reg.get_pt1();
  P2 = reg.get_pt2();
  P3 = reg.get_pt3();
	
  if ( inside_region(*P0) || reg.inside_region(*_pt0) )
  {
    return 0.0;
  }
	
#define EPSILON 0.00000001
	
  Dist[0] = distance_between_line_segments(P0, P1, _pt0, _pt1);
  if (Dist[0] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[1] = distance_between_line_segments(P0, P1, _pt0, _pt2);
  if (Dist[1] < EPSILON )
  {
    return 0.0;
  }
	
	
  Dist[2] = distance_between_line_segments(P0, P1, _pt3, _pt1);
  if (Dist[2] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[3] = distance_between_line_segments(P0, P1, _pt3, _pt2);
  if (Dist[3] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[4] = distance_between_line_segments(P0, P2, _pt0, _pt1);
  if (Dist[4] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[5] = distance_between_line_segments(P0, P2, _pt0, _pt2);
  if (Dist[5] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[6] = distance_between_line_segments(P0, P2, _pt3, _pt1);
  if (Dist[6] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[7] = distance_between_line_segments(P0, P2, _pt3, _pt2);
  if (Dist[7] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[8] = distance_between_line_segments(P3, P1, _pt0, _pt1);
  if (Dist[8] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[9] = distance_between_line_segments(P3, P1, _pt0, _pt2);
  if (Dist[9] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[10] = distance_between_line_segments(P3, P1, _pt3, _pt1);
  if (Dist[10] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[11] = distance_between_line_segments(P3, P1, _pt3, _pt2);
  if (Dist[11] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[12] = distance_between_line_segments(P3, P2, _pt0, _pt1);
  if (Dist[12] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[13] = distance_between_line_segments(P3, P2, _pt0, _pt2);
  if (Dist[13] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[14] = distance_between_line_segments(P3, P2, _pt3, _pt1);
  if (Dist[14] < EPSILON )
  {
    return 0.0;
  }
	
  Dist[15] = distance_between_line_segments(P3, P2, _pt3, _pt2);
  if (Dist[15] < EPSILON )
  {
    return 0.0;
  }
	
#undef EPSILON
	
  return MINFOUR(MINFOUR(Dist[0], Dist[1], Dist[2], Dist[3]), MINFOUR(Dist[4], Dist[5], Dist[6], Dist[7]),
                 MINFOUR(Dist[8], Dist[9], Dist[10], Dist[11]), MINFOUR(Dist[12], Dist[13], Dist[14], Dist[15]));
}


//  This function tells whether the two line segments p0p1 and p2p3 intersect.  Each line segment is
//  parameterized as y=mx+b or x=ym+b, depending on the slope of the line.
bool parall_region::line_segment_intersect(point_t *p0, point_t *p1, 
                                        point_t *p2, point_t *p3) const
{
  static double s1, s2, i1, i2, u1, y1, x1;
  static bool orient1, orient2;
  static type_t X0, X1, X2, X3, Y0, Y1, Y2, Y3;
	
  X0 = p0->x();	Y0 = p0->y();
  X1 = p1->x();	Y1 = p1->y();
  X2 = p2->x();	Y2 = p2->y();
  X3 = p3->x();	Y3 = p3->y();
	
  if ( fabs(X0 - X1) >= fabs(Y0 - Y1) )
  {
    orient1 = true;
  }
  else
  {
    orient1 = false;
  }
	
  if ( fabs(X2 - X3) >= fabs(Y2 - Y3) )
  {
    orient2 = true;
  }
  else
  {
    orient2 = false;
  }
	
  if ( orient1 && orient2 )
  {
    s1 = ( Y0 - Y1 )/( X0 - X1 );
    s2 = ( Y3 - Y2 )/( X3 - X2 );
		
    if ( s1 == s2 )
    {
      return false;
    }
		
    i1 = Y0 - s1*X0;
    i2 = Y2 - s2*X2; 
		
    x1 = (i2 - i1)/(s1 - s2);
    y1 = s1*x1 + i1;
		
    u1 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
		
    return ( u1 <= 1.0 && u1 >= 0.0);
  }
  else if ( orient1 && !orient2 )
  {
    s1 = ( Y0 - Y1 )/( X0 - Y1 );
    i1 = Y0 - s1*X0;
		
    s2 = ( X3 - X2 )/( Y3 - Y2 );
    i2 = X3 - s2*Y3;
		
    y1 = ( s1*i2 + i1 ) / ( 1 - s1*s2 );
    x1 = s2*y1 + i2;
		
    u1 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
		
    return ( u1 <= 1.0 && u1 >= 0.0);
		
  }
  else if ( !orient1 && orient2 )
  {
    s1 = ( Y2 - Y3 )/( X2 - Y3 );
    i1 = Y2 - s1*X2;
		
    s2 = ( X1 - X0 )/( Y1 - Y0 );
    i2 = X0 - s2*Y0;
		
    y1 = ( s1*i2 + i1 ) / ( 1 - s1*s2 );
    x1 = s2*y1 + i2;
		
    u1 = ( ( x1 - X2 )*( X3 - X2 ) + ( y1 - Y2 )*( Y3 - Y2 ) ) / 
      ( ( X3 - X2 )*( X3 - X2 ) + ( Y3 - Y2 )*( Y3 - Y2 ) );
		
    return ( u1 <= 1.0 && u1 >= 0.0);
  }
  else
  {
    s1 = ( X0 - X1 )/( Y0 - Y1 );
    s2 = ( X3 - X2 )/( Y3 - Y2 );
		
    if ( s1 == s2 )
    {
      return false;
    }
		
    i1 = X0 - s1*Y0;
    i2 = X2 - s2*Y2; 
		
    y1 = (i2 - i1)/(s1 - s2);
    x1 = y1*s1 + i1; 
		
    u1 = ( ( x1 - X0 )*( X1 - X0 ) + ( y1 - Y0 )*( Y1 - Y0 ) ) / 
      ( ( X1 - X0 )*( X1 - X0 ) + ( Y1 - Y0 )*( Y1 - Y0 ) );
		
    return ( u1 <= 1.0 && u1 >= 0.0);
  }
}






// ----------------------------------------------------------------------
//  This computes the distance from a point to a line segment.  It is
//  either the orthogonal projection distance, or it is the distance
//  to one of the endpoints.  The line segment is given by p1p2.
//  ----------------------------------------------------------------------
type_t parall_region::distance_from_point_to_linesegment(point_t *p0, point_t *p1, 
                                                         point_t *p2) const
{
  static type_t d1,d2,u,x,y, X0, X1, X2, Y0, Y1, Y2;
	
  X0 = p0->x();	Y0 = p0->y();
  X1 = p1->x();	Y1 = p1->y();
  X2 = p2->x();	Y2 = p2->y();
	
  u = ( ( X0 - X1 )*( X2 - X1 ) + ( Y0 - Y1 )*( Y2 - Y1 ) ) / 
    ( ( X2 - X1 )*( X2 - X1 ) + ( Y2 - Y1 )*( Y2 - Y1 ) );
	
  if ( u >= 0.0 && u <= 1.0 )
  {
    //  This is the orthogonal distance case
    x = X1 + u*(X2 - X1);
    y = Y1 + u*(Y2 - Y1);
    return sqrt( ( X0 - x )*( X0 - x ) + ( Y0 - y )*( Y0 - y ) );
  }
	
  //  This is the case where the min distance is a distance to one of the endpoints.
  d1 = ( X0 - X1 )*( X0 - X1 ) + ( Y0 - Y1 )*( Y0 - Y1 );
  d2 = ( X0 - X2 )*( X0 - X2 ) + ( Y0 - Y2 )*( Y0 - Y2 );
	
  return sqrt( MINTWO(d1, d2) );
}


//  This calculates the distance between two line segments.  It is given by the min of the distances 
//  defined by each endpoint and the other line segment, or 0.0 if they intersect.
type_t parall_region::distance_between_line_segments(point_t *p0, point_t *p1, 
                                                 point_t *p2, point_t *p3) const
{
  static type_t d1, d2, Min;
	
  if ( line_segment_intersect(p0, p1, p2, p3) )
  {
    return 0.0;
  }
  else
  {
    d1 = distance_from_point_to_linesegment(p0, p2, p3);
    d2 = distance_from_point_to_linesegment(p1, p2, p3);
		
    Min = MINTWO(d1, d2);
		
    d1 = distance_from_point_to_linesegment(p2, p0, p1);
    d2 = distance_from_point_to_linesegment(p3, p0, p1);
		
    return MINTWO(Min, MINTWO(d1, d2));
  }
}






// ----------------------------------------------------------------------
//  This computes the distance from this region to
//  Reg (using the function in reg).
// ----------------------------------------------------------------------
type_t parall_region::distance_from_ellip(const ell_region &reg) const
{
  return reg.distance_from(*this);
}
