#ifndef parall_region_h
#define parall_region_h

//----------------------------------------------------------------------------
// .NAME parall_region - Class for parallelogram-shaped (affinely invariant) regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   parall_region.h
// .FILE   parall_region.C
//
// .SECTION Description
//    The class parall_region is an class for
//  parallelogram-shaped (affinely invariant) regions.
//  Their internal representation consists of three points,
//  coined 'pt0', 'pt1' and 'pt2' (three corners of the parallelogram).
//  Also the fourth corner (pt3) can be accessed.
//
//  This class was written to be used with the class regionFinder,
//  which allows to find regions in an affinely invariant way,
//
// .SECTION Author:
//          Tinne Tuytelaars - Andrew Turina - Vittorio Ferrari
//          ESAT / KULeuven           BIWI / ETH Zurich
//
// .SECTION Modifications:
//     07/01 Restructured and resynchronized code      
//
//----------------------------------------------------------------------------

/*******************************************************************************
*
*	Ported to vxl by Aaron Luttman
*	Notes:	This class owns the following memory: _pt3.
*
*******************************************************************************/


#include <regions/region.h>
#include <regions/parall_region_displayer.h>

#include <vcl/vcl_iostream.h>
#include <math.h>
#include <vgl/vgl_homg_point_2d.h>

class vil1_image;

class parall_region: public region
{
public:
	
public:
  parall_region();
  parall_region(const point_t* p0, const point_t* p1, const point_t* p2,
                region_specific_type stype = region::REGION_NO_SPECIFIC_TYPE,
                const point_t* anchor = NULL, vil1_image* image = NULL);
  
  parall_region(const parall_region& reg);
  
  ~parall_region();
	
public:
  //--- Data Access -------------------------------------------
  inline point_t* get_pt3() const { return _pt3; }
	
  inline virtual type_t get_min_x() const
    { return MINFOUR(_pt0->x(), _pt1->x(), _pt2->x(), _pt3->x()); }
  
  inline virtual type_t get_min_y() const
    { return MINFOUR(_pt0->y(), _pt1->y(), _pt2->y(), _pt3->y()); }
  
  inline virtual type_t get_max_x() const
    { return MAXFOUR(_pt0->x(), _pt1->x(), _pt2->x(), _pt3->x()); }
  
  inline virtual type_t get_max_y() const
    { return MAXFOUR(_pt0->y(), _pt1->y(), _pt2->y(), _pt3->y()); }
  
  virtual inline type_t get_area() const 
    {
      return fabs((_pt1->y() - _pt0->y()) * (_pt2->x() - _pt0->x()) -
                  (_pt1->x() - _pt0->x()) * (_pt2->y() - _pt0->y()));
    }

  virtual double get_elongatedness() const;
  inline virtual bool is_centered() const { return false; }
	
  virtual void push_points(vcl_list<point_t* >& pts) const;
	
  //--- Computations ------------------------------------------  
  virtual bool inside_region(type_t x, type_t y) const;
  virtual inline bool inside_region(const point_t& pt) const 
    {
      return inside_region(pt.x(), pt.y());
    }

  virtual type_t distance_from(const region& reg) const;
  inline point_t* center() const;
	
  //--- region Modification -----------------------------------
  virtual inline bool switch_region();
  virtual bool update_region();
	
  //--- region I/O --------------------------------------------
  virtual inline void print(vcl_ostream& str) const;
	
  //--- Safe Down Casting -------------------------------------
  parall_region* cast_to_parallregion(){ return this; }
  const parall_region* cast_to_parallregion() const { return this; }
	
protected:
  virtual inline void create_region_displayer() const;
  bool set_inside_variables();
       
  type_t distance_from_para(const parall_region &reg) const;
  type_t distance_from_ellip(const ell_region &reg) const;
  bool line_segment_intersect(point_t *p0, point_t *p1, 
                            point_t *p2, point_t *p3) const;
  type_t distance_from_point_to_linesegment(point_t *p0, point_t *p1, 
                                        point_t *p2) const;
  type_t distance_between_line_segments(point_t *p0, point_t *p1, 
                                     point_t *p2, point_t *p3) const;
	
  //--- Data members ----------------------------------------
protected:
  point_t* _pt3;
  bool _inside01, _inside02, _inside13, _inside23; // variables used to determine "inside" of region
};




//===========================================================
//=== INLINE FUNCTION DEFINITIONS                         
//===========================================================

//--- computations ------------------------------------------
point_t* parall_region::center() const {
  return new point_t((_pt1->x() + _pt2->x())/2.0, (_pt1->y() + _pt2->y())/2.0);
} 


//--- region I/O --------------------------------------------

void parall_region::print(vcl_ostream& str) const
{
	/*
  str << "Parall. region of type " << _stype << vcl_endl;
  str << "   Pt0: " << *_pt0 << vcl_endl;
  str << "   Pt1: " << *_pt1 << vcl_endl;
  str << "   Pt2: " << *_pt2 << vcl_endl;
  str << "   Pt3: " << *_pt3 << vcl_endl;
  if (_anchor)
    str << "   anchor: " << *_anchor << vcl_endl;
  if (_image) 
    str << "   in image " << _image->GetName() << vcl_endl;  
  if (_inv) 
    str << "   Inv: " << *_inv << vcl_endl;
	*/
}






//--- region Modification ------------------------------------
bool parall_region::switch_region()
{
  type_t x1 = _pt1->x();
  type_t y1 = _pt1->y();
	
  _pt1->set(_pt2->x(), _pt2->y());
  _pt2->set(x1, y1);
	
  update_region();
	
  return true;
}

//--- Protected functions ------------------------------------

// mutable const
void parall_region::create_region_displayer() const
{
  _display = new parall_region_displayer(this);
}




#endif // end of class parall_region
