#include <regions/region.h>
#include <regions/region_displayer.h>

#include <vcl/vcl_iostream.h>
#include <vgl/vgl_line_2d.h>

#include <math.h>

int region::_next_ID = 0;

//-------------------------------------------------------------------
//--- CONSTRUCTORS / DESTRUCTORS ------------------------------------
//-------------------------------------------------------------------

region::region()
  :_anchor(NULL), _image(NULL),
   _type(region::REGION_NO_TYPE), _stype(region::REGION_NO_SPECIFIC_TYPE),
   _inv(NULL), _display(NULL), _stat(NULL)
{
  _pt0 = new point_t;
  _pt1 = new point_t;
  _pt2 = new point_t;
  _ID = _next_ID;
  _next_ID++;
  _sid = 0;
}


region::region(const point_t* p0,
               const point_t* p1,
               const point_t* p2,
               region_type type, region_specific_type stype,
               const point_t* anchor, vil1_image* image)
  : _anchor(anchor)
{
  _pt0 = new point_t(*p0);
  _pt1 = new point_t(*p1);
  _pt2 = new point_t(*p2);

  _image = image;
  _type = type;
  _stype = stype;
  _inv = NULL;
  _display = NULL;
  _stat = NULL;

  _ID = _next_ID;
  _next_ID++;
  _sid = 0;
}
 
region::region(const region& reg)
{
  _pt0 = new point_t(*(reg._pt0));
  _pt1 = new point_t(*(reg._pt1));
  _pt2 = new point_t(*(reg._pt2));
  _anchor = reg._anchor;
  _image = reg._image;
  _type = reg.get_type();
  _stype = reg.get_specific_type();
  _display = NULL;
  _stat = NULL;
   
  _ID = _next_ID;
  _next_ID++;
  _sid = reg.get_sid();


  if (reg.get_invariants_ptr())
  {
    _inv = new vnl_vector<double>(reg.get_invariants());
  }
  else
  {
    _inv = NULL;
  }
}

region::~region()
{
  if (_pt0)
  {
    delete _pt0;
  }
	
  if (_pt1)
  {
    delete _pt1;
  }
	
  if (_pt2)
  {
    delete _pt2;
  }
	
  if (_inv) 
  {
    delete _inv;
  }
	
  if (_stat) 
  {
    delete _stat;
  }
	
  if (_display) 
  {
    delete _display;
  }
}
  

//-------------------------------------------------------------------
//--- Computations --------------------------------------------------  
//-------------------------------------------------------------------
bool region::good_region(double eccentr, double min_area, double max_area) const
{
  double e = get_elongatedness();
  
  if ( e > eccentr )
  {
    //vcl_cout << "Too excentric (" << e << ")" << vcl_endl;
    
    return false;
  }
	
  double area = get_area();
	
  if ( area < min_area )
  {
    //vcl_cout << "Area (" << area << ") too small (" << min_area<<")" << vcl_endl;
    
    return false;
  }
	
  if ( area > max_area )
  {
    //vcl_cout << "Area (" << area << ") too large (" << max_area<<")" << vcl_endl;
    return false;
  }

  if ( _pt0->x() < 2 )
  {
    //vcl_cout << "_pt0->x() < 2 " << vcl_endl;
    return false;
  }

  if ( _pt0->y() < 2 )
  {
    //vcl_cout << "_pt0->y() < 2 " << vcl_endl;
    return false;
  }
	
  return true;
}



// This function returns an approximate distance
// (distance between the centers of Reg and this)
// faster than distance_from
type_t region::approximate_distance_from(const region &reg) const
{
  point_t *P = reg.get_pt0();
	
  return sqrt( (_pt0->x()*P->x())*(_pt0->x()*P->x()) +
               (_pt0->y()*P->y())*(_pt0->y()*P->y()) );
}
