#ifndef region_h
#define region_h

//----------------------------------------------------------------------------
// .NAME region - Basic class for (affinely invariant) regions
// .LIBRARY 
// .HEADER 
// .INCLUDE 
// .FILE   region.h
// .FILE   region.C
//
// .SECTION Description
//    The class region is an abstract base class for
//  (affinely invariant) regions.
//  Their internal representation consists of three points,
//  coined 'pt0', 'pt1' and 'pt2' whose meaning/interpretation 
//  depends on the actual region type (at the moment either
//  parall_region or ell_region.
//
//  This class was written to be used with the class regionFinder,
//  which allows to find regions in an affinely invariant way,
//
// .SECTION Author:
//          Tinne Tuytelaars - Andrew Turina - Vittorio Ferrari
//          ESAT / KULeuven           BIWI / ETH Zurich
//
// .SECTION Modifications:
//     07/01 Restructured and resynchronized code      
//
//----------------------------------------------------------------------------

/******************************************************************************
*
*	Ported to vxl by Aaron Luttman
*	Notes:	This class owns the following memory: _pt0, _pt1, _pt2, _inv, _stat,
*	_display.  The other pointers should not be newed or deleted here.
*
******************************************************************************/

#include <vcl/vcl_ostream.h>
#include <vcl/vcl_iostream.h>
#include <vcl/vcl_list.h>
#include <vbl/vbl_ref_count.h>
#include <vul/vul_timestamp.h>
#include <vgl/vgl_homg_point_2d.h>
#include <vil1/vil1_image.h>
#include <vnl/vnl_vector.h>

#include <misc_utility/defines.h>
#include <regions/region_statistics.h>
#include <regions/region_displayer.h>

class ell_region;
class parall_region;


class region : public vbl_ref_count, vul_timestamp
{
public:
  enum region_type 
    {
      REGION_NO_TYPE=0,
      PARALL_REGIONTYPE,
      ELL_REGIONTYPE,
      NB_OF_TYPES // last element in enum refers to total number of regiontypes
    };
	
  enum region_specific_type 
    {
      REGION_NO_SPECIFIC_TYPE = 0,
      PARALL_STRAIGHT,
      PARALL_STRAIGHT_R,
      PARALL_STRAIGHT_G,
      PARALL_STRAIGHT_B,
      PARALL_CURVED,
      PARALL_CURVED_R,
      PARALL_CURVED_G,
      PARALL_CURVED_B,
      PARALL_HOMOGENEOUS,
      PARALL_HOMOGENEOUS_R,
      PARALL_HOMOGENEOUS_G,
      PARALL_HOMOGENEOUS_B,
      ELL,
      ELL_R,
      ELL_G,
      ELL_B,
      NB_OF_SPECIFIC_TYPES // last element in enum refers to total number of regiontypes
    };
	
protected:
  region();
  
  region(const point_t* p0, const point_t* p1,
         const point_t* p2, 
         region_type type = region::REGION_NO_TYPE,
         region_specific_type stype = region::REGION_NO_SPECIFIC_TYPE,
         const point_t* anchor = NULL, vil1_image* image = NULL);
  
  region(const region& reg);
	
public:
  virtual ~region();
	
public:
  //--- Data Access -------------------------------------------
  inline point_t* get_pt0() const { return _pt0; }
  inline point_t* get_pt1() const { return _pt1; } 
  inline point_t* get_pt2() const { return _pt2; }
  
  inline region_type get_type() const { return _type; }
  inline region_specific_type get_specific_type() const { return _stype; }
  
  inline vnl_vector<double> get_invariants() const { return *_inv; } 
  inline const vnl_vector<double>* get_invariants_ptr() const { return _inv; } 
  inline const point_t* get_anchor_point() const { return _anchor; }
  inline const vil1_image* get_image() const { return _image; }
  inline region_statistics* get_statistics() const { return _stat; }
	
  virtual type_t get_min_x() const = 0;
  virtual type_t get_min_y() const = 0;
  virtual type_t get_max_x() const = 0;
  virtual type_t get_max_y() const = 0;
  virtual type_t get_area() const = 0;
  virtual double get_elongatedness() const = 0;
	
  int get_ID() {return _ID;}
  inline int get_sid() const {return _sid;}

  virtual void push_points(vcl_list<point_t* >& pts) const = 0;


  
  //--- Computations ------------------------------------------  
  virtual bool inside_region(type_t x, type_t y) const = 0;
	
  virtual inline bool inside_region(const point_t& pt) const 
    {
      return inside_region(pt.x(), pt.y());
    }
	
  virtual bool good_region(double eccentr, double min_area, double max_area) const;
  virtual bool is_centered() const = 0;
  virtual type_t distance_from(const region& reg) const = 0;
  virtual type_t approximate_distance_from(const region &reg) const;

  virtual point_t* center() const = 0; // creates a new point_t object
	
  //--- region Modification -----------------------------------
  virtual inline bool set_pt0(point_t* pt);
  virtual inline bool set_pt1(point_t* pt);
  virtual inline bool set_pt2(point_t* pt);
  virtual inline bool set_pt0(float x, float y);
  virtual inline bool set_pt1(float x, float y);
  virtual inline bool set_pt2(float x, float y);
  virtual inline bool set_invariants(const vnl_vector<double>& new_inv);
  virtual inline bool set_invariants(vnl_vector<double>* new_inv);
  virtual inline bool clear_invariants();
  virtual inline bool set_specific_type(region_specific_type stype);
  virtual inline bool set_anchor_point(point_t* pt);
  virtual inline bool set_image(vil1_image* im);
  virtual inline bool set_statistics(region_statistics* stat);
	
  virtual inline bool double_axes();
  virtual inline bool half_axes();
  virtual inline bool switch_region() = 0;

  inline void set_sid(int sid) {_sid = sid;};

  virtual bool update_region() = 0;
	
  //--- region I/O --------------------------------------------
  virtual inline void print(vcl_ostream& str) const;
  virtual inline void show() const; // mutable const 

  //--- Safe Down Casting -------------------------------------
  virtual void* Cast() { return this; }
  virtual inline ell_region* cast_to_ellregion(){ return NULL; }
  virtual inline const ell_region* cast_to_ellregion() const { return NULL; }
  virtual inline parall_region* cast_to_parallregion(){ return NULL; }
  virtual inline const parall_region* cast_to_parallregion() const { return NULL; }
	
protected:
  virtual inline void create_region_displayer() const;
  
	
  //--- Data members ------------------------------------------
protected:
  point_t* _pt0;
  point_t* _pt1;
  point_t* _pt2;
	
  const point_t* _anchor;
  vil1_image* _image;
	
  region_type _type;
  region_specific_type _stype;
	
  vnl_vector<double>* _inv;

  int _ID; // unique region-ID
  static int _next_ID;

  int _sid; // synchronization id (used in multiple view context)

  mutable region_displayer* _display; 
  region_statistics* _stat;     

};




//===========================================================
//=== INLINE FUNCTION DEFINITIONS                         
//===========================================================

//--- region Modification -----------------------------------
bool region::set_pt0(point_t* pt)
{  
  if (!pt)
  {
    return false;
  }

  if ( _pt0 )
  {
    delete _pt0;
  }

  _pt0 = new point_t(*pt);
	
  return true;
}

bool region::set_pt0(float x, float y)
{  
  _pt0->set(x,y);
	
  return true;
}

bool region::set_pt1(point_t* pt)
{
  if (!pt)
  {
    return false;
  }

  if ( _pt1 )
  {
    delete _pt1;
  }

  _pt1 = new point_t(*pt);
	
  return true;
}

bool region::set_pt1(float x, float y)
{  
  _pt1->set(x,y);
	
  return true;
}

bool region::set_pt2(point_t* pt) 
{
  if (!pt)
  {
    return false;
  }

  if ( _pt2 )
  {
    delete _pt2;
  }

  _pt2 = new point_t(*pt);
	
  return true;
}

bool region::set_pt2(float x, float y)
{  
  _pt2->set(x,y);
	
  return true;
}

bool region::set_invariants(const vnl_vector<double>& new_inv)
{
  if ( _inv )
  {
    delete _inv;
  }

  _inv = new vnl_vector<double>(new_inv);

  return true;
}

bool region::set_invariants(vnl_vector<double>* new_inv) 
{
  if (!new_inv) 
  {
    return false;
  }

  if ( _inv )
  {
    delete _inv;
  }

  _inv = new_inv;

  return true;
}

bool region::clear_invariants()
{
  if ( _inv )
  {
    delete _inv;
  }

  _inv = NULL;

  return true;
}

bool region::set_specific_type(region_specific_type stype) 
{
  _stype = stype;
  return true;
}

bool region::set_anchor_point(point_t* pt) 
{
  if (!pt) 
  {
    return false;
  }

  _anchor = pt;

  return true;
}

bool region::set_image(vil1_image* im) 
{
  if (!im) 
  {
    return false;
  }

  _image = im;

  return true;
}

bool region::set_statistics(region_statistics* stat) 
{
  if (!stat) 
    return false;


  if (_stat) 
    delete _stat;


  _stat = stat;

  return true;
}

bool region::double_axes() 
{
  // change pt1
  _pt1->set(_pt0->x() + 2 * (_pt1->x() - _pt0->x()),
            _pt0->y() + 2 * (_pt1->y() - _pt0->y()));
	
  // change pt2
  _pt2->set(_pt0->x() + 2 * (_pt2->x() - _pt0->x()),
            _pt0->y() + 2 * (_pt2->y() - _pt0->y()));
	
  // update region
  update_region();
	
  return true;
}

bool region::half_axes()
{
  // change pt1
  _pt1->set(_pt0->x() + 0.5 * (_pt1->x() - _pt0->x()),
            _pt0->y() + 0.5 * (_pt1->y() - _pt0->y()));
	
  // change pt2
  _pt2->set(_pt0->x() + 0.5 * (_pt1->x() - _pt0->x()), _pt0->y() + 0.5 * (_pt1->y() - _pt0->y()));
	
  // update region
  update_region();
	
  return true;
}


//--- region I/O --------------------------------------------
void region::print(vcl_ostream& str) const
{
  /*str << "region of type " << _type << vcl_endl;
    str << "   Pt0: " << *_pt0 << vcl_endl;
    str << "   Pt1: " << *_pt1 << vcl_endl;
    str << "   Pt2: " << *_pt2 << vcl_endl;
    if (_anchor)
    str << "   anchor: " << *_anchor << vcl_endl;
    if (_vil1_image) 
    str << "   in vil1_image " << _vil1_image->GetName() << vcl_endl;
    if (_inv) 
    str << "   Inv: " << *_inv << vcl_endl;
  */
}

//mutable const
void region::show() const 
{
  if (!_display) 
  {
    create_region_displayer();
  }

  _display->show();
}

//--- Protected functions ----------------------------------

//mutable const
void region::create_region_displayer() const 
{
  _display = new region_displayer(this);
}



#endif // end of class region
